#ifndef _I_EG3000_H
#define _I_EG3000_H

#include "defs.h"

/// \cond EXCLUDED
#ifdef _WIN32
#define _DllImport __declspec( dllimport )
#define _DllExport __declspec( dllexport )
#else
#define _DllImport
#define _DllExport
#endif

#if defined(__compiletime) && !defined(_EG3000_STATIC)
#define _Dll _DllExport
#else
#ifdef _EG3000_DLL
#define _Dll _DllImport
#else
#define _Dll extern
#endif
#endif
/// \endcond

/** \file
  * \brief Interface library for EG3000 controllers
  */

/** \mainpage Interface library for EG3000 controllers
  *
  * Documentation:
  * \link i_eg3000.h
  */

#define VID 0x0403 ///< EG3000 controller vendor id
#define PID 0x89B0 ///< EG3000 controller product id

#if defined __cplusplus
#pragma once
extern "C" {
#endif

#ifndef __return_values
/// \cond EXCLUDED
#define __return_values
/// \endcond
/** \brief All functions return this struct as BYTE or nothing
  */
enum _return_values {
    RET_FALSE = 0, ///< fatal error, like read or malloc functions failed
    RET_TRUE = 1, ///< succesfull completion
    RET_TRUE_ERR = 2 ///< non-fatal error, like wrong input or struct
};
#endif

//low-level interface wrappers
#ifndef _s_info
/// \cond EXCLUDED
#define _s_info
/// \endcond
/** \brief Device information structure
  */
struct s_info {
    void *handle; ///< handle to pass to eg3000_open() function
    DWORD flags; ///< device flags
};

/** \brief Device flags
  */
enum s_info_flags {
    F_NO = 0 ///< Empty flags
};
#endif
/** \brief Change backend
  * \param id index of backend, 0 - ftdi, 1 - libusb
  * \return 0 on success
  */
BYTE _Dll eg3000_set_backend(WORD id);
/** \brief Check if prev r/w op failed
  * \param handle hw device handle
  * \return true if ok
  */
BYTE _Dll eg3000_is_ok(void *handle);
/** \brief Acquires the number of available devices
  *
  * Doesn't count for already opened devices
  * \param vid vendor id
  * \param pid product id
  * \param size pointer to where the number of devices will be stored
  * \return false if any low-level functions failed
  */
BYTE _Dll eg3000_get_number(WORD vid, WORD pid, WORD *size);
/** \brief Acquires the list of available devices
  *
  * Doesn't count for already opened devices
  * \param vid vendor id
  * \param pid product id
  * \param size pointer to where the number of devices in list will be stored
  * \param arr_flags pointer to where the list of devices will be stored, could be NULL to get available devices number
  * \param array_size size of array at arr_flags, could be 0 to get available devices number
  * \return false if any low-level functions failed
  */
BYTE _Dll eg3000_list(WORD vid, WORD pid, WORD *size, struct s_info *arr_flags, WORD array_size);
/** \brief Attempts to open device with handle
  * \param handle device handle from struct s_info that you wish to open
  * \param vid vendor id
  * \param pid product id
  * \param ret_handle pointer to where the hw device handle will be stored
  * \return false if any low-level functions failed, or can't open device
  */
BYTE _Dll eg3000_open(void *handle, WORD vid, WORD pid, void **ret_handle);
/** \brief Closes the device
  * \param handle hw device handle
  * \return false if close failed
  */
BYTE _Dll eg3000_close(void *handle);
/** \brief Clears recieve & transmit buffers assotiated with device
  * \param handle hw device handle
  * \return false if clear failed
  */
BYTE _Dll eg3000_clear(void *handle);
/** \brief Sets transmit and/or recieve timeouts(in ms) assotiated with device
  * \param handle hw device handle
  * \return false if failed
  */
BYTE _Dll eg3000_set_timeout(void *handle, WORD timeout);

/** \brief Sends raw data
  * \param handle hw device handle
  * \param buffer data to send
  * \param data_size size of data to send
  * \return false if send failed
  */
BYTE _Dll eg3000_raw_send(const void *handle, const void *buffer, WORD data_size);
/** \brief Sends packet
  * \param handle hw device handle
  * \param Address destination address
  * \param SubAddress destination subcommand
  * \param buffer data to send
  * \param data_size size of data to send
  * \return false if send failed
  */
BYTE _Dll eg3000_send_packet(const void *handle, BYTE Address, BYTE SubAddress, const void *buffer, WORD data_size);
/** \brief Sends packet with WORD data
  * \param handle hw device handle
  * \param Address destination address
  * \param SubAddress destination subcommand
  * \param Data data to send
  * \return false if send failed
  */
BYTE _Dll eg3000_send_word_packet(const void *handle, BYTE Address, BYTE SubAddress, WORD Data);
/** \brief Sends packet with no data
  * \param handle hw device handle
  * \param Address destination address
  * \param SubAddress destination subcommand
  * \return false if send failed
  */
BYTE _Dll eg3000_send_empty_packet(const void *handle, BYTE Address, BYTE SubAddress);
/** \brief Reads raw data
  * \param handle hw device handle
  * \param buffer buffer to store data
  * \param buf_size size of buffer
  * \param p_bytes_read pointer to where the number of bytes read will be stored
  * \return false if read failed
  */
BYTE _Dll eg3000_read(void *handle, void *buffer, WORD buf_size, DWORD *p_bytes_read);


//packet recv & parsing interface
/** \brief Structure used for eg3000_update_parse() packet parser
  *
  * Should be initialised first with eg3000_init_read_handle() \n
  * and deinitialised with eg3000_close_read_handle()
  */
struct eg3000_read_handle {
    void *buffer; ///< internal
    DWORD buf_data_length; ///< internal
    DWORD buffer_size; ///< internal
    DWORD packet_size;
    BYTE header_read; ///< internal, bool
    BYTE packet_recieved; ///< bool
    BYTE address;
    BYTE sub_address;
    BYTE packet_malformed; ///< bool
};
/** \brief Should be called before use
  * \param p_read_handle pointer to eg3000_read_handle to be initialised
  *
  * Sucessive calls to this function will cause memory leaks
  */
void _Dll eg3000_init_read_handle(struct eg3000_read_handle *p_read_handle);
/** \brief Should be called after use
  * \param p_read_handle pointer to eg3000_read_handle to be closed
  */
void _Dll eg3000_close_read_handle(struct eg3000_read_handle *p_read_handle);
/** \brief Cleares parser state
  * \param p_read_handle pointer to eg3000_read_handle to be cleared
  */
void _Dll eg3000_clear_read_handle(struct eg3000_read_handle *p_read_handle);
/** \brief Removes size bytes from start of the buffer
  * \param p_read_handle pointer to eg3000_read_handle
  * \param size number of bytes to remove
  */
BYTE _Dll eg3000_remove_read_handle(struct eg3000_read_handle *p_read_handle, WORD size);

/** \brief Reads available data from device and parses it
  * \param handle hw device handle
  * \param p_read_handle pointer to eg3000_read_handle to be used
  *
  * eg3000_read_handle::packet_recieved can be used to check for packet available
  * Note, that packet should be either eg3000_get_packet() or eg3000_discard_packet() before next call to this function
  * You should use eg3000_read_handle::address, eg3000_read_handle::sub_address to check which packed arrived
  * For sample packet handling refer to examples
  * You can parse packets directly from eg3000_read_handle::buffer, then eg3000_discard_packet() them
  * \return false if read failed
  */
BYTE _Dll eg3000_update_parse(void *handle, struct eg3000_read_handle *p_read_handle);
/** \brief Gets packet from p_read_handle to buffer, then discards it
  *
  * Buffer should be at least eg3000_read_handle::packet_size bytes
  * Does nothing if no packet available
  * \param p_read_handle pointer to eg3000_read_handle
  */

BYTE _Dll eg3000_get_packet(struct eg3000_read_handle *p_read_handle, BYTE *buffer, DWORD buffersize);
/** \brief Discards packet from p_read_handle
  *
  * Does nothing if no packet available
  * \param p_read_handle pointer to eg3000_read_handle
  */
BYTE _Dll eg3000_discard_packet(struct eg3000_read_handle *p_read_handle);


//bootloader
/** \brief You should never call this function!
  */
BYTE _Dll eg3000_bl_program(void *handle, BYTE *buffer, DWORD size, struct eg3000_read_handle *p_read_handle, BYTE type);
/** \brief You should never call this function!
  */
BYTE _Dll eg3000_bl_verify(void *handle, BYTE *buffer, DWORD size, struct eg3000_read_handle *p_read_handle, BYTE type);
/** \brief You should never call this function!
  */
BYTE _Dll eg3000_bl_erase(void *handle);
/** \brief Boots bootloader to firmware
  *
  * Does nothing if firmware is booted
  */
BYTE _Dll eg3000_bl_run(void *handle);
/** \brief Requests bootloader info
  *
  * Can be used to ping bootloader part
  * Does nothing if firmware is booted
  */
BYTE _Dll eg3000_bl_stats(void *handle);


//simple functions, proxies for address (and on/off byte)
/** \brief Ping firmware
  */
BYTE _Dll eg3000_cpu_loopback(void *handle);
BYTE _Dll eg3000_fb_control(void *handle, BYTE val);
BYTE _Dll eg3000_set_head_dac(void *handle, WORD val);
BYTE _Dll eg3000_set_head_piezotip(void *handle, BYTE val);
BYTE _Dll eg3000_set_head_lfl_current(void *handle, BYTE val);
BYTE _Dll eg3000_set_head_ext_current(void *handle, BYTE val);
BYTE _Dll eg3000_csb_overflow(void *handle, BYTE scanner, BYTE mode);
BYTE _Dll eg3000_csb_double_c_out(void *handle, BYTE scanner, BYTE mode);

BYTE _Dll eg3000_hv_zero(void *handle);
BYTE _Dll eg3000_osc_startstop(void *handle, BYTE on);
BYTE _Dll eg3000_status_onoff(void *handle, BYTE on);

//struct sending functions
/** \brief Sets oscilloscope channels, up to 4
  */
BYTE _Dll eg3000_osc_set_data(void *handle, BYTE ch_a, BYTE ch_b, BYTE ch_c, BYTE ch_d);

/** \brief Send calibration data to controller
  */
BYTE _Dll eg3000_send_caldata(void *handle, int scanner,
                              float arr[48], DWORD mult[3], long shift[3],
                              BYTE nlc_status_cl_xy, BYTE cl_status_xy, float free_positioning_speed,
                              BYTE nlc_status_cl_z, BYTE cl_status_z,
                              BYTE nlc_status_ol_xy, BYTE nlc_status_ol_z);

/** \brief Set x, y, z for head or stage, with close or open loop, nonlinear correction, as in calibration data
  */
BYTE _Dll eg3000_set_point_xyz(void *handle, int scanner, float x_sp, float y_sp, float z_sp);

BYTE _Dll eg3000_scan_speed(void *handle, int scanner, float v);
BYTE _Dll eg3000_scan_params(void *handle, int scanner,
                             WORD NX, WORD NY, float V,
                             float stepx, float stepy, float start_point_x, float start_point_y,
                             BYTE dir_point_type, BYTE scan_a, BYTE scan_b, BYTE scan_aback, BYTE scan_bback,
                             BYTE LmEn, BYTE LmEnA, BYTE LmEnB, BYTE LmEnC, BYTE LmEnD, float LmHeight,
                             BYTE fast_scan_correction, BYTE precise_mode);

BYTE _Dll eg3000_ccd_scan_params(void *handle, int scanner,
                                 int mode, WORD NX, WORD NY,
                                 float stepx, float stepy, float start_point_x, float start_point_y,
                                 BYTE dir_point_type);

BYTE _Dll eg3000_head_params(void *handle, BYTE laser_on_off, BYTE ac_dc);
BYTE _Dll eg3000_fb_pid_params(void *handle, BYTE scanner, float p, float i, float d, BYTE src_channel, BYTE sign);
BYTE _Dll eg3000_kelvin_params(void *handle, BYTE on, float p);
BYTE _Dll eg3000_fb_pid_setpoint(void *handle, float val);

BYTE _Dll eg3000_frequency_scan_start(void *handle, float f_start, float f_end, BYTE channel);
BYTE _Dll eg3000_set_frequency(void *handle, float freq);
BYTE _Dll eg3000_set_frequency2(void *handle, float freq);
BYTE _Dll eg3000_set_cutoff_frequency(void *handle, float cutoff_freq);

BYTE _Dll eg3000_set_ampl_gains(void *handle, int gain1, int gain2, int gain3);
BYTE _Dll eg3000_set_ampl_gains2(void *handle, int gain1, int gain2, int gain3);
BYTE _Dll eg3000_set_phase(void *handle, float phase);
BYTE _Dll eg3000_set_piezo_drive_voltage(void *handle, float voltage);
BYTE _Dll eg3000_set_piezo_drive_voltage_lift(void *handle, float voltage);
BYTE _Dll eg3000_set_cap_mode(void *handle, BYTE isOn);
BYTE _Dll eg3000_set_cap_mode2(void *handle, BYTE isOn);
BYTE _Dll eg3000_set_extra_DAC1(void *handle, float v);
BYTE _Dll eg3000_set_extra_DAC2(void *handle, float v);
BYTE _Dll eg3000_set_extra_DAC3(void *handle, float v);
BYTE _Dll eg3000_set_extra_DAC4(void *handle, float v);

BYTE _Dll eg3000_set_spectroscopy_params(void *handle, BYTE on_off,
                                         float z1, float z2, float Speed,
                                         BYTE Channel, WORD Size);

//step motors functions
BYTE _Dll eg3000_set_motor_leveling(void *handle, BYTE motor_number, BYTE cont_byte, DWORD steps, float speed);
BYTE _Dll eg3000_set_landing(void *handle, BYTE mode, BYTE direction, float speed, float speed_slow, float close_perc);
BYTE _Dll eg3000_set_landing_ex(void *handle, BYTE mode, BYTE direction, float speed, float speed_slow, float close_perc,
                                EG_LandingMCfg *mcfg, int mcfg_elems);

#if defined __cplusplus
} // extern "C" {
#endif

#endif // _I_EG3000_H
