#include "plot_2dplot.h"
#include "plot_2dplot_p.h"
#include "math/plot_MathProcessing.h"

#include "../../src/gui/flowlayout.hpp"
#include <qwt_plot_directpainter.h>
#include <qwt_plot_layout.h>
#include <qwt_plot_marker.h>
#include <qwt_plot_canvas.h>
#include <algorithm>
#include <limits>

#include <QApplication>
#include <QDesktopWidget>
#include <QtEvents>
#include <QColorDialog>
#include <QDrag>
#include <QTimer>
#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>
#include <QGestureEvent>
#include <QDebug>
#include <QWeakPointer>
#include "plot_2dpure.h"
#include "widgets/twodoverlaybuttons.h"

namespace plot {

const char *ManualFitStr = "Manual fit";

//================================================================================================================================
void T2dPlotControl::GUICreate()
{
    TCommonPlotControl::GUICreate();

    BAutoScaleX = new QPushButton;
    BAutoScaleX->setIcon(makeHugeIcon(":/plot/AutoscX_ones.png"));
    BAutoScaleX->setToolTip(QObject::tr("Autoscale in X axis"));

    BAutoScaleY = new QPushButton;
    BAutoScaleY->setIcon(makeHugeIcon(":/plot/AutoscY_ones.png"));
    BAutoScaleY->setToolTip(QObject::tr("Autoscale in Y axis"));

    TBCurveThickness = new QToolButton;
    TBCurveThickness->setIcon(makeHugeIcon(":/plot/line_thickness.png"));
    TBCurveThickness->setToolTip(QObject::tr("Thickness of line"));
    TBCurveThickness->setPopupMode(QToolButton::InstantPopup);
    QMenu *MCurveThickness = new QMenu;
    TBCurveThickness->setMenu(MCurveThickness);

    TBCurveColour = new QToolButton;
    TBCurveColour->setIcon(makeHugeIcon(":/plot/linecolor.png"));
    TBCurveColour->setToolTip(QObject::tr("Colour of line"));
    TBCurveColour->setPopupMode(QToolButton::InstantPopup);
    QMenu *MCurveColour = new QMenu;
    TBCurveColour->setMenu(MCurveColour);

    CreateActionsAndMenus();

    CBChooseActiveCurve = new TComboBox;
    CBChooseActiveCurve->addItem("All");

    BResetMCCurve = new QPushButton;
    BResetMCCurve->setIcon(makeHugeIcon(":/plot/Reset.png"));
    BResetMCCurve->setToolTip(tr("Reset manual curve fit"));
    BResetMCCurve->hide();

    ControlMainLayout->insertWidget(7, BAutoScaleX);
    ControlMainLayout->insertWidget(8, BAutoScaleY);

    addDelimeterH(ControlMainLayout, BAutoScaleX->sizeHint().width());
    ControlMainLayout->addWidget(CBChooseActiveCurve);
    ControlMainLayout->addWidget(TBCurveColour);
    ControlMainLayout->addWidget(TBCurveThickness);
    ControlMainLayout->addWidget(TBFontSize);
    //addDelimeterH(ControlMainLayout, BAutoScaleX->sizeHint().width());

    ControlMainLayout->addWidget(BCancel);
    ControlMainLayout->addWidget(BApply);
    ControlMainLayout->addWidget(BResetMCCurve);

    setTitle("2D");
}

void T2dPlotControl::CreateActionsAndMenus()
{
    AGFitOrders = new QActionGroup(this);
    const QString FitOrders[] = {tr("Fit. None"), tr("Fit. 0 order"), tr("Fit. 1 order"), tr("Fit. 2 order"),
                               tr("Fit. 3 order"), tr("Fit. 4 order"), tr("Fit. 5 order"), tr(ManualFitStr)};
    int i = 0;
    for (auto &fitOrder:FitOrders)
    {
        AGFitOrders->addAction(MMathProcessing->addAction(fitOrder));
        AGFitOrders->actions()[i]->setCheckable(true);
        i++;
    }
    AGFitOrders->actions()[0]->setChecked(true);

    AGCurveColours = new QActionGroup(this);
    const char *CurveColours[] = {"Black", "Red", "Green", "Blue", "Magenta", "Cyan", "Other..."};
    for (unsigned int i = 0; i < sizeof(CurveColours)/sizeof(CurveColours[0]); i++)
    {
        AGCurveColours->addAction(TBCurveColour->menu()->addAction(tr(CurveColours[i])));
        AGCurveColours->actions()[i]->setCheckable(true);
    }
    AGCurveColours->actions()[0]->setChecked(true);

    AGCurveThickness = new QActionGroup(this);
    const char *CurveThickness[] = {"1", "2", "3", "4", "5", "6", "8", "10", "15", "20"};
    for (unsigned int i = 0; i < sizeof(CurveThickness)/sizeof(CurveThickness[0]); i++)
    {
        AGCurveThickness->addAction(TBCurveThickness->menu()->addAction(tr(CurveThickness[i])));
        AGCurveThickness->actions()[i]->setCheckable(true);
    }
    AGCurveThickness->actions()[0]->setChecked(true);
}
T2dPlotControl::~T2dPlotControl()
{
    delete CBChooseActiveCurve;
    CBChooseActiveCurve = nullptr;
}
void T2dPlotControl::ResetAllFilters()
{
    AGFitOrders->actions()[0]->setChecked(true);
    AGFitOrders->actions()[0]->trigger();
}

//================================================================================================================================
T2dPlot::T2dPlot(QObject *_Parent, bool child, QWidget *wparent, bool vertical, TwoDOverlayButtons *ovrlBtns) : QWidget(wparent)
{
    overlayButtons = ovrlBtns;
    depent2dPlot = master2dPlot = nullptr;
    Parent = _Parent;
    GUICreate(child, vertical);

    Pure2dPlot->Vertical = vertical;
    connect(Pure2dPlot, SIGNAL(UpdateAllMarkersInfo()), this, SLOT(UpdateAllMarkersInfo()));

    SyncTimer = new QTimer(this); //TODO replace!!!
    SyncTimer->start(850);
    connect(SyncTimer,  SIGNAL(timeout()),              this, SLOT(on_SyncTimer_timeout()));

    if (overlayButtons)
    {
        connect(Pure2dPlot, SIGNAL(mouseEvent(QMouseEvent*, TPure2dPlot*)),
            overlayButtons, SLOT(plot2DMouseEvent(QMouseEvent*, TPure2dPlot*)));
        connect(Pure2dPlot, SIGNAL(resized(QSize, TPure2dPlot*)),
            overlayButtons, SLOT(plot2DResized(QSize, TPure2dPlot*)));
    }
}

void T2dPlot::CreateOverlayButtons()
{
    #ifndef _ADROID
    overlayButtons = new TwoDOverlayButtons();
#endif
    connect(Pure2dPlot, SIGNAL(mouseEvent(QMouseEvent*, TPure2dPlot*)),
        overlayButtons, SLOT(plot2DMouseEvent(QMouseEvent*, TPure2dPlot*)));
    connect(Pure2dPlot, SIGNAL(resized(QSize, TPure2dPlot*)),
        overlayButtons, SLOT(plot2DResized(QSize, TPure2dPlot*)));
    overlayButtons->set2DPlot(this);
}

void T2dPlot::ConnectPlotControlSlots()
{
    PlotControl->BAutoScale->setChecked(true);
    PlotControl->BFixX->hide();
    PlotControl->BFixY->hide();
    PlotControl->BAutoScaleX->hide();
    PlotControl->BAutoScaleY->hide();

    connect(PlotControl->BAutoScaleX, SIGNAL(clicked()), Pure2dPlot, SLOT(on_BAutoScaleX_clicked()));
    connect(PlotControl->BAutoScaleY, SIGNAL(clicked()), Pure2dPlot, SLOT(on_BAutoScaleY_clicked()));
    connect(PlotControl->BRestoreScale, SIGNAL(clicked()), Pure2dPlot, SLOT(BRestoreScale_clicked()));
    connect(PlotControl->BExpandSelectRect, SIGNAL(clicked()), Pure2dPlot, SLOT(on_BExpandSelectRect_clicked()));

    connect(PlotControl->BMarkers, SIGNAL(toggled(bool)), this, SLOT(on_BMarkers_clicked(bool)));
    connect(PlotControl->BDetailedMarkersInfo, SIGNAL(toggled(bool)), this, SLOT(on_BDetailedMarkersInfo_clicked()));
    connect(PlotControl->BGrid, SIGNAL(toggled(bool)), Pure2dPlot, SLOT(BGrid_clicked(bool)));
    connect(PlotControl->BFixX, SIGNAL(toggled(bool)), this, SLOT(on_BFixX_toggled()));
    connect(PlotControl->BFixY, SIGNAL(toggled(bool)), this, SLOT(on_BFixY_toggled()));

    connect(PlotControl->BSelectByRect, SIGNAL(clicked()), this, SLOT(on_BSelectByRect_clicked()));
    connect(PlotControl->BCancel, SIGNAL(clicked()), this, SLOT(on_BCancel_clicked()));
    connect(PlotControl->BApply, SIGNAL(clicked()), this, SLOT(on_BApply_clicked()));

    connect(PlotControl->BAutoScale, SIGNAL(toggled(bool)), this, SLOT(on_BAutoScale_toggled()));
    connect(PlotControl->BSavePlotPixmap, SIGNAL(clicked()), this, SLOT(on_BSavePlotPixmap_clicked()));
    connect(PlotControl->BSave,  SIGNAL(clicked()), this, SLOT(on_BSave_clicked()));

    connect(PlotControl->CBChooseActiveCurve, SIGNAL(highlighted(int)), this, SLOT(on_CBChooseActiveCurve_highlighted(int)));
    connect(PlotControl->CBChooseActiveCurve, SIGNAL(activated(int)), this, SLOT(on_CBChooseActiveCurve_activated(int)));

    connect(PlotControl->BResetMCCurve, SIGNAL(clicked()), Pure2dPlot, SLOT(ResetMCCurve()));

    connect(PlotControl->AGCurveColours, SIGNAL(triggered(QAction*)), this, SLOT(on_CurveColour_changed()));
    connect(PlotControl->AGCurveThickness, SIGNAL(triggered(QAction*)), this, SLOT(on_CurveThickness_changed()));
    connect(PlotControl->AGFitOrders, SIGNAL(triggered(QAction*)), this, SLOT(on_FitOrder_changed()));
    connect(PlotControl->AGFontSizes, SIGNAL(triggered(QAction*)), this, SLOT(on_FontSize_changed()));
}
void T2dPlot::GUICreate(bool child, bool vertical)
{
    Pure2dPlot = new TPure2dPlot(vertical, Parent, this);
    Pure2dPlot->vertical = vertical;
    PlotControl = new T2dPlotControl;
    MarkersInfo = new TMarkersInfo;
    MainLayout = new QHBoxLayout(this);
    MainLayout->setContentsMargins(0, 3, 3, 0);
    /*if (Pure2dPlot->vertical)
    {
        QVBoxLayout *verticalPlotMarkers = new QVBoxLayout();
        verticalPlotMarkers->addWidget(Pure2dPlot, 1);
        verticalPlotMarkers->addWidget(MarkersInfo, 0);
        MainLayout->addLayout(verticalPlotMarkers);
    }
    else */{
        MainLayout->addWidget(MarkersInfo, 0, Qt::AlignTop);
        MainLayout->addWidget(Pure2dPlot);
    }

    MarkersInfo->GUICreate();
    MarkersInfo->LAngleLbl->setText(tr("dY/dX"));
    MarkersInfo->LAngleUnit->setText("     ");

    if (!child) {
        MainLayout->addWidget(PlotControl);
        PlotControl->GUICreate();

        FillCBChooseActiveCurve();
        ConnectPlotControlSlots();
    } else {
        Pure2dPlot->SetActiveCurve(-1);
    }

    connect(Pure2dPlot, SIGNAL(sync_active_curve(int)), this, SLOT(sync_active_curve(int)), Qt::QueuedConnection);
}
void T2dPlot::FillCBChooseActiveCurve()
{
    PlotControl->CBChooseActiveCurve->blockSignals(true);
    PlotControl->CBChooseActiveCurve->clear();
    if (overlayButtons)
    {
        overlayButtons->CBLineSelector->blockSignals(true);
        overlayButtons->CBLineSelector->clear();
    }

    for (int i = 0; i < Pure2dPlot->QwtCurves.size(); i++)
    {
        PlotControl->CBChooseActiveCurve->addItem(QString::number(i+1), i);
        if (overlayButtons)
            overlayButtons->CBLineSelector->addItem(QString::number(i+1), i);
    }

    PlotControl->CBChooseActiveCurve->addItem(tr("All"), QVariant(-1));
    PlotControl->CBChooseActiveCurve->blockSignals(false);
    if (overlayButtons)
    {
        overlayButtons->CBLineSelector->addItem(tr("All"), QVariant(-1));
        overlayButtons->CBLineSelector->blockSignals(false);
    }

    int idx = (PlotControl->CBChooseActiveCurve->count() <= 2) ? 0 : (PlotControl->CBChooseActiveCurve->count() - 1);
    if (idx >= 0)
    {
        idx = PlotControl->CBChooseActiveCurve->itemData(idx).toInt();
        Pure2dPlot->SetActiveCurve(idx);
    }

    bool Enabled = (Pure2dPlot->QwtCurves.size() != 0);
    PlotControl->CBChooseActiveCurve->setEnabled(Enabled);
    PlotControl->BAutoScaleX->setEnabled(Enabled);
    PlotControl->BAutoScaleY->setEnabled(Enabled);
    PlotControl->TBCurveColour->setEnabled(Enabled);
    PlotControl->TBCurveThickness->setEnabled(Enabled);
    if (overlayButtons)
    {
        overlayButtons->CBLineColor->setEnabled(Enabled);
        overlayButtons->CBLineWidth->setEnabled(Enabled);
        overlayButtons->CBLineSelector->setEnabled(Enabled);
    }
}
void T2dPlot::UpdateAllMarkersInfo()
{
    if (!PlotControl->BMarkers->isChecked())
        return;

    Pure2dPlot->UpdateLabels(PlotControl->BDetailedMarkersInfo->isChecked());
    if (!PlotControl->BDetailedMarkersInfo->isChecked())
        return;

	auto d_func = Pure2dPlot->d_func();
    if (Pure2dPlot->vertical)
    {
        MarkersInfo->LEdXA->setText  (d_func->AMarkerInfo.YInfo.Value);
        MarkersInfo->LXAUnit->setText(d_func->AMarkerInfo.YInfo.Unit);
        MarkersInfo->LEdYA->setText  (d_func->AMarkerInfo.XInfo.Value);
        MarkersInfo->LYAUnit->setText(d_func->AMarkerInfo.XInfo.Unit);
        MarkersInfo->LEdXB->setText  (d_func->BMarkerInfo.YInfo.Value);
        MarkersInfo->LXBUnit->setText(d_func->BMarkerInfo.YInfo.Unit);
        MarkersInfo->LEdYB->setText  (d_func->BMarkerInfo.XInfo.Value);
        MarkersInfo->LYBUnit->setText(d_func->BMarkerInfo.XInfo.Unit);
    }
    else
    {
        MarkersInfo->LEdXA->setText  (d_func->AMarkerInfo.XInfo.Value);
        MarkersInfo->LXAUnit->setText(d_func->AMarkerInfo.XInfo.Unit);
        MarkersInfo->LEdYA->setText  (d_func->AMarkerInfo.YInfo.Value);
        MarkersInfo->LYAUnit->setText(d_func->AMarkerInfo.YInfo.Unit);
        MarkersInfo->LEdXB->setText  (d_func->BMarkerInfo.XInfo.Value);
        MarkersInfo->LXBUnit->setText(d_func->BMarkerInfo.XInfo.Unit);
        MarkersInfo->LEdYB->setText  (d_func->BMarkerInfo.YInfo.Value);
        MarkersInfo->LYBUnit->setText(d_func->BMarkerInfo.YInfo.Unit);
    }

    MarkersInfo->LEdQ->setText(QString::number((int)fabs(0.5*(d_func->AMarkerInfo.XInfo.SrcValue+
                                                              d_func->BMarkerInfo.XInfo.SrcValue)/
                                                         (d_func->AMarkerInfo.XInfo.SrcValue-
                                                          d_func->BMarkerInfo.XInfo.SrcValue)) ));

    double xdist = fabs(d_func->BMarkerInfo.XInfo.SrcValue - d_func->AMarkerInfo.XInfo.SrcValue);
    double ydist = fabs(d_func->BMarkerInfo.YInfo.SrcValue - d_func->AMarkerInfo.YInfo.SrcValue);

    TValueInfoCreator ValueInfo;
    if (Pure2dPlot->vertical)
    {
        ValueInfo(Pure2dPlot->AutoLegendEnabled, Pure2dPlot->YAxisUnit, ydist);
        MarkersInfo->LEdDeltaX->setText(ValueInfo.Value); 
        MarkersInfo->LDeltaXUnit->setText(ValueInfo.Unit);

        ValueInfo(Pure2dPlot->AutoLegendEnabled, Pure2dPlot->XAxisUnit, xdist);
        MarkersInfo->LEdDeltaY->setText(ValueInfo.Value); 
        MarkersInfo->LDeltaYUnit->setText(ValueInfo.Unit);
    }
    else
    {
        ValueInfo(Pure2dPlot->AutoLegendEnabled, Pure2dPlot->XAxisUnit, xdist);
        MarkersInfo->LEdDeltaX->setText(ValueInfo.Value); 
        MarkersInfo->LDeltaXUnit->setText(ValueInfo.Unit);

        ValueInfo(Pure2dPlot->AutoLegendEnabled, Pure2dPlot->YAxisUnit, ydist);
        MarkersInfo->LEdDeltaY->setText(ValueInfo.Value); 
        MarkersInfo->LDeltaYUnit->setText(ValueInfo.Unit);
    }

    ValueInfo(Pure2dPlot->AutoLegendEnabled, Pure2dPlot->XAxisUnit, sqrt(xdist * xdist + ydist * ydist));
    MarkersInfo->LEdDistance->setText(ValueInfo.Value);

    TSIUnit Temp;
    if (Pure2dPlot->YAxisUnit.SIUnit != Pure2dPlot->XAxisUnit.SIUnit) { //prevents both null too
        if (Pure2dPlot->vertical)
        {
            Temp.SetSIUnit(Pure2dPlot->XAxisUnit.SIUnit + " / " + Pure2dPlot->YAxisUnit.SIUnit);
            Temp.UpdateMultiplier(xdist / ydist);
        }
        else
        {
            Temp.SetSIUnit(Pure2dPlot->YAxisUnit.SIUnit + " / " + Pure2dPlot->XAxisUnit.SIUnit);
            Temp.UpdateMultiplier(ydist / xdist);
        }
    }
    if (Pure2dPlot->vertical)
        ValueInfo(Pure2dPlot->AutoLegendEnabled, Temp, xdist / ydist);
    else
        ValueInfo(Pure2dPlot->AutoLegendEnabled, Temp, ydist / xdist);

    MarkersInfo->LEdAngle->setText(ValueInfo.Value);
    MarkersInfo->LAngleUnit->setText(ValueInfo.Unit);
}

void T2dPlot::on_SyncTimer_timeout()
{
    try {
        if (SyncTimer == nullptr)
            return;
        if (PlotControl->CBChooseActiveCurve == nullptr)
            return;
        //bug CBChooseActiveCurve at the end of program is 0xfeefee
        if (Pure2dPlot->QwtCurves.size() + 1 != PlotControl->CBChooseActiveCurve->count())
            FillCBChooseActiveCurve();
    } catch (...) {

    }

}
void T2dPlot::on_FitOrder_changed()
{
    int i = FindCheckedItemIn(PlotControl->AGFitOrders);
    if (PlotControl->AGFitOrders->actions()[i]->text() == QString(ManualFitStr))
        setFitOrder(TPlotEnums::FO_Manual);
    else
        setFitOrder(i - 1);
}
void T2dPlot::setFitOrder(int i)
{
    if (i == TPlotEnums::FO_Manual)
    {
        PlotControl->AGFitOrders->actions().last()->setChecked(true);
        if (Pure2dPlot->isFiltering)
            StartMCCurveConstructing();
        Pure2dPlot->SetFitOrder(TPlotEnums::FO_Manual);
    } else {
        if (Pure2dPlot->FitOrder == TPlotEnums::FO_Manual)
            Pure2dPlot->ExitMCMode();
        PlotControl->AGFitOrders->actions()[i+1]->setChecked(true);
        Pure2dPlot->SetFitOrder(i);
        Refresh();
    }
}
bool T2dPlot::event(QEvent *Event)
{
    if (Event->type() == QEvent::KeyPress)
    {
        if (((QKeyEvent *)Event)->key() == Qt::Key_Return ||
            ((QKeyEvent *)Event)->key() == Qt::Key_Enter)
        {
            if (overlayButtons)
                overlayButtons->event(Event);
        }
    }
    if (Event->type() == QEvent::Resize && isVisible() && Pure2dPlot->vertical)
    {
        if (master2dPlot->parent())
            master2dPlot->setParent(nullptr);
        master2dPlot->resize(height(), width());
        master2dPlot->Refresh();
    }

    return QObject::event(Event);
}
void T2dPlot::setFiltering(bool filter)
{
    Pure2dPlot->isFiltering = filter;
    on_FitOrder_changed();
    Refresh();
}
void T2dPlot::on_CurveColour_changed()
{
    QColor Colour;
    if (PlotControl->AGCurveColours->checkedAction()->text() == "Black") Colour = Qt::black;
    else if (PlotControl->AGCurveColours->checkedAction()->text() == "Red") Colour = Qt::red;
    else if (PlotControl->AGCurveColours->checkedAction()->text() == "Green") Colour = Qt::green;
    else if (PlotControl->AGCurveColours->checkedAction()->text() == "Blue") Colour = Qt::blue;
    else if (PlotControl->AGCurveColours->checkedAction()->text() == "Magenta") Colour = Qt::magenta;
    else if (PlotControl->AGCurveColours->checkedAction()->text() == "Cyan") Colour = Qt::cyan;
    else if (PlotControl->AGCurveColours->checkedAction()->text() == "Other...") {
        Pure2dPlot->VerifyActiveCurve();
        int act = Pure2dPlot->ActiveCurve;
        Colour = Pure2dPlot->ActiveCurve > 0 ? Pure2dPlot->QwtCurves.at(Pure2dPlot->ActiveCurve)->pen().color() : Qt::black;
        Colour = QColorDialog::getColor(Colour);
        if (!Colour.isValid())
            return;
        if (act != Pure2dPlot->ActiveCurve) {
            Pure2dPlot->ActiveCurve = act;
            Pure2dPlot->VerifyActiveCurve();
        }
        if (act != Pure2dPlot->ActiveCurve)
            return;
    }
    setCurveColor(Colour);
}

void T2dPlot::SetTitle(const QString &Ttl)
{
    Pure2dPlot->SetTitle(Ttl);
    if (overlayButtons && isVisible())
    {
        overlayButtons->plot2DResized(size(), Pure2dPlot);
    }
}

void T2dPlot::setCurveColor(QColor color, bool emitEvent)
{
    int line = Pure2dPlot->ActiveCurve;
    if (line >= 0)
        lineColors[line] = color;
    Pure2dPlot->SetAutoScaleEnabled(false, false);
    if (line >= 0 &&
        Pure2dPlot->SourceData.size() > line &&
        Pure2dPlot->SourceData.at(line))
    {
        Pure2dPlot->SourceData.at(line).data()->Pen.setColor(color);

    }
    if (emitEvent)
        emit CurveColourChanged(color, line);
    Refresh();

    if (emitEvent)
    {
        if (depent2dPlot)
            depent2dPlot->setCurveColor(color, false);
        if (master2dPlot)
            master2dPlot->setCurveColor(color, false);
    }

    //Pure2dPlot->SetAutoScaleEnabled(true, false);
}
void T2dPlot::on_BAutoScale_toggled()
{
    if (Pure2dPlot->singleAutoScale)
    {
        Pure2dPlot->SetAutoScaleEnabled(true, true);
        Pure2dPlot->SetAutoScaleEnabled(false, true);
    }
    else
        Pure2dPlot->SetAutoScaleEnabled(PlotControl->BAutoScale->isChecked(), true);
}
void T2dPlot::on_BSave_clicked()
{
    TCurvesDataOwner src;
    TMetaStorage attributes;
    savePlot(src, attributes);
    emit AddDataToContainer(src, attributes);
}
void T2dPlot::savePlot(TCurvesDataOwner &src, TMetaStorage &attributes)
{
    int actCurve = Pure2dPlot->ActiveCurve;
    ResizeOwner(src, actCurve >= 0 ? 1 : Pure2dPlot->SourceData.size());
    if (actCurve >= 0 && actCurve < Pure2dPlot->SourceData.size())
    {
        if (Pure2dPlot->SourceData.at(actCurve))
            *src[0] = *Pure2dPlot->SourceData.at(actCurve).data();
    }
    else
    {
        for (int i = 0; i < Pure2dPlot->SourceData.size(); i++)
        {
            if (Pure2dPlot->SourceData.at(i))
                *src[i] = *Pure2dPlot->SourceData.at(i).data();
        }
    }

    attributes.insert("Preview", QVariant::fromValue(MakePreview()));
    attributes.insert("Fit mode", Pure2dPlot->isFiltering ? FindCheckedItemIn(PlotControl->AGFitOrders) : 0);

    QStringList clrs;
    for (auto &d:src)
        clrs.append(d->Pen.color().name());

    attributes.insert("Curve colour", clrs.join(","));
    attributes.insert("Curve thickness", FindCheckedItemIn(PlotControl->AGCurveThickness));

    attributes.insert("XLabel",
                      (Pure2dPlot->AutoLegendEnabled ? "" : Pure2dPlot->XLegend)+
                      Pure2dPlot->XAxisUnit.Description);

    attributes.insert("XUnit", Pure2dPlot->XAxisUnit.SIUnit);

    attributes.insert("YLabel",
                      (Pure2dPlot->AutoLegendEnabled ? "" : Pure2dPlot->YLegend) +
                      Pure2dPlot->YAxisUnit.Description);

    attributes.insert("YUnit", Pure2dPlot->YAxisUnit.SIUnit);
}
void T2dPlot::SetSelectionModeEnabled(bool Enabled)
{
    PlotControl->BSelectByRect->setChecked(Enabled);
    on_BSelectByRect_clicked();
}
void T2dPlot::on_BSelectByRect_clicked()
{
    bool State = PlotControl->BSelectByRect->isChecked();

    if (State && Parent) QCoreApplication::postEvent(Parent, new TSelectEvent(Pure2dPlot->GetSelectedRect(), TSelectEvent::St_SelectionStarted));

    PlotControl->SetSelectionModeEnabled(State);
    PlotControl->BAutoScaleX->setVisible(!State);
    PlotControl->BAutoScaleY->setVisible(!State);

    Pure2dPlot->EnableSelectMode(State);
}
void T2dPlot::resetAllRilters()
{
    overlayButtons->resetOldFilters();
    PlotControl->ResetAllFilters();

    Pure2dPlot->SetFitOrder(TPlotEnums::FO_None);
    Refresh();
}
void T2dPlot::on_BApply_clicked()
{
    if (overlayButtons && overlayButtons->B2DBack->isVisible())
        overlayButtons->B2DBack->click();
    if (Pure2dPlot->d_func()->State == TPure2dPlot::Mode_Select)
    {
        if (Parent) QCoreApplication::postEvent(Parent, new TSelectEvent(Pure2dPlot->GetSelectedRect(), TSelectEvent::St_SelectionApplied));
        PlotControl->BSelectByRect->setVisible(true);
        PlotControl->BSelectByRect->setChecked(false);
        on_BSelectByRect_clicked();
        Pure2dPlot->SelectionApplied();
    }
    else if (Pure2dPlot->d_func()->State == TPure2dPlot::Mode_ManualConstruct)
    {
        PreparePlotControlForMC(false);
        Pure2dPlot->ExitMCMode();
        emit ManualConstructingDone();
        RestoreEnvironmentState();
        Refresh();
    }
}
void T2dPlot::on_BCancel_clicked()
{
    if (Pure2dPlot->d_func()->State == TPure2dPlot::Mode_Select)
    {
        if (Parent) QCoreApplication::postEvent(Parent, new TSelectEvent(Pure2dPlot->GetSelectedRect(), TSelectEvent::St_SelectionCancled));
        PlotControl->BSelectByRect->setVisible(true);
        PlotControl->BSelectByRect->setChecked(false);
        on_BSelectByRect_clicked();
    }
    else if (Pure2dPlot->d_func()->State == TPure2dPlot::Mode_ManualConstruct)
    {
        PreparePlotControlForMC(false);
        Pure2dPlot->ExitMCMode();
        Pure2dPlot->SetFitOrder(TPlotEnums::FO_None);
        Refresh();
    }
}
void T2dPlot::on_BDetailedMarkersInfo_clicked()
{
    MarkersInfo->setVisible(isVisible() ? PlotControl->BDetailedMarkersInfo->isChecked() : false);
    UpdateAllMarkersInfo();
}
void T2dPlot::on_BMarkers_clicked(bool on)
{
    if (!on)
        PlotControl->BDetailedMarkersInfo->setChecked(false);

    Pure2dPlot->BMarkers_clicked(on);

    PlotControl->BDetailedMarkersInfo->setVisible(on);
}
void T2dPlot::on_BSavePlotPixmap_clicked()
{
    QString SelectedFilter, FileName;

#ifdef _ADROID
    FileName = "/sdcard/";
#endif

    FileName = QFileDialog::getSaveFileName(this, tr("Save plot as picture..."), FileName,
                                            "*.png;;*.jpg;;*.tiff;;*.bmp", &SelectedFilter,
                                            QFileDialog::DontUseNativeDialog);
    if (FileName.isEmpty()) return;
    if (!FileName.endsWith(SelectedFilter.remove(0, 1))) FileName += SelectedFilter;

    Pure2dPlot->grab().toImage().save(FileName, SelectedFilter.remove(0, 1).toAscii().data());
}
void T2dPlot::RestoreEnvironmentState()
{
    if (Pure2dPlot->QwtCurves.size()) {
        int Thickness = Pure2dPlot->QwtCurves.at(qMax(0, Pure2dPlot->ActiveCurve))->pen().width();
        if (Thickness > 0 && Thickness - 1 < PlotControl->AGCurveThickness->actions().size())
        {
            PlotControl->AGCurveThickness->actions()[Thickness - 1]->setChecked(true);
            if (overlayButtons)
                overlayButtons->CBLineWidth->setCurrentIndex(Thickness - 1);
        }
    }

    int FO = Pure2dPlot->FitOrder;
    QList<QAction *> Acts = PlotControl->AGFitOrders->actions();
    if (FO == TPlotEnums::FO_Manual && Acts.size())    Acts[Acts.size() - 1]->setChecked(true);
    else if (FO == TPlotEnums::FO_None && Acts.size()) Acts[0]->setChecked(true);
    else if (FO >= 0 && FO < Acts.size() - 2 && Acts.size() > 2) Acts[FO + 1]->setChecked(true);
}
void T2dPlot::on_CBChooseActiveCurve_highlighted(int Index)
{
    Pure2dPlot->SetAutoScaleEnabled(false, false);
    bool ok = false;
    if (Index == -1 ||
        PlotControl->CBChooseActiveCurve->itemData(Index).toInt(&ok) == -1 || !ok)
        Pure2dPlot->UnselectCurves();
    else
        Pure2dPlot->SelectCurve(PlotControl->CBChooseActiveCurve->itemData(Index).toInt());

    Pure2dPlot->Replot();
    //Pure2dPlot->SetAutoScaleEnabled(true, false);
}
void T2dPlot::on_CBChooseActiveCurve_activated(int Index)
{
    Pure2dPlot->SetAutoScaleEnabled(false, false);
    Pure2dPlot->UnselectCurves();

    Pure2dPlot->ActiveCurve = (Index == -1) ? -1 : PlotControl->CBChooseActiveCurve->itemData(Index).toInt();

    RestoreEnvironmentState();

    Refresh();
    Pure2dPlot->SetAutoScaleEnabled(!Pure2dPlot->singleAutoScale, false);
}
void T2dPlot::sync_active_curve(int active)
{
    for (int i = 0; i < PlotControl->CBChooseActiveCurve->count(); i++) {
        if (active == PlotControl->CBChooseActiveCurve->itemData(i).toInt()) {
            PlotControl->CBChooseActiveCurve->setCurrentIndex(i);
            on_CBChooseActiveCurve_activated(i);
        }
    }
}

void T2dPlot::on_BFixX_toggled()
{
    Pure2dPlot->BFixXIsChecked = PlotControl->BFixX->isChecked();
}

void T2dPlot::on_BFixY_toggled()
{
    Pure2dPlot->BFixYIsChecked = PlotControl->BFixY->isChecked();
}

void T2dPlot::on_FontSize_changed()
{
    Pure2dPlot->SetFontSize(PlotControl->AGFontSizes->checkedAction()->text().toInt());
}

void T2dPlot::on_CurveThickness_changed()
{
    Pure2dPlot->SetAutoScaleEnabled(false, false);
    lineWidth = PlotControl->AGCurveThickness->checkedAction()->text().toInt();
    for (const auto &dt : Pure2dPlot->SourceData)
    {
        if (dt.data())
        {
            dt.data()->Pen.setWidth(lineWidth);
        }
    }
    Refresh();
    //Pure2dPlot->SetAutoScaleEnabled(true, false);
}

void T2dPlot::HideSelectionIfAny()
{
    if (Pure2dPlot->SelectionModeEnabled)
        on_BCancel_clicked();
}

void T2dPlot::Refresh()
{
    Pure2dPlot->Refresh();
}

QImage T2dPlot::MakePreview()
{
    return Pure2dPlot->MakePreview(PlotControl);
}

void T2dPlot::SetGridVisible(const bool Visible)
{
    PlotControl->BGrid->setChecked(Visible);
}
void T2dPlot::SetAutoScalingEnabled(const bool AutoScaling)
{
    PlotControl->BAutoScale->setChecked(AutoScaling);
    on_BAutoScale_toggled();
}
void T2dPlot::PreparePlotControlForMC(bool MCStarts)
{
    PlotControl->BApplyAndCancelSetVisible(MCStarts);
    PlotControl->BResetMCCurve->setVisible(MCStarts);
    PlotControl->CBChooseActiveCurve->setEnabled(!MCStarts);
    for (int i = 0; i < PlotControl->AGFitOrders->actions().size() - 1; i++)
        PlotControl->AGFitOrders->actions()[i]->setEnabled(!MCStarts);
}
void T2dPlot::StartMCCurveConstructing()
{
    if (Pure2dPlot->QwtCurves.empty())
        return;

    Pure2dPlot->StartMCMode();
    PreparePlotControlForMC(true);
}

T2dPlot::~T2dPlot()
{
    if (SyncTimer)
        SyncTimer->stop();
    delete SyncTimer;
}

}
