#include "plot_3dtypes.h"
#include "2d/plot_2dtypes.h"
#include <QPainter>
#include <QPainterPath>
#include <algorithm>
#include <limits>

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtConcurrent>
#else
#include <QtCore>
#endif

namespace plot {

QRectF ConstructRectF(Range XLims, Range YLims)
{
    return QRectF(XLims[0], YLims[0],
                  XLims[1] - XLims[0],
                  YLims[1] - YLims[0]);
}

double DoubleToArrInd(double Pos, int TotalN, const Range &Lims)
{
    return (Pos - Lims[0]) / (Lims[1] - Lims[0]) * (TotalN - 1);
}

TMin3dData::TMin3dData()
{
    XLimits[0] = 0;
    XLimits[1] = 1;
    YLimits[0] = 0;
    YLimits[1] = 1;

    UpdateNxNy();
    UpdateDxDy();
    resetMathLims();
}

void TMin3dData::detach()
{
    Array.detach();
    for (auto &a:Array)
        a.detach();
}

void TMin3dData::setSize(int nx, int ny)
{
    Array = T2dArray(ny, QVector<double>(nx));
    UpdateNxNy();
    UpdateDxDy();
    resetMathLims();
}

void TMin3dData::UpdateDxDy()
{
    dx = (XLimits[1] - XLimits[0]);
    dy = (YLimits[1] - YLimits[0]);

    if (NXElems > 1)
        dx /= (NXElems - 1);
    if (NYElems > 1)
        dy /= (NYElems - 1);
}

void TMin3dData::resetMathLims()
{
    if (NXElems == 0 || NYElems == 0) {
        LimsForMathX[0] = -1; LimsForMathX[1] = -1;
        LimsForMathY[0] = -1; LimsForMathY[1] = -1;
    } else {
        LimsForMathX[0] = 0; LimsForMathX[1] = NXElems;
        LimsForMathY[0] = 0; LimsForMathY[1] = NYElems;
    }
}

bool TMin3dData::IsIn(double x, double y) const
{
    return x > XLimits[0] - dx/2 && x < XLimits[1] + dx/2 &&
           y > YLimits[0] - dy/2 && y < YLimits[1] + dy/2;
}
void TMin3dData::AddPoint(int X, int Y, double Z)
{
    if (Array.size() <= Y) {
        Array.resize(Y + 1);
        UpdateNxNy();
        UpdateDxDy();
    }
    if (Array.at(Y).size() <= X) {
        Array[Y].resize(X + 1);
        UpdateNxNy();
        UpdateDxDy();
    }
    Array[Y][X] = Z;
}

TMin3dData *TMin3dData::Crop(const QRectF &SelectedRect, const Range &XLims, const Range &YLims) const
{
    return CropIndex(qRound(DoubleToArrInd(SelectedRect.left(), NXElems, XLims)),
                     qRound(DoubleToArrInd(SelectedRect.top(), NYElems, YLims)),
                     qRound(DoubleToArrInd(SelectedRect.right(), NXElems, XLims)),
                     qRound(DoubleToArrInd(SelectedRect.bottom(), NYElems, YLims)));
}
TMin3dData *TMin3dData::CropIndex(int minx, int miny, int maxx, int maxy) const
{
    TMin3dData *Cropped = new TMin3dData;

    if (maxx < minx)
    {
        int t = minx;
        minx = maxx;
        maxx = t;
    }

    if (maxy < miny)
    {
        int t = miny;
        miny = maxy;
        maxy = t;
    }

    miny = qBound<int>(0, miny, NYElems-1);
    maxy = qBound<int>(0, maxy, NYElems-1);
    minx = qBound<int>(0, minx, NXElems-1);
    maxx = qBound<int>(0, maxx, NXElems-1);

    Cropped->XLimits[0] = XLimits[0] + minx*dx;
    Cropped->XLimits[1] = XLimits[0] + maxx*dx;
    Cropped->YLimits[0] = YLimits[0] + miny*dy;
    Cropped->YLimits[1] = YLimits[0] + maxy*dy;
    Cropped->setSize(maxx - minx + 1, maxy - miny + 1);

    for (int YC = miny; YC <= maxy; YC++)
        for (int XC = minx; XC <= maxx; XC++)
            Cropped->Array[YC - miny][XC - minx] = Array[YC][XC];

    return Cropped;
}
bool TMin3dData::LimsForMathCorrect() const
{
    if (LimsForMathX[0] < 0 || LimsForMathX[1] < 0 || LimsForMathY[0] < 0 || LimsForMathY[1] < 0) return false;
    if (LimsForMathX[1] > NXElems) return false;
    if (LimsForMathY[1] > NYElems) return false;
    if (LimsForMathX[0] >= LimsForMathX[1]) return false;
    if (LimsForMathY[0] >= LimsForMathY[1]) return false;

    return true;
}
int FindDataWithPlotsLimits(const T3dDataOwner &data, const Range &XLimits, const Range &YLimits)
{
    QRectF ThisRect = ConstructRectF(XLimits, YLimits);
    for (int i = 0; i < data.size(); i++)
        if (data[i] && ThisRect == ConstructRectF(data[i]->XLimits, data[i]->YLimits))
            return i;

    return -1;
}

//================================================================================================================================
inline int TCompleteInfoFor3dPlot::FindNecessaryData(double x, double y, int c) const
{
    while (c < PaintData.size()) {
        if (PaintData[c] && PaintData[c].data()->IsIn(x, y))
            return c;
        c++;
    }

    return -1;
}
double TCompleteInfoFor3dPlot::value(double x, double y) const
{
    const TMin3dData *Temp3dData;
    double XPos, YPos;
    int i = -1;

    do {
        i = FindNecessaryData(x, y, i+1);
        if (i < 0)
            return PlotsZLimits[0];

        Temp3dData = PaintData[i].data();

        XPos = (x - Temp3dData->XLimits[0]) / Temp3dData->dx;
        YPos = (y - Temp3dData->YLimits[0]) / Temp3dData->dy;
    } while (XPos < Temp3dData->LimsForMathX[0] || XPos > Temp3dData->LimsForMathX[1]-1 ||
             YPos < Temp3dData->LimsForMathY[0] || YPos > Temp3dData->LimsForMathY[1]-1);

    if (YPos >= Temp3dData->Array.size() || XPos >= Temp3dData->Array[0].size())
        return PlotsZLimits[0];

    return valueBiLinear(XPos, YPos, Temp3dData);
}
double TCompleteInfoFor3dPlot::valueBiLinear(double xPos, double yPos, const TMin3dData *Temp3dData) const
{
    int XPos = (int)xPos;
    int YPos = (int)yPos;
    double dx = xPos - XPos;
    double dy = yPos - YPos;

    int XMax = XPos + 1;
    int YMax = YPos + 1;

    if (XPos == Temp3dData->LimsForMathX[1] - 1)
        XMax = XPos;
    if (YPos == Temp3dData->LimsForMathY[1] - 1)
        YMax = YPos;

    return Temp3dData->Array[YMax][XMax]*dx*dy +
            Temp3dData->Array[YMax][XPos]*(1-dx)*dy +
            Temp3dData->Array[YPos][XMax]*dx*(1-dy) +
            Temp3dData->Array[YPos][XPos]*(1-dx)*(1-dy);
}

//--------------------------
class TMinMax : public QPair<double, double>
{
public:
    TMinMax(double min = std::numeric_limits<double>::max(), double max = -std::numeric_limits<double>::max()) : QPair<double, double>(min, max) { }
};

void reduceMinMax(TMinMax &val, const TMinMax &next)
{
    if (val.first > next.first)
        val.first = next.first;
    if (val.second < next.second)
        val.second = next.second;
}
class FMapMinMax
{
public:
    const int xmin, xmax;
    FMapMinMax(int min = -1, int max = -1) : xmin(min), xmax(max) { }

    typedef TMinMax result_type; //for qtconcurrent::mapreduce
    result_type operator() (const QVector<double> &val) const
    {
        if (!val.size())
            return TMinMax();

        int xmin_local = xmin;
        if (xmin_local < 0)
            xmin_local = 0;
        int xmax_local = xmax;
        if (xmax_local < 0)
            xmax_local = val.size();

        std::pair<QVector<double>::const_iterator, QVector<double>::const_iterator> minmax;
        minmax = std::minmax_element(val.constBegin() + xmin_local, val.constBegin() + xmax_local);

        return TMinMax(*minmax.first, *minmax.second);
    }
};

//--------------------------
template<int N>
void reduceHistogram(THistogram<N> &val, const THistogram<N> &next)
{
    for (int i = 0; i < N; i++)
        val.bins[i] += next.bins[i];
}
template<int N = DEFAULT_HIST_SIZE>
class FMapHistogram
{
public:
    const int xmin, xmax;
    const double zmin, zrangeinv;
    FMapHistogram(double _zmin, double _zrangeinv, int min = -1, int max = -1) : xmin(min), xmax(max), zmin(_zmin), zrangeinv(_zrangeinv) { }

    typedef THistogram<N> result_type; //for qtconcurrent::mapreduce
    result_type operator() (const QVector<double> &val) const
    {
        if (!val.size())
            return THistogram<N>();

        int xmin_local = xmin;
        if (xmin_local < 0)
            xmin_local = 0;
        int xmax_local = xmax;
        if (xmax_local < 0)
            xmax_local = val.size();

        THistogram<N> hist;
        for (int i = xmin_local; i < xmax_local; i++) {
            int idx = qBound<int>(0, int((val.constData()[i] - zmin) * zrangeinv), N-1);
            hist.bins[idx]++;
        }
        return hist;
    }
};


void TCompleteInfoFor3dPlot::UpdatePlotsZLimits()
{
    if (!PaintData.size() || !PaintData[0]) {
        ZAxisUnit.UpdateMultiplier(qMax(fabs(PlotsZLimits[1]), fabs(PlotsZLimits[0])));
        return;
    }

    const TMin3dData &Data = *PaintData[0].data();
    const T2dArray &Array = Data.Array;

    TMinMax vminmax;
    if (Data.LimsForMathCorrect())
        vminmax = QtConcurrent::blockingMappedReduced(
                    Array.constBegin() + (int)Data.LimsForMathY[0], Array.constBegin() + (int)Data.LimsForMathY[1],
                FMapMinMax(Data.LimsForMathX[0], Data.LimsForMathX[1]), reduceMinMax);
    else
        vminmax = QtConcurrent::blockingMappedReduced(Array.constBegin(), Array.constEnd(), FMapMinMax(), reduceMinMax);
    PlotsZLimits[0] = vminmax.first;
    PlotsZLimits[1] = vminmax.second;

    if (!qIsFinite(PlotsZLimits[0]))
        PlotsZLimits[0] = 0;
    if (!qIsFinite(PlotsZLimits[1]))
        PlotsZLimits[1] = 1;

    ZAxisUnit.UpdateMultiplier(qMax(fabs(PlotsZLimits[1]), fabs(PlotsZLimits[0])));

    //histogram
    double zrangeinv = 1. * DEFAULT_HIST_SIZE / qFabs(PlotsZLimits[1] - PlotsZLimits[0]);
    if (Data.LimsForMathCorrect())
        zhistogram = QtConcurrent::blockingMappedReduced(
                    Array.constBegin() + (int)Data.LimsForMathY[0], Array.constBegin() + (int)Data.LimsForMathY[1],
                FMapHistogram<>(PlotsZLimits[0], zrangeinv, Data.LimsForMathX[0], Data.LimsForMathX[1]), reduceHistogram<DEFAULT_HIST_SIZE>);
    else
        zhistogram = QtConcurrent::blockingMappedReduced(Array.constBegin(), Array.constEnd(), FMapHistogram<>(PlotsZLimits[0], zrangeinv), reduceHistogram<DEFAULT_HIST_SIZE>);
}

//================================================================================================================================
TCompleteInfoFor3dPlot::TCompleteInfoFor3dPlot()
{
    PlotsZLimits[0] = 0;
    PlotsZLimits[1] = 1;

    QSpectrogram = new QwtPlotSpectrogram();
    QSpectrogram->setData(this);
    colormap = new QwtLinearColorMap();
    QSpectrogram->setColorMap(colormap);
}

void TCompleteInfoFor3dPlot::SetPalette(const TPalette &palette, double min, double max)
{
    active_palette = palette;
    if (fabs(min - max) < 1e-3)
        (min > 0.5) ? (min -= 1e-3) : (max += 1e-3);

    colormap->setColorInterval(palette.first(), palette.last());
    for (int i = 0; i < palette.size(); i++)
        colormap->addColorStop(min + i * (max - min) / (palette.size() - 1), palette.at(i));
    QSpectrogram->setColorMap(colormap); //notify updated
}

}
