#pragma once

#include <QHash>
#include "plot_basetypes.h"
#include "plot_events.h"
#include <qwt_plot.h>
#include <qwt_symbol.h>
#include <qwt_plot_curve.h>
#include <qwt_scale_draw.h>
#include <qwt_scale_engine.h>
#include <qwt_plot_rasteritem.h>
#include <QDebug>

//int FindCheckedItemIn(const QActionGroup *AG);
#define FindCheckedItemIn(AG) (AG->actions().indexOf(AG->checkedAction()))

class QwtPlotMarker;
class QwtPlotGrid;
class QGestureEvent;

namespace plot {

class TLineSegment : public QwtPlotCurve
{
public:
    TLineSegment()
    {
        data.resize(2);
        setSamples(data);
        setPen(QPen(QColor(0, 0, 255), 0, Qt::SolidLine));
        setRenderHint(QwtPlotItem::RenderAntialiased, true);
    }
    void SetData(const QPointF &A, const QPointF &B)
    {
        data[0] = A;
        data[1] = B;
        setSamples(data);
    }

private:
    QVector<QPointF> data;
};

class ImageItem : public QwtPlotRasterItem
{
public:
    QImage image;
    QRectF targetRect; //full-plot image if empty
    QImage renderImage (const QwtScaleMap &xMap,
                        const QwtScaleMap &yMap, const QRectF &area, const QSize &imageSize) const override;
};



class TPlotEnums
{
public:
    enum TFitOrders {FO_None = -1, FO_Manual = -2};
    enum TAxis {A_XAxis, A_YAxis, A_Both};
};

class TSIUnit
{
public:
    double Multiplier;
    QString StrMultiplier;
    QString Description;
    QString SIUnit;
    QString Legend;

    TSIUnit()
    {
        Multiplier = 1;
    }
    void SetDescription(const QString &Desc)
    {
        Description = Desc;
        UpdateLegend();
    }
    void SetSIUnit(const QString &Unit)
    {
        SIUnit = Unit;
        UpdateLegend();
    }
    void SetDescAndSIUnit(const QString &Desc, const QString &Unit = QString())
    {
        Description = Desc;
        SIUnit = Unit;
        UpdateLegend();
    }
    void UpdateMultiplier(double Value);
    void UpdateLegend();
    void UpdateLegend(double Value) { UpdateMultiplier(Value); UpdateLegend(); }
    void UpdateLegend(double FirstBound, double SecondBound) { UpdateLegend(qMax(fabs(FirstBound), fabs(SecondBound))); }
};

class TValueInfoCreator
{
public:
    double SrcValue;
    QString Value;
    QString Unit;
    void operator()(bool AutoLegendEnabled, const TSIUnit &UnitInfo, double Val);
};

class TGeneralMarkerInfoHandler
{
public:
    TValueInfoCreator XInfo;
    TValueInfoCreator YInfo;

    QString FormUniteString() { return "X = " + XInfo.Value + XInfo.Unit + "\nY = " + YInfo.Value + YInfo.Unit; }
    void operator()(bool AutoLegendEnabled, const TSIUnit &XUnitInfo, const TSIUnit &YUnitInfo, double XVal, double YVal)
    { XInfo(AutoLegendEnabled, XUnitInfo, XVal);  YInfo(AutoLegendEnabled, YUnitInfo, YVal); }
};

class TInfoForAllPlots
{
public:
    bool AutoLegendEnabled;

    int FontSize;

    QString XLegend;
    QString YLegend;
    QString Title;

    Range CurrentXLimits;
    Range CurrentYLimits;
    Range PlotsXLimits;
    Range PlotsYLimits;
    Range DefaultXLimits;
    Range DefaultYLimits;

    #ifdef _SCAN_SLICE
    bool sliceBounds = false;
    Range sliceX;
    Range sliceY;
    double slicePercent = 0.0;
#endif

    TInfoForAllPlots();
#ifndef Q_CC_MSVC
    virtual ~TInfoForAllPlots() { } //fuck warnings
#endif
    void SetAutoLegendEnabled(bool Enable) { if (AutoLegendEnabled == Enable) return; AutoLegendEnabled = Enable; UpdateAxisLegends(); }
    virtual void UpdateAxisLegends() { }
};

typedef QHash<QString, QVariant> TMetaStorage;


class TPlotContainer : public QWidget
{
public:
    void Replot(){ QPlot->replot(); }

    void showEvent(QShowEvent *Event) override;
    void resizeEvent(QResizeEvent *Event) override;
    virtual void ResizeQPlot(bool = false) { QPlot->resize(size()); UpdateScalesDiv(); }
    void UpdateScalesDiv();

    QPoint mapToCanvas(const QPoint &Pos) const;
    double ScreenXToPlotX(int X) const { return QPlot->invTransform(QwtPlot::xBottom, X); }
    double ScreenYToPlotY(int Y) const { return QPlot->invTransform(QwtPlot::yLeft, Y); }
    double PlotXToScreenX(double X) const { return QPlot->transform(QwtPlot::xBottom, X); }
    double PlotYToScreenY(double Y) const { return QPlot->transform(QwtPlot::yLeft, Y); }
    QPointF ScreenToPlot(const QPoint &Pos) const { return QPointF(ScreenXToPlotX(Pos.x()), ScreenYToPlotY(Pos.y())); }
    QPointF PlotToScreen(const QPointF &Pos) const { return QPoint(PlotXToScreenX(Pos.x()), PlotYToScreenY(Pos.y())); }

    QwtPlot *QPlot;
};


template <class T> //last ancestor of Draw* iface
class TPlotDecorated : public TPlotContainer
{
public:
    void DrawBase(QPainter &) { }
    void DrawSource(QPainter &) { }
    void paintEvent(QPaintEvent *) override;

    QPixmap m_base_buffer;
};

struct TPlotSelectionData
{
    int SelectAction;
    mutable qreal kx, ky;
    QPointF Offset;
    QRectF Rectangle;
    QRectF CanvasRectInArtFr;
    bool SelectionModeEnabled;
    int SelectRectEventOrigin;
    bool Vertical;
    enum TActions {None, MoveLeft, MoveRight, MoveTop, MoveBottom, Panning};

    TPlotContainer *link;
    virtual const Range *GetCurXLims() const = 0;
    virtual const Range *GetCurYLims() const = 0;
    TPlotSelectionData();

    void DefineAction(QMouseEvent *Event);
    void ManageCursorShape(QMouseEvent *Event);

    void SetSelectionModeEnabled(const bool State);
    void ReInitSelectRectGeometry(const double k);
    void ExpandSelectionRect() { ReInitSelectRectGeometry(1.); }

    void Updatekxky() const;
    void UpdateCanvasRect();
    void ResizeEvent(QRectF prevsel);

    void SetSelect(QPointF center, qreal width, qreal height); //px
    void SetSelectCenter(QPointF Point);
    void SetSelectWidthRelCenter(qreal W);
    void SetSelectHeightRelCenter(qreal H);
    void SetSelectWidth(qreal W);
    void SetSelectHeight(qreal H);
    QRectF GetSelectedRect() const;
    QRectF TransformForeignRectToRectangle(QRectF fFrect) const;

    void ReDrawSelectedRectangle(int from = TSelectEvent::SRO_Unknown);
    virtual void SelectRectChanged(const QRectF &/*Rect*/) { }
};
template <class T>// = TPlotDecorated<TPlotSelection> >
class TPlotSelection : public T, public TPlotSelectionData
{
public:
    void DrawBase(QPainter &p);
    void DrawSource(QPainter &p);

    void mouseMoveEvent(QMouseEvent *) override;
    void mousePressEvent(QMouseEvent *) override;
    void mouseReleaseEvent(QMouseEvent *) override;
    void resizeEvent(QResizeEvent *Event) override;

    void SetSelectionModeEnabled(const bool State)
    {
        if (SelectionModeEnabled == State) return;
        m_base_buffer = QPixmap();
        TPlotSelectionData::SetSelectionModeEnabled(State);
    }
    TPlotSelection() {
        link = (TPlotContainer *)this;
    }

    using T::m_base_buffer;
    using T::rect;
    using T::size;
    using T::QPlot;
};
template <class T>// = TPlotDecorated<TPlotEnable> >
class TPlotEnable : public T
{
public:
    TPlotEnable() : Enabled(true) { this->setAttribute(Qt::WA_OpaquePaintEvent); }
    void DrawBase(QPainter &p);
    void DrawSource(QPainter &p);
    void SetPlotEnabled(const bool State) { if (Enabled == State) return; Enabled = State; m_base_buffer = QPixmap(); this->update(); }
    bool Enabled;

    using T::m_base_buffer;
    using T::rect;
    using T::size;
    using T::QPlot;
};
template <class T>// = TPlotDecorated<TPlotActive> >
class TPlotActive : public T
{
public:
    TPlotActive() : PlotIsActive(false) { this->setAttribute(Qt::WA_OpaquePaintEvent); }
    void DrawBase(QPainter &p);
    void SetPlotActive(const bool State) { if (PlotIsActive == State) return; PlotIsActive = State; m_base_buffer = QPixmap(); this->update(); }
    bool PlotIsActive;

    using T::m_base_buffer;
    using T::rect;
};
template <class T>// = TPlotDecorated<TPlotBackground> >
class TPlotBackground : public T
{
public:
    TPlotBackground() { this->setAttribute(Qt::WA_OpaquePaintEvent); }
    void DrawBase(QPainter &p);

    using T::rect;
};
template <class T>// = TPlotDecorated<TPlotBackground> >
class TPlotRoundedBorder : public T
{
public:
    void DrawBase(QPainter &p);

    using T::rect;
};

//plot + rounded corners + selection, active, enabled, background
//+ markers A, B, C, M, Angle + limits management
class TPlotCore : public TPlotRoundedBorder< TPlotSelection< TPlotActive< TPlotEnable< TPlotBackground< TPlotDecorated<TPlotCore> > > > > >,
        virtual public TInfoForAllPlots
{
    void InitQwtItems();
public:
    virtual const Range *GetCurXLims() const override { return &CurrentXLimits; }
    virtual const Range *GetCurYLims() const override { return &CurrentYLimits; }
public:
    TPlotCore(QWidget *parent = nullptr);
    ~TPlotCore() override;

    void ReAssignPlotOptions();
    void PlotWasChanged()                      { if (!SelectionModeEnabled && Enabled) return; m_base_buffer = QPixmap(); update(); }

    void SetMarkerMCoords();
    void UpdateABCCurves();
    void ReSyncMarkers();

    virtual void HandleAxisMultiplier() { }
    void ReSetAxisScales(const TPlotEnums::TAxis ApplyTo = TPlotEnums::A_Both);

    void SetCurrentXScales()                                   { QPlot->setAxisScale(QwtPlot::xBottom, CurrentXLimits[0], CurrentXLimits[1]); }
    void SetCurrentYScales()                                   { QPlot->setAxisScale(QwtPlot::yLeft, CurrentYLimits[0], CurrentYLimits[1]); }
    void SetCurrentXScales(const double Min, const double Max) { CurrentXLimits[0] = Min; CurrentXLimits[1] = Max; SetCurrentXScales(); HandleAxisMultiplier(); }
    void SetCurrentYScales(const double Min, const double Max) { CurrentYLimits[0] = Min; CurrentYLimits[1] = Max; SetCurrentYScales(); HandleAxisMultiplier(); }
    void SetPlotsXLimits(const double Min, const double Max)   { PlotsXLimits[0] = Min; PlotsXLimits[1] = Max; }
    void SetPlotsYLimits(const double Min, const double Max)   { PlotsYLimits[0] = Min; PlotsYLimits[1] = Max; }
    void SetDefaultXLimits(const double Min, const double Max) { DefaultXLimits[0] = Min; DefaultXLimits[1] = Max; }
    void SetDefaultYLimits(const double Min, const double Max) { DefaultYLimits[0] = Min; DefaultYLimits[1] = Max; }

    QwtPlotMarker *MarkerC;
    QwtPlotMarker *MarkerM;
    QwtPlotMarker *MarkerA;
    QwtPlotMarker *MarkerB;
    QwtPlotGrid   *PlGrid;
    TLineSegment  *CurveAB;
    TLineSegment  *CurveAC;

};

//base functions - pan, zoom, markers movement
class TCommonPlotControl;
class TCommonPlot: public TPlotCore
{
    Q_OBJECT
public slots:
    void BRestoreScale_clicked();
    void BMarkers_clicked(bool IsChecked);
    void BAngle_clicked(bool IsChecked);
    void BGrid_clicked(bool IsChecked);

public:
    void SetClickParams(const QPoint &Pos);
    QPoint ClickPos;
    bool singleAutoScale = false;
    double singleAutoScaleXLimit = 10;
    double singleAutoScaleYLimit[2] = {-10, 10};
    double ClickMinX, ClickMaxX, ClickMinY, ClickMaxY;

    TCommonPlot(QWidget *parent = nullptr): TPlotCore(parent) {
        Action = ExternalAct;
        BFixXIsChecked = BFixYIsChecked = false;
        blockRefresh = false;

        connect(QPlot, SIGNAL(pinchSig(QPoint,double)), SLOT(pinchEvent(QPoint,double)));
        connect(QPlot, SIGNAL(isGestureActive(bool)), SLOT(setGestureActive(bool)));
    }

    bool blockRefresh;
    void ReplotChanged() { if (blockRefresh) return; Replot(); PlotWasChanged(); }
    virtual void Refresh() { ReplotChanged(); }

    void SetFontSize(int);
    double CalcAngle() const;
    void UpdateLabels() { emit UpdateAllMarkersInfo(); }

    virtual void DefineClickInfo(const QPoint &Pos);
    void MoveMarker(QwtPlotMarker *Marker, double x, double y);
    void MoveMarker(const double x, const double y, bool IsPosCoordInPlot);
    void PanPlot(const QPointF &Delta, Range *XLimits = nullptr, Range *YLimits = nullptr);
    void ZoomPlot(const QPointF &Pos, int Increment, Range *XLimits = nullptr, Range *YLimits = nullptr);

    const QImage MakePreview(TCommonPlotControl *Control = nullptr, double width = 150., double height = 150.);
    QImage MakeRawPreview(double width, double height);
    void PreparePlotForPreview();
    void RestorePlotAfterPreview(TCommonPlotControl *Control);

    enum TActions {MoveA, MoveB, MoveC, Panning, Zooming, Gesture, CrossAct, ExternalAct};
    int Action;

    bool BFixXIsChecked, BFixYIsChecked;
    bool KeyFixXPressed = false, KeyFixYPressed = false;

    void SetAllXScales(const double Min, const double Max)   { SetDefaultXLimits(Min, Max); SetPlotsXScales(Min, Max); }
    void SetAllYScales(const double Min, const double Max);
    void SetPlotsXScales(const double Min, const double Max) { SetPlotsXLimits(Min, Max); ReSetAxisScales(TPlotEnums::A_XAxis); ReplotChanged(); }
    void SetPlotsYScales(const double Min, const double Max) { SetPlotsYLimits(Min, Max); ReSetAxisScales(TPlotEnums::A_YAxis); ReplotChanged(); }

    void SetXLegend(const QString &Legend) { if (AutoLegendEnabled) return; XLegend = Legend; Refresh(); }
    void SetYLegend(const QString &Legend) { if (AutoLegendEnabled) return; YLegend = Legend; Refresh(); }

    void SetTitle(const QString &Ttl) { Title = Ttl; Refresh(); }

    void HandleZeroDiff(double &Max, double &Min) const;
    void FindSizeWithAspectRatio(double MaxWidth, double MaxHeight, double *ResultX, double *ResultY) const;

signals:
    void UpdateAllMarkersInfo();

public slots:
    void pinchEvent(QPoint, double);
    void setGestureActive(bool);
};


class QwtPlot_ : public QwtPlot
{
    Q_OBJECT
public:
    QwtPlot_(QWidget *parent = nullptr) : QwtPlot(parent)
    {
        grabGesture(Qt::TapAndHoldGesture);
        grabGesture(Qt::PinchGesture);
    }

    void emitGestureActive(bool active) { emit isGestureActive(active); }

signals:
    void doubleClick();
    void pinchSig(QPoint, double);
    void isGestureActive(bool);

protected:
    bool event(QEvent *) override;
    bool gestureEvent(QGestureEvent *event);

};


class TScale : public QwtScaleDraw
{
public:
    void drawLabel( QPainter *painter, double value ) const override;
    mutable QFont font;
};


class TScaleDiv : public QwtLinearScaleEngine
{
public:
    TScaleDiv(TScale *scale = nullptr);
    QwtScaleDiv divideScale( double x1, double x2, int numMajorSteps, int numMinorSteps, double stepSize = 0.0 ) const override;

    TScale *scale;
    mutable int prev_counter;
    mutable int prev_len;
    mutable double prev_range;
    mutable int prev_steps;
};


template <class T>
void ResizeOwner(QVector<QSharedPointer<T> > &data, int new_size)
{
    int old_size = data.size();

    data.resize(new_size);
    for (int i = old_size; i < new_size; i++)
#if QT_VERSION < QT_VERSION_CHECK(5, 1, 0)
        data[i] = QSharedPointer<T>(new T);
#else
        data[i] = QSharedPointer<T>::create();
#endif
}

template <class T>
QVector<QSharedPointer<T> > getLock(const QVector<QWeakPointer<T> > &data)
{
    QVector<QSharedPointer<T> > lock(data.size());
    for (int i = 0; i < data.size(); i++)
        lock[i] = data.at(i);

    return lock;
}

template <class T>
QVector<QWeakPointer<T> > getUnLock(const QVector<QSharedPointer<T> > &data)
{
    QVector<QWeakPointer<T> > lock(data.size());
    for (int i = 0; i < data.size(); i++)
        lock[i] = data.at(i);

    return lock;
}

}
