#include "videoplayer.h"
#include "gui/status.hpp"
#include "tools.hpp"

#include "oscilloscope.h"
#include "remote_nst_dataid.hpp"
#include "mobile/remote_mobile.hpp"
#include "motors.hpp"
#include "scanplot.h"
#include <QCheckBox>
#include <QBuffer>
#include <QSettings>
#include <QThread>
#include <QList>

void VideoPlayer::on_Camera_ImgReceived(QImage Img)
{
    if (refresh_limit_timer.isValid() && refresh_limit_timer.elapsed() < 20 + refresh_limit_last_elapsed) //~50 refreshes per second max
        return;
    refresh_limit_timer.start();

    Img.detach();
    QRectF cropp(Img.width()*(1 - cw) / 2,
                 Img.height()*(1 - ch) / 2,
                 Img.width()*cw,
                 Img.height()*ch);

    if (!Img.rect().isValid() || !cropp.isValid()) {
        qDebug() << "invalid frame";
        return;
    }

    if (!rotxy->isChecked())
        ritem->image = Img.copy(cropp.toRect()).mirrored(flipx->isChecked(), flipy->isChecked());
    else {
        QTransform m;
        m.rotate(90);
        m.scale(flipx->isChecked() ? -1 : 1, flipy->isChecked() ? -1 : 1);
        ritem->image = Img.copy(cropp.toRect()).transformed(m);
    }
    emit ImgReceivedAfter(ritem->image);

    int minSizeWidth = ritem->image.size().width() / 32;
    int minSizeHeight = ritem->image.size().height() / 32;
    if (minSizeWidth < 0) minSizeWidth = 0;
    if (minSizeHeight < 0) minSizeHeight = 0;
    QPlot->setMinimumSize(minSizeWidth, minSizeHeight);
    QPlot->replot();

    frames++;

    if (inst_idx == 0)
        if (auto mob = MobileServer::GetIt()) {
            if (mob->any_need_data("video_need")) {
                QByteArray vdata;
                QBuffer buf(&vdata);
                buf.open(QBuffer::WriteOnly);
                ritem->image.save(&buf, "bmp");
                buf.close();

                mob->setPropIdVar(DataID::VideoData, vdata, 1);
            }
        }

    refresh_limit_last_elapsed = refresh_limit_timer.elapsed() / 0.2; //time budget, %
}



void VideoPlayer::calcOscContrast(QByteArray data)
{
    //magic: (const char*) to (const Struct*) WTF??? (comment by Oplkill)
    const EG_OscData *sdata = (const EG_OscData *)data.constData();

    if (sdata->Channel >= 4)
        return;

    osc_contrast = 0;
    for (int i = 0; i < sdata->Size; i++)
        osc_contrast += sdata->Data[i];
    osc_contrast /= sdata->Size;

    new_osc_contrast = true;
}

void VideoPlayer::on_BAutoFocusOsc_clicked()
{
    double speed = 1.0; //mkm/s
    double z1 = 0;
    double z2 = zPos->edit->maximum();//mkm

    double contrast = 0;
    double old = 0;
    double max = 0;
    double zmax = 0;
    int counter = 0;

    auto osc = Oscilloscope::GetIt();
    if (!osc)
        return;

    setEnabled(false);
    osc->setEnabled(false);

    int old_ch[4];
    for (int i = 0; i < 4; i++) {
        old_ch[i] = osc->CBChannel[i]->currentIndex();
        osc->CBChannel[i]->setCurrentIndex(srcNone);
    }

    int srcch = srcADC4;
    {
        QSettings settings("Param/ccd.conf", QSettings::IniFormat);
        if (settings.value("use_pmt", false).toBool())
            srcch = srcADCHR1;
    }
    osc->CBChannel[0]->setCurrentIndex(osc->CBChannel[0]->findData(srcch));

    osc->start->change_state_man_with_sig(true);

    DoubleChanger *extra_dac = osc->DacsWidget->Extra_Dac_1;
    extra_dac->setValue(0);
    extra_dac->setNewValue(true);

    isInterrupted = false;

    QPointer<StatusBar> bar = new StatusBar(tr("PMT Autofocus"), 4);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    zPos->setValue(0);
    zPos->setNewValue(true);
    SleepExport::msleep_gui(50);

    int time = 50;
    QElapsedTimer timer;

    const int daclow = 35000;
    const int dachigh = 55000;
    double dacspeed = 0.1;
    int dacsign = 1;

    counter = 0;
    do {
        extra_dac->setValue(extra_dac->value() + dacsign*dacspeed);
        extra_dac->setNewValue(true);
        SleepExport::msleep_gui(50);

        new_osc_contrast = false;
        connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

        time = 500;
        timer.start();
        while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
        if (timer.hasExpired(time)) {
            qDebug() << "AFO: no osc signal";
            new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
            doBeep();

            destructStatus(bar);
            isInterrupted = true;
        }

        disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
        if (isInterrupted)
            continue;

        old = contrast;
        contrast = osc_contrast;

        if ((old > daclow && contrast < daclow) ||
                (old < dachigh && contrast > dachigh)) {
            dacsign *= -1;
            dacspeed /= 2;
            qDebug() << "dac overflow" << old << contrast << dacspeed;
        }
        counter++;
    } while ((contrast < daclow || contrast > dachigh) && counter < 50 && extra_dac->value() < 0.9 && !isInterrupted);

    if (counter >= 50 || extra_dac->value() >= 0.9 || contrast != contrast || isInterrupted) {
        destructStatus(bar);

        setEnabled(true);
        osc->setEnabled(true);

        for (int i = 0; i < 4; i++)
            osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

        osc->start->change_state_man_with_sig(false);

        extra_dac->setValue(0);
        extra_dac->setNewValue(true);

        if (!isInterrupted) {
            qDebug() << "AFO: no signal";
            new Notification(tr("PMT Autofocus") + ": " + tr("No PMT signal"), Notification::Warn, 10);
            doBeep();
        }

        return;
    }

    qDebug() << "AFO dac" << contrast << extra_dac->value();

    counter = 0;
    do {
        bar->setValue(1+counter);
        qDebug() << "AFO pass" << counter << z1 << z2;

        zPos->setValue(z1);
        zPos->setNewValue();
        SleepExport::msleep_gui(50);

        do {
            if (contrast > max) {
                max = contrast;
                zmax = zPos->value();
            }

            new_osc_contrast = false;
            connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

            time = 500;
            timer.start();
            while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            if (timer.hasExpired(time)) {
                qDebug() << "AFO: no osc signal2";
                new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
                doBeep();

                destructStatus(bar);
                isInterrupted = true;
            }

            disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
            if (isInterrupted)
                continue;

            old = contrast;
            contrast = osc_contrast;

            if (contrast > dachigh) {
                dacsign = -1;
                dacspeed = 0.05;
                double olddac = extra_dac->value();

                do {
                    extra_dac->setValue(extra_dac->value() + dacsign*dacspeed);
                    extra_dac->setNewValue();
                    SleepExport::msleep_gui(50);

                    new_osc_contrast = false;
                    connect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));

                    time = 500;
                    timer.start();
                    while (!new_osc_contrast && !isInterrupted && !timer.hasExpired(time)) {
                        qApp->processEvents();
                        SleepExport::msleep2(10);
                    }
                    if (timer.hasExpired(time)) {
                        qDebug() << "AFO: no osc signal3";
                        new Notification(tr("PMT Autofocus") + ": " + tr("No data from controller"), Notification::Warn, 10);
                        doBeep();

                        destructStatus(bar);
                        isInterrupted = true;
                    }

                    disconnect(osc, SIGNAL(newOscData_(QByteArray)), this, SLOT(calcOscContrast(QByteArray)));
                    if (isInterrupted)
                        continue;

                    old = contrast;
                    contrast = osc_contrast;

                    if ((old > daclow && contrast < daclow) ||
                            (old < dachigh && contrast > dachigh)) {
                        dacsign *= -1;
                        dacspeed /= 2;
                        qDebug() << "dac overflow" << old << contrast << dacspeed;
                    }
                } while ((contrast < daclow || contrast > dachigh) && !isInterrupted);
                max *= extra_dac->value() / olddac;

                qDebug() << "AFO dac" << contrast << extra_dac->value() << max;
            }

            qDebug() << "AFO_0" << contrast << zPos->value();

            zPos->setValue(zPos->value() + speed);
            zPos->setNewValue();
            SleepExport::msleep_gui(50);
        } while ((zPos->value() + speed) <= z2 && !isInterrupted);

        double dz = (z2 - z1)/8;
        double zrange = zPos->edit->maximum();

        z1 = qBound(0., zmax - dz/2, zrange);
        z2 = qBound(0., zmax + dz/2, zrange);

        speed /= 4;
        counter++;
    } while (speed >= 0.2 && counter < 3 && !isInterrupted);

    qDebug() << "AFO" << counter << zmax;

    zPos->setValue(zmax);
    zPos->setNewValue(true);

    destructStatus(bar);

    setEnabled(true);
    osc->setEnabled(true);

    for (int i = 0; i < 4; i++)
        osc->CBChannel[i]->setCurrentIndex(old_ch[i]);

    osc->start->change_state_man_with_sig(false);

    extra_dac->setValue(0);
    extra_dac->setNewValue(true);
}

double VideoPlayer::autoFocusGetAverageContrast(int numFrames, int maxTimerWait)
{
    double av = 0.0;
    QList<double> contrasts;
    int frame = frames;
    int prevFrame = 0;
    QElapsedTimer timer;
    timer.start();
    while (frames < frame + numFrames && !isInterrupted && !timer.hasExpired(maxTimerWait))
    {
        qApp->processEvents();
        //SleepExport::msleep2(1);
        if (prevFrame != frames)
        {
            double contrast = getContrast();
            //qDebug()<<"av"<<contrast;
            contrasts.append(contrast);
        }
        prevFrame = frames;
    }
    if (autoFocusStopped || isInterrupted)
        return 0.;
    if (timer.hasExpired(maxTimerWait))
    {
        qDebug()<<"No frames1";
        new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
        doBeep();

        destructStatus(autoFocusBar);
        isInterrupted = true;
        BAutoFocus_clicked(false);
        return 0.;
    }
    for (auto d : contrasts)
        av = av + d;
    av = av / contrasts.count();
    //qDebug()<<"===="<<av;
    return av;
}

void VideoPlayer::autoFocusStages(int mZfocus, bool atEnd, bool active)
{
    if (autoFocusStopped)
        return;
    auto motors = Motors::GetIt();
    if (!motors || mZfocus != motors->getRoleMotor(Motors::MotorRoles::RoleZ2Focus))
        return;
    static double prevContrast = 0.0;
    static double prevPrevContrast = 0.0;
    static int autoFocusContrast0 = 0;
    static bool autoFocusScanning = false;
    static bool autoFocusStopScan = false;
    static bool autoFocusEndedSucsessfull = false;
    QElapsedTimer timer;
    const int maxTimerWait = 6000;

    qDebug()<<mZfocus<<"stage "<<autoFocusStage<<atEnd<<active;

    if (autoFocusStage == -10)
    {
        destructStatus(autoFocusBar);
        autoFocusStopped = true;
        qDebug()<<"AutoFocus Stopping";
        disconnect(motors, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(autoFocusStages(int, bool, bool)));
        motors->goStop(mZfocus);
        prevContrast = 0.0;
        prevPrevContrast = 0.0;
        autoFocusStopScan = true;
        autoFocusScanning = false;
        emit AutoFocusEnded(autoFocusEndedSucsessfull);
        autoFocusEndedSucsessfull = false;
        return;
    }

    if (isInterrupted || autoFocusStopped)
        return;

    if (autoFocusStage == 0) //upped
    {
        if (active) return;
        if (!atEnd) return;

        autoFocusEndedSucsessfull = false;
        if (!autoFocusBar.isNull())
            autoFocusBar->setValue(1);
        auto scan = ScanPlot::GetIt();
//        qDebug()<<autoFocusDist1<<"(-"<<
//                  scan->comboMaterialThin->currentData().toDouble()*1000<<")"<<
//                  autoFocusV1<<autoFocusV2<<autoFocusMagicMultiply[autoFocusMagicMultiplyCurrentId]<<
//                  autoFocusVLeftStPr;
//        qDebug()<<"Upped";
        autoFocusStage++;
        //int dist = autoFocusDist1 - scan->comboMaterialThin->currentData().toDouble()*1000;
        int dist = autoFocusDist1 - scan->comboMaterialThin->itemData(scan->comboMaterialThin->currentIndex()).toDouble()*1000.0;
        if (dist < 0)
            dist = 1;
        qDebug()<<dist;
        motors->goUp(mZfocus, 9999, dist);
        return;
    }

    if (autoFocusStage == 1) //downed to start scan position
    {
        qDebug()<<"call with stage"<<autoFocusStage;
        if (active) return;

        if (!autoFocusBar.isNull())
            autoFocusBar->setValue(2);
        qDebug()<<"In start position";
        autoFocusContrast0 = autoFocusGetAverageContrast(10, maxTimerWait);
        if (isInterrupted || autoFocusStopped)
            return;
        if (autoFocusContrast0 < 1.0)
            autoFocusContrast0 = 1.0;
        qDebug()<<"C0="<<autoFocusContrast0;
        autoFocusStage = 2;
        if (autoFocusDirection == 0)
            motors->goUp(mZfocus, autoFocusV1, autoFocusDist2);
        else
            motors->goDown(mZfocus, autoFocusV1, autoFocusDist2);
        qDebug()<<"autoFocusDist2"<<autoFocusDist2;
        return;
    }

    if (autoFocusStage == 2)
    {
        qDebug()<<"2222222222STAGEEEEEEEEEEEEEEEE------------------------------------------------------------------";
        if (active)
        {
            if (!autoFocusBar.isNull())
                autoFocusBar->setValue(3);
            bool findedBestContrast = false;
            autoFocusStopScan = false;
            autoFocusScanning = true;
            while (autoFocusScanning && !autoFocusStopScan && !isInterrupted)
            {
                int frame = frames;
                timer.start();
                while (frame == frames && !autoFocusStopScan && !isInterrupted && !timer.hasExpired(maxTimerWait)) {
                    qApp->processEvents();
                    SleepExport::msleep2(1);
                }
                if (autoFocusStopped || isInterrupted)
                    return;
                if (timer.hasExpired(maxTimerWait))
                {
                    qDebug()<<"No frames2";
                    new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
                    doBeep();

                    destructStatus(autoFocusBar);
                    isInterrupted = true;
                    BAutoFocus_clicked(false);
                    return;
                }

                double contrast = getContrast();
                qDebug()<<contrast;
                qDebug()<<"++++++++++++++++++++++++++++++++++++++++++++++"<<autoFocusMagicMultiply[autoFocusMagicMultiplyCurrentId];
                prevContrast = contrast;
                if (contrast > autoFocusContrast0 * autoFocusMagicMultiply[autoFocusMagicMultiplyCurrentId])
                {
                    findedBestContrast = true;
                    motors->goStop(mZfocus);
                    break;
                }
            }
            autoFocusScanning = false;
            if (findedBestContrast)
                qDebug()<<"scanning1 is done";
            else
            {
                qDebug()<<"Didnt finded best contrast in 1 scan!";
                BAutoFocus_clicked(false);
                return;
            }
        }
        else
        {
            autoFocusStopScan = true;
            if (autoFocusScanning)
            {
                qDebug()<<"Didnt finded best contrast in 1 scan! And stopped";
                BAutoFocus_clicked(false);
                return;
            }

            if (!autoFocusBar.isNull())
                autoFocusBar->setValue(4);
            prevContrast = autoFocusGetAverageContrast(10, maxTimerWait);
            if (isInterrupted || autoFocusStopped)
                return;
            qDebug()<<"Getting control contrast = "<<prevContrast;
            qDebug()<<"Going down 1 point";
            autoFocusStage++;
            if (autoFocusDirection == 0)
                motors->goUp(mZfocus, 1, 1);
            else
                motors->goDown(mZfocus, 1, 1);
        }
        return;
    }

    if (autoFocusStage == 3)
    {
        qDebug()<<active;
        if (active) return;

        double contrast = autoFocusGetAverageContrast(10, maxTimerWait);
        if (isInterrupted || autoFocusStopped)
            return;

        if (autoFocusScanVariant == 0)
        {
            qDebug()<<"Choosing side with contrast="<<contrast;
            autoFocusStage++;
            if (contrast > prevContrast)
            { // Left side
                qDebug()<<"Left side";
                if (autoFocusDirection == 0)
                    motors->goUp(mZfocus, autoFocusV2, autoFocusDist3);
                else
                    motors->goDown(mZfocus, autoFocusV2, autoFocusDist3);
            }
            else { // Right side
                qDebug()<<"Right side";
                if (autoFocusDirection == 0)
                    motors->goDown(mZfocus, autoFocusV2, autoFocusDist3);
                else
                    motors->goUp(mZfocus, autoFocusV2, autoFocusDist3);
            }
        }
        else if (autoFocusScanVariant == 1)
        {
            if (contrast > prevContrast)
            { // Left side
                qDebug()<<"still left side"<<contrast;
                prevContrast = contrast;
                if (autoFocusDirection == 0)
                    motors->goUp(mZfocus, autoFocusV1, autoFocusVLeftStPr);
                else
                    motors->goDown(mZfocus, autoFocusV1, autoFocusVLeftStPr);
                return;
            }

            autoFocusStage++;
            qDebug()<<"Right side"<<contrast;
            if (autoFocusDirection == 0)
                motors->goDown(mZfocus, autoFocusV2, autoFocusDist3);
            else
                motors->goUp(mZfocus, autoFocusV2, autoFocusDist3);
        }

        return;
    }

    if (autoFocusStage == 4)
    {
        if (active)
        {
            if (!autoFocusBar.isNull())
                autoFocusBar->setValue(5);
            qDebug()<<"startig scanning2";
            prevPrevContrast = prevContrast = getContrast();
            bool findedBestContrast = false;
            autoFocusStopScan = false;
            autoFocusScanning = true;
            while (autoFocusScanning && !autoFocusStopScan && !isInterrupted)
            {
                int frame = frames;
                timer.start();
                while (frame == frames && !autoFocusStopScan && !isInterrupted && !timer.hasExpired(maxTimerWait)) {
                    qApp->processEvents();
                    SleepExport::msleep2(1);
                }
                if (autoFocusStopped || isInterrupted)
                    return;
                if (timer.hasExpired(maxTimerWait))
                {
                    qDebug()<<"No frames3";
                    new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
                    doBeep();

                    destructStatus(autoFocusBar);
                    isInterrupted = true;
                    BAutoFocus_clicked(false);
                    return;
                }

                double contrast = getContrast();
                qDebug()<<contrast;
                double d1 = prevContrast - prevPrevContrast;
                double d2 = contrast - prevContrast;
                qDebug()<<(d1 < 0)<<(d2 < 0)<<prevPrevContrast<<prevContrast<<contrast;
                if (d1 < 0 && d2 < 0)
                {
                    findedBestContrast = true;
                    autoFocusStage++;
                    motors->goStop(mZfocus);
                    break;
                }
                else
                {
                    prevPrevContrast = prevContrast;
                    prevContrast = contrast;
                }
            }
            autoFocusScanning = false;
            if (findedBestContrast)
            {
                qDebug()<<"scanning2 is done";
                autoFocusStage++;
                autoFocusStages(mZfocus, 0, 0);
            }
            else
            {
                qDebug()<<"Didnt finded best contrast in 2 scan!";
                BAutoFocus_clicked(false);
            }

        }
        return;
    }

    if (autoFocusStage == 5)
    {
        qDebug()<<"Finded best contrast";
        autoFocusEndedSucsessfull = true;
        BAutoFocus_clicked(false);
        return;
    }

    BAutoFocus_clicked(false);
}

double VideoPlayer::getContrast()
{
    QImage img = ritem->image;
    img.detach();
    return getContrast(&img);
}

#define DIFF(a, b) (abs(long(a & 0xFF) - long(b & 0xFF)) + abs(long((a >> 8) & 0xFF) - long((b >> 8) & 0xFF)) + abs(long((a >> 16) & 0xFF) - long((b >> 16) & 0xFF)))
double VideoPlayer::getContrast(QImage *img)
{
    double contrast = 0.;
    for (int i = 1; i < img->width(); i++)
        for (int j = 1; j < img->height(); j++) {
            int diff1 = DIFF(img->pixel(i, j), img->pixel(i-1, j));
            if (abs(diff1) > 32)
                contrast += abs(diff1)/128.0;

            int diff2 = DIFF(img->pixel(i, j), img->pixel(i, j-1));
            if (abs(diff2) > 32)
                contrast += abs(diff2)/128.0;
        }
    return contrast;
}

void VideoPlayer::BAutoFocus_clicked(bool turnOn)
{
    //static bool old_start = false;
    auto motors = Motors::GetIt();
    auto mZfocus = motors->getRoleMotor(Motors::MotorRoles::RoleZ2Focus);

    if (turnOn)
    {
        autoFocusStopped = false;
        //old_start = m_BStart->on_off;
        //qDebug()<<"cur camera"<<old_start;
        if (!m_BStart->isOn)
            m_BStart->change_state_man_with_sig(true);
        //-----
        autoFocusMagicMultiplyCurrentId = 0;
        autoFocusDirection = 0;
        isInterrupted = false;
        autoFocusBar = new StatusBar(tr("Optical Autofocus"), 6);
        connect(autoFocusBar, SIGNAL(canceled()), SLOT(cancelOperation()));
        autoFocusBar->add();
        int frm = frames;
        int time = 6000;
        QElapsedTimer timer;
        timer.start();
        while (frm == frames && !isInterrupted && !timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
        if (timer.hasExpired(time)) {
            qDebug()<<"No frames4";
            new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
            doBeep();

            destructStatus(autoFocusBar);
            isInterrupted = true;
            return;
        }
        //-----
        autoFocusStage = 0;
        disconnect(motors, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(autoFocusStages(int, bool, bool)));
        //qDebug()<<"------------------------";
        SleepExport::msleep_gui(300);
        connect(motors, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(autoFocusStages(int, bool, bool)));
        motors->goDownE(mZfocus, 9999);
    }
    else {
        autoFocusSettingsChanged();
        autoFocusStage = -10;
        //qDebug()<<"set camera"<<old_start;
        //if (!old_start)
        //    m_BStart->change_state_man_with_sig(old_start);
        autoFocusStages(mZfocus,0,0);
        disconnect(motors, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(autoFocusStages(int, bool, bool)));
    }
}

void VideoPlayer::on_BAutoFocus_clicked()
{
    double speed = 2.0;//mkm/s
    double z1 = 0;
    double z2 = zPos->edit->maximum();//mkm

    double max = 0;
    double zmax = 0;
    int counter = 0;

    setEnabled(false);
    //bool old_start = m_BStart->on_off;
    if (!m_BStart->isOn)
        m_BStart->change_state_man_with_sig(true);

    isInterrupted = false;
    autoFocusStopped = false;

    QPointer<StatusBar> bar = new StatusBar(tr("Optical Autofocus"), 3);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    int time;
    QElapsedTimer timer;

    do {
        bar->setValue(counter);
        qDebug() << "AF pass" << counter << z1 << z2;

        zPos->setValue(z1);
        zPos->setNewValue(true);
        SleepExport::msleep_gui(200);

        double contrast = 0;

        do {
            if (contrast > max) {
                max = contrast;
                zmax = zPos->value();
            }

            int frm = frames;
            time = 6000;
            timer.start();
            while (frm == frames && !isInterrupted && !timer.hasExpired(time)) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
            if (timer.hasExpired(time)) {
                new Notification(tr("Autofocus") + ": " + tr("No frames from camera"), Notification::Warn, 10);
                doBeep();

                destructStatus(bar);
                isInterrupted = true;
            }

            if (isInterrupted)
                continue;

            //async  move & calc
            zPos->setValue(zPos->value() + speed);
            zPos->setNewValue(true);

            contrast = autoFocusGetAverageContrast(10, 6000);

            qDebug() << "=AF_=" << contrast << zPos->value() - speed;
        } while (zPos->value() < z2 && !isInterrupted);

        double dz = (z2 - z1)/16;
        double zrange = zPos->edit->maximum();

        z1 = qBound(0., zmax - dz/2, zrange);
        z2 = qBound(0., zmax + dz/2, zrange);

        speed /= 8;
        counter++;
    } while (speed >= 0.0005 && counter < 3 && !isInterrupted);

    qDebug() << "AF" << counter << zmax;

    zPos->setValue(zmax);
    zPos->setNewValue(true);
    autoFocusStopped = true;

    destructStatus(bar);

    //qDebug()<<"set camera"<<old_start;
    //if (!old_start)
    //    m_BStart->change_state_man_with_sig(false);

    setEnabled(true);
    emit AutoFocus2Ended(!isInterrupted);
}

void VideoPlayer::cancelOperation()
{
    isInterrupted = true;
    if (autoFocusStage >= 0)
        BAutoFocus_clicked(false);
}
