#include "autocalib.hpp"
#include "oscilloscope.h"
#include "gui/status.hpp"
#include "tools.hpp"
#include "modules.hpp"
#include "scanplot.h"
#include "math/plot_SubtPlane.h"
#include <qwt_plot_marker.h>
#include <QSettings>
#include <QtEvents>
#include <QApplication>
#include <QCheckBox>
#include <QDesktopWidget>
#include <QFormLayout>
#include <QTimer>

AutoCalibPlot::AutoCalibPlot(QWidget *parent) : DelayedInitWidget(parent)
{
    coeffOk = nullptr;
    CBLevel = nullptr;
    LEPeriod = nullptr;
    LELevel = nullptr;
    Source = nullptr;
    levelMarker = nullptr;
    osc = nullptr;
    CalibCurve = nullptr;
    CBScanner = 0;
    activemarker = nullptr;
    activepoint = -1;

    calib_min = 0;
    calib_coeff = 1;

    CBValue = 0;
    CBAxis = 0;

    for (int i = 0; i < 6; i++)
        lineC[i] = 0;
    lineC[2] = 1;
}

#include <limits>
bool AutoCalibPlot::eventFilter(QObject *obj, QEvent *Event)
{
    if (Event->type() == QEvent::KeyPress && ((QKeyEvent *)Event)->key() == Qt::Key_Delete) {
        QVector<int> rm;
        if (activepoint != -1) {
            rm.append(activepoint);
            activepoint = -1;
        } else if (activemarker) {
            rm.append(markers.indexOf(activemarker));
            activemarker = nullptr;
        }

        foreach (int i, rm) {
            markers.at(i)->detach();
            delete markers.at(i);
            markers.remove(i);
        }
        if (rm.size()) {
            Source->Refresh();
            ReCalcFromSourceMan();
        }
    }

    if (obj == Source->Pure2dPlot) {
        if (Event->type() == QEvent::Resize) {
            //we need it because plot core does not syncs font size for this axis, and not resets axis layout for different size
            Source->Pure2dPlot->QPlot->axisScaleDiv(QwtPlot::xTop)->invalidate();

            QFont font = qApp->font(); font.setPointSize(Source->Pure2dPlot->FontSize);
            Source->Pure2dPlot->QPlot->setAxisFont(QwtPlot::xTop, font);
        }
        if (Event->type() == QEvent::MouseButtonPress) {
            QPoint Pos = Source->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            Source->Pure2dPlot->SetClickParams(Pos);

            if (markers.size() == 0)
                return false;

            double mindist = std::numeric_limits<double>::max();
            int minidx = -1;
            for (int i = 0; i < markers.size(); i++) {
                double dist = (Pos - Source->Pure2dPlot->PlotToScreen(markers[i]->value())).manhattanLength();
                if (dist < mindist) {
                    mindist = dist;
                    minidx = i;
                }
            }

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
            double scale = qApp->desktop()->logicalDpiX();
#else
            double scale = qApp->desktop()->physicalDpiX();
#endif

            int Range = 1.5*qMax<int>(20, 0.25 * scale);
            activemarker = nullptr;
            if (mindist < Range)
                activemarker = markers[minidx];

            return activemarker != nullptr;
        }
        else if (Event->type() == QEvent::MouseButtonRelease) activemarker = nullptr;
        else if (Event->type() == QEvent::MouseMove && activemarker) {
            QPoint Pos = Source->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            activemarker->setValue(Source->Pure2dPlot->ScreenToPlot(Pos));
            Source->Pure2dPlot->Replot();
            ReCalcFromSourceMan();
            return true;
        }

        return false;
    }
    else
    if (obj == CalibCurve->Pure2dPlot) {
        if (Event->type() == QEvent::MouseButtonPress) {
            QPoint Pos = CalibCurve->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            CalibCurve->Pure2dPlot->SetClickParams(Pos);

			auto curveDt = curveDataCalib.at(0);

            if (curveDataCalib.size() == 0 || curveDt->size() == 0)
                return false;

            double mindist = std::numeric_limits<double>::max();
            int minidx = -1;
            for (int i = 0; i < curveDt->size(); i++) {
				auto point = QPointF(curveDt->getXat(i), curveDt->YSeq.at(i));
                double dist = (Pos - CalibCurve->Pure2dPlot->PlotToScreen(point)).manhattanLength();
                if (dist < mindist) {
                    mindist = dist;
                    minidx = i;
                }
            }

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
            double scale = qApp->desktop()->logicalDpiX();
#else
            double scale = qApp->desktop()->physicalDpiX();
#endif

            int Range = 1.5*qMax<int>(20, 0.25 * scale);
            activepoint = -1;
            if (mindist < Range)
                activepoint = minidx;

            return activepoint != -1;
        }
        else if (Event->type() == QEvent::MouseButtonRelease) activepoint = -1;
        else if (Event->type() == QEvent::MouseMove && activepoint != -1) {
            QPoint Pos = CalibCurve->Pure2dPlot->mapToCanvas(((QMouseEvent *)Event)->pos());
            curveDataCalib[0]->YSeq[activepoint] = CalibCurve->Pure2dPlot->ScreenYToPlotY(Pos.y());

            CalibCurve->Refresh();
            ReCalcFromCalibCurve();
            return true;
        }

        return false;
    }
    else
        return QWidget::eventFilter(obj, Event);
}

void AutoCalibPlot::BGetClicked()
{
    auto scan = ScanPlot::GetIt();
    if (!scan) return;

    CBScanner = scan->scanner_scan;
    osc->CalibWidget->Scanner_type->setCurrentIndex(CBScanner);
    CBValue = (CBScanner ? osc->CalibWidget->CL2_status_xy :
                           osc->CalibWidget->CL_status_xy) ? 1 : 0;
    CBAxis = scan->scan_direction ? 1 : 0;

    plot::TPure2dPlot *plot = scan->Plot->TwoDPlot->Pure2dPlot;

    curveData = plot::getLock(plot->SourceData);
    plot::TSIUnit xunit = plot->XAxisUnit;
    plot::TSIUnit yunit = plot->YAxisUnit;

    int active = qBound(0, plot->ActiveCurve, curveData.size() - 1);
    while (curveData.size() && !(curveData.at(active) && curveData.at(active)->size())) {
        curveData.remove(active);
        active = 0;
    }
    if (!curveData.size()) {
        curveData.clear();
        new Notification(tr("Auto calibration") + ": " + tr("Source is empty!"), Notification::Warn, 5);
        doBeep();
        return;
    }
    curveData = (plot::TCurvesDataOwner() << curveData.at(active));
    curveDataCalib.clear();

    Source->Pure2dPlot->XAxisUnit = xunit;
    Source->Pure2dPlot->YAxisUnit = yunit;
    Source->Pure2dPlot->SourceData = plot::getUnLock(curveData);
    Source->Refresh();

    ReCalcFromSource();
}

void AutoCalibPlot::BSendClicked()
{
    new Notification(tr("Autocalib") + ": " + tr("Coefficients set succesfully"), Notification::Good, 5);
    doBeep();


    void *calib_arr = nullptr;
    NLCWidget **calib_wid = nullptr;
    if (CBScanner == 0) {
        switch (CBValue) {
        case 0:
            calib_arr = osc->CalibWidget->linePiezoCalibration;
            calib_wid = osc->CalibWidget->viewPiezoCalibration;
            break;
        case 1:
            calib_arr = osc->CalibWidget->lineSensorCalibration;
            calib_wid = osc->CalibWidget->viewSensorCalibration;
            break;
        }
    } else {
        switch (CBValue) {
        case 0:
            calib_arr = osc->CalibWidget->linePiezoCalibration2;
            calib_wid = osc->CalibWidget->viewPiezoCalibration2;
            break;
        case 1:
            calib_arr = osc->CalibWidget->lineSensorCalibration2;
            calib_wid = osc->CalibWidget->viewSensorCalibration2;
            break;
        }
    }
    if (calib_arr) {
        double *calib_arr2 = &((double *)calib_arr)[CBAxis * 6];
        for (int i = 0; i < 6; i++)
            calib_arr2[i] = lineC[5-i];
        calib_wid[CBAxis]->SetPic(buildNlcPic(calib_arr2));
    }

    osc->CalibWidget->Scanner_type->setCurrentIndex(CBScanner);
    osc->CalibWidget->s_sendData();
}

void AutoCalibPlot::ReCalcCoefficent()
{
    int npower = qMin(4, adcvalues.size()-1);
    if (npower < 0)
        return;

    double coeffs[5];
    memset(coeffs, 0, sizeof(double)*5);
    plot::math::dataLinePartFitPolynom(adcvalues.data(), npower, coeffs, 0, adcvalues.size());

    double scale = LEPeriod->value();

    coeffs[4] /= scale*scale*scale*scale;
    coeffs[3] /= scale*scale*scale;
    coeffs[2] /= scale*scale;
    coeffs[1] /= scale;
    //coeffs[0] /= 1;

    for (int i = 0; i < 5; i++)
        lineC[i+1] = coeffs[i];
    lineC[0] = 0;
}

void AutoCalibPlot::ReCalcFromCalibCurve()
{
    adcvalues.clear();

    for (int i = 0; i < curveDataCalib.at(0)->size(); i++) {
        adcvalues.push_back(curveDataCalib.at(0)->YSeq.at(i));
        markers[i]->setXValue((adcvalues.back() / calib_coeff) + calib_min);
    }

    ValuesUpdate();
}

void AutoCalibPlot::ReCalcFromSourceMan()
{
    adcvalues.clear();

    for (int i = 0; i < markers.size(); i++)
        adcvalues.push_back((markers[i]->xValue() - calib_min)*calib_coeff);

    ValuesUpdate();
}

void AutoCalibPlot::ReCalcFromSource()
{
    while (!Source->Pure2dPlot->isMathReady()) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }
    Source->Pure2dPlot->s_Refresh();

    if (!Source->Pure2dPlot->QwtCurves.size() || !Source->Pure2dPlot->QwtCurves.at(0)->dataSize()) {
        //new Notification(tr("Auto calibration") + ": " + tr("Source is empty!"), Notification::Warn, 5);
        //doBeep();
        return;
    }

    const QwtPlotCurve &curve = *Source->Pure2dPlot->QwtCurves.at(0);

    bool edge_retry = false;
    bool falling_edge = CBLevel->isChecked();
    double lvl = LELevel->value() / 100;
    double level = (1-lvl)*curve.minYValue() + lvl*curve.maxYValue();
    double prev = curve.sample(0).y();

    calib_min = curve.sample(0).x();
    calib_coeff = 65535 / ((curve.sample(curve.dataSize() - 1).x()) - calib_min);

retry:
    adcvalues.clear();
    foreach (QwtPlotMarker *marker, markers) {
        marker->detach();
        delete marker;
    }
    markers.clear();

    levelMarker->setYValue(level);

    for (int i = 1; i < curve.dataSize(); i++) {
		auto y = curve.sample(i).y();
        if ((falling_edge && prev >= level && y < level) ||
                (!falling_edge && prev <= level && y > level)) {
            double k = (level - y) / (prev - y);
            double x = QLineF(curve.sample(i), curve.sample(i-1)).pointAt(k).x();

            QwtPlotMarker *marker = new QwtPlotMarker;
            marker->attach(Source->Pure2dPlot->QPlot);
            marker->setSymbol(
                        new QwtSymbol(QwtSymbol::Diamond, QColor(255,255,255), QColor(255,255,255), QSize(9, 9)));
            marker->setXValue(x);
            marker->setYValue(level);
            marker->setLineStyle(QwtPlotMarker::VLine);
            marker->setLinePen(QPen(Qt::white, 1, Qt::DashLine));
            markers.push_back(marker);

            adcvalues.push_back((x - calib_min)*calib_coeff);
        }

        prev = curve.sample(i).y();
    }

    if (!edge_retry && adcvalues.size() < 2) {
        edge_retry = true;
        falling_edge = !falling_edge;
        goto retry;
    }

    ValuesUpdate();
}

inline
double polynom_at_pos(double pos, double coeff[], int count = 4) {
    double acc = coeff[0];
    for (int i = 1; i <= count; i++)
        acc = acc * pos + coeff[i];

    return acc;
}

void AutoCalibPlot::ValuesUpdate()
{
    CalibCurve->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
    CalibCurve->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Counts"));
    CalibCurve->Pure2dPlot->Title = tr("Calibration curve");
    Source->Pure2dPlot->Title = tr("Source curve");

    curveDataCalib.clear();
    plot::ResizeOwner(curveDataCalib, 2);
    curveDataCalib[0]->Pen.setColor(Qt::black);
    curveDataCalib[1]->Pen.setColor(Qt::red);

    curveDataCalib[0]->setXCalib(0, LEPeriod->value() * 1e-6);
    curveDataCalib[0]->YSeq = adcvalues;

    ReCalcCoefficent();

    double coeff[6];
    for (int i = 0; i < 6; i++)
        coeff[i] = lineC[5 - i];
    double val = -1;

    curveDataCalib[1]->setXCalib(-5e-6, 1e-6);
    for (int i = -5; i < adcvalues.size() * LEPeriod->value(); i++) {
        val = polynom_at_pos(i, coeff);
        curveDataCalib[1]->AddY(val);
    }

    Source->Refresh();
    CalibCurve->Pure2dPlot->SourceData = plot::getUnLock(curveDataCalib);
    CalibCurve->Refresh();
    CalibCurve->Pure2dPlot->QwtCurves[0]->setSymbol(
            new QwtSymbol(QwtSymbol::Ellipse, QBrush(Qt::gray), QPen(Qt::blue), QSize(4, 4)));

    bool ok = osc->CalibWidget->checkNLCform(coeff);
    if (!ok) {
        coeffOk->setText(tr("Invalid curve"));
        coeffOk->setIcon(makeHugeIcon(":/off.png"));
    } else {
        coeffOk->setText(tr("Valid curve") + ", " + tr("click to apply"));
        coeffOk->setIcon(makeHugeIcon(":/onn.png"));
    }
}

void AutoCalibPlot::GuiCreate()
{
    QHBoxLayout *mainL = new QHBoxLayout(this);
    mainL->setContentsMargins(0, 0, 0, 0);

    QVBoxLayout *rightL1 = new QVBoxLayout;
    QVBoxLayout *plotSplit = new QVBoxLayout;
    plotSplit->setSpacing(9);

    mainL->addLayout(rightL1);
    mainL->addLayout(plotSplit, 1);


    Source = new plot::T2dPlot(nullptr, false, this);
    Source->layout()->setContentsMargins(0, 0, 6, 0);

    CalibCurve = new plot::T2dPlot(nullptr, false, this);
    CalibCurve->layout()->setContentsMargins(0, 0, 6, 0);

    plotSplit->addWidget(Source);
    plotSplit->addWidget(CalibCurve);


    QHBoxLayout *highL11 = new QHBoxLayout();
    highL11->setSpacing(6);

    QGroupBox *L2grBox = new QGroupBox(this);
    QFormLayout *highLopt = makeFormLayout(L2grBox);

    rightL1->addLayout(highL11);
    rightL1->addWidget(L2grBox);
    rightL1->addStretch();

    { //fixed scale for source data
        plot::TScale *sc;
        Source->Pure2dPlot->QPlot->setAxisScaleDraw(QwtPlot::xTop, sc = new plot::TScale);
        Source->Pure2dPlot->QPlot->setAxisScaleEngine(QwtPlot::xTop, new plot::TScaleDiv(sc));
        Source->Pure2dPlot->QPlot->enableAxis(QwtPlot::xTop);

        QFont font = qApp->font(); font.setPointSize(Source->Pure2dPlot->FontSize);
        QwtText tt("Counts"); tt.setFont(font);
        Source->Pure2dPlot->QPlot->setAxisTitle(QwtPlot::xTop, tt);
        Source->Pure2dPlot->QPlot->setAxisFont(QwtPlot::xTop, font);
        ((plot::TScale *)Source->Pure2dPlot->QPlot->axisScaleDraw(QwtPlot::xTop))->font = font;
        Source->Pure2dPlot->QPlot->setAxisScale(QwtPlot::xTop, 0, 65535);
    }

    Source->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Counts"));
    Source->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Z"));
    Source->Pure2dPlot->Title = tr("Source curve");

    CalibCurve->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
    CalibCurve->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("Counts"));
    CalibCurve->Pure2dPlot->Title = tr("Calibration curve");

    Source->PlotControl->hide();
    CalibCurve->PlotControl->hide();

    Source->Refresh();
    CalibCurve->Refresh();

    coeffOk = new QPushButton(tr("Send"));
    coeffOk->setFlat(true);
    connect(coeffOk, SIGNAL(clicked()), this, SLOT(BSendClicked()));

    QPushButton *BGet = new QPushButton(makeHugeIcon(":/AutoCalib/CalImport.png"), tr("Get scan section"));
    BGet->setFlat(true);

    LEPeriod = new Edit(0.1, 100, 2);
    LEPeriod->setValue(3);
    LEPeriod->setImmediateMode();
    LELevel = new Edit(10, 90, -1);
    LELevel->setValue(50);
    LELevel->setImmediateMode();

    CBLevel = new QCheckBox(tr("Falling edge"));

    QVBoxLayout *settL = new QVBoxLayout();
    settL->setSpacing(6);
    QVBoxLayout *plotL = new QVBoxLayout();
    plotL->setSpacing(6);
    QHBoxLayout *settL0 = new QHBoxLayout();
    settL0->addLayout(settL, 1);
    settL0->addLayout(plotL);

    highLopt->addRow(BGet);

    {
        QGroupBox *delim = new QGroupBox();
        delim->setFixedHeight(1);
        highLopt->addRow(delim);
    }

    highLopt->addRow(settL0);
    highLopt->addRow(tr("Level")+", %", LELevel);
    highLopt->addRow(tr("Period, ")+ tr("micro")+("m"), LEPeriod);

    {
        QGroupBox *delim = new QGroupBox();
        delim->setFixedHeight(1);
        highLopt->addRow(delim);
    }

    highLopt->addRow(coeffOk);


    plotL->addWidget(Source->PlotControl->TBMathProcessing);
    plotL->addWidget(Source->PlotControl->BResetMCCurve);
    plotL->addWidget(Source->PlotControl->BApply);
    plotL->addWidget(Source->PlotControl->BCancel);

    settL->addWidget(CBLevel);



    QTimer *delayUpdate = new QTimer(this);
    delayUpdate->setInterval(100);
    delayUpdate->setSingleShot(true);

    levelMarker = new QwtPlotMarker;
    levelMarker->setYValue(0);
    levelMarker->setLinePen(QPen(QColor(0, 255, 255), 0, Qt::SolidLine));
    levelMarker->setLineStyle(QwtPlotMarker::HLine);
    levelMarker->attach(Source->Pure2dPlot->QPlot);

    connect(BGet, SIGNAL(clicked()), SLOT(BGetClicked()));

    connect(delayUpdate, SIGNAL(timeout()), this, SLOT(ReCalcFromSource()));
    connect(LEPeriod, SIGNAL(newValue()), this, SLOT(ReCalcFromSource()));
    connect(LELevel, SIGNAL(newValue()), this, SLOT(ReCalcFromSource()));
    connect(CBLevel, SIGNAL(clicked()), this, SLOT(ReCalcFromSource()));
    connect(Source->PlotControl->BApply, SIGNAL(clicked()), delayUpdate, SLOT(start()));
    for (int i = 0; i < Source->PlotControl->AGFitOrders->actions().size()-1; i++)
        connect(Source->PlotControl->AGFitOrders->actions()[i], SIGNAL(triggered()), delayUpdate, SLOT(start()));

    Source->Pure2dPlot->installEventFilter(this);
    CalibCurve->Pure2dPlot->installEventFilter(this);
}
