#include "ccd.h"

#include <plot.h>
#include "ccdmatrix.h"

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtConcurrent>
#else
#include <QtCore>
#endif

namespace ccd {

void GetIntersections(const TSpectrum &spectra, int start, int end, double level, QVector<bool> &dir, QVector<int> &pos)
{
    for (int i = start+1; i <= end; i++) {
        if (spectra[i-1] <= level && spectra[i] >= level) {
            dir.push_back(true);
            pos.push_back(i);
        }
        if (spectra[i-1] >= level && spectra[i] <= level) {
            dir.push_back(false);
            pos.push_back(i);
        }
    }
}

class ThreadForCalculations
{
public:
    double raman_shift(double laser_wl, double wl)
    {
        if (qIsNull(wl) || !qIsFinite(wl)) return 0;
        return 1e7 * (1 / laser_wl - 1 / wl);
    }
    double wavelength(double central_wl, double disp, double pixelsize, int /*nx*/, double pixel_num, int cp)
    {
        return central_wl + disp * pixelsize * (pixel_num - cp);
    }
    void CheckMinMax(const QVector<double> &Seq)
    {
        if (Seq.size() <= 1) { StartIndex = 0; StopIndex = 0; return; }

        StartIndex = (int)MinForFuncApply - (int)MinForRecord;
        StopIndex = (int)MaxForFuncApply - (int)MinForRecord - 1;

        if ((int)StartIndex < 0) StartIndex = 0;
        if ((int)StopIndex < 0) StopIndex = 0;
        if ((int)StartIndex >= Seq.size()) StartIndex = Seq.size() - 1;
        if ((int)StopIndex >= Seq.size()) StopIndex = Seq.size() - 1;
        if (StopIndex < StartIndex) qSwap(StartIndex, StopIndex);
    }
    double scale(double in) {
        switch (ConvXscaleUnits) {
        case pixels :
            return in;
            break;
        default:
        case nm:
            return wavelength(CalibInfo[1], CalibInfo[2], CalibInfo[3], (int)CalibInfo[4], in, (int)CalibInfo[5]);
            break;
        case back_cm:
            return raman_shift(CalibInfo[0], wavelength(CalibInfo[1], CalibInfo[2], CalibInfo[3], (int)CalibInfo[4], in, (int)CalibInfo[5]));
            break;
        }
        return 0.0;
    }

    double Integrate(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        double Result = 0;
        for (int i = StartIndex; i <= StopIndex; i++)
            Result += Spectrum[i];

        if ((StopIndex - StartIndex + 1) > 0)
            Result /= (StopIndex - StartIndex + 1);

        return Result;
    }
    double Maximum(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        double Result = Spectrum[StartIndex];
        for (int i = StartIndex; i <= StopIndex; i++)
            Result = qMax(Result, Spectrum[i]);

        return Result;
    }
    double MaximumPosition(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        double maxvalue = Spectrum[StartIndex];
        int maxpos = StartIndex;

        for (int i = StartIndex; i <= StopIndex; i++)
            if  (Spectrum[i] > maxvalue) {
                maxvalue = Spectrum[i];
                maxpos = i;
            }

        return scale(maxpos);
    }
    double MassCenterPosition(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        double Result = 0;
        double x_sum = 0;
        for (int i = StartIndex; i <= StopIndex; i++) {
            Result += Spectrum[i] * i;
            x_sum += Spectrum[i];
        }

        return scale(Result / x_sum);
    }
    double PeakWidth(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        double min = Spectrum[StartIndex];
        for (int i = StartIndex + 1; i <= StopIndex; i++)
            min = qMin(min, Spectrum[i]);

        double max = Maximum(Spectrum);

        QVector<bool> intersect;
        QVector<int> intersectx;
        GetIntersections(Spectrum, (min + max) * 0.5, intersect, intersectx);

        if (intersect.size() != 2 || !intersect.first() || intersect.last())
            return 0;

        return scale(intersectx.last()) - scale(intersectx.first());
    }

    int FixJMCurveCrop(const TSpectrum &Spectrum)
    {
        int mid = (StartIndex + StopIndex) / 2;
        while (Spectrum[StartIndex] == -10 && StartIndex < StopIndex)
            StartIndex++;
        while (Spectrum[StopIndex] == -10 && StartIndex < StopIndex)
            StopIndex--;

        if (mid < StartIndex || mid > StopIndex)
            mid = (StartIndex + StopIndex) / 2;

        return mid;
    }
    double Stiffness(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        int mid = FixJMCurveCrop(Spectrum);
        if (StopIndex - StartIndex <= 2)
            return 0;

        double minvalue = Spectrum[StartIndex];
        int minpos = StartIndex;

        for (int i = StartIndex; i <= mid + 1; i++)
            if  (Spectrum[i] <= minvalue) {
                minvalue = Spectrum[i];
                minpos = i;
            }

        QVector<bool> intersect;
        QVector<int> intersectx;
        GetIntersections(Spectrum, (Spectrum[StartIndex] + minvalue) * 0.5, intersect, intersectx);

        double k = 0;
        ConvXscaleUnits = ccd::nm;
        for (int i = 0; i < intersectx.size(); i++)
            k += qFabs(Spectrum[qBound(StartIndex, (intersectx.at(i) + minpos) / 2, StopIndex)] - Spectrum[qBound(StartIndex, (intersectx.at(i)*3 - minpos) / 2, StopIndex)]) /
                    qFabs(scale(minpos) - scale(intersectx.at(i))) /
                    intersectx.size();

        return qIsFinite(k) ? k : 0;
    }
    double Adhesion(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        int mid = FixJMCurveCrop(Spectrum);
        if (StopIndex - StartIndex <= 2)
            return 0;

        double max = Spectrum[mid];
        for (int i = mid + 1; i <= StopIndex; i++)
            max = qMax(max, Spectrum[i]);

        return max;
    }
    double EnergyD(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        FixJMCurveCrop(Spectrum);
        if (StopIndex - StartIndex <= 2)
            return 0;

        double minvalue = Spectrum[StartIndex];
        int minpos = StartIndex;

        for (int i = StartIndex; i <= StopIndex; i++)
            if  (Spectrum[i] < minvalue) {
                minvalue = Spectrum[i];
                minpos = i;
            }

        double Result = 0;
        for (int i = qMax(StartIndex, 2*minpos - StopIndex); i <= minpos; i++)
            if (Spectrum[minpos + (minpos - i)] > Spectrum[i]) //if (minpos + (minpos - i) <= StopIndex)
                Result += (Spectrum[minpos + (minpos - i)] - Spectrum[i]);

        return Result;
    }
    double Baseline(const TSpectrum &Spectrum)
    {
        CheckMinMax(Spectrum);

        if (StartIndex == 0 && StopIndex <= 1)
            return 0;

        FixJMCurveCrop(Spectrum);
        if (StopIndex - StartIndex <= 2)
            return 0;

        double min = Spectrum[StartIndex];
        double max = Spectrum[StartIndex];
        for (int i = StartIndex; i <= StopIndex; i++) {
            min = qMin(min, Spectrum[i]);
            max = qMax(max, Spectrum[i]);
        }

        const int size = 15;
        int bins[size];
        for (int i = 0; i < size; i++) bins[i] = 0;
        max -= min; //range
        for (int i = StartIndex; i <= StopIndex; i++) {
            int bin = qRound((size-1) * ((Spectrum[i] - min) / max));
            bins[bin]++;
        }

        double avg = 0;
        int avg_count = 0;
        bool bin_on[size];
        bin_on[0] = (bins[0] >= (StopIndex - StartIndex) / (size-1) / 2);
        bin_on[size-1] = (bins[size-1] >= (StopIndex - StartIndex) / (size-1) / 2);
        for (int i = 1; i < size-1; i++)
            bin_on[i] = (bins[i] >= (StopIndex - StartIndex) / (size-1));
        for (int i = StartIndex; i <= StopIndex; i++) {
            int bin = qRound((size-1) * (1. - (Spectrum[i] - min) / max));
            if (bin_on[bin]) {
                avg += Spectrum[i];
                avg_count++;
            }
        }

        return avg_count ? avg / avg_count : 0;
    }
    void GetIntersections(const TSpectrum &Spectrum, double level, QVector<bool> &dir, QVector<int> &pos)
    {
        CheckMinMax(Spectrum);

        ccd::GetIntersections(Spectrum, StartIndex, StopIndex, level, dir, pos);
    }

public:
    int ConvXscaleUnits;
    typedef double (ThreadForCalculations::*FuncPtr)(const TSpectrum &);
    FuncPtr AllFuncPtrs[9];
    FuncPtr CurrentFunc;
    QVector<double> CalibInfo;
    int StartIndex, StopIndex, MinForRecord;
    double MinForFuncApply, MaxForFuncApply;
    const TInfoForSetOfSpectrums *Info;
    plot::T2dArray *Results;
    int WorkStartsAt, WorkEndsWith;

    ThreadForCalculations() {
        Results = nullptr;
        StartIndex = 0;
        StopIndex = 0;
        MinForRecord = 0;
        MinForFuncApply = MaxForFuncApply = 0.;
        ConvXscaleUnits = 0;
        Info = nullptr;
        AllFuncPtrs[0]=&ThreadForCalculations::Integrate;
        AllFuncPtrs[1]=&ThreadForCalculations::Maximum;
        AllFuncPtrs[2]=&ThreadForCalculations::MaximumPosition;
        AllFuncPtrs[3]=&ThreadForCalculations::MassCenterPosition;
        AllFuncPtrs[4]=&ThreadForCalculations::PeakWidth;
        AllFuncPtrs[5]=&ThreadForCalculations::Stiffness;
        AllFuncPtrs[6]=&ThreadForCalculations::Adhesion;
        AllFuncPtrs[7]=&ThreadForCalculations::EnergyD;
        AllFuncPtrs[8]=&ThreadForCalculations::Baseline;
        CurrentFunc = AllFuncPtrs[0];
        WorkStartsAt = WorkEndsWith = 0;
        CalibInfo.resize(6);
    }
    void run()
    {
        MinForFuncApply = Info->MinForFuncApply;
        MaxForFuncApply = Info->MaxForFuncApply;
        MinForRecord = (int)Info->MinForRecord;
        CurrentFunc = AllFuncPtrs[Info->CurrentFunc];

        int StartY, StartX, StopY, StopX;
		auto firstArraySize = Info->Array.at(0).size();
        StartY = WorkStartsAt / firstArraySize;
        StopY = WorkEndsWith / firstArraySize;
        StartX = WorkStartsAt % firstArraySize;
        StopX = WorkEndsWith % firstArraySize;

        for (int Y = StartY; Y <= StopY; Y++)
        {
            for (int X = StartX; X < firstArraySize; X++)
            {
                if (Y == StopY && X >= StopX) return;
                auto arr = Info->Array.at(Y);
                if (X >= arr.size()) continue;
                QPoint Point = Info->Translate(X, Y);
                //(*Results)[Point.y()][Point.x()]
                const_cast<double &>(Results->at(Point.y()).at(Point.x())) = (this->*CurrentFunc)(arr.at(X));
            }
            StartX = 0;
        }
    }
    //void operator()() { run(); } typedef void result_type;
};


int TSetOfSpectrum::CalcResults(int from)
{
    if (!Array.size() || !Array.at(0).size())
        return 0;

    int TotalNumOfPoint = 0; //NXElems * NYElems
    for (int i = 0; i < Array.size(); i++)
        TotalNumOfPoint += Array.at(i).size();

    bool incr = false;
    if (from <= 0 || from >= TotalNumOfPoint)
        from = 0;
    else incr = true;

    plot::T3dDataOwner newResults;
    if (!incr) {
        plot::ResizeOwner(newResults, 1);
		auto result = newResults.at(0);
		result->XLimits = XLimits;
		result->YLimits = YLimits;
		result->setSize(NXElems, NYElems);
		result->detach();
    } else
        Results.at(0)->detach();

    int NumOfThreads = qMin<int>(QThread::idealThreadCount(), TotalNumOfPoint - from);
    int NumOfPointPerThread = (TotalNumOfPoint - from) / NumOfThreads;
    QVector<ThreadForCalculations> Threads(NumOfThreads);
    QVector<QFuture<void> > futures(NumOfThreads);
    TInfoForSetOfSpectrums data = *this;
    for (int i = 0; i < NumOfThreads; i++)
    {
        Threads[i].ConvXscaleUnits = XscaleUnit;
        Threads[i].CalibInfo = CalibInfo;
        Threads[i].Info = &data;
        Threads[i].Results = &(incr ? Results : newResults).at(0)->Array;
        Threads[i].WorkStartsAt = from + i * NumOfPointPerThread;
        Threads[i].WorkEndsWith = from + (i + 1) * NumOfPointPerThread;
        if (i == NumOfThreads - 1)
            Threads[i].WorkEndsWith = TotalNumOfPoint;

        futures[i] = QtConcurrent::run(&Threads[i], &ThreadForCalculations::run); //QtConcurrent::run(Threads[i]);
    }

    //TODO use async completion signals
    for (int i = 0; i < NumOfThreads; i++)
        futures[i].waitForFinished();

    if (!incr)
        *Results.at(0) = *newResults.at(0);

    emit finished();

    return TotalNumOfPoint;
}

void TSetOfSpectrumConstructor::CalcSection()
{
    QPoint point;
    Section[0]->clear();
    if ((bool)data->Direction)
    {
        //Section[0]->setXCalib(0, 0);
        int nx = CurrentXPos;
        int ny = CurrentYPos;
        if (nx == 0 && ny > 0) {
            nx = data->NXElems;
            ny--;
        }
        for (int i = 0; i < nx; i++)
        {
            point = data->Translate(i, ny);

            double x = double(i) / (data->NXElems - 1);
            switch (data->StartPoint) {
            case TInfoForSetOfSpectrums::TopRight:
            case TInfoForSetOfSpectrums::BottomRight:
                x = 1 - x;
                break;
            }
            Section[0]->AddPoint(x * (data->XLimits[1] - data->XLimits[0]), data->Results.at(0)->Array.at(point.y()).at(point.x()));
        }
    }
    else
    {
        int lineFull=1;
        int counterX1=0;
        int counterX2=0;
        int counter=0;
        for (int counterY = YlinesCounter * data->NYElems / data->NXElems; counterY <= CurrentYPos; counterY++)
        {
            counterX1 = lineFull * YlinesCounter * data->NYElems % data->NXElems;
            counterX2 = data->NXElems - 1;
            lineFull=0;
            if (counterY == CurrentYPos)
                counterX2 = CurrentXPos - 1;

            for (int counterX=counterX1; counterX<=counterX2; counterX++)
            {
                point = data->Translate(counterX, counterY);

                double y = double(counter) / (data->NYElems - 1);
                switch (data->StartPoint) {
                case TInfoForSetOfSpectrums::TopLeft:
                case TInfoForSetOfSpectrums::TopRight:
                    y = 1 - y;
                    break;
                }
                Section[0]->AddPoint(y * (data->YLimits[1] - data->YLimits[0]), data->Results.at(0)->Array.at(point.y()).at(point.x()));
                counter++;
            }
        }
    }
}


QFuture<void> TSetOfSpectrumSaveOpen::Open(QIODevice *Dataf, TSetOfSpectrum *Src)
{
    Spectrum = Src;
    Data = Dataf;
    Action = TSetOfSpectrumSaveOpen::Act_Open;

    return QtConcurrent::run(this, &TSetOfSpectrumSaveOpen::run);
}

QFuture<void> TSetOfSpectrumSaveOpen::Save(QIODevice *Dataf, TSetOfSpectrum *Src)
{
    if (Src->CalibInfo.size() != 6) {
        qDebug() << "Wrong calib size";
        return QFuture<void>();
    }

    Spectrum = Src;
    Data = Dataf;
    Action = TSetOfSpectrumSaveOpen::Act_Save;

    return QtConcurrent::run(this, &TSetOfSpectrumSaveOpen::run);
}

struct TMetaDataForFile
{
    int Direction, StartPoint;
    unsigned int NX, NY;
    double XMin, XMax, YMin, YMax, MinForF, MaxForF, MinForRec, MaxForRec;
    double Laser_WL, Cent_WL, Disp, pixel_Xsize, Nx, Cent_Pixel;
};


void TSetOfSpectrumSaveOpen::SaveMetaData(QIODevice &File, TMetaDataForFile &MetaData)
{
    MetaData.Direction = Spectrum->Direction;
    MetaData.StartPoint = Spectrum->StartPoint;

    MetaData.NX = Spectrum->NXElems;
    MetaData.NY = Spectrum->NYElems;

    MetaData.XMin = Spectrum->XLimits[0];
    MetaData.XMax = Spectrum->XLimits[1];
    MetaData.YMin = Spectrum->YLimits[0];
    MetaData.YMax = Spectrum->YLimits[1];

    MetaData.MinForRec = Spectrum->MinForRecord;
    MetaData.MaxForRec = Spectrum->MaxForRecord;

    MetaData.MinForF = Spectrum->MinForFuncApply;
    MetaData.MaxForF = Spectrum->MaxForFuncApply;

    MetaData.Laser_WL = Spectrum->CalibInfo[0];
    MetaData.Cent_WL = Spectrum->CalibInfo[1];

    MetaData.Disp = Spectrum->CalibInfo[2];
    MetaData.pixel_Xsize = Spectrum->CalibInfo[3];

    MetaData.Nx = Spectrum->CalibInfo[4];
    MetaData.Cent_Pixel = Spectrum->CalibInfo[5];

    File.write((char *)&MetaData, sizeof(MetaData));
}
void TSetOfSpectrumSaveOpen::SaveSpectrum(QIODevice &File, const TSpectrum &Spec)
{
    quint32 NumberOfPoints = Spec.size();
    File.write((const char *)&NumberOfPoints, sizeof(NumberOfPoints));
    File.write((const char *)Spec.constData(), NumberOfPoints * sizeof(double));
}
bool TSetOfSpectrumSaveOpen::GetMetaDataFromFile(QIODevice &File, TMetaDataForFile &MetaData)
{
    if (File.read((char *)&MetaData, sizeof(MetaData)) != sizeof(MetaData)) {
        qDebug() << "metadata read error";
        return false;
    }

    Spectrum->CalibInfo.resize(6);
    Spectrum->CalibInfo[0] = MetaData.Laser_WL;
    Spectrum->CalibInfo[1] = MetaData.Cent_WL;
    Spectrum->CalibInfo[2] = MetaData.Disp;
    Spectrum->CalibInfo[3] = MetaData.pixel_Xsize;
    Spectrum->CalibInfo[4] = MetaData.Nx;
    Spectrum->CalibInfo[5] = MetaData.Cent_Pixel;

    return true;
}
TSpectrum TSetOfSpectrumSaveOpen::GetSpectrumFromFile(QIODevice &File)
{
    TSpectrum Spec;

    quint32 NumberOfPoints;
    File.read((char *)&NumberOfPoints, sizeof(NumberOfPoints));

    Spec.reserve(NumberOfPoints);
    Spec.resize(NumberOfPoints);
    File.read((char *)Spec.data(), NumberOfPoints * sizeof(double));
    Spec.squeeze();

    return Spec;
}
void TSetOfSpectrumSaveOpen::_Open(QIODevice &Data)
{
    TMetaDataForFile MetaData;
    TSetOfSpectrumConstructor I4dAdd;
    I4dAdd.data = Spectrum;
    if (!GetMetaDataFromFile(Data, MetaData)) {
        return;
    }

    Range XLims, YLims;
    XLims[0] = MetaData.XMin; XLims[1] = MetaData.XMax;
    YLims[0] = MetaData.YMin; YLims[1] = MetaData.YMax;

    I4dAdd.PrepareForNewScan(MetaData.NX, MetaData.NY, XLims, YLims, MetaData.MinForRec, MetaData.MaxForRec, MetaData.MinForF, MetaData.MaxForF);
    Spectrum->SetDirStPoint(MetaData.Direction, MetaData.StartPoint);

    long TotalWork = MetaData.NX * MetaData.NY;

    QElapsedTimer t; t.start();
    try {
        for (int C = 0; C < TotalWork && !(Data.bytesAvailable() == 0 && Data.atEnd()); C++) {
            TSpectrum spec = GetSpectrumFromFile(Data);
            I4dAdd.AddNextSpectrum(spec);
            if ((C % 10) == 0 && t.elapsed() > 100) {
                t.restart();
                emit ChangeProgress(qMax(qMin(1., double(Data.pos()) / Data.size()), double(C) / TotalWork));
            }
        }
    } catch (...) {
        //Mem fail
    }
}
void TSetOfSpectrumSaveOpen::_Save(QIODevice &Data)
{
    TMetaDataForFile MetaData;
    SaveMetaData(Data, MetaData);

    int C = 0;
    double TotalWork = 0; //MetaData.NX * MetaData.NY

	for (int YInd = 0; YInd < Spectrum->Array.size(); YInd++)
    {
		auto spec = Spectrum->Array.at(YInd);
		for (int XInd = 0; XInd < spec.size(); XInd++)
			if (spec.at(XInd).size())
				TotalWork++;
	}

    QElapsedTimer t; t.start();
    for (int YInd = 0; YInd < Spectrum->Array.size(); YInd++)
	{
		auto spec = Spectrum->Array.at(YInd);
        for (int XInd = 0; XInd < spec.size(); XInd++) {
            SaveSpectrum(Data, spec.at(XInd));
            if (spec.at(XInd).size())
                C++;
            if ((C % 10) == 0 && t.elapsed() > 100) {
                t.restart();
                emit ChangeProgress(double(C) / TotalWork);
            }
        }
    }
}
void TSetOfSpectrumSaveOpen::run()
{
    if (!Data) {
        qDebug() << "CCDSaveOpen no data, command" << Action;
        return;
    }

    try {
        if (Action == Act_Save) _Save(*Data);
        else if (Action == Act_Open) _Open(*Data);
    } catch (...) {
        qDebug() << "CCDSaveOpen exception!";
    }

    if (Data && Data->isOpen())
        Data->close();
    if (Data && Data->parent() == this)
        Data->deleteLater();
    Data = nullptr;

    emit finished();
}

} //namespace ccd




QString T4dDataHelper::getConvZUnits(int func, int scale, const QString &srcYunit)
{
    switch (func) {
    case ccd::TSetOfSpectrum::F_MaxPosition:
    case ccd::TSetOfSpectrum::F_MassCenterPozition:
    case ccd::TSetOfSpectrum::F_PeakWidth:
        return getScaleUnits(scale);

    case ccd::TSetOfSpectrum::F_Integrate:
    case ccd::TSetOfSpectrum::F_Maximum:
        return !srcYunit.isNull() ? srcYunit : QObject::tr("Counts");

    case ccd::TSetOfSpectrum::F_Adhesion_F:
    case ccd::TSetOfSpectrum::F_Energy:
    case ccd::TSetOfSpectrum::F_Baseline:
        return !srcYunit.isNull() ? srcYunit : QObject::tr("V");

    case ccd::TSetOfSpectrum::F_Stiffness: {
        QString u = !srcYunit.isNull() ? srcYunit : QObject::tr("V");
        QString suf = "/" + QObject::tr("n") + QObject::tr("m");
        return u + (u.isNull() ? "1" : QString()) + suf;
    }
    }

    return QObject::tr("Counts");
}

double T4dDataHelper::ReCalibrate(double X, const QVector<double> &CalibInfo, int idx)
{
    if (idx != 0)
        X = 1e-9 * wavelength(CalibInfo[1], CalibInfo[2], CalibInfo[3], (int)X, (int)CalibInfo[5]);
    if (idx == 2)
        X = raman_shift(CalibInfo[0], 1e9 * X);
    return X;
}
double T4dDataHelper::ReCalibrate_inverse(double X, const QVector<double> &CalibInfo, int idx)
{
    if (idx == 2)
        X = 1e-9 * raman_shift_Inverse(X, CalibInfo[0]);
    if (idx != 0)
        X = wavelength_Inverse(1e9 * X, CalibInfo[1], CalibInfo[2], CalibInfo[3], (int)CalibInfo[5]);
    return X;
}

void T4dDataHelper::ShowSpectr(plot::T2dPlot *Plot, plot::T2dPlot *PlotVert, ccd::TSetOfSpectrum *Info, int XInd, int YInd, int sc_ind)
{
    if (XInd < 0)
        XInd = 0;
    if (YInd < 0)
        YInd = 0;

    if (YInd >= Info->Array.size() || XInd >= Info->Array[YInd].size())
        return;

	auto firstCurve = Info->pointCurve[0];
    firstCurve->clear();
    firstCurve->YSeq = Info->Array[YInd][XInd];
    for (int i = 0; i < Info->Array[YInd][XInd].size(); i++)
        firstCurve->XSeq.append(ReCalibrate(i + Info->MinForRecord, Info->CalibInfo, sc_ind));

    QString xl, xu;
    switch (sc_ind) {
    case 0:
        xl = QObject::tr("pixels");
        break;
    case 1:
        xl = QObject::tr("WL");
        xu = QObject::tr("m");
        break;
    case 2:
        xl = QObject::tr("Raman Shift, 1/cm");
        break;
    }
    Plot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(xl, xu);
    Plot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(!Info->YscaleUnit.isNull() ? QString() : QObject::tr("Counts"), Info->YscaleUnit);
    Plot->Pure2dPlot->Title = QObject::tr("Point Spectrum");
    if (PlotVert)
    {
        PlotVert->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(xl, xu);
        PlotVert->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(!Info->YscaleUnit.isNull() ? QString() : QObject::tr("Counts"), Info->YscaleUnit);
        PlotVert->Pure2dPlot->Title = QObject::tr("Point Spectrum");
    }

    Plot->Pure2dPlot->SourceData = plot::getUnLock(Info->pointCurve);
    if (PlotVert)
    {
        PlotVert->Pure2dPlot->clearSouceData();
        PlotVert->Pure2dPlot->SourceData = *Plot->Pure2dPlot->cloneSourceData();
        Plot->Refresh();
        PlotVert->Refresh();
    }
    else
        Plot->Refresh();
}

void T4dDataHelper::SetSelectGeometry(plot::T2dPlot *Plot, double L, double R)
{

    if (Plot->Pure2dPlot->vertical)
    {
        Plot->Pure2dPlot->SetSelectCenter(QPointF((Plot->Pure2dPlot->CurrentXLimits[1]/2. +
                                                  Plot->Pure2dPlot->CurrentXLimits[0]) / 2,
                                                (L + R) / 2));
        Plot->Pure2dPlot->SetSelectHeight(L - R);
        Plot->Pure2dPlot->SetSelectWidth(Plot->Pure2dPlot->CurrentXLimits[1] - Plot->Pure2dPlot->CurrentXLimits[0]);
    }
    else
    {
        Plot->Pure2dPlot->SetSelectCenter(QPointF((L + R) / 2,
                                                  (Plot->Pure2dPlot->CurrentYLimits[1]/2. +
                                                  Plot->Pure2dPlot->CurrentYLimits[0]) / 2));
        Plot->Pure2dPlot->SetSelectWidth(R - L);
        Plot->Pure2dPlot->SetSelectHeight(Plot->Pure2dPlot->CurrentYLimits[1] - Plot->Pure2dPlot->CurrentYLimits[0]);
    }
}
