#pragma once

#include <2d/plot_2dtypes.h>
#include <3d/plot_3dtypes.h>
#include "eg3000/defs.h"
#include <QFuture>
#include <QFile>

namespace ccd {

enum ZUnits {pixels,nm,back_cm};

typedef QVector<double> TSpectrum;

class GaussSpectrum: public TSpectrum
{
    double Sigma, d, U, A, pw;
    int NumberOfSteps;
    double GaussF(double x, double r2) { return A * ((sin(r2 / 5.) + 1.7) + 1) * exp( -(qPow(x - U, 2.*pw) + r2) / (2*Sigma*Sigma)) / (Sigma * qSqrt(2*M_PI)); }
    void Generate(const QPointF &P)
    {
        for (int i = 0; i < NumberOfSteps; i++)
            append(GaussF(i*d, 0.6*(P.x()*P.x() + P.y()*P.y()) ) );
    }
public:
    GaussSpectrum(const int &Nx, const QPointF &C, const QPointF &CP, const double &u = 0, const double &SigmaMult = 1., const double &AmplitMult = 1., double _pw = 1.0)
    { Sigma = 10. * SigmaMult; A = 1e5 * AmplitMult; U = u; NumberOfSteps = Nx; d = 1.; pw = _pw; Generate(C - CP); }
};

void GetIntersections(const TSpectrum &spectra, int start, int end, double level, QVector<bool> &dir, QVector<int> &pos);


template <class T>
class TAbstract3dData
{
public:
    QVector<QVector<T> > Array;

    void Add(int X, int Y, const T &Val)
    {
        if (Array.size() <= Y) Array.resize(Y + 1);
        if (Array.at(Y).size() <= X) Array[Y].resize(X + 1);
        Array[Y][X] = Val;
    }
};
//TODO common with 3dplot

class TInfoForSetOfSpectrums: public TAbstract3dData<TSpectrum>
{
public:
    enum TFunctions {F_Integrate, F_Maximum, F_MaxPosition, F_MassCenterPozition, F_PeakWidth, F_Stiffness, F_Adhesion_F, F_Energy, F_Baseline };
    enum TDirections { Vert, Horiz };
    enum TStartPoints { BottomLeft, BottomRight, TopLeft, TopRight };

    Range XLimits, YLimits;
    int NXElems, NYElems;
    double dx, dy;

    double MinForFuncApply, MaxForFuncApply, MinForRecord, MaxForRecord;

    int Direction;
    int StartPoint;
    int CurrentFunc;

    TInfoForSetOfSpectrums()
    {
        NXElems = NYElems = 0;
        dx = dy = 0.;
        XLimits[0] = 0;
        YLimits[0] = 0;
        XLimits[1] = 100;
        YLimits[1] = 100;
        Direction = Horiz;
        StartPoint = TopLeft;
        CurrentFunc = F_Integrate;
        MinForRecord = MinForFuncApply = 0;
        MaxForRecord = MaxForFuncApply = 1;
    }

    QPoint Translate(int CurrX, int CurrY) const
    {
        int NumberOfProcessedPoints = CurrY * NXElems + CurrX;
        int nxy = (Direction == TInfoForSetOfSpectrums::Horiz) ? NXElems : NYElems;
        int DeltaX = NumberOfProcessedPoints / nxy;
        int DeltaY = NumberOfProcessedPoints % nxy;
        if (Direction == TInfoForSetOfSpectrums::Horiz)
            qSwap(DeltaX, DeltaY);

        QPoint Result;
        Result.setY((StartPoint == TInfoForSetOfSpectrums::BottomLeft || StartPoint == TInfoForSetOfSpectrums::BottomRight) ?
                        DeltaY : NYElems - DeltaY - 1);
        Result.setX((StartPoint == TInfoForSetOfSpectrums::BottomLeft || StartPoint == TInfoForSetOfSpectrums::TopLeft) ?
                        DeltaX : NXElems - DeltaX - 1);
        return Result;
    }
    void SetLimits(double Xmin, double Xmax, double Ymin, double Ymax)
    {
        XLimits[0] = Xmin; XLimits[1] = Xmax;
        YLimits[0] = Ymin; YLimits[1] = Ymax;
    }
    void SetDirStPoint(int dir, int stPoint) { Direction = dir; StartPoint = stPoint; }
    void SetMinMaxForFuncApply(double Min, double Max) { MinForFuncApply = Min; MaxForFuncApply = Max; }
    void SetMinMaxForRecord(double Min, double Max) { MinForRecord = Min; MaxForRecord = Max; }
    void UpdateDxDy() { dx = (XLimits[1] - XLimits[0]) / (NXElems - 1); dy = (YLimits[1] - YLimits[0]) / (NYElems - 1); }
    void SetNxNy(int Nx, int Ny) { NXElems = Nx; NYElems = Ny; UpdateDxDy(); }
    //double GetSizeOfScan(double nx, double ny) const { return fabs(MaxForRecord - MinForRecord) * sizeof(double) * nx * ny / 1024. / 1024.; }

};

class TSetOfSpectrum: public QObject, public TInfoForSetOfSpectrums
{
    Q_OBJECT

signals:
    void finished();

public:
    int XscaleUnit;
    QString YscaleUnit;
    QVector<double> CalibInfo;

    plot::T3dDataOwner Results;
    plot::TCurvesDataOwner pointCurve;

public:

    TSetOfSpectrum()
    {
        CalibInfo.resize(6);
        XscaleUnit = ccd::nm;
        plot::ResizeOwner(pointCurve, 1);
        plot::ResizeOwner(Results, 1);
    }
    int CalcResults(int from = 0); //can calc incrementally

    bool IsIn(const Range &Lims, double Val) const { return Lims[0] <= Val && Val <= Lims[1]; }
    void GetIntegerIndexes(double PlotX, double PlotY, int &XInd, int &YInd) const
    {
        XInd = 0; YInd = 0;

        if (!IsIn(XLimits, PlotX) || !IsIn(YLimits, PlotY))
            return;

        switch (StartPoint) {
        case TopLeft:
        case BottomLeft:
            XInd = qRound((PlotX - XLimits[0]) / dx);
            break;
        case TopRight:
        case BottomRight:
            XInd = qRound((XLimits[1] - PlotX) / dx);
            break;
        }
        switch (StartPoint) {
        case TopLeft:
        case TopRight:
            YInd = qRound((YLimits[1] - PlotY) / dy);
            break;
        case BottomRight:
        case BottomLeft:
            YInd = qRound((PlotY - YLimits[0]) / dy);
            break;
        }
        if (!Direction)
            qSwap(XInd, YInd);
    }
};

class TSetOfSpectrumConstructor {
public:
    TSetOfSpectrum *data;

    int CurrentXPos, CurrentYPos;
    int YlinesCounter;

    plot::TCurvesDataOwner Section;
    void CalcSection();

    TSetOfSpectrumConstructor()
    {
        data = nullptr;
        YlinesCounter = 0;
        CurrentXPos = CurrentYPos = 0;
        plot::ResizeOwner(Section, 1);
    }
    void PrepareForNewScan(unsigned int Nx, unsigned int Ny, const Range &XLims, const Range &YLims, double MinForRec, double MaxForRec, double MinForF, double MaxForF)
    {
        data->Array.clear();

        data->XLimits = XLims;
        data->YLimits = YLims;
        data->SetNxNy(Nx, Ny);
        data->SetMinMaxForFuncApply(MinForF, MaxForF);
        data->SetMinMaxForRecord(MinForRec, MaxForRec);

        CurrentXPos = 0;
        CurrentYPos = 0;
        YlinesCounter = 0;
    }
    void AddNextSpectrum(TSpectrum Sp)
    {
        data->Add(CurrentXPos, CurrentYPos, Sp);

        CurrentXPos++;
        if (CurrentXPos == data->NXElems) { CurrentXPos = 0; CurrentYPos++; }

        if (CurrentXPos - 1 + CurrentYPos * data->NXElems == (YlinesCounter + 1) * data->NYElems)
            YlinesCounter++;
    }
    int GetCurrentLineNumber()
    {
        return ((bool)data->Direction) ? (CurrentYPos - (CurrentXPos == 0 && CurrentYPos > 0 ? 1 : 0)) : YlinesCounter;
    }
    //bool IsFinished() const { return CurrentXPos >= data->NXElems || CurrentYPos >= data->NYElems; }
};

struct TMetaDataForFile;
class TSetOfSpectrumSaveOpen: public QObject
{
    Q_OBJECT
signals:
    void ChangeProgress(double);
    void finished();

private:
    QIODevice *Data;
    TSetOfSpectrum *Spectrum;

    enum TActions { Act_Save, Act_Open };
    TActions Action;

public:
    TSetOfSpectrumSaveOpen() { Spectrum = nullptr; }
    QFuture<void> Save(QIODevice *Dataf, TSetOfSpectrum *Src);
    QFuture<void> Open(QIODevice *Dataf, TSetOfSpectrum *Src);
    QFuture<void> Open(const QString &File, TSetOfSpectrum *Src)
    {
        QFile *Dataf = new QFile(File, this);
        Dataf->open(QIODevice::ReadOnly);

        return Open(Dataf, Src);
    }

private:
    void SaveMetaData(QIODevice &File, TMetaDataForFile &MetaData);
    void SaveSpectrum(QIODevice &File, const TSpectrum &Spec);
    bool GetMetaDataFromFile(QIODevice &File, TMetaDataForFile &MetaData);
    TSpectrum GetSpectrumFromFile(QIODevice &File);
    void _Open(QIODevice &Data);
    void _Save(QIODevice &Data);
    void run();

};

} //namespace ccd;


namespace plot { class T2dPlot; }
class T4dDataHelper
{
public:
    static double getScaleUnity(int scale)
    {
        switch (scale) {
        case 0: return 1;
        case 1: return 1e-9;
        case 2: return 1;
        }

        return 1;
    }
    static QString getScaleUnits(int scale)
    {
        switch (scale) {
        case 0: return QObject::tr("pixels");
        case 1: return QObject::tr("nm");
        case 2: return QObject::tr("1/cm");
        }

        return QString();
    }
    static QString getConvZUnits(int func, int scale, const QString &srcYunit = QString());

    static double ReCalibrate(double X, const QVector<double> &CalibInfo, int idx);
    static double ReCalibrate_inverse(double X, const QVector<double> &CalibInfo, int idx);

    static void ReCalcRamanOrWavelengthToPixels(const QVector<double> &CalibInfo, double &Min, double &Max, int sc)
    {
        Min = ReCalibrate_inverse(Min, CalibInfo, sc);
        Max = ReCalibrate_inverse(Max, CalibInfo, sc);
    }
    static void ReCalcRamanOrWavelengthFromPixels(const QVector<double> &CalibInfo, double &Min, double &Max, int sc)
    {
        Min = ReCalibrate(Min, CalibInfo, sc);
        Max = ReCalibrate(Max, CalibInfo, sc);
    }

    static void ShowSpectr(plot::T2dPlot *Plot, plot::T2dPlot *PlotVert, ccd::TSetOfSpectrum *Info, int XInd, int YInd, int sc_ind);

    static void SetSelectGeometry(plot::T2dPlot *Plot, double L, double R);
};
