#include "ccdplot.h"
#include "gui/status.hpp"
#include "tools.hpp"
#include "modules.hpp"

#include <oscilloscope.h>
#include <autocalib.hpp>
#include <hardware/EG3000.h>
#include <motors.hpp>

#include "ccdmatrix.h"
#include "hardware/coboltcontroller.hpp"
#include "hardware/shamrock.hpp"
#include "hardware/solart2.hpp"
#include "hardware/nspectr.hpp"
#include "hardware/nspectr_proxy.h"
#include "remote_nst_dataid.hpp"
#include "mobile/remote_mobile.hpp"
#include "widgets/twodoverlaybuttons.h"

#include <QSettings>
#include <QCheckBox>
#include <QFileDialog>
#include <QApplication>
#include <QRadioButton>
#include <QTextEdit>
#include <QMessageBox>
#include <QTimer>
#include <QGridLayout>
#include <QFormLayout>


void CCDPlot::fillLaserConf()
{
    QSettings settings("Param/laser.conf", QSettings::IniFormat);

    settings.beginGroup("port");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        QString addr = settings.value(key, "COM1").toString();
        qDebug() << "cobolt laser" << key << "at" << addr;
        cobolt_gui_list.push_back(new CoboltGui(this, key, addr));
    }
    settings.endGroup();

    settings.beginGroup("max_power");
    int i = 0;
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        double maxv = settings.value(key, 50).toUInt();
        cobolt_gui_list[i]->power->setMainRangeAndStep(0, maxv, 1);
        cobolt_gui_list[i]->power_ew->setMaximum(maxv);
        if (++i >= cobolt_gui_list.size())
            break;
    }
    settings.endGroup();

    settings.beginGroup("wavelength");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_wavelength.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("edgefilter");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_edgefilter_idx.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("beamexpander_wl_idx");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_beamexpander_wl_idx.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("beamexpander_diam_idx");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_beamexpander_diam_idx.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("monoch_x");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_monoch_x.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("monoch_y");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_monoch_y.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("monoch_z");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_monoch_z.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("confocal_pmt_x");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_confocal_pmt_x.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("confocal_pmt_y");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_confocal_pmt_y.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();

    settings.beginGroup("confocal_pmt_z");
    foreach (QString key, settings.allKeys()) {
        if (key == "dummy")
            continue;

        laser_list_confocal_pmt_z.push_back(settings.value(key, 0).toDouble());
    }
    settings.endGroup();
}



void CCDPlot::intensity_calib_enabled()
{
    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    settings.setValue("ccd/intensity_calib_on_off", calibEn->isChecked());
    settings.setValue("ccd/intensity_calib_pixel_on_off", pixelCalibEn->isChecked());

    ccdMatrix->intensity_calib_on_off = calibEn->isChecked();
    ccdMatrix->intensity_calib_pixel_on_off = pixelCalibEn->isChecked();
}
void CCDPlot::intensity_calib_open_wl(QString name)
{
    if (name.isEmpty())
        name = QFileDialog::getOpenFileName(this, QString(), nullptr, nullptr, nullptr,
                                            QFileDialog::DontUseNativeDialog);

    if (name.isEmpty())
        return;

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    settings.setValue("ccd/intensity_calib_name", name);

    QVector<QPair<double, double> > calib;
    double mult = 1;

    QFile fl(name);
    fl.open(QIODevice::ReadOnly);
    QTextStream st(&fl);
    while (!st.atEnd()) {
        QString str = st.readLine();
        if (str.startsWith('#')) {
            if (str == "# X units = WL, m") mult = 1e9; //ease of use hack for our ascii export files
            continue;
        }

        QTextStream st1; st1.setString(&str);
        double pos, val;
        st1 >> pos >> val;

        if (st1.status() != QTextStream::Ok)
            continue;

        calib.append(QPair<double, double>(pos * mult, val));
    }

    if (calib.size())
        calibLbl->setText(QFileInfo(name).fileName() + ": " +
                          QString::number(calib.first().first) + " - " +
                          QString::number(calib.last().first) + " nm");
    else
        calibLbl->setText(QFileInfo(name).fileName() + ": " + "no data");

    ccdMatrix->intensity_calib_on_off = false;
    ccdMatrix->intensity_calib_data = calib;
    ccdMatrix->intensity_calib_on_off = calibEn->isChecked();
}
void CCDPlot::intensity_calib_open_px(QString name)
{
    if (name.isEmpty())
        name = QFileDialog::getOpenFileName(this, QString(), nullptr, nullptr, nullptr,
                                            QFileDialog::DontUseNativeDialog);

    if (name.isEmpty())
        return;

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    settings.setValue("ccd/intensity_calib_pixel_name", name);

    QVector<QPair<int, double> > calib;

    QFile fl(name);
    fl.open(QIODevice::ReadOnly);
    QTextStream st(&fl);
    while (!st.atEnd()) {
        QString str = st.readLine();
        if (str.startsWith('#'))
            continue;

        QTextStream st1; st1.setString(&str);
        double pos, val;
        st1 >> pos >> val;

        if (st1.status() != QTextStream::Ok)
            continue;

        calib.append(QPair<int, double>(pos, val));
    }

    if (calib.size())
        pixelCalibLbl->setText(QFileInfo(name).fileName() + ": " +
                               QString::number(calib.first().first) + " - " +
                               QString::number(calib.last().first) + " px");
    else
        pixelCalibLbl->setText(QFileInfo(name).fileName() + ": " + "no data");

    ccdMatrix->intensity_calib_pixel_on_off = false;
    ccdMatrix->intensity_calib_pixel_data = calib;
    ccdMatrix->intensity_calib_pixel_on_off = pixelCalibEn->isChecked();
}

void CCDPlot::SetManualBaseLine()
{
    plot::TCurveData Base = MyPl->TwoDPlot->Pure2dPlot->GetMCCurve();
    if (Base.size() < 2)
        return;

    /*curveCalib.insert("Laser Wavelength, nm", ccdMatrix->Laser_WL);
    curveCalib.insert("Central Wavelength, nm", ccdMatrix->Cent_WL_real);
    curveCalib.insert("Dispersion", ccdMatrix->Disp_real);
    curveCalib.insert("pixelsize", ccdMatrix->pixel_Xsize);
    curveCalib.insert("Nx", ccdMatrix->Nx_real);
    curveCalib.insert("Central Pixel", ccdMatrix->Cent_Pixel_real);*/

    double CalibInfo[5];
    CalibInfo[0]=ccdMatrix->Laser_WL;
    CalibInfo[1]=ccdMatrix->Cent_WL_real;
    CalibInfo[2]=ccdMatrix->Disp_real;
    CalibInfo[3]=(double)ccdMatrix->pixel_Xsize;
    CalibInfo[4]=(double)ccdMatrix->Cent_Pixel_real;

    int size = qMin(80000, 2*curveData[0]->size());
    if (size < 2)
        return;
    long *Base_Line = new long[size];
    for (int i = 0; i < size; i++) {
        double x = 0.;
        switch (currentRelativeScale)
        {
        case 0: x=(double)i;
            break;
        case 1: x=1e-9*wavelength(CalibInfo[1], CalibInfo[2], CalibInfo[3], i, (int)CalibInfo[4]);
            break;
        case 2: x=raman_shift(CalibInfo[0], wavelength(CalibInfo[1], CalibInfo[2], CalibInfo[3], i, (int)CalibInfo[4]));
            break;
        }
        if (x <= Base.XSeq.data()[0])
            Base_Line[i] = Base.YSeq.data()[0];
        else if (x >= Base.XSeq.data()[Base.size()-1])
            Base_Line[i] = Base.YSeq.data()[Base.size()-1];
        else
            for (int k=0; k < (int)(Base.size()-1); k++) {
                if ((x <= Base.XSeq.data()[k+1]) && (x >= Base.XSeq.data()[k])) {
                    Base_Line[i] = Base.YSeq.data()[k] +
                            (x - Base.XSeq.data()[k]) * (Base.YSeq.data()[k+1] -
                                                        Base.YSeq.data()[k]) / (Base.XSeq.data()[k+1] -
                                                                                Base.XSeq.data()[k]);
                }
            }
    }

    ccdMatrix->SetManualBaseLine(Base_Line);
    delete [] Base_Line;
}



void CCDPlot::initOptics()
{
    if (ccdMatrix->m_spectrInterface)
        return;

    Monochr->setEnabled(false);
    B3DMonochr->setEnabled(false);

    I_SPECTR *spectr = nullptr;
    nspectr_mutex = nullptr;
    Nspec_Initalized = false;

    void *hWnd = nullptr;
#ifdef _WIN32
    hWnd = (void *)parentWidget()->winId();
#endif

    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
#ifdef USE_SOLART2
    if (settings.value("use_solart2", false).toBool())
        spectr = new ISolarT2(hWnd);
    else
#endif
#ifdef USE_SHAMROCK
    if (settings.value("use_shamrock", false).toBool())
        spectr = new IShamrock();
    else
#endif
    {
        spectr = new INspectr(hWnd);
        nspectr_mutex = &((INspectr *)spectr)->guiThreadLock;
    }
    setDeviceStatus2(EDplay); //busy
    spectr->initialize();

    setDeviceStatus2(spectr->initialized ? EDok : EDno);
    if (spectr->initialized) {
        if (nspectr_mutex)
            Nspec_Initalized = true;
        else {
            Nspec_Initalized = settings.value("use_ars", true).toBool();
            if (Nspec_Initalized)
                ARS3000::init();
        }

        ccdMatrix->m_spectrInterface = spectr;
        deviceStatus2->label->setToolTip(spectr->getName());
    } else {
        nspectr_mutex = nullptr;

        delete spectr;
        Monochr->setEnabled(true);
        B3DMonochr->setEnabled(true);
        return;
    }

    if (spectr->getGratingCount() > 0) {
        CBgrating_select->blockSignals(true);
        CBgrating_select->clear();
        QList<QVariant> list;
        for (int i = 0; i < spectr->getGratingCount(); i++)
        {
            QString name = spectr->getGratingName(i);
            list << name;
            CBgrating_select->addItem(name);
        }
        auto mob = MobileServer::GetIt();
        if (mob && mob->need_send())
            mob->setPropIdVar(DataID::setMonochrGratingList, list);
        CBgrating_select->blockSignals(false);

        CBgrating_select->setCurrentIndex(spectr->getActiveGrating());
    } else {
        CBgrating_select->disconnect();
        CBgrating_select->clear();
        CBgrating_select->addItem(tr("Not supported"));
        CBgrating_select->setEnabled(false);
    }

    AutoCalibration->change_state_man_with_sig(true);
    Monochr->setEnabled(true);
    B3DMonochr->setEnabled(true);
}

void CCDPlot::sl_show_Monochromator()
{
    if (!ccdMatrix->m_spectrInterface)
        initOptics();
    if (!ccdMatrix->m_spectrInterface)
        return;

    if (Nspec_Initalized && nspectr_mutex) {
        QMutexLocker lck(nspectr_mutex);
        Proxy_ShowWnd();
    }
}
void CCDPlot::opticsReset()
{
    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (!settings.value("autoinit_ccd", false).toBool())
            return;
    }

    if (!ccdMatrix->m_spectrInterface)
        initOptics();
    if (!ccdMatrix->m_spectrInterface || !ModuleManager::instance->isRunning)
        return;

    Monochr->setEnabled(false);
    B3DMonochr->setEnabled(false);
    I_SPECTR *spectr = ccdMatrix->m_spectrInterface;

    StatusBar *bar = new StatusBar(tr("Optics reset: edge filter"), 4);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    isInterrupted = false;

    //reset commands
    if (Nspec_Initalized && nspectr_mutex) {
        QMutexLocker lck(nspectr_mutex);
        NProxy_SetEdgeFilterPos(NProxy_GetEdgeFilterPos());
        while (Proxy_GetStatus() != STATUS_READY && !isInterrupted) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }

    if (isInterrupted) {
        Monochr->setEnabled(true);
        B3DMonochr->setEnabled(true);
        return;
    }

    bar->statusLabel->setText(tr("Optics reset: monochromator grating"));
    bar->setValue(1);

    spectr->setActiveGrating(spectr->getActiveGrating(), true);
    while (!spectr->isReady() && !isInterrupted) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }

    if (isInterrupted) {
        Monochr->setEnabled(true);
        B3DMonochr->setEnabled(true);
        return;
    }

    bar->statusLabel->setText(tr("Optics reset: monochromator wavelength"));
    bar->setValue(2);

    spectr->setWl(spectr->getWl(), true);
    while (!spectr->isReady() && !isInterrupted) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }

    if (isInterrupted) {
        Monochr->setEnabled(true);
        B3DMonochr->setEnabled(true);
        return;
    }

    bar->statusLabel->setText(tr("Optics reset: laser configuration"));
    bar->setValue(3);

    QSettings settings("Param/laser.conf", QSettings::IniFormat);
    if (settings.value("index", -1).toInt() >= 0 && settings.value("index", -1).toInt() < CBPreMonochWl->count() && Nspec_Initalized) {
        CBPreMonochWl->setCurrentIndex(settings.value("index", -1).toInt());
        if (nspectr_mutex) {
            QMutexLocker lck(nspectr_mutex);
            while (Proxy_GetStatus() != STATUS_READY && !isInterrupted) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
        } else {
            while (!ARS3000::isReady() && !isInterrupted) {
                qApp->processEvents();
                SleepExport::msleep2(10);
            }
        }
    }

    destructStatus(bar);
    Monochr->setEnabled(true);
    B3DMonochr->setEnabled(true);
}

void CCDPlot::cancelOperation()
{
    isInterrupted = true;
}

void CCDPlot::PMTScan()
{
#ifdef USE_CCD_EMULATOR
    QSettings settings_("Param/Emulator.conf", QSettings::IniFormat);
    if (!settings_.value("ccd", false).toBool())
#endif
    if (!ccdMatrix->m_spectrInterface || !ModuleManager::instance->isRunning) {
        new Notification(tr("PMT Scan") + ": " + tr("Initialize optics before use."), Notification::Warn, 10);
        doBeep();

        return;
    }
    auto osc = Oscilloscope::GetIt();
    if (!osc)
        return;


    //params
    QDialog *w = new QDialog(this);
    Edit *EStartPos = new Edit(0, 1000, 1); EStartPos->setImmediateMode();
    Edit *EEndPos = new Edit(0, 1000, 1); EEndPos->setImmediateMode();
    Edit *EStepsCount = new Edit(2, 10000); EStepsCount->setImmediateMode();
    Edit *EAvgCount = new Edit(1, 10000); EAvgCount->setImmediateMode();
    QComboBox *CBch = new QComboBox;
    fillCBchannels_(CBch, false);
    {
        QPushButton *BOk = new QPushButton(tr("Start"));
        QPushButton *BCancel = new QPushButton(tr("Cancel"));


        connect(BOk, SIGNAL(clicked()), w, SLOT(accept()));
        connect(BCancel, SIGNAL(clicked()), w, SLOT(reject()));

        QFormLayout *layout = makeFormLayout(w);
        layout->setSizeConstraint(QLayout::SetFixedSize);

        layout->addRow(new QLabel(tr("Start wavelength: ")), EStartPos);
        layout->addRow(new QLabel(tr("End wavelength: ")), EEndPos);
        layout->addRow(new QLabel(tr("Points: ")), EStepsCount);
        layout->addRow(new QLabel(tr("Averaging: ")), EAvgCount);
        layout->addRow(new QLabel(tr("Channel: ")), CBch);

        layout->addRow(BOk);
        layout->addRow(BCancel);

        w->setWindowTitle(tr("PMT Scan Options"));
    }
    QSettings settings("Param/ccd.conf", QSettings::IniFormat);
    EStartPos->setValue(settings_double(settings, "pmt_scan_start", 450));
    EEndPos->setValue(settings_double(settings, "pmt_scan_end", 650));
    EStepsCount->setValue(settings_double(settings, "pmt_scan_steps", 1000));
    EAvgCount->setValue(settings_double(settings, "pmt_scan_avgs", 1));
    CBch->setCurrentIndex(CBch->findData(settings.value("pmt_scan_ch", srcADCHR1).toInt()));

    int res = w->exec();

    double start = EStartPos->value();
    double end = EEndPos->value();
    int steps = EStepsCount->value();
    int avgs = EAvgCount->value();
    int ch = CBch->itemData(CBch->currentIndex()).toInt();
    QString chs = CBch->currentText();
    delete w;
    if (res != 1)
        return;

    settings.setValue("pmt_scan_start", start);
    settings.setValue("pmt_scan_end", end);
    settings.setValue("pmt_scan_steps", steps);
    settings.setValue("pmt_scan_avgs", avgs);
    settings.setValue("pmt_scan_ch", ch);
    settings.sync();

    if (!PMTPlot) {
        PMTPlot = new plot::T2dPlot();
        PMTPlot->SetTitle(tr("PMT Spectra"));
        PMTPlot->layout()->setContentsMargins(0, 0, 6, 0);
        ((QBoxLayout *)MyPl->parentWidget()->layout())->addWidget(PMTPlot, 1);

        plot::ResizeOwner(PMTcurveData, 1);
        PMTPlot->Pure2dPlot->SourceData = plot::getUnLock(PMTcurveData);

        connect(PMTPlot, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)),
                this, SLOT(AddDataToContainer_slotPMT(plot::TCurvesDataOwner,plot::TMetaStorage)));
    }
    MyPl->hide();
    PMTPlot->show();


    //scan
    setEnabled(false);
    I_SPECTR *spectr = ccdMatrix->m_spectrInterface;

    StatusBar *bar = new StatusBar(tr("PMT Scan"), steps);
    connect(bar, SIGNAL(canceled()), SLOT(cancelOperation()));
    bar->add();

    isInterrupted = false;

    PMTPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Wavelength"), QObject::tr("m"));
    PMTPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(chs, input_units(ch));

    double step = (end - start) / steps;
    PMTcurveData[0]->YSeq.clear();
    PMTcurveData[0]->setXCalib(start * 1e-9, step * 1e-9);

    QElapsedTimer tr; tr.start();
    for (int i = 0; i < steps; i++) {
        if (spectr)
            spectr->setWl(start + step * i, i == 0);
        while (spectr && !spectr->isReady() && !isInterrupted) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }

        if (isInterrupted) {
            setEnabled(true);
            return;
        }

        SleepExport::msleep_gui(10); //steady

        double avg = 0;
        for (int ii = 0; ii < avgs && !isInterrupted; ii++)
            avg += input_calib(ch, osc->acf->read_direct(ch));
        PMTcurveData[0]->YSeq << (avg / avgs);

        if (isInterrupted) {
            setEnabled(true);
            return;
        }

        //bar->statusLabel->setText(tr(" "));
        bar->setValue(i);
        qApp->processEvents();
        if (tr.elapsed() > 100) {
            PMTPlot->Refresh();
            tr.restart();
        }
    }
    PMTPlot->Refresh();

    destructStatus(bar);
    setEnabled(true);

    lineCentralWlMonoch->setNewValue(true); //restore previous wl

    if (AutoSend) {
        PMTPlot->on_SyncTimer_timeout();
        PMTPlot->PlotControl->BSave->animateClick();
    }
}
void CCDPlot::AddDataToContainer_slotPMT(plot::TCurvesDataOwner src, plot::TMetaStorage attributes)
{
    attributes.insert("Name", PMTPlot->Pure2dPlot->Title);
    attributes.insert("Source", "CCD");

    auto osc = Oscilloscope::GetIt();
    if (osc) {
        attributes.insert("Stage X, um", osc->DacsWidget->X2->value());
        attributes.insert("Stage Y, um", osc->DacsWidget->Y2->value());
    }
    if (ccdMatrix->m_spectrInterface)
    attributes.insert("Grating", CBgrating_select->currentText());
    if (Nspec_Initalized && nspectr_mutex)
    attributes.insert("ND filter", NDEdit->value());

    foreach (CoboltGui *it, cobolt_gui_list)
        attributes.insert("Laser Power, mW - " + it->windowTitle(), it->cpower->text());

    emit AddDataToContainer(src, attributes);
}



static double ars_slit_pos[2] = {-1, -1};
static double ars_laser_pos = -1;
bool ARS3000::init()
{
    return true;
}

bool ARS3000::setSlitWidth(int idx, double width)
{
    if (idx >= 2)
        return false;

    auto motors = Motors::GetIt();
    if (!motors)
        return false;

#ifdef _NEW_MOTORS_CHANGES
    int csb = motors->csb_motor_count[0];
    if (csb < 0) csb = 6;
    else csb += motors->csb_motor_count[1] +  motors->csb_motor_count[2];
#else
    int csb = motors->csb_motor_count;
    if (csb < 0) csb = 6;
#endif

    int midx0 = 1+idx; //target motor
    int midx = csb+1+midx0;
    midx = motors->mapPhysToVirt(midx);


    QSettings settings("Param/ars.conf", QSettings::IniFormat);
    double k = settings.value("motor_k/"+QString::number(midx0), 1).toDouble();
    k *= settings.value("slit_k/"+QString::number(idx), 1).toDouble();

    double speed = 200;// * k;
    double zeropos = settings.value("slit_zeropos/"+QString::number(idx), 0).toDouble();
    width += zeropos;

    if (ars_slit_pos[idx] < 0 || width == ars_slit_pos[idx] || width == 0) {
        ars_slit_pos[idx] = 0;
        motors->goDownE(midx, speed);

        SleepExport::msleep_gui(300);
        while (!isReady()) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }

    if (width == ars_slit_pos[idx])
        return true;

    motors->go(midx, 1, speed, 1e3 * (width - ars_slit_pos[idx]) * k);
    ars_slit_pos[idx] = width;
    return true;
}

bool ARS3000::setLaser(int idx)
{
    auto motors = Motors::GetIt();
    if (!motors)
        return false;

#ifdef _NEW_MOTORS_CHANGES
    int csb = motors->csb_motor_count[0];
    if (csb < 0) csb = 6;
    else csb += motors->csb_motor_count[1] +  motors->csb_motor_count[2];
#else
    int csb = motors->csb_motor_count;
    if (csb < 0) csb = 6;
#endif

    int midx0 = 0; //target motor
    int midx = csb+1+midx0;
    midx = motors->mapPhysToVirt(midx);


    QSettings settings("Param/ars.conf", QSettings::IniFormat);
    double k = settings.value("motor_k/"+QString::number(midx0), 1).toDouble();

    double speed = 700 * k;
    double pos = settings.value("laser_select_pos/"+QString::number(idx), 0).toDouble();

    if (ars_laser_pos < 0 || ars_laser_pos == pos || pos == 0) {
        ars_laser_pos = 0;
        motors->goDownE(midx, speed);

        SleepExport::msleep_gui(300);
        while (!isReady()) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }

    if (pos == ars_laser_pos)
        return true;

    motors->go(midx, 1, speed, 1e3 * (pos - ars_laser_pos) * k);
    ars_laser_pos = pos;
    return true;
}

bool ARS3000::isReady()
{
    auto motors = Motors::GetIt();
    if (!motors)
        return true;

#ifdef _NEW_MOTORS_CHANGES
    int csb = motors->csb_motor_count[0];
    if (csb < 0) csb = 6;
    else csb += motors->csb_motor_count[1] +  motors->csb_motor_count[2];
#else
    int csb = motors->csb_motor_count;
    if (csb < 0) csb = 6;
#endif

    for (int i = csb; i < csb + 4, i < 16; i++) { //fixme first condition will be ignored
        if (motors->Motor_on[i])
            return false;
    }

    return true;
}



CCDNLCEditor::CCDNLCEditor(CCDPlot *_plot) : ccdplot(_plot), eng(nullptr)
{
    setWindowTitle("NSpec - CCD NLC");

    timer = new QTimer(this);
    timer->setSingleShot(true);
    timer->setInterval(500);
    connect(timer, SIGNAL(timeout()), this, SLOT(testFormula()));

    QGridLayout *layout = new QGridLayout(this);

    QPushButton *pb;
    layout->addWidget(pb = new QPushButton(tr("Open")), 0, 0);
    pb->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    connect(pb, SIGNAL(clicked()), this, SLOT(open()));

    layout->addWidget(pb = new QPushButton(tr("Save")), 0, 1);
    pb->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    connect(pb, SIGNAL(clicked()), this, SLOT(save()));

    layout->addWidget(pb = new QPushButton(tr("Update")), 0, 2);
    pb->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    connect(pb, SIGNAL(clicked()), timer, SLOT(start()));
    connect(pb, SIGNAL(clicked()), this, SLOT(setStatus()));

    layout->addWidget(LFormulaStatus = new QLabel, 0, 3);

    layout->addWidget(BSpline = new QCheckBox(tr("Use cubic spline")), 0, 4);
    connect(BSpline, SIGNAL(clicked()), timer, SLOT(start()));
    connect(BSpline, SIGNAL(clicked()), this, SLOT(setStatus()));
    BSpline->setChecked(true);
    BSpline->hide();

    layout->addWidget(edit = new QTextEdit, 2, 0, 1, 5);
    layout->addWidget(plot2d = new plot::T2dPlot, 3, 0, 1, 5);
    plot2d->PlotControl->hide();

    plot2d->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(tr("Pixels"));
    plot2d->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(tr("WL"), tr("m"));
    plot2d->Pure2dPlot->Title = tr("Calibration curve");

    edit->setAcceptRichText(false);
    connect(edit, SIGNAL(textChanged()), timer, SLOT(start()));
    connect(edit, SIGNAL(textChanged()), this, SLOT(setStatus()));

    mode = -1;

    plot::ResizeOwner(curveData, 1);
    plot2d->Pure2dPlot->SourceData = plot::getUnLock(curveData);
    
    dataFormula = "function calibration(pixel, central_pixel, central_wl, disp, pixelsize) \n{"
            "\n\t return central_wl + disp * pixelsize * (pixel - central_pixel)"
            "\n\t * Math.abs(pixel - central_pixel) / 512"
            "; \n} \n";
    dataFile = "#pixel - wavelength \n10-532 \n512 - 544.2 \n\n870 569.01";
}

void CCDNLCEditor::open()
{
    QString name = QFileDialog::getOpenFileName(this, QString(), nullptr,
                                                QString("CCD calibration(*.txt)"),
                                                nullptr,
                                                QFileDialog::DontUseNativeDialog);

    if (name.isEmpty())
        return;

    if (!name.endsWith(".txt"))
        name.append(".txt");

    QFile fl(name);
    fl.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream st(&fl);
    st.setCodec("Windows-1251");
    edit->setPlainText(st.readAll());
}

void CCDNLCEditor::save()
{
    QString name = QFileDialog::getSaveFileName(this, QString(), "",
                                                QString("CCD calibration(*.txt)"),
                                                nullptr,
                                                QFileDialog::DontUseNativeDialog);

    if (name.isEmpty())
        return;

    if (!name.endsWith(".txt"))
        name.append(".txt");

    QFile fl(name);
    fl.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text);
    QTextStream st(&fl);
    st.setCodec("Windows-1251");
    st << edit->toPlainText();
}

void CCDNLCEditor::setMode()
{
    if (mode == 0)
        dataFormula = edit->toPlainText();
    if (mode == 1)
        dataFile = edit->toPlainText();

    if (ccdplot->BNLCFormula->isChecked()) {
        mode = 0;
        edit->setText(dataFormula);
        ccdplot->BNLCCfg->setEnabled(true);
        BSpline->hide();
    } else if (ccdplot->BNLCFile->isChecked()) {
        mode = 1;
        edit->setText(dataFile);
        ccdplot->BNLCCfg->setEnabled(true);
        BSpline->show();
    } else if (ccdplot->BNLCAuto->isChecked()) {
        mode = -1;
        ccdplot->ccdMatrix->nonlinearMode = 1;
        ccdplot->BNLCCfg->setEnabled(false);
    } else {//if (ccdplot->BNLCNone->isChecked()) {
        mode = -2;
        ccdplot->ccdMatrix->nonlinearMode = 0;
        ccdplot->BNLCCfg->setEnabled(false);
    }
}

void CCDNLCEditor::abort()
{
    if (eng)
        eng->abortEvaluation();
}

#include <qwt_spline.h>
void CCDNLCEditor::testFormula()
{
    QString val = edit->toPlainText();

    if (mode == 1) {
        QVector<QPointF> lst;
        QVector<QStringList> lst_2;
        bool simple_form = true; //values
        QString prologue;
        QString delim = "-";

        QTextStream st(&val, QIODevice::ReadOnly);
        while (!st.atEnd()) {
            QString s = st.readLine();
            if (s.startsWith("#code:"))
                prologue += s.right(s.length() - 6) + ";";
            if (s.startsWith("#scripted"))
                simple_form = false;
            if (s.startsWith("#delimeters:"))
                delim = s.right(s.length() - 12);
            if (s.startsWith('#'))
                continue;

            if (simple_form) {
                QVector<double> v;
                QStringList l = s.split(QRegExp("[ -]"), QString::SkipEmptyParts);
                foreach (QString s1, l) {
                    bool ok;
                    double val = s1.remove(' ').toDouble(&ok);
                    if (ok)
                        v.append(val);
                }
                if (v.size() != 2)
                    continue;
                lst.append(QPointF(v.first(), v.last()));
            } else {
                QStringList l = s.split(QRegExp("[" + delim + "]"), QString::SkipEmptyParts);
                if (l.size() != 2)
                    continue;
                lst_2.append(l);
            }
        }

        //#code:function cm_to_wl(raman_shift, laser_wl) { return 1 / (1 / laser_wl - raman_shift * 1e-7); }
        //100 - cm_to_wl(0, 544.2)
        if (!simple_form) {
            QScriptEngine myEngine;
            myEngine.setProcessEventsInterval(500);

            QTimer *timer = new QTimer;
            connect(timer, SIGNAL(timeout()), this, SLOT(abort()));
            eng = &myEngine;

            lst.clear();
            foreach(QStringList p, lst_2) {
                timer->start(300);
                QScriptValue res = myEngine.evaluate(prologue + p.first());
                QScriptValue res2 = myEngine.evaluate(prologue + p.last());
                if (!res.isNumber() || !res2.isNumber())
                    continue;
                lst.append(QPointF(res.toNumber(), res2.toNumber()));
            }
            lst_2.clear();
            if (myEngine.hasUncaughtException())
                lst.clear();

            eng = nullptr;
            timer->stop();
            timer->deleteLater();
        }

        if (lst.size() < 2) {
            setStatus(2);
            return;
        }

        val = "var datax = [%1]; var datay = [%2];"
                "function calibration(pixel, central_pixel, central_wl, disp, pixelsize) \n{"
                "j = 0; while (datax[j+1] <= pixel) j++;"
                "dx = (pixel - datax[j]) / (datax[j+1] - datax[j]);"
                "dy = datay[j+1] - datay[j];"
                "return datay[j] + dy * dx;"
                "}";

        QStringList l1, l2;
        if (BSpline->isChecked()) {
            QwtSpline spline;
            spline.setPoints(QPolygonF(lst));

            for (int i = 0; i < ccdplot->ccdMatrix->Nx; i++) {
                double val = spline.value(i);
                l1.append(QString::number(i));
                l2.append(QString::number(val));
            }
            for (int i = ccdplot->ccdMatrix->Nx; i < ccdplot->ccdMatrix->Nx*2; i+=10) {
                double val = spline.value(i);
                l1.append(QString::number(i));
                l2.append(QString::number(val));
            }
        } else {
            if (lst.first().x() >= 0) {
                QLineF l(lst.first(), lst.at(1));
                double dx = (-1 - l.p1().x()) / (l.p2().x() - l.p1().x());
                lst.prepend(l.pointAt(dx));
            }
            {
                QLineF l(lst.at(lst.size() - 2), lst.last());
                double dx = (ccdplot->ccdMatrix->Nx_real*2 - l.p1().x()) / (l.p2().x() - l.p1().x());
                lst.append(l.pointAt(dx));
            }

            foreach(QPointF p, lst) {
                l1.append(QString::number(p.x()));
                l2.append(QString::number(p.y()));
            }
        }

        val = val.arg(l1.join(", "), l2.join(", "));
    }

    if (mode == 0 || mode == 1) {
        QScriptEngine myEngine;
        myEngine.setProcessEventsInterval(500);

        QTimer *timer = new QTimer;
        connect(timer, SIGNAL(timeout()), this, SLOT(abort()));
        eng = &myEngine;
        timer->start(300);

        myEngine.evaluate(val);
        QScriptValue fun = myEngine.globalObject().property("calibration");
        QScriptValue res = myEngine.evaluate("calibration(1,1,1,1,1)");
        //QScriptValue res = fun.call(QScriptValue(), QScriptValueList() << 1 << 1 << 1 << 1 << 1);

        eng = nullptr;
        timer->stop();
        timer->deleteLater();
        myEngine.setProcessEventsInterval(-1);

        curveData[0]->clear();
        if (!fun.isFunction() || fun.isError() || res.isError() || !res.isValid() || myEngine.hasUncaughtException())
            setStatus(2);
        else {
            double prev = -1e10;
            bool ok = true;
            for (int i = 0; i < ccdplot->ccdMatrix->Nx; i++) {
                double val = fun.call(QScriptValue(),
                                      QScriptValueList() << i
                                      << ccdplot->ccdMatrix->Cent_Pixel
                                      << ccdplot->ccdMatrix->Cent_WL
                                      << ccdplot->ccdMatrix->Disp
                                      << ccdplot->ccdMatrix->pixel_Xsize).toNumber();

                curveData[0]->AddPoint(i, 1e-9 * val);

                if (val < prev)
                    ok = false;
                prev = val;
            }

            ccdplot->ccdMatrix->nonlinearCalibFormula = val;
            setStatus(ok ? 0 : 3);
        }

        plot2d->Pure2dPlot->Refresh();
        return;
    }

    setStatus(2);
}

void CCDNLCEditor::setStatus(int st)
{
    if (st == 0) {
        LFormulaStatus->setText(tr("Ok"));
        LFormulaStatus->setStyleSheet("background-color: green;");
        if (mode >= 0) ccdplot->ccdMatrix->nonlinearMode = 2;
    } else if (st == 1) {
        LFormulaStatus->setText(tr("Validating..."));
        LFormulaStatus->setStyleSheet("background-color: grey;");
        if (mode >= 0) ccdplot->ccdMatrix->nonlinearMode = 0;
    } else if (st == 2) {
        LFormulaStatus->setText(tr("Bad formula"));
        LFormulaStatus->setStyleSheet("background-color: red;");
        if (mode >= 0) ccdplot->ccdMatrix->nonlinearMode = 0;
    } else if (st == 3) {
        LFormulaStatus->setText(tr("Ok. Warning: non-monotonically increasing function"));
        LFormulaStatus->setStyleSheet("background-color: orange;");
        if (mode >= 0) ccdplot->ccdMatrix->nonlinearMode = 2;
    }
}

CCDPlot* CCDPlot::GetIt() noexcept
{
    auto ccd = ModuleManager::instance->find1VoidPtr("ccdplot");
    return static_cast<CCDPlot*>(ccd);
}
