#include "cont_nstfile.h"
#include "ccd/ccd.h"

#include <QElapsedTimer>
#include <QThread>
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtConcurrent>
#else
#include <QtCore>
#endif
#include <quazip.h>
#include <quazipfile.h>

const QString Label_2d("2d");
const QString Label_3d("3d");
const QString Label_4d("4d");
const QString Label_Image("Image");

const QString ItemBegin("[BeginOfItem]");
const QString ItemEnd("[EndOfItem]");

const QString AttributesField("Attributes");

TNstFileHandler::TNstFileHandler()
{
    DataInput1 = nullptr;
    readDataSize = 0.;
    NumberOfFiles = CurrentFileNumber = -1;
}

void TNstFileHandler::progress4d(double val)
{
    emit FileProccessed(NumberOfFiles, CurrentFileNumber / 2. + val);
}

void TNstFileHandler::setLoadProgress(int idx, double val)
{
    loadProgress[idx] = val;
    double progress = 0;
    for (int i = 0; i < loadProgress.size(); i++)
        progress += loadProgress.at(i);
    emit FileProccessed(NumberOfFiles, progress);
}

void TNstFileHandler::Save(const QString &FileName, const TAbstractInfos &Data, bool oldFormat)
{
    QuaZip Archive(FileName);
    if (!Archive.open(QuaZip::mdCreate)) throw "Error creating file!";

    emit FileProccessed(NumberOfFiles = Data.size(), CurrentFileNumber = 0);

    for (int i = 0; i < Data.size(); i++)
    {
        QuaZipFile DataOutput(&Archive);
        QString textFileName = "";
        if (!oldFormat)
            textFileName += "data";
        textFileName += QString::number(i);
        textFileName += (oldFormat)? ".lsdlsd" : ".txt";
        QuaZipNewInfo DataInfo(textFileName);

        if (!DataOutput.open(QIODevice::WriteOnly, DataInfo)) throw "Error saving data!";

        Stream.setDevice(&DataOutput);
        Stream.setCodec("UTF-8");
        Stream << QChar(QChar::ByteOrderMark); //so load() decoder could use UTF-8 instead of win1251
        switch (Data[i].InfoType) {
        case TAbstractInfo::_2d:
            SaveData((plot::TCurvesDataOwner *)Data[i].Info, Data[i].attributes);
            break;
        case TAbstractInfo::_3d:
            SaveData((plot::T3dDataOwner *)Data[i].Info, Data[i].attributes);
            break;
        case TAbstractInfo::_4d:
            SaveData((ccd::TSetOfSpectrum *)Data[i].Info, Data[i].attributes);
            break;
        case TAbstractInfo::_Image:
            SaveData((QImage *)Data[i].Info, Data[i].attributes);
            break;
        }

        if (DataOutput.isOpen())
            DataOutput.close();
        emit FileProccessed(NumberOfFiles, ++CurrentFileNumber / 2.);

        QuaZipFile PreviewOutput(&Archive);
        QString imgFileName = "";
        if (!oldFormat)
            imgFileName += "data";
        imgFileName += QString::number(i) + ".png";
        QuaZipNewInfo PreviewInfo(imgFileName);
        if (!PreviewOutput.open(QIODevice::WriteOnly, PreviewInfo)) //, 0, 0, Z_DEFLATED, Z_NO_COMPRESSION
            throw "Error saving preview!";

        QImage preview = Data[i].attributes.value("Preview").value<QImage>();
        if (preview.isNull()) {
            preview = QImage(50, 50, QImage::Format_ARGB32);
            preview.fill(Qt::white);
        }
        preview.save(&PreviewOutput, "png", 100);

        PreviewOutput.close();

        emit FileProccessed(NumberOfFiles, ++CurrentFileNumber / 2.);
    }
    emit FileProccessed(NumberOfFiles = -1, CurrentFileNumber = -1);
    Archive.close();
}

void TNstFileHandler::SaveData(plot::TCurvesDataOwner *Info, plot::TMetaStorage attributes)
{
    Stream << Label_2d << endl;

    QStringList attriblist;
    for (QHash<QString, QVariant>::const_iterator it = attributes.constBegin(); it != attributes.constEnd(); it++)
        if (it.value().canConvert<QString>()) {
            attriblist.append(it.key());
            attriblist.append(it.value().toString());
        }
    Stream << AttributesField << " " << attriblist.join("*_*|^_^") << endl;

    for (int i = 0; i < Info->size(); i++) {
        Stream << ItemBegin << endl;
        SaveItem(Info->at(i).data());
        Stream << ItemEnd << endl;
    }
}
void TNstFileHandler::SaveItem(const plot::TCurveData *Item)
{
    for (int i = 0; i < Item->size(); i++) {
        if (Item->isCompressedXCurve())
            Stream << (Item->XSeq.constData()[0] + i * Item->XSeq.constData()[1]) << " " << Item->YSeq.constData()[i];
        else
            Stream << Item->XSeq.constData()[i] << " " << Item->YSeq.constData()[i];
        Stream << endl;
    }
}
void TNstFileHandler::SaveData(plot::T3dDataOwner *Info, plot::TMetaStorage attributes)
{
    if (Info->size() < 1)
        return;

    Stream << Label_3d << endl;

    Info->at(0)->UpdateNxNy();
    attributes.insert("XPoints", Info->at(0)->NXElems);
    attributes.insert("YPoints", Info->at(0)->NYElems);

    bool rawbin = false;
    if (Info->at(0)->NXElems * Info->at(0)->NYElems > 1000 * 1000) {
        attributes.insert("RawBinData", true);
        rawbin = true;
    }

    QStringList attriblist;
    for (QHash<QString, QVariant>::const_iterator it = attributes.constBegin(); it != attributes.constEnd(); it++)
        if (it.value().canConvert<QString>()) {
            attriblist.append(it.key());
            attriblist.append(it.value().toString());
        }
    Stream << AttributesField << " " << attriblist.join("*_*|^_^") << endl;

    Stream << ItemBegin << endl;
    QElapsedTimer t; t.start();
    QDataStream bst;
    if (rawbin) {
        bst.setDevice(Stream.device());
        bst.setByteOrder(QDataStream::LittleEndian);
        bst << (quint32)Info->at(0)->NYElems;
    }
    for (int YCounter = 0; YCounter < Info->at(0)->Array.size(); YCounter++) {
        if (rawbin) {
            bst << (quint32)Info->at(0)->Array[YCounter].size();
            bst.writeRawData((const char *)Info->at(0)->Array[YCounter].constData(), Info->at(0)->Array[YCounter].size() * sizeof(double));
        } else
            for (int XCounter = 0; XCounter < Info->at(0)->Array[YCounter].size(); XCounter++) {
                Stream << XCounter << " " << YCounter << " " << Info->at(0)->Array[YCounter][XCounter];
                Stream << endl;
            }

        if (t.elapsed() > 100) {
            t.restart();
            emit FileProccessed(NumberOfFiles, CurrentFileNumber / 2. + double(YCounter) / Info->at(0)->Array.size());
        }
    }
}
void TNstFileHandler::SaveData(ccd::TSetOfSpectrum *Info, plot::TMetaStorage attributes)
{
    Stream << Label_4d << endl;

    QStringList attriblist;
    for (QHash<QString, QVariant>::const_iterator it = attributes.constBegin(); it != attributes.constEnd(); it++)
        if (it.value().canConvert<QString>()) {
            attriblist.append(it.key());
            attriblist.append(it.value().toString());
        }
    Stream << AttributesField << " " << attriblist.join("*_*|^_^") << endl;

    Stream << ItemBegin << endl;
    ccd::TSetOfSpectrumSaveOpen ccdSaveOpen;
    connect(&ccdSaveOpen, SIGNAL(ChangeProgress(double)), this, SLOT(progress4d(double)));
    ccdSaveOpen.Save(Stream.device(), Info).waitForFinished();
}
void TNstFileHandler::SaveData(QImage * /*Info*/, plot::TMetaStorage attributes)
{
    Stream << Label_Image << endl;

    QStringList attriblist;
    for (QHash<QString, QVariant>::const_iterator it = attributes.constBegin(); it != attributes.constEnd(); it++)
        if (it.value().canConvert<QString>()) {
            attriblist.append(it.key());
            attriblist.append(it.value().toString());
        }
    Stream << AttributesField << " " << attriblist.join("*_*|^_^") << endl;

    Stream << ItemBegin << endl;
    //QImageWriter writer(Stream.device(), "png");
}

static void Add2dInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj);
static void Add3dInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj);
static void Add4dInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj);
static void AddImageInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj);
static void LoadFile(QString filename, QPair<QString, QString> cfilename, TAbstractInfo *Data, int CurrentIndex, QObject *progressObj)
{
    try {
        if (cfilename.second.endsWith(".png", Qt::CaseInsensitive))
        {
            QuaZipFile file(filename, cfilename.second);
            file.open(QIODevice::ReadOnly);

            QImage img;
            if (!img.loadFromData(file.readAll()))
                img = QImage(":/container/question.png");

            Data->attributes.insert("Preview", QVariant::fromValue(img));
        }
        if (cfilename.first.endsWith(".lsdlsd", Qt::CaseInsensitive)
                || cfilename.first.endsWith(".txt", Qt::CaseInsensitive))
        {
            QuaZipFile file(filename, cfilename.first);
            file.open(QIODevice::ReadOnly);

            QTextStream Stream;
            Stream.setDevice(&file);
            Stream.setCodec("Windows-1251");

            QString Type; Stream >> Type;

            if (Type == Label_2d)
                Add2dInfo(Stream, &file, CurrentIndex, Data, progressObj);
            else if (Type == Label_3d) {
                QuaZipFile filebuf(filename, cfilename.first); //3d raw (re)seek iodevice
                filebuf.open(QIODevice::ReadOnly);

                //file.close();
                Add3dInfo(Stream, &filebuf, CurrentIndex, Data, progressObj);
            } else if (Type == Label_4d) {
                QuaZipFile filebuf(filename, cfilename.first); //4d raw (re)seek iodevice
                filebuf.open(QIODevice::ReadOnly);

                //file.close();
                Add4dInfo(Stream, &filebuf, CurrentIndex, Data, progressObj);
            } else if (Type == Label_Image) {
                QuaZipFile filebuf(filename, cfilename.first); //image raw (re)seek iodevice
                filebuf.open(QIODevice::ReadOnly);

                //file.close();
                AddImageInfo(Stream, &filebuf, CurrentIndex, Data, progressObj);
            }
        }
    } catch (...) {
        qDebug() << "nst load fail" << filename << cfilename;
    }
}

void TNstFileHandler::Load(const QString &FileName, TAbstractInfos &Data)
{
    static bool useMultithreading = true;
    try {
        QuaZip Archive(FileName);
        if (!Archive.open(QuaZip::mdUnzip)) throw "Error opening file!";

        NumberOfFiles = CalcNumOfInfos(&Archive);
        emit FileProccessed(NumberOfFiles, 0);
        emit DoAddBlanks(NumberOfFiles);

        Data.resize(NumberOfFiles);
        loadProgress.resize(NumberOfFiles);

        QVector<QString> datalist(NumberOfFiles);
        QVector<QString> previewlist(NumberOfFiles);
        for (bool More = Archive.goToFirstFile(); More; More = Archive.goToNextFile())
        {
            const QString Name = Archive.getCurrentFileName();

            QString Index = Name.right(Name.size() - (Name.contains("data") ? 4:0));
            Index.truncate(Index.indexOf('.'));
            int CurrentIndex = Index.toLong();
            if (CurrentIndex >= NumberOfFiles) {
                qDebug() << "NSTDAT nameorder bug" << FileName << Name;
                continue;
            }

            if (Name.endsWith(".png", Qt::CaseInsensitive))
                previewlist[CurrentIndex] = Name;
            if (Name.endsWith(".lsdlsd", Qt::CaseInsensitive)
                    || Name.endsWith(".txt", Qt::CaseInsensitive))
                datalist[CurrentIndex] = Name;
        }
        Archive.close();

        QVector<QFuture<void> > files;
        for (int i = 0; i < NumberOfFiles; i++) {
            if (useMultithreading && QThread::idealThreadCount() > 1) //use threaded load
                files.append(QtConcurrent::run(LoadFile, FileName, QPair<QString, QString>(datalist.at(i), previewlist.at(i)), &Data[i], i, this));
            else
                LoadFile(FileName, QPair<QString, QString>(datalist.at(i), previewlist.at(i)), &Data[i], i, this);
        }
        foreach (QFuture<void> f, files)
            f.waitForFinished();

        loadProgress.clear();
        emit FileProccessed(-1, -1);
    } catch (char *str) {
        qDebug() << "nst load exception: " << str;
        emit Error(tr("Open error"), str);
    } catch (...) {
        qDebug() << "nst load exception";
        emit Error(tr("Open error"), tr("File corrupted."));
    }

    emit AllDataLoaded(&Data);
}
int TNstFileHandler::CalcNumOfInfos(QuaZip *Archive)
{
    int NumberOfInfos = 0;

    for (bool More = Archive->goToFirstFile(); More; More = Archive->goToNextFile())
        ++NumberOfInfos;

    Archive->goToFirstFile();
    return NumberOfInfos / 2;
}

static void AddItem(QTextStream &Stream, plot::TCurvesDataOwner *Src);
void Add2dInfo(QTextStream &Stream, QuaZipFile * /*filebuf*/, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj)
{
    plot::TCurvesDataOwner *Info = new plot::TCurvesDataOwner;
    Data->Info = Info;
    Data->InfoType = TAbstractInfo::_2d;

    QString Str;
    for (;;) {
        Str = Stream.readLine();
        if (Str.startsWith(ItemBegin))
            break;

        if (Str.startsWith(AttributesField)) {
            QStringList serial = Str.remove(AttributesField + " ").split("*_*|^_^");
            for (int i = 0; i < serial.size()-1; i += 2) {
                //TODO convert to source type (int, double, datetime, string)
                Data->attributes.insert(serial.at(i), serial.at(i+1));
            }
        }
    }

    for (; !(Stream.device()->atEnd() && Stream.atEnd()); ) {
        if (Str.startsWith(ItemBegin))
            AddItem(Stream, Info);
        Str = Stream.readLine();
    }
    QMetaObject::invokeMethod(progressObj, "setLoadProgress", Qt::DirectConnection, Q_ARG(int, CurrentIndex), Q_ARG(double, 1.));
}

void AddItem(QTextStream &Stream, plot::TCurvesDataOwner *Src)
{
    plot::ResizeOwner(*Src, Src->size() + 1);

    while( !(Stream.device()->atEnd() && Stream.atEnd()) )
    {
        QString Str = Stream.readLine();
        if (Str.startsWith(ItemEnd))
            break;
        else {
            QStringList lst = Str.split(" ");
            if (lst.size() != 2)
                continue;

            double x, y;
            bool ok;
            x = lst.first().toDouble(&ok);
            if (!ok)
                continue;
            y = lst.last().toDouble(&ok);
            if (!ok)
                continue;

            Src->last()->AddPoint(x, y);
        }
    }
    auto xsec = Src->last()->XSeq;
    if (xsec.size() == 0)
        return;
    double dx0 = xsec.at(1) - xsec.first(); //TODO check each dx
    double dxall = (xsec.last() - xsec.first()) / xsec.size();
    if (qFuzzyCompare(dx0, dxall)) {
        //qDebug() << "2d compressed x " << dx0 << dxall;
        //Src->last()->setXCalib(Src->last()->XSeq.first(), dxall); //TODO fix ccd/container X recalc, autocalib, etc
    }
}
void Add3dInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj)
{
    plot::T3dDataOwner *Info = new plot::T3dDataOwner;
    Data->Info = Info;
    Data->InfoType = TAbstractInfo::_3d;

    plot::ResizeOwner(*Info, 1);

    QString Str;
    bool utf = false;
    int len = 2;//type len
    if (utf = Stream.codec()->name().contains("UTF-8"))
        len += 3; //utf-8 bom
    for (;;)
    {
        Str = Stream.readLine();
        len += (utf ? Str.toUtf8().length() : Str.length()) + 1; // 1 = srtlen("\n")
        if (Str.startsWith(ItemBegin))
            break;

        if (Str.startsWith(AttributesField)) {
            QStringList serial = Str.remove(AttributesField + " ").split("*_*|^_^");
            for (int i = 0; i < serial.size()-1; i += 2) {
                //TODO convert to source type (int, double, datetime, string)
                Data->attributes.insert(serial.at(i), serial.at(i+1));
            }
        }
    }

    if (Data->attributes.contains("XPoints") && Data->attributes.contains("YPoints")) {
        Info->at(0)->setSize(Data->attributes.value("XPoints").toInt(), Data->attributes.value("YPoints").toInt());
        Data->attributes.remove("XPoints");
        Data->attributes.remove("YPoints");
    }
    bool rawbin = false;
    if (Data->attributes.contains("RawBinData")) {
        Data->attributes.remove("RawBinData");
        rawbin = true;
    }

    int readDataSize = filebuf->size();

    QElapsedTimer t; t.start();
    int c = 0;
    Stream.setIntegerBase(10);
    if (rawbin) {
        filebuf->read(len);
        QDataStream bst(filebuf);
        bst.setByteOrder(QDataStream::LittleEndian);
        quint32 nx, ny;
        bst >> ny;

        for (int i = 0; i < (int)ny && !(filebuf->atEnd() && bst.atEnd()); i++) {
            bst >> nx;
            Info->at(0)->AddPoint(nx-1, i, 0);
            bst.readRawData((char *)Info->at(0)->Array[i].data(), nx * sizeof(double));

            if ((c++ % 10) == 0 && t.elapsed() > 100) {
                t.restart();
                QMetaObject::invokeMethod(progressObj, "setLoadProgress", Qt::DirectConnection, Q_ARG(int, CurrentIndex), Q_ARG(double, qMin(1., double(i) / ny)));
            }
        }
    } else
    while (!(Stream.device()->atEnd() && Stream.atEnd())) {
        quint32 x, y; double z;
        Stream >> x >> y >> z;
        if (Stream.status() != QTextStream::Ok) continue;
        Info->at(0)->AddPoint(x, y, z);

        if ((c++ % 10000) == 0 && t.elapsed() > 100 && readDataSize > 0) {
            t.restart();
            QMetaObject::invokeMethod(progressObj, "setLoadProgress", Qt::DirectConnection, Q_ARG(int, CurrentIndex), Q_ARG(double, qMin(1., double(Stream.device()->pos()) / readDataSize)));
        }
    }

    Info->at(0)->UpdateNxNy();
    Info->at(0)->UpdateDxDy();
    Info->at(0)->resetMathLims();
}

class T4dProgressWrapper : public QObject
{
    Q_OBJECT
public:
    int idx = 0;
public slots:
    void set4dprogress(double val) {
        emit progress(idx, val);
    }
signals:
    void progress(int, double);
};
#include <cont_nstfile.moc>

void Add4dInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj)
{
    ccd::TSetOfSpectrum *Info = new ccd::TSetOfSpectrum;
    Data->Info = Info;
    Data->InfoType = TAbstractInfo::_4d;

    QString Str;
    bool utf = false;
    int len = 2;//type len
    if (utf = Stream.codec()->name().contains("UTF-8"))
        len += 3; //utf-8 bom
    for (;;)
    {
        Str = Stream.readLine();
        len += (utf ? Str.toUtf8().length() : Str.length()) + 1; // 1 = srtlen("\n")
        if (Str.startsWith(ItemBegin))
            break;

        if (Str.startsWith(AttributesField)) {
            QStringList serial = Str.remove(AttributesField + " ").split("*_*|^_^");
            for (int i = 0; i < serial.size()-1; i += 2) {
                //TODO convert to source type (int, double, datetime, string)
                Data->attributes.insert(serial.at(i), serial.at(i+1));
            }
        }
    }

    if (!Data->attributes.value("YUnit").toString().isNull())
        Info->YscaleUnit = Data->attributes.value("YUnit").toString();

    filebuf->read(len);
    ccd::TSetOfSpectrumSaveOpen ccdSaveOpen;
    T4dProgressWrapper wrap4d; wrap4d.idx = CurrentIndex;
    QObject::connect(&ccdSaveOpen, SIGNAL(ChangeProgress(double)), &wrap4d, SLOT(set4dprogress(double)), Qt::DirectConnection);
    QObject::connect(&wrap4d, SIGNAL(progress(int,double)), progressObj, SLOT(setLoadProgress(int,double)), Qt::DirectConnection);
    ccdSaveOpen.Open(filebuf, Info).waitForFinished();
}

void AddImageInfo(QTextStream &Stream, QuaZipFile *filebuf, int CurrentIndex, TAbstractInfo *Data, QObject *progressObj)
{
    //QImage *Info = new QImage;
    //Data->Info = Info;
    Data->Info = nullptr;
    Data->InfoType = TAbstractInfo::_Image;

    QString Str;
    bool utf = false;
    int len = 2;//type len
    if (utf = Stream.codec()->name().contains("UTF-8"))
        len += 3; //utf-8 bom
    for (;;)
    {
        Str = Stream.readLine();
        len += (utf ? Str.toUtf8().length() : Str.length()) + 1; // 1 = srtlen("\n")
        if (Str.startsWith(ItemBegin))
            break;

        if (Str.startsWith(AttributesField)) {
            QStringList serial = Str.remove(AttributesField + " ").split("*_*|^_^");
            for (int i = 0; i < serial.size()-1; i += 2) {
                //TODO convert to source type (int, double, datetime, string)
                Data->attributes.insert(serial.at(i), serial.at(i+1));
            }
        }
    }

    filebuf->read(len);
    //QImageReader reader(filebuf, "png");

    QMetaObject::invokeMethod(progressObj, "setLoadProgress", Qt::DirectConnection, Q_ARG(int, CurrentIndex), Q_ARG(double, 1.));
}
