#include "container.h"
#include "datalist.hpp"
#include "gui/t3dcontrol.hpp"
#include "gui/viewer3d.hpp"
#include "gui/status.hpp"

#ifndef _ADROID
#include "modules.hpp"
#include "tools.hpp"
#endif

#include "dataplot.h"
#include "ccd/ccd.h"
#include "ccd/ccdmatrix.h"
#include "gwynetio.hpp"
#include "spcfile.h"
#include "ascifile.h"
#include <QFileDialog>
#include <QVBoxLayout>
#include <QSplitter>
#include <QTimer>
#include <QApplication>
#include "gui/flowlayout.hpp"
#include "options/options.h"
#include "3d/plot_3dplot.h"
#include "widgets/threedoverlaybuttons.h"
#include "widgets/twodoverlaybuttons.h"
#include "options_window.h"
#include "container/cont_nstfile.h"
#include "cont_filesproccessing.h"

#ifdef _ADROID
#include <QThread>
#include <QtGlobal>
class SleepExport : public QThread
{
public:
    static void msleep2(unsigned long ms) { msleep(ms); }
};
#endif

void TContainer::btnToGwyddion_clicked()
{
    #ifndef _ADROID
    if (!CurrentDataList)
        return;

    GwyddionExport *exp = new GwyddionExport();
    connect(exp, SIGNAL(no_gwyddion()), this, SLOT(gwyddion_error()));

    if (CurrentDataList->ClickedItem)
        exp->AddData(CurrentDataList->ClickedItem->Info, CurrentDataList->ClickedItem->InfoType, CurrentDataList->ClickedItem->attributes);

    exp->accept_and_send();
    exp->start();
#endif
}
void TContainer::gwyddion_error()
{
    #ifndef _ADROID
    new Notification(tr("Can't start Gwyddion."), Notification::Bad, 10);
    doBeep();
#endif
}

void TContainer::btnToDevice_clicked()
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    emit restoreAttributes(CurrentDataList->ClickedItem->attributes);
}

void TContainer::btnToOldNspec_clicked()
{
    SaveAs(true);
}

void TContainer::btnToAscii_clicked()
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    QString exr_str, FileName;
    if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_2d)
        FileName = QFileDialog::getSaveFileName(this, tr("Export"), FileName, "CSV (*.csv);;SPC (*.spc)",
                                                &exr_str, QFileDialog::DontUseNativeDialog);
    else
        FileName = QFileDialog::getSaveFileName(this, tr("Export"), FileName, "CSV (*.csv)",
                                                &exr_str, QFileDialog::DontUseNativeDialog);

    if (FileName.isEmpty())
        return;

    if (exr_str == "CSV (*.csv)" && FileName.endsWith(".csv", Qt::CaseInsensitive)) FileName.chop(4);
    if (exr_str == "SPC (*.spc)" && FileName.endsWith(".spc", Qt::CaseInsensitive)) FileName.chop(4);

    if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_2d) {
        int ActiveCurve = Plot->TwoDPlot->Pure2dPlot->ActiveCurve;
        QString suffix;
        plot::TCurvesDataOwner Elements_export = plot::getLock(Plot->TwoDPlot->Pure2dPlot->SourceData);

        for (int counter = 0; counter < Elements_export.size(); counter++) {
            auto elements = Elements_export.at(counter);
            if (!elements || !elements->size())
                    continue;

            if (ActiveCurve == -1 || ActiveCurve == counter) {
                if (ActiveCurve == -1 && Elements_export.size() != 1)
                    suffix = QString::number(counter+1);

                if (exr_str == "CSV (*.csv)") {
                    ascifile ascif (FileName+suffix+".csv");
                    ascif.set_instrument("Data CSV export");
                    ascif.set_data_name(CurrentDataList->ClickedItem->attributes.value("Name", tr("2d curve")).toString());
                    Plot->TwoDPlot->Pure2dPlot->XAxisUnit.StrMultiplier.clear();
                    Plot->TwoDPlot->Pure2dPlot->XAxisUnit.UpdateLegend();
                    Plot->TwoDPlot->Pure2dPlot->YAxisUnit.StrMultiplier.clear();
                    Plot->TwoDPlot->Pure2dPlot->YAxisUnit.UpdateLegend();
                    ascif.set_x_type(Plot->TwoDPlot->Pure2dPlot->XAxisUnit.Legend);
                    ascif.set_y_type(Plot->TwoDPlot->Pure2dPlot->YAxisUnit.Legend);
                    ascif.set_xpts(elements->size());

                    QVector<double> xseq = elements->XSeq;
                    if (elements->isCompressedXCurve()) {
                        xseq.resize(elements->size());
                        for (int i = 0; i < xseq.size(); i++)
                            xseq[i] = elements->getXat(i);
                    }
                    ascif.set_x_data(xseq.constData());
                    ascif.set_y_data(elements->YSeq.constData());
                    ascif.write_file();
                } else {
                    QVector<float> xseq(elements->size());
                    QVector<float> yseq(elements->size());

                    spcfile spc(FileName+suffix+".spc");
                    spc.set_npts(elements->size());
                    spc.set_y_type(YCOUNT);
                    spc.set_exper(SPCRMN);

                    int Scale = FindCheckedItemIn(AG_Scale_select);
                    if (!Scale_select->isVisibleTo(this)) {
                        if (Plot->TwoDPlot->Pure2dPlot->XAxisUnit.Description == QObject::tr("pixels"))
                            Scale = 0;
                        if (Plot->TwoDPlot->Pure2dPlot->XAxisUnit.Description == QObject::tr("WL, nm") ||
                            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.Description == QObject::tr("WL"))
                            Scale = 1;
                        if (Plot->TwoDPlot->Pure2dPlot->XAxisUnit.Description == QObject::tr("Raman Shift, 1/cm"))
                            Scale = 2;
                        if (Scale == -1) {
                            qDebug() << "spc export not supported x axis scale";
                            continue;
                        }
                    }

                    switch (Scale) {
                    case 0:
                        spc.set_x_type(XPOINT);
                        break;
                    case 1:
                        spc.set_x_type(XMETERS);
                        break;
                    case 2:
                        spc.set_x_type(XRAMANS);
                        break;
                    }
                    for (int i = 0; i < elements->size(); i++) {
                        xseq[i] = elements->getXat(i);
                        yseq[i] = elements->YSeq.at(i);
                    }

                    spc.set_x_data(xseq.constData());
                    spc.set_y_data(yseq.constData());
                    spc.write_file();
                }
            }
        }
    } else if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_Image) {
        CurrentDataList->ClickedItem->attributes.value("Preview").value<QImage>().save(FileName+".png");
    } else if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_3d) {
        plot::T3dDataOwner *info = (plot::T3dDataOwner *)CurrentDataList->ClickedItem->Info;

        ascifile ascif (FileName+".csv");
        ascif.set_instrument("Data CSV export");
        ascif.set_data_name(CurrentDataList->ClickedItem->attributes.value("Name", tr("Scan")).toString());
        ascif.set_x_type(CurrentDataList->ClickedItem->attributes.value("XYUnit", tr("m")).toString());
        ascif.set_y_type(CurrentDataList->ClickedItem->attributes.value("ZUnit", tr("Counts")).toString());
		auto firstInfo = info->at(0);
        ascif.set_xsz(firstInfo->XLimits);
        ascif.set_ysz(firstInfo->YLimits);
        ascif.set_xpts(firstInfo->NXElems);
        ascif.set_ypts(firstInfo->NYElems);

        ascif.set_array(firstInfo->Array);

        ascif.write_file_3d();
    } else if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_4d) {
        ccd::TSetOfSpectrum *info = (ccd::TSetOfSpectrum *)CurrentDataList->ClickedItem->Info;

        QDir::current().mkpath(FileName + "/");

        {
            ascifile ascif (FileName+"/3d.csv");
            ascif.set_instrument("Data CSV export");
            ascif.set_data_name(CurrentDataList->ClickedItem->attributes.value("Name", tr("Scan")).toString());
            ascif.set_x_type(CurrentDataList->ClickedItem->attributes.value("XYUnit", tr("m")).toString());
            ascif.set_y_type(CurrentDataList->ClickedItem->attributes.value("ZUnit", tr("Counts")).toString());
            ascif.set_xsz(info->XLimits);
            ascif.set_ysz(info->YLimits);
            ascif.set_xpts(info->NXElems);
            ascif.set_ypts(info->NYElems);

            ascif.set_array(info->Results.at(0)->Array);

            ascif.write_file_3d();
        }

        if (info->CalibInfo.size() >= 6) {
            QFile dataFile(FileName+"/spectra_calibrations.txt");
            dataFile.open(QIODevice::WriteOnly | QIODevice::Text);
            QTextStream outData(&dataFile);

            outData << "# Nano Scan Technologies Ltd." << endl;
            outData << "# CSV Data" << endl << endl;
            outData << "# 4d Calibration Data" << endl << endl;
            outData << "# Laser Wavelength, nm = " << info->CalibInfo[0] << endl;
            outData << "# Central Wavelength, nm = " << info->CalibInfo[1] << endl;
            outData << "# Dispersion, nm/mm = " << info->CalibInfo[2] << endl;
            outData << "# pixelsize, mm = " << info->CalibInfo[3] << endl;
            outData << "# Nx = " << info->CalibInfo[4] << endl;
            outData << "# Central Pixel = " << info->CalibInfo[5] << endl;

            dataFile.close();
        }

        QDir::current().mkpath(FileName + "/data/");
        for (int i = 0; i < (int)info->NYElems && i < info->Array.size(); i++) {
            for (int j = 0; j < (int)info->NXElems && j < info->Array[i].size(); j++) {
                QString suffix = "/data/spectra_at_" + QString::number(i) + "_" + QString::number(j);
                ccd::TSpectrum &sp = info->Array[i][j];


                ascifile ascif (FileName+suffix+".csv");
                ascif.set_instrument("Data CSV export");
                ascif.set_data_name(suffix);
                Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->XAxisUnit.StrMultiplier.clear();
                Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->XAxisUnit.UpdateLegend();
                Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.StrMultiplier.clear();
                Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.UpdateLegend();
                ascif.set_x_type(Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->XAxisUnit.Legend);
                ascif.set_y_type(tr("Counts"));
                ascif.set_xpts(sp.size());

                QVector<double> xseq(sp.size());
                int Scale = FindCheckedItemIn(AG_Scale_select);
                for (int i = 0; i < sp.size(); i++) {
                    double val = i + info->MinForRecord;
                    if (info->CalibInfo.size() == 6) {
                        if (Scale == 1) {
                            val = 1e-9*wavelength(info->CalibInfo[1],
                                    info->CalibInfo[2], info->CalibInfo[3],
                                    val, (int)info->CalibInfo[5]);
                        } else if (Scale == 2) {
                            val = raman_shift(info->CalibInfo[0], wavelength(info->CalibInfo[1],
                                        info->CalibInfo[2], info->CalibInfo[3],
                                        val, (int)info->CalibInfo[5]));
                        }
                    }
                    xseq[i] = val;
                }

                ascif.set_x_data(&xseq[0]);
                ascif.set_y_data(&sp[0]);
                ascif.write_file();
            }
        }
    }
}



void TContainer::CrossEvent(QPointF pt)
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem || CurrentDataList->ClickedItem->InfoType != TAbstractInfo::_4d)
        return;

    ccd::TSetOfSpectrum *Info = (ccd::TSetOfSpectrum *)CurrentDataList->ClickedItem->Info;

    int XInd, YInd;
    Info->GetIntegerIndexes(pt.x(), pt.y(), XInd, YInd);

    T4dDataHelper::ShowSpectr(Plot->ThreeDPlot->TwoDPlot,
                              Plot->ThreeDPlot->TwoDPlotVertical,
                              Info, lastXind = XInd, lastYind = YInd, FindCheckedItemIn(AG_Scale_select));

    Plot->ThreeDPlot->PlotControl->BMarkers->setChecked(false);
    Plot->ThreeDPlot->Set2dPlotVisible(true);
}

bool TContainer::SelectEvent(QRectF SelectedRect, int State, int Origin)
{
    if (!ActiveHBL)
        return false;

    if (!CurrentDataList || !CurrentDataList->ClickedItem || CurrentDataList->ClickedItem->InfoType != TAbstractInfo::_4d)
        return false;

    ccd::TSetOfSpectrum *Info = (ccd::TSetOfSpectrum *)CurrentDataList->ClickedItem->Info;

    double Min = SelectedRect.left(), Max = SelectedRect.right();

    if (State == plot::TSelectEvent::St_SelectionStarted) {
        plot::TCurvesDataOwner Elements_temp = plot::getLock(Plot->TwoDPlot->Pure2dPlot->SourceData);
        if (Elements_temp.size() > 0 && Elements_temp.at(0) && Elements_temp.at(0)->XSeq.size() != 0) {
            Min = Elements_temp.at(0)->XSeq.first();
            Max = Elements_temp.at(0)->XSeq.last();

            Min = (Min+0.25*(Max-Min));
            Max = (Min+0.75*(Max-Min));
        }
        Min = (func_min ? T4dDataHelper::ReCalibrate(func_min, Info->CalibInfo, FindCheckedItemIn(AG_Scale_select)) : Min);
        Max = (func_max ? T4dDataHelper::ReCalibrate(func_max, Info->CalibInfo, FindCheckedItemIn(AG_Scale_select)) : Max);
    }

    if (Origin == plot::TSelectEvent::SRO_Mouse || Origin == plot::TSelectEvent::SRO_Unknown)
    {
        T4dDataHelper::SetSelectGeometry(Plot->ThreeDPlot->TwoDPlot, Min, Max);
        T4dDataHelper::SetSelectGeometry(Plot->ThreeDPlot->TwoDPlotVertical, Min, Max);
        T4dDataHelper::ReCalcRamanOrWavelengthToPixels(Info->CalibInfo, Min, Max, FindCheckedItemIn(AG_Scale_select));
        func_min = Min;
        func_max = Max;

        Info->SetMinMaxForFuncApply(Min, Max);
        Info->CalcResults();

        Plot->ThreeDPlot->ThreeDPlots[0]->Refresh();
    }

    if (State == plot::TSelectEvent::St_SelectionApplied ||
            State == plot::TSelectEvent::St_SelectionCancled) //TODO restore prev conv.range on cancel
    {
        Func_interval->change_state_man_with_sig(false);
        BFuncInterval->change_state_man(false);
    }

    return true;
}

void TContainer::sl_Func_range_select(bool st)
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    if (st) {
        Plot->ThreeDPlot->PlotControl->BCrossection->setChecked(false);
        Plot->ThreeDPlot->PlotControl->BMarkers->setChecked(false);
        Plot->ThreeDPlot->PlotControl->BMarkers->setEnabled(false);

        if (Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->Title == QObject::tr("section")) {
            ccd::TSetOfSpectrum *Info = (ccd::TSetOfSpectrum *)CurrentDataList->ClickedItem->Info;

            #ifndef _ADROID
            T4dDataHelper::ShowSpectr(Plot->ThreeDPlot->TwoDPlot, Plot->ThreeDPlot->TwoDPlotVertical, Info, lastXind, lastYind, FindCheckedItemIn(AG_Scale_select));
#endif
            Plot->ThreeDPlot->Set2dPlotVisible(true);
            Plot->Refresh();
        }
    } else
        Plot->ThreeDPlot->PlotControl->BMarkers->setEnabled(true);

    ActiveHBL = st;
    Plot->ThreeDPlot->TwoDPlot->SetSelectionModeEnabled(st);
    Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->SetZoomBySelection(!st);
    Plot->ThreeDPlot->TwoDPlotVertical->SetSelectionModeEnabled(st);
    Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->SetZoomBySelection(!st);
}
void TContainer::onAG_Scale_select_trigered()
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    int Scale = FindCheckedItemIn(AG_Scale_select);
    AGScaleSelect->actions()[Scale]->setChecked(true);
    if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_4d) {
        ccd::TSetOfSpectrum *Info = (ccd::TSetOfSpectrum *)CurrentDataList->ClickedItem->Info;

        switch (Scale) {
        case 0:
            Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
            Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
            break;
        case 1:
            Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
            Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
            break;
        case 2:
            Plot->ThreeDPlot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
            Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
            break;
        }

        Info->XscaleUnit = FindCheckedItemIn(AG_Scale_select);

        Plot->Refresh();

        T4dDataHelper::ShowSpectr(Plot->ThreeDPlot->TwoDPlot, Plot->ThreeDPlot->TwoDPlotVertical, Info, lastXind, lastYind, Scale);
    } else if (CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_2d)  {
        switch (Scale) {
        case 0:
            CurrentDataList->ClickedItem->attributes.insert("XLabel", QObject::tr("pixels"));
            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
            Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
            break;
        case 1:
            CurrentDataList->ClickedItem->attributes.insert("XLabel", QObject::tr("WL"));
            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
            Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
            break;
        case 2:
            CurrentDataList->ClickedItem->attributes.insert("XLabel", QObject::tr("Raman Shift, 1/cm"));
            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
            Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
            break;
        }

        for (int i = 0; i < Plot->TwoDPlot->Pure2dPlot->SourceData.size(); i++)
            set_spectra_calib(*Plot->TwoDPlot->Pure2dPlot->SourceData[i].data(), CurrentDataList->ClickedItem->attributes, Scale);

        #ifdef _ADROID
        Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->SourceData = *Plot->TwoDPlot->Pure2dPlot->cloneSourceData();
#endif
        Plot->Refresh();
    }
}

void TContainer::onAG_ConvFunc_trigered()
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem || CurrentDataList->ClickedItem->InfoType != TAbstractInfo::_4d)
        return;

    ccd::TSetOfSpectrum *Info = (ccd::TSetOfSpectrum *)CurrentDataList->ClickedItem->Info;

    int idx = FindCheckedItemIn(AG_Convol_func);
    Plot->ThreeDPlot->SetZDescAndSIUnit(QString(),
                                        T4dDataHelper::getConvZUnits(idx, FindCheckedItemIn(AG_Scale_select), Info->YscaleUnit), 0);

    Info->CurrentFunc = idx;

    Info->CalcResults();
    onItemClicked(Info);
}



QWidget *TContainer::create3dPlot(QWidget *parent)
{
    QWidget *Wid3d = new QWidget(parent);

    #ifndef _ADROID
    Wid3d->setStyleSheet("QWidget {border: 1px solid #B4B4B4; border-style: outset; border-width: 2px; border-radius: 10px; border-color: #B4B4B4; background-color: #E3E3EC;}");
    if (parent && qobject_cast<QBoxLayout *>(parent->layout()))
        qobject_cast<QBoxLayout *>(parent->layout())->insertWidget(0, Wid3d, 1);

    QHBoxLayout *pl_l = new QHBoxLayout(Wid3d);
    pl_l->setSpacing(0);
    pl_l->setContentsMargins(10, 10, 10, 10);


    pl3 = new T3dViewer(Wid3d);
    pl3->invertY = true;
    pl3->showAxes = true;
    pl3->setMinimumWidth(400);

    T3DControl *pl3Control = new T3DControl(pl3, Wid3d);

    pl_l->addWidget(pl3, 1);
    pl_l->addWidget(pl3Control);

    connect(pl3Control, SIGNAL(exit(bool)), this, SLOT(btn3d_toggled(bool)));
    #endif

    return Wid3d;
}
T3dEnchancement::T3dEnchancement(QObject *parent) : QObject(parent)
{
    #ifndef _ADROID
    pl3 = nullptr;
#endif
    sourceplot = nullptr;
    upd = nullptr;
}
bool T3dEnchancement::eventFilter(QObject *, QEvent *e)
{
    if (e->type() == QEvent::DragEnter)
        e->accept();
    if (e->type() == QEvent::Drop) {
        plot::T3dPlot plot;
        plot.hide();
        plot.ThreeDPlots[0]->dropEvent((QDropEvent *)e);

        plot::T3dPlot *prev = sourceplot;
        sourceplot = &plot;
        updateTexture();
        sourceplot = prev;

        e->accept();
        return true;
    }
    return false;
}

QImage TContainer::get3dTexture(plot::T3dDataOwner Info, double zmin, double zmax, QwtLinearColorMap *colorMap)
{
	auto firstInfo = Info.at(0);
    if (!firstInfo)
        return QImage();

    int nx = firstInfo->NXElems;
    int ny = firstInfo->NYElems;
    int xstride = 1, ystride = 1;
    while (nx / xstride > 1500)
        xstride++;
    while (ny / ystride > 1500)
        ystride++;
    nx /= xstride;
    ny /= ystride;

    QImage texture(nx, ny, QImage::Format_ARGB32);
    for (int i = 0; i < ny; i++) {
        QRgb *p = (QRgb *)texture.scanLine(i);
        for (int j = 0; j < nx; j++)
            *(p++) = colorMap->rgb(QwtInterval(zmin, zmax), firstInfo->Array[i*ystride][j*xstride]);
    }

    return texture;
}
void T3dEnchancement::queueUpdate()
{
    #ifndef _ADROID
    if (!pl3)
        return;
#endif

    if (!sourceplot)
        return;

    if (!upd) {
        upd = new QTimer(this);
        upd->setSingleShot(true);
        upd->setInterval(50);
        connect(upd, SIGNAL(timeout()), this, SLOT(updateTexture()));
    }

    if (!upd->isActive())
        upd->start();
}
void T3dEnchancement::updateTexture()
{
    #ifndef _ADROID
    if (!pl3)
        return;
#endif

    if (!sourceplot)
        return;

	auto plot = sourceplot->ThreeDPlots.at(0);
    plot::T3dDataOwner Info = plot::getLock(plot->PaintData);
    double min = plot->PlotsZLimits[0];
    double max = plot->PlotsZLimits[1];

    #ifndef _ADROID
    pl3->updateTexture(TContainer::get3dTexture(Info, min, max, plot->colormap));
    pl3->updateGL();
#endif
}

void TContainer::on3dUpdate()
{
    #ifndef _ADROID
    if (BShow3DView->isChecked()) {
        parentWidget()->setEnabled(false);
        while (!Plot->ThreeDPlot->ThreeDPlots.at(0)->isMathReady()) {
            qApp->processEvents();
            SleepExport::msleep2(3);
        }
        parentWidget()->setEnabled(true);

        btn3d_toggled(true);
    }
#endif
}
void TContainer::btn3d_toggled(bool checked)
{
    #ifndef _ADROID

    BShow3DView->setChecked(checked);
    foreach (QObject *obj, Plot->ThreeDPlot->PlotControl->children()) {
        QWidget *wid = qobject_cast<QWidget *>(obj);
        if (wid && wid->isVisible() &&
            wid != Plot->ThreeDPlot->PlotControl->TBMathProcessing &&
            wid != Plot->ThreeDPlot->PlotControl->TBMath2 &&
            wid != Plot->ThreeDPlot->PlotControl->TBPalettes)
            wid->setEnabled(!checked);
    }

    if (!checked) {
        BShow3DView->setIcon(makeHugeIcon(":/container/3dimage.png"));

        if (pl3) pl3->parentWidget()->hide();
        Plot->show();
        Plot->ThreeDPlot->ResizeSliders();

        float data[4];
        if (pl3) pl3->updateData(data, 2, 2);
        if (pl3) pl3->updateTexture(QImage());
        t3dench->pl3 = nullptr;
        return;
    }

    if (!CurrentDataList->ClickedItem ||
            CurrentDataList->ClickedItem->InfoType == TAbstractInfo::Blank ||
            CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_2d ||
            CurrentDataList->ClickedItem->InfoType == TAbstractInfo::_Image)
        return;

    auto plot = Plot->ThreeDPlot->ThreeDPlots.at(0);
    plot::T3dDataOwner Info = plot::getLock(plot->PaintData);
    auto firstInfo = Info.at(0);
    double min = plot->PlotsZLimits[0];
    double max = plot->PlotsZLimits[1];
    int nx = firstInfo->NXElems;
    int ny = firstInfo->NYElems;
    int xstride = 1, ystride = 1;
    while (nx / xstride > 1500)
        xstride++;
    while (ny / ystride > 1500)
        ystride++;
    nx /= xstride;
    ny /= ystride;

    float *data = nullptr;
try {
    data = new float[nx * ny];

    for (int i = 0; i < ny; i++)
        for (int j = 0; j < firstInfo->Array[i].size() / xstride; j++)
            data[j*ny + i] = (firstInfo->Array[i*ystride][j*xstride] - min) / (max - min) - 0.5;

    if (!pl3) {
        auto data = DataPlot::GetIt();
        create3dPlot(data->Splitter->widget(1));
    }
    if (!t3dench->pl3) //do not overwrite on detach
        t3dench->pl3 = pl3;

    pl3->updateData(data, nx, ny,
                    (firstInfo->XLimits[1] - firstInfo->XLimits[0]) * xstride /
                    (firstInfo->YLimits[1] - firstInfo->YLimits[0]) / ystride);
    pl3->updateTexture(get3dTexture(Info, min, max, plot->colormap));

} catch(...) {
    qDebug() << "3d oom";
    float data[4];
    if (pl3) pl3->updateData(data, 2, 2);
    if (pl3) pl3->updateTexture(QImage());
}

    delete [] data;

    Plot->hide();
    Plot->ThreeDPlot->Slider->setMaximumHeight(100000);
    pl3->parentWidget()->show();
    BShow3DView->setIcon(makeHugeIcon(":/container/3dimage_on.png"));

    if (new3d) {
        pl3->x_rot = 30;
        pl3->y_rot = 0;
        pl3->z_rot = 10;
        pl3->x_pos = pl3->y_pos = pl3->z_pos = 0;
        pl3->z_scale = 0.2;
        pl3->scale = 1;
    }
    new3d = false;

    /*
    double scale = 1/plot->XYAxisUnit.Multiplier;
    double zscale = 1/plot->ZAxisUnit.Multiplier;
    pl3->coordinates()->axes[2].setLimits(min * zscale, max * zscale);

    pl3->coordinates()->axes[i].setLabelString(plot->XYAxisUnit.Legend);
    pl3->coordinates()->axes[2].setLabelString(plot->ZAxisUnit.Legend);

    QFont font = plot->QPlot->axisFont(QwtPlot::xBottom);
    pl3->coordinates()->setLabelFont(font);
    */

    QFont font = qApp->font();
    font.setPointSize(plot->FontSize);
    pl3->setTitle(plot->Title, font);

    plot::TValueInfoCreator ValueInfo;

    ValueInfo(plot->AutoLegendEnabled,
		plot->XYAxisUnit,
		plot->PlotsXLimits[1] -
		plot->PlotsXLimits[0]);
    pl3->setAxisTitle(0, ValueInfo.Value + ValueInfo.Unit, font);

    ValueInfo(plot->AutoLegendEnabled,
		plot->XYAxisUnit,
		plot->PlotsYLimits[1] -
		plot->PlotsYLimits[0]);
    pl3->setAxisTitle(1, ValueInfo.Value + ValueInfo.Unit, font);

    ValueInfo(plot->AutoLegendEnabled,
		plot->ZAxisUnit,
		plot->PlotsZLimits[1] -
		plot->PlotsZLimits[0]);
    pl3->setAxisTitle(2, ValueInfo.Value + ValueInfo.Unit, font);

    pl3->resizeGL(pl3->width(), pl3->height());
    pl3->updateGL();

#endif
}



void TContainer::DelayedPreviewCreation()
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem || !CurrentDataList->ClickedItem->UpdatePreview)
        return;

    CurrentDataList->ClickedItem->UpdatePreview = false;

    SavePreview();
    CurrentDataList->setFlow(QListView::LeftToRight);
}

void TContainer::onItemClicked(const plot::TCurvesDataOwner *Info)
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    SetDownButtonsEnabled(true);
    if (Plot->TwoDPlot->overlayButtons->showBack)
    {
        Plot->TwoDPlot->overlayButtons->B2DBack->click();
    }
    if (Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->showBack)
    {
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->B2DBack->click();
    }

    Plot->ThreeDPlot->Slider->hide();
    Plot->SwitchPlots(plot::TPlot::_2d);
    Plot->TwoDPlot->Pure2dPlot->blockRefresh = true;
    Plot->TwoDPlot->Pure2dPlot->isFiltering = true;
    Plot->TwoDPlot->overlayButtons->hideFilterButton = true;
    Plot->TwoDPlot->overlayButtons->setShowingSendToDataButton(false);
    Plot->TwoDPlot->overlayButtons->setVisibleOldFilters(true);
    Plot->ThreeDPlot->TwoDPlotVertical->Pure2dPlot->isFiltering = true;
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->hideFilterButton = true;
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->setShowingSendToDataButton(false);
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->setVisibleOldFilters(true);

    QStringList clrs = CurrentDataList->ClickedItem->attributes.value("Curve colour").toString().split(",");
    for (int i = 0; i < Info->size(); i++) {
        QString clr = "#000000";
        if (clrs.size() > i)
            clr = clrs.at(i);
        Info->at(i)->Pen.setColor(QColor(clr));
    }

    Plot->TwoDPlot->Pure2dPlot->Title = CurrentDataList->ClickedItem->attributes.value("Name", tr("2d curve")).toString();
    Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(CurrentDataList->ClickedItem->attributes.value("XLabel").toString(),
                                                           CurrentDataList->ClickedItem->attributes.value("XUnit").toString());
    Plot->TwoDPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(CurrentDataList->ClickedItem->attributes.value("YLabel").toString(),
                                                           CurrentDataList->ClickedItem->attributes.value("YUnit").toString());

    Plot->TwoDPlot->SetAutoScalingEnabled(true);
    Plot->TwoDPlot->Pure2dPlot->AutoLegendEnabled = true;
    Plot->TwoDPlot->Pure2dPlot->UpdateAxisLegends();

    Plot->TwoDPlot->Pure2dPlot->SourceData = plot::getUnLock(*Info);
#ifdef _ADROID
    Plot->TwoDPlotVertical->Pure2dPlot->blockRefresh = true;
    Plot->TwoDPlotVertical->Pure2dPlot->isFiltering = true;
    Plot->TwoDPlotVertical->overlayButtons->hideFilterButton = true;
    Plot->TwoDPlotVertical->overlayButtons->setShowingSendToDataButton(false);
    Plot->TwoDPlotVertical->overlayButtons->setVisibleOldFilters(true);
    Plot->TwoDPlotVertical->Pure2dPlot->Title = CurrentDataList->ClickedItem->attributes.value("Name", tr("2d curve")).toString();
    Plot->TwoDPlotVertical->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(CurrentDataList->ClickedItem->attributes.value("XLabel").toString(),
                                                                   CurrentDataList->ClickedItem->attributes.value("XUnit").toString());
    Plot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(CurrentDataList->ClickedItem->attributes.value("YLabel").toString(),
                                                                   CurrentDataList->ClickedItem->attributes.value("YUnit").toString());
    Plot->TwoDPlotVertical->SetAutoScalingEnabled(true);
    Plot->TwoDPlotVertical->Pure2dPlot->AutoLegendEnabled = true;
    Plot->TwoDPlotVertical->Pure2dPlot->UpdateAxisLegends();
    Plot->TwoDPlotVertical->Pure2dPlot->SourceData = *Plot->TwoDPlot->Pure2dPlot->cloneSourceData();
#endif
    //Plot->ThreeDPlot->ThreeDPlots[0]->MathData.clear(); //ensure no math running, no event fuck-up

    Plot->TwoDPlot->PlotControl->AGFitOrders->actions()[CurrentDataList->ClickedItem->attributes.value("Fit mode", 0).toInt()]->trigger();
    Plot->TwoDPlot->PlotControl->AGCurveThickness->actions()[CurrentDataList->ClickedItem->attributes.value("Curve thickness", 0).toInt()]->trigger();
    Plot->TwoDPlot->PlotControl->CBChooseActiveCurve->setCurrentIndex(Plot->TwoDPlot->PlotControl->CBChooseActiveCurve->count()-1);

    while (!Plot->TwoDPlot->Pure2dPlot->isMathReady()) { //lots of dependencies\breakage with events\preview_update
        //qApp->processEvents();
        SleepExport::msleep2(3);
    }

    Plot->TwoDPlot->Pure2dPlot->SetAutoScaleEnabled(true, false);
#ifdef _ADROID
    Plot->TwoDPlotVertical->Pure2dPlot->SetAutoScaleEnabled(true, false);
    Plot->TwoDPlotVertical->Pure2dPlot->blockRefresh = false;
#endif
    Plot->TwoDPlot->Pure2dPlot->blockRefresh = false;
    Plot->Refresh();

    if (CurrentDataList->ClickedItem->UpdatePreview)
    {
        CurrentDataList->ClickedItem->UpdatePreview = false;
        SavePreview();
    }

    if (BShow3DView->isChecked())
        BShow3DView->toggle();

    Func_interval->hide();
    Convol_func->hide();
    gb4DButtons->hide();

    Plot->ThreeDPlot->TwoDPlot->overlayButtons->removeButton(BConvolFuncSelect);
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->removeButton(BConvolFuncSelect);
    Plot->TwoDPlot->overlayButtons->removeButton(BConvolFuncSelect);
    BConvolFuncSelect->hide();
    Plot->ThreeDPlot->TwoDPlot->overlayButtons->removeButton(BFuncInterval);
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->removeButton(BFuncInterval);
    Plot->TwoDPlot->overlayButtons->removeButton(BFuncInterval);
    BFuncInterval->hide();

    //x scale recalc for spectrum, if available
    bool spectra_calib =
            CurrentDataList->ClickedItem->attributes.contains("Laser Wavelength, nm") &&
            CurrentDataList->ClickedItem->attributes.contains("Central Wavelength, nm") &&
            CurrentDataList->ClickedItem->attributes.contains("Dispersion") &&
            CurrentDataList->ClickedItem->attributes.contains("pixelsize") &&
            CurrentDataList->ClickedItem->attributes.contains("Nx") &&
            CurrentDataList->ClickedItem->attributes.contains("Central Pixel");
    if (spectra_calib) {
        int scale = -1;
		auto label = CurrentDataList->ClickedItem->attributes.value("XLabel").toString();
        if (label == QObject::tr("pixels"))
            scale = 0;
        if (label == QObject::tr("WL, nm") || label == QObject::tr("WL"))
            scale = 1;
        if (label == QObject::tr("Raman Shift, 1/cm"))
            scale = 2;
        
        //who thought that saving LOCALIZED string as a identificator of scale is a good idea?...
        if (scale == -1)
        {
            if (label.contains(","))
            {
                if (label.contains("1"))
                    scale = 2;
                else
                    scale = 1;
            }
            else
                scale = 0;
        }

        AG_Scale_select->actions()[scale]->setChecked(true);
        AGScaleSelect->actions()[scale]->setChecked(true);
        AGScaleSelect->actions()[scale]->trigger();
        switch (scale) {
        case 0:
            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("pixels"));
            break;
        case 1:
            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("WL"), QObject::tr("m"));
            break;
        case 2:
            Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QObject::tr("Raman Shift, 1/cm"));
            break;
        }
    }

    if (prev2DLayoutScaleSelect)
        prev2DLayoutScaleSelect->removeWidget(Scale_select);
    prev2DLayoutScaleSelect = Plot->TwoDPlot->PlotControl->ControlMainLayout;
    prev2DLayoutScaleSelect->addWidget(Scale_select);
    Scale_select->setVisible(spectra_calib);
    if (spectra_calib)
    {
        #ifndef _ADROID
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->addLeftButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->addLeftButton(BScaleSelect);
        Plot->TwoDPlot->overlayButtons->addLeftButton(BScaleSelect);
#endif
    }
    else
    {
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->removeButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->removeButton(BScaleSelect);
        Plot->TwoDPlot->overlayButtons->removeButton(BScaleSelect);
    }

    if (Plot->TwoDPlot->PlotControl->BMarkers->isChecked())
        Plot->TwoDPlot->PlotControl->BMarkers->click();

    updateAttributes();

    Plot->TwoDPlot->showPlotControl = showRightMenu;
    Plot->ThreeDPlot->TwoDPlot->showPlotControl = showRightMenu;
    Plot->ThreeDPlot->TwoDPlotVertical->showPlotControl = showRightMenu;
    Plot->ThreeDPlot->showPlotControl = showRightMenu;
    if (!showRightMenu)
    {
        Plot->TwoDPlot->PlotControl->hide();
        Plot->ThreeDPlot->TwoDPlot->PlotControl->hide();
        Plot->ThreeDPlot->TwoDPlotVertical->PlotControl->hide();
        Plot->ThreeDPlot->PlotControl->hide();
    }

    Plot->TwoDPlot->overlayButtons->plot2DResized();
    auto e = new QEvent(QEvent::Leave);
    Plot->TwoDPlot->overlayButtons->plot2DMouseEvent((QMouseEvent*)e, Plot->TwoDPlot->Pure2dPlot);

    static bool firstTime = true;
    if (firstTime)
    {
        Plot->TwoDPlot->resize(Plot->TwoDPlot->size()-QSize(1,1));
        firstTime = false;
    }
}
void TContainer::onItemClicked(const QImage img)
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    SetDownButtonsEnabled(true);
    if (BShow3DView->isChecked())
        BShow3DView->toggle();

    //fake info - background
    plot::T3dDataOwner *Info = new plot::T3dDataOwner;
    plot::ResizeOwner(*Info, 1);
    auto firstInfo = Info->at(0);
	firstInfo->setSize(1, 1);
	firstInfo->UpdateNxNy();
	firstInfo->resetMathLims();

	firstInfo->XLimits[0] = CurrentDataList->ClickedItem->attributes.value("XMin", 0).toDouble();
	firstInfo->XLimits[1] = CurrentDataList->ClickedItem->attributes.value("XMax", 1).toDouble();
	firstInfo->YLimits[0] = CurrentDataList->ClickedItem->attributes.value("YMin", 0).toDouble();
	firstInfo->YLimits[1] = CurrentDataList->ClickedItem->attributes.value("YMax", 1).toDouble();
	firstInfo->UpdateDxDy();

    show3d(plot::getUnLock(*Info));
    Plot->ThreeDPlot->Slider->hide();

    imageitem->image = img;
    imageitem->targetRect = QRectF(firstInfo->XLimits[0], firstInfo->YLimits[0],
		firstInfo->XLimits[1] - firstInfo->XLimits[0],
		firstInfo->YLimits[1] - firstInfo->YLimits[0]);
    imageitem->show();
}
void TContainer::onItemClicked(const plot::T3dDataOwner *Info)
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    SetDownButtonsEnabled(true);
	auto firstInfo = Info->at(0);
	firstInfo->XLimits[0] = CurrentDataList->ClickedItem->attributes.value("XMin", 0).toDouble();
	firstInfo->XLimits[1] = CurrentDataList->ClickedItem->attributes.value("XMax", 1).toDouble();
	firstInfo->YLimits[0] = CurrentDataList->ClickedItem->attributes.value("YMin", 0).toDouble();
	firstInfo->YLimits[1] = CurrentDataList->ClickedItem->attributes.value("YMax", 1).toDouble();
	firstInfo->UpdateDxDy();

    show3d(plot::getUnLock(*Info));
}
void TContainer::onItemClicked(const ccd::TSetOfSpectrum *Info)
{
    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    SetDownButtonsEnabled(true);
    func_min = Info->MinForFuncApply;
    func_max = Info->MaxForFuncApply;
    ((ccd::TSetOfSpectrum *)Info)->XscaleUnit = FindCheckedItemIn(AG_Scale_select);

    show3d(plot::getUnLock(Info->Results), true);
    Plot->ThreeDPlot->SetZDescAndSIUnit(QString(), T4dDataHelper::getConvZUnits(FindCheckedItemIn(AG_Convol_func), FindCheckedItemIn(AG_Scale_select), Info->YscaleUnit), 0);
    Plot->Refresh();

    T4dDataHelper::ShowSpectr(Plot->ThreeDPlot->TwoDPlot,
                              Plot->ThreeDPlot->TwoDPlotVertical,
                              (ccd::TSetOfSpectrum *)Info, lastXind = 0, lastYind = 0, FindCheckedItemIn(AG_Scale_select));
}
void TContainer::show3d(const plot::T3dData data, bool is4d)
{
    Plot->ThreeDPlot->lastSplitPlotSizes[0] = Plot->height()*0.75;
    Plot->ThreeDPlot->lastSplitPlotSizes[1] = Plot->height()*0.25;
    //deactivate image display
    imageitem->hide();
    imageitem->image = QImage();

    if (!CurrentDataList || !CurrentDataList->ClickedItem)
        return;

    new3d = true;
    #ifndef _ADROID
    t3dench->pl3 = nullptr;
#endif
    if (!preview_update)
        CurrentDataList->ClickedItem->UpdatePreview = true;

    if (Plot->ThreeDPlot->overlayButtons->visibilityMode[0] != plot::ThreeDOverlayButtons::SpecialVisibilityMode::svm_none)
    {
        Plot->ThreeDPlot->overlayButtons->resetBackBtnWainting();
    }
    
    Plot->SwitchPlots(plot::TPlot::_3d);
    if (Plot->ThreeDPlot->showSlider)
        Plot->ThreeDPlot->Slider->show();
	auto plot = Plot->ThreeDPlot->ThreeDPlots[0];
	plot->blockRefresh = true;

    if (prev2DLayoutScaleSelect)
        prev2DLayoutScaleSelect->removeWidget(Scale_select);
    if (is4d)
        prev2DLayoutScaleSelect = ((plot::TCommonPlotControl*)Plot->ThreeDPlot->ControlLayout->itemAt(1)->widget())->ControlMainLayout;
    else
        prev2DLayoutScaleSelect = Plot->TwoDPlot->PlotControl->ControlMainLayout;
    prev2DLayoutScaleSelect->addWidget(Scale_select);
    if (Plot->ThreeDPlot->SplitPlotTab->sizes().last() != 0)
        Plot->ThreeDPlot->SplitPlotTab->setSizes(QList<int>() << Plot->ThreeDPlot->lastSplitPlotSizes[0] << Plot->ThreeDPlot->lastSplitPlotSizes[1]);
    Plot->ThreeDPlot->Set2dPlotVisible(is4d);
    Plot->ThreeDPlot->PlotControl->BSave->setVisible(showRightMenu && is4d);
    Plot->ThreeDPlot->PlotControl->BCross->setVisible(showRightMenu && is4d);
    if (Plot->ThreeDPlot->PlotControl->BCross->isOn)
        Plot->ThreeDPlot->PlotControl->BCross->click();
    Plot->ThreeDPlot->PlotControl->BMarkers->setChecked(false);

    if (Func_interval->isOn)
        Func_interval->change_state_man_with_sig(false);
    BFuncInterval->change_state_man(Func_interval->isOn);


	plot->SourceData = data;
    //plot->MathData.clear(); //ensure no math running, no event fuck-up

    if (is4d && !CurrentDataList->ClickedItem->attributes.contains("ZViewAuto"))
        CurrentDataList->ClickedItem->attributes.insert("ZViewAuto", true);

    Plot->ThreeDPlot->restoreAttributes(CurrentDataList->ClickedItem->attributes);


    while (!Plot->ThreeDPlot->ThreeDPlots.at(0)->isMathReady()) { //lots of dependencies\breakage with events\preview_update
        //qApp->processEvents();
        SleepExport::msleep(3);
    }

	plot->blockRefresh = false;
    Plot->Refresh();
	plot->ResizeQPlot();

    Func_interval->setVisible(showRightMenu && is4d);
    Convol_func->setVisible(showRightMenu && is4d);
    Scale_select->setVisible(showRightMenu && is4d);
    gb4DButtons->setVisible(showRightMenu && is4d);

    if (is4d)
    {
        #ifndef _ADROID
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->addLeftButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->addLeftButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->addLeftButton(BFuncInterval);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->addLeftButton(BConvolFuncSelect);
        Plot->TwoDPlot->overlayButtons->addLeftButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->addLeftButton(BFuncInterval);
        Plot->TwoDPlot->overlayButtons->addLeftButton(BFuncInterval);
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->addLeftButton(BConvolFuncSelect);
        Plot->TwoDPlot->overlayButtons->addLeftButton(BConvolFuncSelect);
#endif
#ifdef _ADROID
//        Plot->TwoDPlotVertical->overlayButtons->addLeftButton(BScaleSelect);
//        Plot->TwoDPlotVertical->overlayButtons->addLeftButton(BFuncInterval);
//        Plot->TwoDPlotVertical->overlayButtons->addLeftButton(BConvolFuncSelect);
#endif
    }
    else
    {
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->removeButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->removeButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->removeButton(BFuncInterval);
        Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->removeButton(BConvolFuncSelect);
        Plot->TwoDPlot->overlayButtons->removeButton(BScaleSelect);
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->removeButton(BFuncInterval);
        Plot->TwoDPlot->overlayButtons->removeButton(BFuncInterval);
        Plot->ThreeDPlot->TwoDPlot->overlayButtons->removeButton(BConvolFuncSelect);
        Plot->TwoDPlot->overlayButtons->removeButton(BConvolFuncSelect);
#ifdef _ADROID
        if (Plot->TwoDPlotVertical)
        {
//            Plot->TwoDPlotVertical->overlayButtons->removeButton(BScaleSelect);
//            Plot->TwoDPlotVertical->overlayButtons->removeButton(BFuncInterval);
//            Plot->TwoDPlotVertical->overlayButtons->removeButton(BConvolFuncSelect);
        }
#endif
    }

    Plot->ThreeDPlot->overlayButtons->addUpButton(B3DImageCrop);

    Plot->ThreeDPlot->TwoDPlot->overlayButtons->setVisibleOldFilters(true);
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->setVisibleOldFilters(true);
    #ifndef _ADROID
    Plot->ThreeDPlot->overlayButtons->addUpButton(BShow3DView);
    #endif

	plot->ResizeQPlot();

    updateAttributes();

    Plot->ThreeDPlot->PlotControl->show();
    Plot->ThreeDPlot->TwoDPlot->PlotControl->hide();

    if (!BShow3DView->isChecked())
        foreach (QObject *obj, Plot->ThreeDPlot->PlotControl->children()) {
            QWidget *wid = qobject_cast<QWidget *>(obj);
            if (wid && wid->isVisible())
                wid->setEnabled(true);
        }

    Plot->ThreeDPlot->TwoDPlot->overlayButtons->setShowingSendToDataButton(true);
    #ifdef _ADROID
    Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->setShowingSendToDataButton(true);
#endif
    #ifndef _ADROID
    if (is4d)
    {
        auto opt = options::GetIt();
        opt->optWin->applyFilterSettings();
        //opt->optWin->applyFilter2DSettings();
    }

    //Plot->ThreeDPlot->TwoDPlotVertical->overlayButtons->setShowingSendToDataButton(true);
    Plot->TwoDPlot->showPlotControl = showRightMenu;
    Plot->ThreeDPlot->TwoDPlot->showPlotControl = showRightMenu;
    Plot->ThreeDPlot->TwoDPlotVertical->showPlotControl = showRightMenu;
    Plot->ThreeDPlot->showPlotControl = showRightMenu;
#endif

    if (!showRightMenu)
    {
        Plot->TwoDPlot->PlotControl->hide();
        Plot->ThreeDPlot->TwoDPlot->PlotControl->hide();
        Plot->ThreeDPlot->TwoDPlotVertical->PlotControl->hide();
        Plot->ThreeDPlot->PlotControl->hide();
    }
    if (!is4d)
    {
        Plot->ThreeDPlot->ThreeDPlots.first()->isFiltering = true;
        Plot->ThreeDPlot->ResetAllFilters();
    }
    Plot->ThreeDPlot->TwoDPlot->resetAllRilters();
    Plot->ThreeDPlot->TwoDPlotVertical->resetAllRilters();
    bool oldFilters = !is4d;
    #ifdef _ADROID
    oldFilters = true;
    #endif
    Plot->ThreeDPlot->overlayButtons->blockResizeTimes = 0;
    Plot->ThreeDPlot->overlayButtons->setVisibleOldFilters(oldFilters);
    Plot->ThreeDPlot->overlayButtons->setVisibleFilterButton(!oldFilters);
    Plot->ThreeDPlot->overlayButtons->setVisiblePointSelection(!oldFilters);
    Plot->ThreeDPlot->overlayButtons->resetBackBtnWainting();
    Plot->ThreeDPlot->overlayButtons->setVisibleIfUnderMouse(0);
    Plot->ThreeDPlot->overlayButtons->blockResizeTimes = 0;
    Plot->ThreeDPlot->overlayButtons->setAttributes(!oldFilters ? nullptr : &CurrentDataList->ClickedItem->attributes);

    Plot->ThreeDPlot->lastSplitPlotSizes[0] = Plot->height()*0.75;
    Plot->ThreeDPlot->lastSplitPlotSizes[1] = Plot->height()*0.25;

    Plot->ThreeDPlot->TwoDPlotVertical->resetAllRilters();

    emit on3DPlotShowed(is4d);
}
