#include "flowlayout.hpp"

#include <QGroupBox>

void addDelimeterV(QLayout *lay, short size)
{
    QGroupBox *delim = new QGroupBox();
    if (size < 0) size = 0;
    if (size > 9999) size = 9999;
    delim->setFixedSize(1, size);
    lay->addWidget(delim);
}

void addDelimeterH(QLayout *lay, short size)
{
    QGroupBox *delim = new QGroupBox();
    if (size < 0) size = 0;
    if (size > 9999) size = 9999;
    delim->setFixedSize(size, 1);
    lay->addWidget(delim);
}

VFlowLayout::VFlowLayout(QWidget *parent) : QLayout(parent) {}
VFlowLayout::~VFlowLayout()
{
    QLayoutItem *item;
    while ( (item = takeAt(0)) != nullptr )
        delete item;
}

void VFlowLayout::addItem(QLayoutItem *item)
{
    itemList.append(item);
}

void VFlowLayout::insertWidget(int pos, QWidget *wid)
{
    addWidget(wid);

    itemList.insert(pos, itemList.takeLast());
}

int VFlowLayout::count() const
{
    return itemList.size();
}

QLayoutItem *VFlowLayout::itemAt(int index) const
{
    if (index >= 0 && index < itemList.size())
        return itemList.at(index);
    else
        return nullptr;
}

QLayoutItem *VFlowLayout::takeAt(int index)
{
    if (index >= 0 && index < itemList.size())
        return itemList.takeAt(index);
    else
        return nullptr;
}
void VFlowLayout::setGeometry(const QRect &rect)
{
    QLayout::setGeometry(rect);
    doLayout(rect, false);
}

QSize VFlowLayout::minimumSize() const
{
    QSize size;
    QLayoutItem *item;
    foreach (item, itemList)
        size = size.expandedTo(item->minimumSize());

    int left, top, right, bottom;
    getContentsMargins(&left, &top, &right, &bottom);
    size += QSize(left + right, top + bottom);
    return size;
}

QSize VFlowLayout::maximumSize() const
{
    QSize size;
    QLayoutItem *item;
    foreach (item, itemList)
        size += item->minimumSize();

    int left, top, right, bottom;
    getContentsMargins(&left, &top, &right, &bottom);
    size += QSize(left + right, top + bottom);
    return size;
}

QSize VFlowLayout::sizeHint() const
{
    return minimumSize();
}

Qt::Orientations VFlowLayout::expandingDirections() const
{
    return Qt::Horizontal;
}

bool VFlowLayout::hasHeightForWidth() const
{
    return true;
}

int VFlowLayout::heightForWidth(int width) const
{
    int height = doLayout(QRect(0, 0, width, 0), true);
    return height;
}

int VFlowLayout::doLayout(const QRect &rect, bool testOnly) const
{
    int left, top, right, bottom;
    getContentsMargins(&left, &top, &right, &bottom);
    QRect effectiveRect = rect.adjusted(+left, +top, -right, -bottom);
    int x = effectiveRect.x();
    int y = effectiveRect.y();
    int lineHeight = 0;

    QLayoutItem *item;

    foreach (item, itemList) {
        if (item->widget() && item->widget()->isHidden())
            continue;

        int nextX = x + item->sizeHint().width() + spaceX;
        if (nextX - spaceX > effectiveRect.right() && lineHeight > 0) {
            x = effectiveRect.x();
            y = y + lineHeight + spaceY;
            nextX = x + item->sizeHint().width() + spaceX;
            lineHeight = 0;
        }

        if (!testOnly)
            item->setGeometry(QRect(QPoint(x, y), item->sizeHint()));

        x = nextX;
        lineHeight = qMax(lineHeight, item->sizeHint().height());
    }

    return y + lineHeight - effectiveRect.y() + bottom + top;
}

//HFlowLayout

HFlowLayout::HFlowLayout(QWidget *parent) : VFlowLayout(parent) {}

bool HFlowLayout::hasHeightForWidth() const
{
    return false;
}

Qt::Orientations HFlowLayout::expandingDirections() const
{
    return Qt::Vertical;
}

void HFlowLayout::setGeometry(const QRect &rect)
{
    if (prev_height != rect.height()) {
        invalidate();
        prev_height = rect.height();
    }

    QLayout::setGeometry(rect);
    doLayout(rect, false);
}

QSize HFlowLayout::minimumSize() const
{
    QSize size;
    QLayoutItem *item;
    foreach (item, itemList)
        size = size.expandedTo(item->minimumSize());

    int left, top, right, bottom;
    getContentsMargins(&left, &top, &right, &bottom);
    size += QSize(left + right, top + bottom);

    if (!parentWidget() || !parentWidget()->isVisible() || !geometry().isValid())
        return size;

    //hack for correct expansion
    int width = doLayout(QRect(0, 0, geometry().width(), geometry().height()), true);

    return QSize(width, size.height());
}

int HFlowLayout::doLayout(const QRect &rect, bool testOnly) const
{
    int left, top, right, bottom;
    getContentsMargins(&left, &top, &right, &bottom);
    QRect effectiveRect = rect.adjusted(+left, +top, -right, -bottom);
    int x = effectiveRect.x();
    int y = effectiveRect.y();
    int lineWidth = 0;

    QLayoutItem *item;

    foreach (item, itemList) {
        if (item->widget()->isHidden())
            continue;

        int nextY = y + item->sizeHint().height() + spaceY;
        if (nextY - spaceY > effectiveRect.bottom() && lineWidth > 0) {
            y = effectiveRect.y();
            x = x + lineWidth + spaceX;
            nextY = y + item->sizeHint().height() + spaceY;
            lineWidth = 0;
        }

        if (!testOnly)
            item->setGeometry(QRect(QPoint(x, y), item->sizeHint()));

        y = nextY;
        lineWidth = qMax(lineWidth, item->sizeHint().width());
    }

    return x - effectiveRect.left() + lineWidth  + left + right;
}

