#include "coboltcontroller.hpp"
#include "gui/status.hpp"
#include "tools.hpp"

#include <QTimer>
#include <QFormLayout>
#include <QThread>
#include <QApplication>
#include <QDebug>

void CoboltGui::set_power(double power)
{
    m_interface->set_power(power*1e-3);
}

void CoboltGui::laser_onoff(bool st)
{
    if (!st)
        m_interface->laser_on();
    else
        m_interface->laser_off();
}

void CoboltGui::update()
{
    char ch;
    QString str;

    if (m_interface->hSerial == nullptr) {
        setDeviceStatus(EDconn);
        m_interface->Connect();

        if (m_interface->hSerial == nullptr) {
            upd->stop();
            upd->start(10000);//restart timer for reconnect check after very long time
            setEnabled(false);
            setDeviceStatus(EDno);
            return;
        } else {
            setDeviceStatus(EDok);
            if (!Laser_onoff->isOn)
                m_interface->laser_on();

            upd->stop();
            upd->start(500);//restart timer for normal update
        }
    }

    setEnabled(true);

    m_interface->Write("pa?\r", sizeof("pa?\r") - 1);

    ch = 0;
    str.clear();
    while (ch != '\r') {
        m_interface->Read(&ch, 1);
        str.append(ch);
    }
    str.chop(1);

    double power = str.toDouble() * 1e3;
    cpower->setText(QString::number(power));

    m_interface->Write("i?\r", sizeof("i?\r") - 1);

    ch = 0;
    str.clear();
    while (ch != '\r') {
        m_interface->Read(&ch, 1);
        str.append(ch);
    }
    str.chop(1);

    double current = str.toDouble();
    ccurrent->setText(QString::number(current));

    bool laser_onn = (power >= 0.2 || current >= 0.001);
    setDeviceStatus(laser_onn ? EDplay : EDok);
}

CoboltGui::CoboltGui(QObject *parent, QString suffix, QString port) :
    QWidget((QWidget *)parent)
{
    m_interface = new CoboltController(this, port);

    deviceStatus = new StatusWidget1(suffix);
    setDeviceStatus(EDno);

    setWindowTitle(tr("Laser control") + " " + suffix);
    setParent(nullptr);

    power_ew = new Edit(0, 50, 1);
    power_ew->setValue(1);

    power = new NSTSlider(0, 50, 1, this);
    power->setValue(1);

    Laser_onoff = new TwoStateButton(":/onn.png", ":/off.png");
    Laser_onoff->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

    cpower = new QLabel("0.00");
    ccurrent = new QLabel("0.00");

    Laser_onoff->change_state_man(true);

    connect(Laser_onoff, SIGNAL(state_changed(bool)), this, SLOT(laser_onoff(bool)));
    connect(power_ew, SIGNAL(newValue(double)), power, SLOT(setValue(double)));
    connect(power_ew, SIGNAL(newValue(double)), this, SLOT(set_power(double)));
    connect(power, SIGNAL(valueChanged(double)), power_ew, SLOT(setValue(double)));
    connect(power, SIGNAL(newValue(double)), power_ew, SLOT(setValue(double)));
    connect(power, SIGNAL(newValue(double)), this, SLOT(set_power(double)));

    upd = new QTimer(this);
    connect(upd, SIGNAL(timeout()), this, SLOT(update()));

    if (m_interface->hSerial != nullptr)
        upd->start(500);
    else
        upd->start(10000);

    QVBoxLayout *mainLay = new QVBoxLayout(this);

    QHBoxLayout *lay0 = new QHBoxLayout();
    QVBoxLayout **lay0Sub = new QVBoxLayout*[2];
    for (int i=0; i<2; i++)
    {
        lay0Sub[i] = new QVBoxLayout();
        lay0->addLayout(lay0Sub[i]);
    }
    lay0->setAlignment(Qt::AlignVCenter);
    mainLay->addLayout(lay0);

    lay0Sub[0]->addWidget(new QLabel(tr("Laser power, mW")), 0, Qt::AlignVCenter);
    lay0Sub[1]->addWidget(power_ew, 0, Qt::AlignVCenter);

    lay0Sub[0]->addWidget(Laser_onoff, 0, Qt::AlignVCenter);
    lay0Sub[1]->addWidget(power, 0, Qt::AlignVCenter);

    QHBoxLayout *lay1 = new QHBoxLayout();
    QVBoxLayout **lay1Sub = new QVBoxLayout*[4];
    for (int i=0; i<4; i++)
    {
        lay1Sub[i] = new QVBoxLayout();
        lay1->addLayout(lay1Sub[i]);
    }
    mainLay->addLayout(lay1);

    lay1Sub[0]->addWidget(new QLabel(tr("Power, mW")), 0, Qt::AlignVCenter);
    lay1Sub[1]->addWidget(cpower, 0, Qt::AlignVCenter);

    lay1Sub[2]->addWidget(new QLabel(tr("Current, A")), 0, Qt::AlignVCenter);
    lay1Sub[3]->addWidget(ccurrent, 0, Qt::AlignVCenter);
}

CoboltGui::~CoboltGui()
{
    delete m_interface;
}

void CoboltGui::setDeviceStatus(int idx)
{
    if (!deviceStatus)
        return;

	auto status = deviceStatus->status;
    setEnabled(idx >= EDok);
    switch (idx) {
    case EDno:
        status->set(StatusWidget::Cross, Qt::red);
        status->setToolTip(tr("No device"));
        break;
    case EDconn:
        status->set(StatusWidget::Circle, QColor(255, 128, 0));
        status->setToolTip(tr("Connecting"));
        break;
    case EDok:
        status->set(StatusWidget::Circle, Qt::green);
        status->setToolTip(tr("Connected"));
        break;
    case EDplay:
        status->set(StatusWidget::Play, Qt::green);
        status->setToolTip(tr("On"));
        break;
    }
}

CoboltController::CoboltController(QObject *parent, QString _port) :
    QObject(parent)
{
    hSerial = nullptr;
    port = _port;
    if (!port.contains("COM")) {
        qDebug() << "cobolt_controller: wrong port" << port;
        port = "COM1";
    }
    Connect();
}

CoboltController::~CoboltController()
{
    Write("l0\r", sizeof("l0\r") - 1);
    //read ok
    char ch = 0;
    while (ch != '\r') {
        Read(&ch, 1);
    }

    Disconnect();
}

void CoboltController::Connect()
{
#ifdef _WIN32
    hSerial = CreateFileA(("\\\\.\\" + port).toStdString().data(),
                        GENERIC_READ | GENERIC_WRITE,
                        0, nullptr, OPEN_EXISTING,
                        0, nullptr);

    if (hSerial == INVALID_HANDLE_VALUE) {
        //qDebug() << "com init1 error" << GetLastError();
        hSerial = nullptr;
        return;
    }

    DCB dcbSerialParams = {0};
    dcbSerialParams.DCBlength = sizeof(dcbSerialParams);
    if (!GetCommState(hSerial, &dcbSerialParams)) {
        //error getting state
        hSerial = nullptr;
        return;
    }
    dcbSerialParams.BaudRate=CBR_115200;
    dcbSerialParams.ByteSize=8;
    dcbSerialParams.StopBits=ONESTOPBIT;
    dcbSerialParams.Parity=NOPARITY;
    if (!SetCommState(hSerial, &dcbSerialParams)) {
        //qDebug() << "com init2 error" << GetLastError();
        hSerial = nullptr;
        return;
    }

    COMMTIMEOUTS timeouts={0};
    timeouts.ReadIntervalTimeout=50;
    timeouts.ReadTotalTimeoutConstant=50;
    timeouts.ReadTotalTimeoutMultiplier=10;
    timeouts.WriteTotalTimeoutConstant=50;
    timeouts.WriteTotalTimeoutMultiplier=10;

    if (!SetCommTimeouts(hSerial, &timeouts)) {
        //qDebug() << "com init3 error" << GetLastError();
        hSerial = nullptr;
        return;
    }

    char ch;
    QString str;
    Write("@cobas?\r", sizeof("@cobas?\r") - 1);

    ch = 0;
    str.clear();
    while (ch != '\r') {
        Read(&ch, 1);
        str.append(ch);
    }
    str.chop(1);

    bool ok = false;
    str.toInt(&ok);
    if (!ok || str.toInt() != 0) {
        //disable autostart
        //qDebug() << "cobolt disable autostart";

        Write("@cobas 0\r", sizeof("@cobas 0\r") - 1);
        //read ok
        char ch = 0;
        while (ch != '\r') {
            Read(&ch, 1);
        }
    }

    /*Write("l0\r", sizeof("l0\r") - 1);
    //read ok
    ch = 0;
    while (ch != '\r') {
        Read(&ch, 1);
    }*/

    //qDebug() << "com init ok";
#endif
}

void CoboltController::Disconnect()
{
#ifdef _WIN32
    //qDebug() << "com disconnected";
    CloseHandle(hSerial);
    hSerial = nullptr;
#endif
}

void CoboltController::set_power(double power)
{
#ifdef _WIN32
    //qDebug() << "power" << power;

    Write("p ", sizeof("p ") - 1);
    Write(QString::number(power).toAscii().data(),
          QString::number(power).toAscii().size());
    Write("\r", sizeof("\r") - 1);

    //read ok
    char ch = 0;
    while (ch != '\r') {
        Read(&ch, 1);
    }
#endif
}

void CoboltController::clear_fault()
{
#ifdef _WIN32
    Write("cf\r", sizeof("cf\r") - 1);

    //read ok
    char ch = 0;
    while (ch != '\r') {
        Read(&ch, 1);
    }
#endif
}

void CoboltController::laser_on()
{
    qDebug() << "cobolt on";
#ifdef _WIN32
    Write("l1\r", sizeof("l1\r") - 1);

    //read ok
    char ch = 0;
    while (ch != '\r') {
        Read(&ch, 1);
    }
#endif
}

void CoboltController::laser_off()
{
    qDebug() << "cobolt off";
#ifdef _WIN32
    Write("l0\r", sizeof("l0\r") - 1);

    //read ok
    char ch = 0;
    while (ch != '\r') {
        Read(&ch, 1);
    }
#endif
}

class ReadThread : public QThread {
public:
    HANDLE hSerial;
    void *ptr;
    int size;
    bool result;
    int read;

    ReadThread(HANDLE _hSerial, void *_ptr, int _size) {
        hSerial = _hSerial;
        ptr = _ptr;
        size = _size;
        result = false;
        read = 0;
    }

    void run() override {
#ifdef _WIN32
        while (read < size) {
            DWORD dwBytesRead = 0;
            if (!ReadFile(hSerial, ptr, size - read, &dwBytesRead, nullptr)) {
                qDebug() << "com read error" << GetLastError();
                return;
            }

            //qDebug() << "com read" << dwBytesRead;
            read += dwBytesRead;
            msleep(1);
        }

        result = true;
#endif
    }
};

void CoboltController::Read(void *ptr, int size)
{
    if (hSerial == nullptr) {
        ((char *)ptr)[0] = '\r';
        return;
    }

    ReadThread read(hSerial, ptr, size);

    QElapsedTimer tm;
    read.start();
    tm.start();
    while (read.isRunning() && tm.elapsed() < 100) {
        qApp->processEvents(); //CCDPLOT_CTR->LASER_GUI_CTR->LASER_CONTROL_CTR->Connect->(connected)->Read->processEvents->possible crashes. TODO fix
        SleepExport::msleep2(10);
    }

    if (read.isRunning())
        read.terminate();

    if (!read.result)
        Disconnect();
}

void CoboltController::Write(const void *ptr, int size)
{
#ifdef _WIN32
    if (hSerial == nullptr)
        return;

    DWORD dwBytesWrite = 0;
    if (!WriteFile(hSerial, ptr, size, &dwBytesWrite, nullptr)) {
        qDebug() << "com write error" << GetLastError();
        Disconnect();
    }

    //qDebug() << "com write" << size << ' ' << dwBytesWrite;
#endif
}
