#include "shuttle.h"
#include <QSettings>
#include <QApplication>
#include <QWidget>
#include <QDebug>


ShuttleControllerNative *scninst = nullptr;
#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
bool myEventFilter(void *message, long *)
{
    MSG *p = (MSG *)message;
    if (p->message == WM_INPUT) {
        UINT dataSize = 256;
        BYTE buf[256];
        GetRawInputData((HRAWINPUT)p->lParam, RID_INPUT, buf, &dataSize, sizeof(RAWINPUTHEADER));
        const RAWINPUT *raw = (const RAWINPUT*)buf;
        if (raw->header.dwType == RIM_TYPEHID) {
            if (raw->data.hid.dwCount == 1 && raw->data.hid.dwSizeHid == 6) {
                if (scninst)
                    scninst->set_hid((signed char)raw->data.hid.bRawData[1], raw->data.hid.bRawData[2], raw->data.hid.bRawData[4] + (raw->data.hid.bRawData[5] << 8));
                return true;
            } else
                qDebug() << "invalid raw input" << raw->data.hid.dwCount << raw->data.hid.dwSizeHid;
        }
    }
    return false;
}
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QAbstractNativeEventFilter>
class efilt : public QAbstractNativeEventFilter
{
public:
     bool nativeEventFilter(const QByteArray &, void * message, long * result) override
     {
         return myEventFilter(message, result);
     }
};
efilt nativefilter;
#endif
#endif

#if defined(Q_OS_LINUX)
#include <QDir>
#include <QThread>
#include <linux/input.h>
#include <fcntl.h>
#include <errno.h>
#include <windows.h>
static volatile bool evrun;
class EvReader : public QThread
{
    void run() {
        while (evrun) {
            QStringList l = QDir("/dev/input/by-id", "*Contour_Design_Shuttle*").entryList(QDir::Files);
            if (!l.size()) {
                for (int i = 0; i < 10 && evrun; i++)
                    msleep(50);
                continue;
            }

            QFile f("/dev/input/by-id/" + l.first());
            if (!f.open(QIODevice::ReadWrite | QIODevice::Unbuffered)) {
                qDebug() << l.first() << "open fail";
                for (int i = 0; i < 10 && evrun; i++)
                    msleep(50);
                continue;
            }
            int flags = fcntl(f.handle(), F_GETFL, 0);
            fcntl(f.handle(), F_SETFL, flags | O_NONBLOCK);
            struct {timeval t; WORD type; WORD code; uint val;} ev;
            int wheel_abs = 0;
            int shuttle = 0;
            long buttons = 0;
            do {
                int read = f.read((char *)&ev, sizeof(ev));
                if (read != sizeof(ev)) {
                    if (errno != EWOULDBLOCK)
                        break;
                    msleep(10);
                    continue;
                }
                //qDebug() << ev.type << ev.code << ev.val;
                if (ev.type == EV_KEY) {
                    if (ev.val)
                        buttons |= 1 << (ev.code - 256);
                    else
                        buttons &= ~(1 << (ev.code - 256));
                }
                if (ev.type == EV_REL && ev.code == 7)
                    wheel_abs = ev.val;
                if (ev.type == EV_REL && ev.code == 8)
                    shuttle = (signed char)ev.val;
                if (ev.type == EV_SYN) {
                    if (scninst)
                        scninst->set_hid(shuttle, wheel_abs, buttons);

                    wheel_abs = 0;
                    shuttle = 0;

                    //fake press, so we always get sync report even with rel wheel and shuttle == 0
                    //button still works as this changes kernel button state too
                    ev.type = EV_KEY; ev.code = BTN_0; ev.val = 1;
                    ev.t.tv_usec++;
                    f.write((char *)&ev, sizeof(ev));
                    ev.type = EV_KEY; ev.code = BTN_0; ev.val = 0;
                    ev.t.tv_usec++;
                    f.write((char *)&ev, sizeof(ev));
                }
            } while (evrun);
            f.close();
        }
    }
};
EvReader evreader;
#endif

ShuttleControllerNative::ShuttleControllerNative(QObject *parent) : QObject(parent)
{
    if (scninst) throw scninst;
    scninst = this;

    QSettings settings("Param/shuttle.conf", QSettings::IniFormat);
    small_remap = settings.value("shuttle_xpress", false).toBool();
    prev_wheel = 127;
    prev_shuttle = 0;
    for (int i = 0; i < 16; i++)
        prev_buttons[i] = 0;

#if defined(Q_OS_LINUX)
    QStringList l = QDir("/dev/input/by-id", "*Contour_Design_Shuttle*").entryList(QDir::Files);
    if (l.size())
        small_remap = !l.first().contains("PRO_v2");

    evrun = true;
    evreader.start(QThread::LowPriority);
#endif

#ifdef _WIN32
    QWidgetList l;
    foreach (QWidget *w, qApp->topLevelWidgets())
        if (!w->windowTitle().isNull())
            l << w;
    if (l.size() == 0) {
        qDebug() << "no window found for shuttle";
        return;
    }

    RAWINPUTDEVICE device;
    device.usUsagePage = 12;
    device.usUsage = 1;
    device.dwFlags = RIDEV_INPUTSINK;
    device.hwndTarget = (HWND)l.last()->winId();
    RegisterRawInputDevices(&device, 1, sizeof device);

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
    qApp->installNativeEventFilter(&nativefilter);
#else
    qApp->setEventFilter(&myEventFilter);
#endif
#endif
}
#include <tools.hpp>
ShuttleControllerNative::~ShuttleControllerNative()
{
    scninst = nullptr;
#if defined(Q_OS_LINUX)
    evrun = false;
    while (evreader.isRunning()) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }
#endif
}
void ShuttleControllerNative::set_hid(int shuttle, int wheel_abs, long rawbuttons)
{
    if (shuttle != prev_shuttle)
        emit shuttle_position(prev_shuttle = shuttle);

    int delta = (signed char)((wheel_abs - prev_wheel + 256)%256);
    if (qAbs(delta) == 1)
        emit wheel_event(delta);
    prev_wheel = wheel_abs;

    for (int i = 0; i < 16; i++) {
        bool ebuttons = rawbuttons & (1 << i);
        int idx = i;
        if (small_remap) idx += 4; //xpress remap
        if (i == 9) idx = 10; if (i == 13) idx = 9; if (i == 10) idx = 13; //pro v2 remap
        if (ebuttons && ebuttons != prev_buttons[i])
            emit button_press(idx);
        prev_buttons[i] = ebuttons;
    }
}
