#include "modules.hpp"
#include "mainwindow.h"
#include "hardware/relay.hpp"
#include "ccd/ccdplot.h"
#include <QApplication>
#include <QSettings>
#include <QTranslator>
#include <QDesktopWidget>
#include <QSplashScreen>
#include <QMessageBox>
#include <QTimer>
#include <QFile>
#include <QDir>
#include <QPluginLoader>
#include "plugins/moduleplugin.hpp"
#include "scanplot.h"
#include <QTextStream>
#include "gui/splashaboutwindow.h"



#ifdef Q_OS_UNIX
#include <sys/types.h>
#include <sys/stat.h>
#endif

#define NOMINMAX
#include "windows.h"

#ifdef NST_DELAYLOAD
#include <delayimp.h>
FARPROC WINAPI __pfnDliFailureHook2__(unsigned dliNotify, PDelayLoadInfo pdli)
{
    if (dliNotify == dliFailLoadLib)
        qDebug() << "dll delayload failed" << pdli->szDll;
    if (dliNotify == dliFailGetProc) {
        qDebug() << "dll function delayload failed" << pdli->szDll;
        if (pdli->dlp.fImportByName)
            qDebug() << "name" << pdli->dlp.szProcName;
        else
            qDebug() << "ordinal" << pdli->dlp.dwOrdinal;
    }
    throw 0;
    return nullptr;
}
#endif

#include <QSharedMemory>
#define UNIQUE_KEY "nspec_signleapp_key"

class SingleApp : public I_MODULE
{
public:
    SingleApp() {
        singleapp_sharedmem.setKey(UNIQUE_KEY);
        if (!(is_running = singleapp_sharedmem.attach())) {
            if (!singleapp_sharedmem.create(1024)) {
                qDebug() << "Unable to create single instance.";
                is_running = true;
                return;
            }

            QByteArray msg("0\0"); // default value to note that no message is available.
            setMessage(msg);
        }
    }
    ~SingleApp() override {
        if (singleapp_sharedmem.isAttached())
            singleapp_sharedmem.detach();
    }
    bool setMessage(const QByteArray ba) {
        if (!singleapp_sharedmem.lock())
            return false;

        memcpy((char *)singleapp_sharedmem.data(), ba.constData(),
               qMin(singleapp_sharedmem.size(), ba.size()));

        return singleapp_sharedmem.unlock();
    }
    QByteArray getMessage() {
        if (!singleapp_sharedmem.lock())
            return QByteArray();

        QByteArray byteArray = QByteArray((char*)singleapp_sharedmem.constData(), singleapp_sharedmem.size());

        singleapp_sharedmem.unlock();
        return byteArray;
    }

    static const QLatin1String name() { return QLatin1String("singleapp"); }
    void *basecast(const QLatin1String name) override { return name.latin1() != nullptr ? nullptr : static_cast<SingleApp *>(this); }
    void setRunning(bool on) override {
        if (!on && singleapp_sharedmem.isAttached())
            singleapp_sharedmem.detach();
    }

    QSharedMemory singleapp_sharedmem;
    bool is_running;

    static SingleApp* GetIt() noexcept
    {
        auto sa = ModuleManager::instance->find1VoidPtr("singleapp");
        return static_cast<SingleApp*>(sa);
    }
};
QByteArray singleappGet()
{
    auto *sa = SingleApp::GetIt();
    return sa->getMessage();
}
bool singleappSet(const QByteArray ba)
{
    auto *sa = SingleApp::GetIt();
    return sa->setMessage(ba);
}


#ifdef NST_DELAYLOAD
#if _MSC_VER >= 1900 //vc2015
const
PfnDliHook   __pfnDliFailureHook2 = __pfnDliFailureHook2__;
#endif
#endif

int main(int argc, char *argv[])
{
#ifdef NST_DELAYLOAD
#if _MSC_VER < 1900 //vc2015
    __pfnDliFailureHook2 = __pfnDliFailureHook2__;
#endif
#endif
#ifdef Q_OS_UNIX
    umask(0);
#endif

    //qputenv("QT_DEBUG_PLUGINS", QByteArray("1"));

    QApplication::setEffectEnabled(Qt::UI_AnimateMenu, false);
    QApplication::setEffectEnabled(Qt::UI_FadeMenu, false);
    QApplication::setEffectEnabled(Qt::UI_AnimateCombo, false);
    QApplication::setEffectEnabled(Qt::UI_AnimateTooltip, false);
    QApplication::setEffectEnabled(Qt::UI_FadeTooltip, false);
    QApplication::setEffectEnabled(Qt::UI_AnimateToolBox, false);

    QLocale::setDefault(QLocale::system());

    QApplication app(argc, argv);
    app.setQuitOnLastWindowClosed(false);
#ifndef _WIN32
    QApplication::setWindowIcon(QIcon(":/MainToolbar/About_nst.png"));
#endif

    InitPlotIO();

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
    QTimer *precision_timer = new QTimer(qApp);
    precision_timer->setTimerType(Qt::PreciseTimer);
    precision_timer->start(666);
#elif defined(_WIN32)
    if (timeBeginPeriod(1) != TIMERR_NOERROR)
        qDebug() << "no high-res timer";
#endif

    QElapsedTimer startup;
    startup.start();

    //change working directory to app dir, else data won't be found
    QDir::setCurrent(app.applicationDirPath());
#ifndef _WIN32
    QString cfgpath = QDir::homePath() + "/.config/nspec";
    QDir::current().mkpath(cfgpath);
    QDir::setCurrent(cfgpath);
#endif

    //parses command line for file open
    bool open_file_on_load = false;
    QString filename;
    for (int i = 1; i < app.arguments().size(); i++) {
        QString arg = app.arguments().at(i);

        if (arg.startsWith("--")) {
            continue;
        }

        if (arg.endsWith(".nstdat") &&
                QFileInfo(arg).exists())
        {
            open_file_on_load = true;
            filename = arg;
            qDebug() << "command line open file" << filename;
            break;
        }
    }

    SingleApp *singleapp = new SingleApp;
    if (singleapp->is_running) {
        if (open_file_on_load) {
            QByteArray msg("2");
            msg.append(filename.toUtf8());
            msg.append('\0');
            singleapp->setMessage(msg);
        } else {
            QByteArray msg("1\0");
            singleapp->setMessage(msg);
        }

        qDebug() << "already running, my pid =" << qApp->applicationPid();
        delete singleapp;
        return 0;
    }


    QDir::current().mkpath("Param");
    QDir::current().mkpath(TContainer::getDataDir());

    QTranslator qtTranslator;
    QTranslator translator;
    {//gui settings
        QSettings settings("Param/gui.conf", QSettings::IniFormat);

        QFont font;
        font.fromString(settings.value("font", "Arial,10,-1,5,50,0,0,0,0,0").toString());
        qApp->setFont(font);

        QString translation = settings.value("translation", QString(":/translation_base")).toString();
        translator.load(translation);
        app.installTranslator(&translator);
        app.setProperty("translation", translation);

        qtTranslator.load(QString("qt_") + (translation.contains("russian") ? "ru" : "en"));
        app.installTranslator(&qtTranslator);

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)) && !defined(_ADROID)
        double scale = qApp->desktop()->logicalDpiX() / 101.;
#else
        double scale = qApp->desktop()->physicalDpiX() / 101.;
#endif

        int Bsize = settings.value("buttonsize", 25).toInt();
        QString style = "QAbstractButton, QComboBox, QTabWidget, QTabBar, QToolBar { qproperty-iconSize: " //, QAbstractItemView
                + QString::number(Bsize)
                + "px "
                + QString::number(Bsize)
                + "px; }"
                + "QGroupBox { border: 1px solid gray; border-radius: 5px; margin-top: 5px; margin-bottom: 0px; padding-top: 5px; padding-bottom: 1px; }"
                + "QGroupBox::title {top:" + QString::number(int(-7*scale)) + " px;left: 10px; subcontrol-origin: border; }"
                + "QComboBox { border: 1px solid #0000DF; border-radius: 3px; padding: 1px " + QString::number(int(5*scale)) + "px 1px 3px; }"
                + "QComboBox:focus { border: 2px solid #5050DF; border-radius: 3px; padding: 1px " + QString::number(int(5*scale)) + "px 1px 3px; }"
                + "QLineEdit { border: 1px solid #0000DF; border-radius: 3px; }" //, QDoubleSpinBox
                + "QTabBar::tab:only-one[documentMode=\"true\"] { height: 0 px; width: 0 px; }";

        app.setStyleSheet(style);

        app.setProperty("buttonsize", Bsize); //TODO remove
    }


    QSplashScreen *splash = makeSplash();

    ModuleManager modules;
    modules.addToFactory<ManagedThreadedCleaner>();

    modules.add(singleapp);

    modules.load("eg3000");

#ifdef USE_RELAY
    modules.add(new RelayModule);
#endif

    MainWindow window;
    modules.add(QLatin1String("mainwindow"), &window);

    CCDPlot *ccdplot = nullptr;
    QWidget *parentw = window.ContentStackedWidget;
    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);

        if (settings.value("display/data", true).toBool())
            modules.load("data", parentw);

        if (settings.value("display/spectroscopy", false).toBool())
            ccdplot = static_cast<CCDPlot *>(modules.load("ccdplot", parentw)->basecast());

        if (settings.value("display/scan", true).toBool())
        {
            modules.load("scan", parentw);

        }

        if (settings.value("display/oscilloscope", true).toBool())
            modules.load("oscilloscope", parentw);

        //bool spm = false;
        //if (spm = settings.value("display/spm", true).toBool())
            modules.load("spm", parentw);

        if (settings.value("display/video", true).toBool())
            modules.load("video", parentw);
        if (settings.value("display/video2", false).toBool())
            modules.load("video", parentw);

        if (settings.value("display/lithography", false).toBool()) {
            QString prefix, suffix;
#if (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
    #ifdef Q_OS_WIN
            suffix = ".dll";
    #else
            prefix = "lib";
        #ifdef Q_OS_DARWIN
            suffix = ".dylib";
        #else //linux
            suffix = ".so";
        #endif
            if (QFileInfo(app.applicationDirPath() + "/" + prefix + "litho" + suffix).exists())
                prefix = app.applicationDirPath() + "/" + prefix;
            if (QFileInfo("/usr/lib/nspec"  "/" + prefix + "litho" + suffix).exists())
                prefix = "/usr/lib/nspec"  "/" + prefix;
    #endif
#else
    #ifndef Q_OS_WIN
            app.addLibraryPath("/usr/lib/nspec");
    #endif
#endif
            IModulePlugin *p = qobject_cast<IModulePlugin *>((new QPluginLoader(prefix + "litho" + suffix))->instance());
            if (p) {
                p->init(&modules, &ModuleManager::unload);
                modules.load("litho", parentw);
            } else
                qDebug() << "litho" << "plugin not found";
        }

        if (settings.value("net/srv_on", false).toBool())
            modules.add(new MobileServer);

        modules.load("options", parentw);
        modules.load("motors", parentw);

        //if (settings.value("display/motors", false).toBool() || spm)
            //modules.load("motors");

        if (settings.value("display/about", true).toBool())
            modules.add(new TAboutModule);
    }


    splash->close();
    splash->deleteLater();
    window.show();

    qDebug() << "startup time" << startup.elapsed();


    if (open_file_on_load)
        if (auto data = DataPlot::GetIt()) {
            data->Container->OpenFile(filename, true);
        }

    modules.setRunning();
    if (ccdplot)
        ccdplot->opticsReset();

    int retv = app.exec();

#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#elif defined(_WIN32)
    timeEndPeriod(1);
#endif

    return retv;
}
