#include "remote_mobile.hpp"
#include "gui/status.hpp"
#include "remote_nst_dataid.hpp"
#include "networkserver.hpp"

#include <QTcpSocket>
#include <QUdpSocket>
#include <QTimer>
#include <QSettings>
#include <QDateTime>
#include <QFile>

struct RemoteOPCode {
    enum _ {
        DataVersion,
        IdValue,
        KeepAlive
    };
};


void MobileServer::ban(const QHostAddress &ip)
{
    ban_map.insert(ip.toString(), 10);
    if (!ban_timer) {
        ban_timer = new QTimer(this);
        connect(ban_timer, SIGNAL(timeout()), this, SLOT(banStep()));
    }
    ban_timer->start(1000);
}

void MobileServer::banStep()
{
    foreach (QString k, ban_map.keys()) {
        int c = ban_map.value(k) - 1;
        if (c <= 0)
            ban_map.remove(k);
        else
            ban_map.insert(k, c);
    }
    if (ban_map.isEmpty() && ban_timer) {
        ban_timer->stop();
        delete ban_timer;
        ban_timer = nullptr;
    }
}

void MobileServer::newDiscovery()
{
    while (discovery_socket->hasPendingDatagrams()) {
        char buf[4];
        QHostAddress host;
        quint16 port;
        int sz = discovery_socket->readDatagram(buf, 4, &host, &port);
        if (sz != 4 || buf[0] != 1 || buf[1] != 2 || ban_map.contains(host.toString()))
            continue;

        char buf1[4] = {1, 2, 0, 0};
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (settings.value("net/srv_port", 33331).toInt() == 33331)
            discovery_socket->writeDatagram(buf1, 4, host, port);
        else {
            QUdpSocket *discovery_socket = new QUdpSocket;
            discovery_socket->bind(settings.value("net/srv_port", 33331).toInt(), QUdpSocket::ShareAddress | QUdpSocket::ReuseAddressHint);
            connect(discovery_socket, SIGNAL(error(QAbstractSocket::SocketError)), discovery_socket, SLOT(deleteLater()));
            connect(discovery_socket, SIGNAL(bytesWritten(qint64)), discovery_socket, SLOT(deleteLater()));
            discovery_socket->writeDatagram(buf1, 4, host, port);
        }
    }
}

void MobileServer::newConnection(INetworkClient *socket)
{
    QSettings settings("Param/gui.conf", QSettings::IniFormat);

    MobileInstance *inst = new MobileInstance;
    inst->auth = !settings.value("net/srv_pwd_on", false).toBool();
    inst->dropped_count = 0;
    inst->setParent(this);
    inst->serv = this;
    inst->socket = (BaseNetworkClient *)socket;

    inst->keepalive_times = -1;
    inst->keepalive_timer = new QTimer(inst);
    connect(inst->keepalive_timer, SIGNAL(timeout()), inst, SLOT(keepalive()));

    connect(inst->socket, SIGNAL(newData()), inst, SLOT(newData()));
    connect(inst->socket, SIGNAL(closed()), inst->keepalive_timer, SLOT(stop()));
    connect(inst->socket, SIGNAL(closed()), this, SLOT(newDisConnection()));
    connect(inst->socket, SIGNAL(closed()), inst, SLOT(deleteLater()));

    clients.append(inst);
    if (networkStatus) {
        networkStatus->status->set(StatusWidget::Play, Qt::green);
        networkStatus->label->setText(tr("Net. connected") + ": " + QString::number(clients.size()));
    }

    if (settings.value("net/srv_log", true).toBool()) {
        QFile fl("network_access.log");
        if (fl.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
            QTextStream st(&fl);
            st << "con> " << (void *)inst << " " << QDateTime::currentDateTime().toString() << " " << socket->socket->peerAddress().toString() << endl;
        }
    }

    quint32 ver = inst->socket->version;
    QByteArray out;
    QDataStream st(&out, QIODevice::WriteOnly);
    st << ver;
    st.setDevice(nullptr);
    inst->socket->sendData(RemoteOPCode::DataVersion, (quint8 *)out.data(), out.size());

    if (settings.value("net/srv_pwd_on", false).toBool())
        setPropIdVarS(inst->socket, DataID::Password, true);

    if (ban_map.contains(socket->socket->peerAddress().toString())) {
        if (settings.value("net/srv_log", true).toBool()) {
            QFile fl("network_access.log");
            if (fl.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
                QTextStream st(&fl);
                st << "err> " << (void *)inst << " " << QDateTime::currentDateTime().toString() << " " << "in temp ban list" << endl;
            }
        }
        socket->socket->abort();
    }

    inst->socket->sendData(RemoteOPCode::KeepAlive, nullptr, 0);
}

void MobileServer::newDisConnection()
{
    if (!sender())
        return;

    for (int i = 0; i < clients.size(); i++)
        if (clients.at(i)->socket == sender()) {
            QSettings settings("Param/gui.conf", QSettings::IniFormat);
            if (settings.value("net/srv_log", true).toBool()) {
                QFile fl("network_access.log");
                if (fl.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
                    QTextStream st(&fl);
                    st << "dis> " << (void *)clients.at(i) << " " << QDateTime::currentDateTime().toString() << endl;
                }
            }

            clients.removeAt(i);
            break;
        }

    if (networkStatus) {
        networkStatus->status->set(clients.size() ? StatusWidget::Play : StatusWidget::Circle, Qt::green);
        networkStatus->label->setText(tr("Net. connected") + ": " + QString::number(clients.size()));
    }
}

void MobileInstance::keepalive()
{
    if (keepalive_times >= 2) {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        if (settings.value("net/srv_log", true).toBool()) {
            QFile fl("network_access.log");
            if (fl.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
                QTextStream st(&fl);
                st << "timeout> " << (void *)this << " " << QDateTime::currentDateTime().toString() << endl;
            }
        }
        keepalive_times = 0;
        socket->socket->abort();
    }
    keepalive_times++;
    socket->sendData(RemoteOPCode::KeepAlive, nullptr, 0);
}

MobileInstance::~MobileInstance()
{
    qDebug()<<"~~~~~DESTRUCTIVE mobile instance~~~~"<<this;
//    disconnect(keepalive_timer, SIGNAL(timeout()), this, SLOT(keepalive()));
//    keepalive_times = 99;
//    //keepalive_timer->stop();
//    delete keepalive_timer;
//    delete socket;
//    delete serv;

//    keepalive_timer = nullptr;
//    socket = nullptr;
//    serv = nullptr;
}

void MobileInstance::newData()
{
    QSettings settings("Param/gui.conf", QSettings::IniFormat);

    for (int i = 0; i < socket->data.size() && i < socket->address.size(); i++) {
        switch (socket->address.at(i)) {
        case RemoteOPCode::KeepAlive:
            if (keepalive_times == -1)
                keepalive_timer->start(2500);
            keepalive_times = 0;
            break;
        case RemoteOPCode::DataVersion: {
            QDataStream st(socket->data.at(i));
            quint32 remote_ver;
            st >> remote_ver;

            if (st.status() != QDataStream::Ok)
                break;

            socket->version = qMin<int>(socket->version, remote_ver);

            serv->setPropIdVarS(socket, DataID::GuestMode, settings.value("net/srv_guest", true).toBool());

            foreach (I_MODULE *m, ModuleManager::instance->modules)
                if (I_NET_MODULE_BASE *m0 = (I_NET_MODULE_BASE *)m->basecast("net"))
                    m0->initialSend(this);
        }
            break;
        case RemoteOPCode::IdValue: {
            QDataStream st(socket->data.at(i));
            st.setVersion(socket->version);
            int id;
            QVariant val;
            st >> id >> val;

            if (st.status() != QDataStream::Ok)
                break;

            if (!auth && (id != DataID::Password || val.toString() != settings.value("net/srv_pwd", "").toString())) {
                if (settings.value("net/srv_log", true).toBool()) {
                    QFile fl("network_access.log");
                    if (fl.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
                        QTextStream st(&fl);
                        if (id != DataID::Password)
                            st << "err> " << (void *)this << " " << QDateTime::currentDateTime().toString() << " " << "no auth" << endl;
                        else
                            st << "err> " << (void *)this << " " << QDateTime::currentDateTime().toString() << " " << "wrong auth: " << val.toString() << endl;
                    }
                }
                if (socket->socket) {
                    serv->ban(socket->socket->peerAddress());
                    socket->socket->abort();
                }
                break;
            }
            auth = true;

            foreach (I_MODULE *m, ModuleManager::instance->modules)
                if (I_NET_MODULE_BASE *m0 = (I_NET_MODULE_BASE *)m->basecast("net"))
                    m0->setValue(id, val, sender());
        }
            break;
        }
    }
    socket->data.clear();
    socket->address.clear();
}


MobileServer::MobileServer()
{
    ban_timer = nullptr;
    max_dropped_count = 10;

    networkStatus = new StatusWidget1();
    networkStatus->status->set(StatusWidget::Circle, Qt::green);
    networkStatus->label->setText(tr("Net. connected") + ": " + QString::number(0));

    discovery_socket = new QUdpSocket;
    discovery_socket->bind(33331, QUdpSocket::ShareAddress | QUdpSocket::ReuseAddressHint);
    connect(discovery_socket, SIGNAL(readyRead()), SLOT(newDiscovery()));

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    NetworkServer *server = new NetworkServer(settings.value("net/srv_port", 33331).toInt());
    server->setParent(this);
    connect(server, SIGNAL(newClient(INetworkClient*)), this, SLOT(newConnection(INetworkClient*)));
}

bool MobileServer::any_need_data(const QString &name)
{
    return any_need_data(name.toLocal8Bit().constData());
}

bool MobileServer::any_need_data(const char *name)
{
    foreach (MobileInstance *ch, clients)
        if (ch->socket->property(name).toBool())
            return true;

    return false;
}

void MobileServer::sendData(int id, QVariant val)
{
    auto mob = MobileServer::GetIt();
    if (!mob)
        return;

    mob->setPropIdVar(id, val);
}
void MobileServer::setPropIdVar(int id, QVariant val, int can_drop_id)
{
    if (can_drop_id == 1)
        can_drop_id = id;

    QMultiMap<int, MobileInstance *> clients_map;

    foreach (MobileInstance *ch, clients)
        if (ch->socket->socket)
            clients_map.insert(ch->socket->version, ch);

    foreach (int key, clients_map.uniqueKeys()) {
        QByteArray out;
        QDataStream st(&out, QIODevice::WriteOnly);
        st.setVersion(key);
        st << id << val;
        st.setDevice(nullptr);

        foreach (MobileInstance *ch, clients_map.values(key)) {
            if (can_drop_id) {
                bool overfl = ch->socket->socket->bytesToWrite()*2 >= out.size();
                if (overfl || (!ch->dropped_list.isEmpty() && ch->dropped_list.head() != can_drop_id)) {
                    if (!ch->dropped_list.contains(can_drop_id))
                        ch->dropped_list.enqueue(can_drop_id);

                    if (!overfl)
                        ch->dropped_count++;
                    else
                        ch->dropped_count = 0;

                    if (overfl || ch->dropped_count < max_dropped_count)
                        continue;
                }

                if (!ch->dropped_list.isEmpty()) {
                    ch->dropped_list.removeFirst();
                    ch->dropped_list.removeOne(can_drop_id);
                }

                ch->dropped_count = 0;
            }

            ch->socket->sendData(RemoteOPCode::IdValue, (quint8 *)out.data(), out.size());
        }
    }
}

void MobileServer::setPropIdVarS(BaseNetworkClient *socket, int id, QVariant val)
{
/*
    bool ok = false;
    foreach (MobileInstance *ch, clients)
        if (ch->socket == socket && socket->socket) {
            ok = true;
            break;
        }
    if (!ok)
        return;
*/

    QByteArray out;
    QDataStream st(&out, QIODevice::WriteOnly);
    st.setVersion(socket->version);
    st << id << val;
    st.setDevice(nullptr);

    socket->sendData(RemoteOPCode::IdValue, (quint8 *)out.data(), out.size());
}

MobileServer* MobileServer::GetIt() noexcept
{
    auto mob = ModuleManager::instance->find1VoidPtr("net");
    return static_cast<MobileServer*>(mob);
}
