#include "motors.hpp"
#include "gui/status.hpp"
#include "gui/mslider.hpp"
#include "remote_nst_dataid.hpp"
#include "mobile/remote_mobile.hpp"
#include "gui/ttabwidget.hpp"

#include <qmath.h>
#include <QSettings>
#include <QCheckBox>
#include <QGroupBox>
#include <QComboBox>
#include <QRadioButton>
#include <QTimer>
#include <QHBoxLayout>
#include <QFormLayout>
#include <QApplication>
#include <QToolButton>
#include <QMenu>
#include <QDebug>
#include <QTableWidget>
#include <QHeaderView>
#include <QCloseEvent>
#include "controls.h"
#include "edit.hpp"
#include "tools.hpp"
#include "spm_motors.hpp"
#include "options_window.h"
#include "USBVideo/videoplayer.h"
#include "scanplot.h"

//see also MotorRoles enum
static const char *axes_names[] = { QT_TR_NOOP("-"), QT_TR_NOOP("X Head"), QT_TR_NOOP("Y Head"), QT_TR_NOOP("Z Landing"),
                                    QT_TR_NOOP("X Stage"), QT_TR_NOOP("Y Stage"), QT_TR_NOOP("Z Focus"),
                                    QT_TR_NOOP("X3"), QT_TR_NOOP("Y3"), QT_TR_NOOP("Z3"),
                                    QT_TR_NOOP("Unused") };
static const int axes_names_count = 11;


void Motors::go(int num, int dir, double speed, double steps)
{
    if (num == -1)
        num = active_motor;
    int phys_num = mapVirtToPhys(num);

    if (steps < 0) {
        steps *= -1;
        dir *= -1;
    }

    BYTE direction = smStop;
    if (dir != 0) {
        bool morient = num == 250 || num == 0 ? orient[0] : orient[num-1];
        if (dir < 0) morient = !morient;
        if (!morient) direction = steps == 0 ? smDownToLimit : smDown;
        else direction = steps == 0 ? smUpToLimit : smUp;
    }
    if (num == 250)
        num = mapPhysToVirt(1);

    double min_speed = EMinSpeed[num-1]->value();
    double max_speed = EMaxSpeed[num-1]->value();
    if (min_speed > 0)
        speed = qMax(speed, min_speed);
    if (max_speed > 0)
        speed = qMin(speed, max_speed);

    double k = Estepsrev[num-1]->value() / Epitch[num-1]->value() / 1e3;
    set_step_motors_leveling(phys_num, direction, steps * k, speed * k / 16); //to mkm
}

void Motors::goUpE(int num, double speed)
{
    if (num == -1) {
        if (last_command == 2) return goStop();
        last_command = 2;
    }
    if (speed == 0) speed = Espeed->value();
    go(num, 1, speed, 0);
}
void Motors::goUp(int num, double speed, double dist)
{
    if (num == -1) {
        if (last_command == 1) return goStop();
        last_command = 1;
    }
    if (speed == 0) speed = Espeed->value();
    if (dist == 0) dist = Edist->value();
    go(num, 1, speed, dist);
}
void Motors::goDownE(int num, double speed)
{
    if (num == -1) {
        if (last_command == -2) return goStop();
        last_command = -2;
    }
    if (speed == 0) speed = Espeed->value();
    go(num, -1, speed, 0);
}
void Motors::goDown(int num, double speed, double dist)
{
    if (num == -1) {
        if (last_command == -1) return goStop();
        last_command = -1;
    }
    if (speed == 0) speed = Espeed->value();
    if (dist == 0) dist = Edist->value();
    go(num, -1, speed, dist);
}
void Motors::goStop(int num)
{
    last_command = 0;
    go(num, 0, 0, 0);
}

void Motors::goShuttleInt(int speedidx, int num)
{
    goShuttle(speedidx, num);
}
void Motors::goShuttle(double speedidx, int num)
{
    if (num == -1)
        num = active_motor;

    int dir = 0;
    if (speedidx > 0)
        dir = 1;
    else if (speedidx < 0)
        dir = -1;
    double speed = getShuttleSpeed(speedidx, num);
    if (qAbs((int)speedidx) == 2)
        speed = minSpeed->value();
    if (qAbs((int)speedidx) == 3)
        speed = minSpeed->value()*2.0;
    if (qAbs((int)speedidx) == 4)
        speed = minSpeed->value()*3.0;

    if (qAbs((int)speedidx) <= 1) //HACK for fixing, not stopping shuttle
        go(num, 0, 0, 0);
    else
        go(num, dir, speed, 0);
    //qDebug()<<speed;

    int id = mapVirtToPhys(num);
    if (id == 250)
        id = 0;
    if (id >= 0 && id <= 4)
        MobileServer::sendData(DataID::Motor_123 + id, qBound(-5, qRound(speedidx), 5));
}
double Motors::getShuttleSpeed(int speed, int num)
{
    speed = qAbs(speed);
    const int maxShuttleStep = 7;
    double delta = EMaxSpeed[num-1]->value() - EMinSpeed[num-1]->value();

    double newSpeed = EMinSpeed[num-1]->value();
    newSpeed += delta*((double)speed/maxShuttleStep);
    return newSpeed;
}

void Motors::goRoleUp(int role, double speed, double dist)
{
    goUp(getRoleMotor(role), speed, dist);
}
void Motors::goRoleDown(int role, double speed, double dist)
{
    goDown(getRoleMotor(role), speed, dist);
}
void Motors::SetCurrent(int num)
{
    if (sender())
    for (int i = 0; i < 16; i++)
        if (sender() == PBselector[i]) {
            num = i;
            break;
        }

    if (last_command) goStop();

    PBselector[active_motor-1]->setChecked(false);
    active_motor = num+1;
    last_command = 0;
    PBselector[active_motor-1]->setChecked(true);
}
void Motors::FillMUI(int num)
{
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    QString stridx = QString::number(num+1);
    orient[num] = !settings.value("invert_dir/" + stridx, true).toBool();
    CBdir[num]->setChecked(orient[num]);
    CBinv[num]->setChecked(settings.value("invert_ends/" + stridx, false).toBool());
    LErole[num]->setCurrentIndex(settings_double(settings, "role/" + stridx, 0));
    int idx = settings_double(settings, "remap_motor/" + stridx, 0);
    if (idx <= 0) idx = num+1;
    Eidx[num]->setValue(idx);
    EBLdist[num]->setValue(settings_double(settings, "backlash_dist/" + stridx, 0));
    EMinSpeed[num]->setValue(settings_double(settings, "min_speed/" + stridx, 1));
    EMaxSpeed[num]->setValue(settings_double(settings, "max_speed/" + stridx, 500));
    Epitch[num]->setValue(settings_double(settings, "pitch/" + stridx, 0.25));
    Estepsrev[num]->setValue(settings_double(settings, "steps_rev/" + stridx, 3200));
    QSettings settings2("Param/shuttle.conf", QSettings::IniFormat);
    ShuttleType->setCurrentIndex(settings2.value("shuttle_xpress").toBool());
}

void Motors::save_current()
{
    if (blocksave) return;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    for (int i = 0; i < 16; i++) {
        QString stridx = QString::number(i+1);
        settings.setValue("role/" + stridx, LErole[i]->currentIndex());
        settings.setValue("invert_dir/" + stridx, !(orient[i] = CBdir[i]->isChecked()));
        settings.setValue("invert_ends/" + stridx, CBinv[i]->isChecked());
        settings.setValue("remap_motor/" + stridx, Eidx[i]->value());
        settings.setValue("backlash_dist/" + stridx, EBLdist[i]->value());
        settings.setValue("min_speed/" + stridx, EMinSpeed[i]->value());
        settings.setValue("max_speed/" + stridx, EMaxSpeed[i]->value());
        settings.setValue("pitch/" + stridx, Epitch[i]->value());
        settings.setValue("steps_rev/" + stridx, Estepsrev[i]->value());
    }
}

void Motors::save_shuttleType()
{
    if (blocksave) return;
    QSettings settings("Param/shuttle.conf", QSettings::IniFormat);
    settings.setValue("shuttle_xpress", (bool)ShuttleType->currentIndex());
}

void Motors::save_minSpeed()
{
    if (blocksave) return;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    settings.setValue("min_speedShuttle", minSpeed->value());
}

void Motors::save_params()
{
    if (blocksave) return;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);

    settings.setValue("mv_dist", Edist->value());
    settings.setValue("mv_speed", Espeed->value());
    settings.setValue("backlash_speed", EBLspeed->value());

    settings.setValue("motors_count", Enum->value());

    ShowHideRows();
}
void Motors::ShowHideRows()
{
    for (int r = 0; r < TableMotors->rowCount(); r++)
        TableMotors->setRowHidden(r, r >= Enum->value());
    if ((int)Enum->value() != 1)
        TableMotors->setRowHidden(0, true);
}

#include <numeric>
void Motors::init()
{
    motorRst = 0; //wait for rst
    blocksave = true;
    last_command = 0;

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    Espeed->setValue(settings.value("mv_speed", 1000).toDouble());
    Edist->setValue(settings.value("mv_dist", 1000).toDouble());


    int csb_addr = settings_double(settings, "csb_motors_addr", -1);
    if (csb_addr > 0) {
        csb_motor_addr = csb_addr;
        updateCsbMotorAddr = true;
    }

#ifdef _NEW_MOTORS_CHANGES
    if (csb_motor_count[0] != 255) {
#else
    if (csb_motor_count >= 0) {
#endif
        updateCsbMotorCount = true;
    }

    motorRst = 1; //do rst

    for (int i = 0; i < 16; i++) {
        Motor[i] = smStopped;
        Motor_on[i] = false;

        if (i >= Enum->value())
            continue;

        int idx = mapVirtToPhys(i+1) - 1;
        if (idx >= 16)
            continue;

        motor_inv_byte[idx] = settings.value("invert_ends/" + QString::number(i+1), false).toBool();
        update_inv_end[idx] = true;

        double k = Estepsrev[i]->value() / Epitch[i]->value();
        motor_bl_steps[idx] = settings_double(settings, "backlash_dist/" + QString::number(i+1), 0) * k;
        send_bl_steps[idx] = true;
    }

    double k = Estepsrev[3-1]->value() / Epitch[3-1]->value();
    motor_land_safe_calib = qRound(k / 1e3);
    sendLandSafeCalib = true;

    EBLspeed->setValue(settings.value("backlash_speed", 10).toDouble());
    setBLspeed();

    motorStatusLimit = qMax<int>(Enum->value(), 4);
    Slot_Set_Motor_Status();
    blocksave = false;
}

void Motors::setBLspeed()
{
    double k = Estepsrev[0]->value() / 16 / Epitch[0]->value();
    motor_bl_speed = EBLspeed->value() * k;
    sendBLspeed = true;
}

Motors::Motors(QWidget *parent): QWidget(parent)
{
    motor_bl_speed = 0;
    motor_land_safe_calib = 0;
    rawmotoraddr = 0;
    rawmotordata = 0;
    sendLMheight = false;
    csb_motor_addr = 0;
    motorRst = 0;
    last_command = 0;
    reqsent = false;
    blocksave = true;
    active_motor = 1;
    QSettings settings("Param/motor.conf", QSettings::IniFormat);

#ifdef _NEW_MOTORS_CHANGES
    csb_motor_count[0] = settings_double(settings, "csb_motors_count", 255);
    csb_motor_count[1] = settings_double(settings, "csb_motors_count2", 0);
    csb_motor_count[2] = settings_double(settings, "csb_motors_count3", 0);
    csb_motor_count[3] = settings_double(settings, "csb_motors_count4", 0);
#else
    csb_motor_count = settings_double(settings, "csb_motors_count", -1);
#endif

    setWindowTitle(tr("Options"));

    motorStatusLimit = 4;
    SWmotor1 = new StatusWidget1("1234");
    SWmotor1->status->set(StatusWidget::Square, Qt::gray);

    for (int i = 0; i < 16; i++) {
        Motor[i] = smStopped;
        Motor_on[i] = false;
    }

    motorSync = new QTimer(this);
    motorSync->setSingleShot(true);
    connect(motorSync, SIGNAL(timeout()), this, SLOT(Slot_Set_Motor_Status()));


    //QHBoxLayout *mlay = new QHBoxLayout(this);
    //mlay->setSizeConstraint(QLayout::SetMinAndMaxSize);


    //TTabs = new TTabWidget;
    //((TTabWidget *)TTabs)->setAutoHide(false);
    //mlay->addWidget(TTabs);

    mainw = new QWidget(this);
    //TTabs->addTab(mainw, tr("Motors"));

    QHBoxLayout *hMainLay = new QHBoxLayout(mainw);
    QVBoxLayout *lay1 = new QVBoxLayout(); //Containing motor test and table gbox
    hMainLay->addLayout(lay1, 1);

    QVBoxLayout *shutLandLay = new QVBoxLayout();
    hMainLay->addLayout(shutLandLay, 0);
    QGroupBox *shuttleGB = new QGroupBox(tr("Shuttle"));
    ShuttleLay = new QVBoxLayout();
    shuttleGB->setLayout(ShuttleLay);
    QGroupBox *landingGB = new QGroupBox(tr("SPM/Landing"));
    LandingLay = new QHBoxLayout();
    landingGB->setLayout(LandingLay);

    QHBoxLayout *incShuttleTypeLay = new QHBoxLayout();
    QGroupBox *shutTypeGB = new QGroupBox(tr("Shuttle type"));
    QVBoxLayout *ShuttleTypeLay = new QVBoxLayout(shutTypeGB);
    incShuttleTypeLay->addWidget(shutTypeGB, 0, Qt::AlignTop);
    ShuttleType = new QComboBox();
    ShuttleTypeLay->addWidget(ShuttleType, 0);
    ShuttleType->addItem(tr("Shuttle Pro V2"));
    ShuttleType->addItem(tr("Xpress Shuttle"));
    connect(ShuttleType, SIGNAL(currentIndexChanged(int)), this, SLOT(save_shuttleType()));

    ShuttleLay->addWidget(shutTypeGB);
    shutLandLay->addWidget(shuttleGB, 1);


    auto gbFirstSpeed = new QGroupBox(tr("Joystick min speed"));
    auto layFirstSpeed = new QHBoxLayout(gbFirstSpeed);
    ShuttleLay->addWidget(gbFirstSpeed);
    minSpeed = new Edit(1, 1000);
    {
        QSettings settings("Param/motor.conf", QSettings::IniFormat);
        minSpeed->setValue(settings.value("min_speedShuttle", 1).toDouble());
    }
    connect(minSpeed, SIGNAL(newValue()), this, SLOT(save_minSpeed()));
    layFirstSpeed->addWidget(new QLabel(tr("Min speed (mkm/s)")));
    layFirstSpeed->addWidget(minSpeed);

    QHBoxLayout *gbMTLay = new QHBoxLayout;
    QHBoxLayout *inGbMTLay = new QHBoxLayout;
    QVBoxLayout *slayb = new QVBoxLayout;
    QGroupBox *groupMotorTest = new QGroupBox(tr("Motor test"));
    gbMTLay->addWidget(groupMotorTest);
    lay1->addLayout(gbMTLay);
    groupMotorTest->setLayout(inGbMTLay);
    inGbMTLay->addLayout(slayb);


    QString buttonsStyleSheet = "QPushButton {background-color:RGBT(0,0,0,0); border-width:0; margin:0; padding:0; outline: none; border: none; } "
                                "QPushButton:hover { background-color:RGBT(174,79,173,155); border-width:0; margin:0; padding:0; outline: none; border: none;"
                                  "border-color:RGBT(0,0,0,0); }"
                                "QPushButton:pressed { background-color:RGBT(174,79,173,195); border-width:0; margin:0; padding:0; outline: none; border: none;}";


    QHBoxLayout *tuningButsLay = new QHBoxLayout();

    QPushButton *pb;
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/FastUp.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goUpE()));
    tuningButsLay->addWidget(pb);
    pb->setStyleSheet(buttonsStyleSheet);
    pb->setToolTip(tr("Fast up"));
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/Up.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goUp()));
    tuningButsLay->addWidget(pb);
    pb->setStyleSheet(buttonsStyleSheet);
    pb->setToolTip(tr("Up"));
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/approach/landing_done.png"), QString());
    pb->setShortcut(QKeySequence(Qt::Key_Escape));
    connect(pb, SIGNAL(clicked()), this, SLOT(goStop()));
    tuningButsLay->addWidget(pb);
    pb->setStyleSheet(buttonsStyleSheet);
    pb->setToolTip(tr("Stop"));
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/Down.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goDown()));
    tuningButsLay->addWidget(pb);
    pb->setStyleSheet(buttonsStyleSheet);
    pb->setToolTip(tr("Down"));
    pb = new QPushButton(makeHugeIcon(":/SPMTuning_window/FastDown.png"), QString());
    connect(pb, SIGNAL(clicked()), this, SLOT(goDownE()));
    tuningButsLay->addWidget(pb);
    pb->setStyleSheet(buttonsStyleSheet);
    pb->setToolTip(tr("Fast down"));
    slayb->addLayout(tuningButsLay);

    slayb->addWidget(new QLabel(tr("Speed (")+QObject::tr("micro")+tr("m/s)")));
    slayb->addWidget(Espeed = new Edit(1, 10000));
    slayb->addWidget(new QLabel(tr("Distance (")+QObject::tr("micro")+tr("m)")));
    slayb->addWidget(Edist = new Edit(1, 100000));
    Espeed->setImmediateMode();
    Edist->setImmediateMode();
    connect(Espeed, SIGNAL(newValue()), this, SLOT(save_params()));
    connect(Edist, SIGNAL(newValue()), this, SLOT(save_params()));

    slayb->addWidget(new QLabel(tr("Backslash speed(mm/s)") + ": "));
    slayb->addWidget(EBLspeed = new Edit(0.1, 999, 1));
    connect(EBLspeed, SIGNAL(newValue()), this, SLOT(setBLspeed()));




    QHBoxLayout *lay = new QHBoxLayout;
    lay1->addLayout(lay);

    QVBoxLayout *sliderLay = new QVBoxLayout;
    MotorSlider *Mslider = new MotorSlider(true);
    connect(Mslider, SIGNAL(newPos(int)), this, SLOT(goShuttleInt(int)));
    sliderLay->addWidget(Mslider, 1000);
    inGbMTLay->addLayout(sliderLay);

    QWidget* WsettingsLeft = new QWidget();
    lay->addWidget(WsettingsLeft, 2);
    QFormLayout *slay0 = makeFormLayout(WsettingsLeft);
    slay0->setContentsMargins(0, 0, 0, 0);

    QHBoxLayout *gbMTableLay = new QHBoxLayout;
    QVBoxLayout *inGbMTableLay = new QVBoxLayout;
    QGroupBox *groupMotorTable = new QGroupBox(tr("Motor table"));
    groupMotorTable->setLayout(inGbMTableLay);
    gbMTableLay->addWidget(groupMotorTable);
    lay1->addLayout(gbMTableLay, 1);

    //Num motors edit and Save device button
    QHBoxLayout *Lheader = new QHBoxLayout;
    inGbMTableLay->addLayout(Lheader);

    Lheader->addWidget(new QLabel(tr("Active motors: ")));
    Enum = new Edit(1, 16);
    Enum->setImmediateMode();
    Enum->setValue(settings_double(settings, "motors_count", 8));
    Enum->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    connect(Enum, SIGNAL(newValue()), this, SLOT(save_params()));
    Lheader->addWidget(Enum);

#ifdef _NEW_MOTORS_CHANGES
    motorsDebug = new QLabel*[3];
    for (int i=0; i<3; i++)
    {
        motorsDebugNumMot[i] = -1;
        motorsDebug[i] = new QLabel("");
        motorsDebug[i]->setStyleSheet("QLabel {"
                                      "border-style: solid;"
                                      "border-width: 1px;"
                                      "border-color: black; "
                                      "}");
        Lheader->addWidget(motorsDebug[i]);
    }
    motorsDebug[0]->setText(tr("UART1 = none"));
    motorsDebug[1]->setText(tr("UART2 = none"));
    motorsDebug[2]->setText(tr("UART3 = none"));
#endif

    connect(EBLspeed, SIGNAL(newValue()), this, SLOT(save_params()));

    Lheader->addStretch(1);

    BSave = new QPushButton(makeHugeIcon(":/SPMTuning_window/savePribor.png"), QString());
    connect(BSave, SIGNAL(clicked(bool)), this, SLOT(save_device()));
    BSave->hide();
    Lheader->addWidget(BSave, 0, Qt::AlignRight);
    BSave->setStyleSheet(buttonsStyleSheet);
    BSave->setToolTip(tr("Save to device"));

    BClearDevice = new QPushButton(makeHugeIcon(":/SPMTuning_window/clearMemory.png"), QString());
    connect(BClearDevice, SIGNAL(clicked(bool)), this, SLOT(clear_device()));
    BClearDevice->hide();
    BClearDevice->setToolTip(tr("Clear device mem"));
    Lheader->addWidget(BClearDevice, 0, Qt::AlignRight);
    BClearDevice->setStyleSheet(buttonsStyleSheet);

    TableMotors = new QTableWidget(16, 10);
    TableMotors->verticalHeader()->setVisible(false);
    inGbMTableLay->addWidget(TableMotors);

    TableMotors->setDragEnabled(false);
    TableMotors->setFocusPolicy(Qt::NoFocus);
    TableMotors->horizontalHeader()->setClickable(false);
    TableMotors->horizontalHeader()->setMovable(false);
    TableMotors->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
    TableMotors->setSortingEnabled(false);
    TableMotors->setAlternatingRowColors(true);

    lay1->addWidget(landingGB);

    QStringList headList;
    //header
    headList.append(tr(""));
    headList.append(tr("Role"));
    headList.append(tr("Motor#"));
    headList.append(tr("Limit switch mode"));
    headList.append(tr("Direction"));
    headList.append(tr("Min speed (")+QObject::tr("micro")+tr("m/s)"));
    headList.append(tr("Max speed (")+QObject::tr("micro")+tr("m/s)"));
    headList.append(tr("Backlash(mm)"));
    headList.append(tr("Thread pitch(mm)"));
    headList.append(tr("Steps/revolution"));
    TableMotors->setHorizontalHeaderLabels(headList);

    for (int i = 0; i < 16; i++) {
        QString stridx = QString::number(i);
        PBselector[i] = new QRadioButton(stridx);
        LErole[i] = new QComboBox();
        for (int u = 0; u < axes_names_count; u++)
            LErole[i]->addItem(tr(axes_names[u]));

        TableMotors->setCellWidget(i, 0, PBselector[i]);
        TableMotors->setCellWidget(i, 1, LErole[i]);
        //LErole[i]->setStyleSheet("border: 0px solid black;");
        LErole[i]->setStyleSheet("QComboBox {border: 0px solid white;} "
                                 "QComboBox:!editable:on, QComboBox::drop-down:editable:on {border: 1px solid blue;}");

        TableMotors->setCellWidget(i, 2, Eidx[i] = new Edit(1, 16));
        QWidget *CBinvWidget = new QWidget();
        CBinv[i] = new QCheckBox();
        QHBoxLayout *CBinvLayout = new QHBoxLayout(CBinvWidget);
        CBinvLayout->addWidget(CBinv[i]);
        CBinvLayout->setAlignment(Qt::AlignCenter);
        CBinvLayout->setContentsMargins(0,0,0,0);
        TableMotors->setCellWidget(i, 3, CBinvWidget);
        QWidget *CBdirWidget = new QWidget();
        CBdir[i] = new QCheckBox();
        QHBoxLayout *CBdirLayout = new QHBoxLayout(CBdirWidget);
        CBdirLayout->addWidget(CBdir[i]);
        CBdirLayout->setAlignment(Qt::AlignCenter);
        CBdirLayout->setContentsMargins(0,0,0,0);
        TableMotors->setCellWidget(i, 4, CBdirWidget);
        TableMotors->setCellWidget(i, 5, EMinSpeed[i] = new Edit(1, 10000));
        TableMotors->setCellWidget(i, 6, EMaxSpeed[i] = new Edit(1, 10000));
        TableMotors->setCellWidget(i, 7, EBLdist[i] = new Edit(0, 999, 2));
        EBLdist[i]->setSpecialValueText(tr("none"));
        TableMotors->setCellWidget(i, 8, Epitch[i] = new Edit(0.01, 9, 2));
        TableMotors->setCellWidget(i, 9, Estepsrev[i] = new Edit(1, 99000));

        FillMUI(i);

        connect(PBselector[i], SIGNAL(clicked()), this, SLOT(SetCurrent()));
        connect(LErole[i], SIGNAL(currentIndexChanged(int)), this, SLOT(save_current()));
        connect(Eidx[i], SIGNAL(newValue()), this, SLOT(save_current()));
        connect(CBinv[i], SIGNAL(clicked()), this, SLOT(save_current()));
        connect(CBdir[i], SIGNAL(clicked()), this, SLOT(save_current()));
        connect(EMinSpeed[i], SIGNAL(newValue()), this, SLOT(save_current()));
        connect(EMaxSpeed[i], SIGNAL(newValue()), this, SLOT(save_current()));
        connect(EBLdist[i], SIGNAL(newValue()), this, SLOT(save_current()));
        connect(Epitch[i], SIGNAL(newValue()), this, SLOT(save_current()));
        connect(Estepsrev[i], SIGNAL(newValue()), this, SLOT(save_current()));
        connect(Eidx[i], SIGNAL(newValue()), this, SLOT(init()));
        connect(CBinv[i], SIGNAL(clicked()), this, SLOT(init()));
        connect(CBdir[i], SIGNAL(clicked()), this, SLOT(init()));
        connect(EBLdist[i], SIGNAL(newValue()), this, SLOT(init()));
    }
    ShowHideRows();

    lay->addStretch(1);

    int minWidth = TableMotors->horizontalHeader()->length()+
            TableMotors->verticalHeader()->width();
    TableMotors->setMinimumWidth(minWidth);

    TableMotors->setRowHidden(0, true);

    init();
}

void Motors::addLandingSett(QWidget *w)
{
    LandingLay->addWidget(w);
}

void Motors::addShuttleSett(QWidget *w, int stretch)
{
    ShuttleLay->addWidget(w, stretch);
}

int Motors::mapVirtToPhys(int motor)
{
    if (motor == 250)
        return 250;

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    QString stridx = QString::number(motor);

    int idx = settings_double(settings, "remap_motor/" + stridx, 0);
    if (idx <= 0) idx = motor;
    return idx;
}

int Motors::mapPhysToVirt(int motor)
{
    if (motor == 250)
        return 250;

    for (int i = 0; i < Enum->value(); i++) {
        int idx = Eidx[i]->value();
        if (idx <= 0) idx = i + 1;
        if (idx == motor)
            return i + 1;
    }
    return 0 + 1;//todo here an error that all unknown motors are at 1st slot
}

QList<int> Motors::getRoleMotors(int role) //see MotorRoles enum
{
    QList<int> list;
    if (role >= axes_names_count)
        return list;

    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    for (int i = 0; i < settings_double(settings, "motors_count", 8); i++) {
        QString stridx = QString::number(i+1);
        if (role != settings_double(settings, "role/" + stridx, 0)) //todo why its double?
            continue;
        list << (i+1);
    }

    return list;
}
int Motors::getRoleMotor(int role) //see MotorRoles enum
{
    QList<int> l = getRoleMotors(role);
    if (l.size())
        return l.first();

    //default fallbacks
    switch(role) {
    case RoleX1:
        return 1;
    case RoleY1:
        return 2;
    case RoleZ1Land:
        return 250;
    case RoleStageX:
        return 7;
    case RoleStageY:
        return 8;
    case RoleZ2Focus:
        return 4;
    case RoleX3:
    case RoleY3:
    case RoleZ3:
    default:
        return 15;
    };
}

QString motorStr(QString name, bool active, bool atEnd)
{
    QString str;

    str += "<span";
    if (atEnd) {
        str += " style='color:red'";
    } else {
        if (!active)
            str += " style='color:#999999'";
        else
            str += " style='color:green'";
    }
    str += ">";
    str += name;
    str += "</span>";

    return str;
}
bool Motors::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD )
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_STEP_MOTOR_STATUS) {
        EG_StepMotorStatus *motor_satus = (EG_StepMotorStatus *)data;
        int mnumber = motor_satus->MotorN;
        int mstatus = motor_satus->Status;

        if (mnumber >= 16)
            return true;

        Motor[mnumber - 1] = mstatus;
        if (mnumber == active_motor && mstatus == smStopped)
            last_command = 0;

        QMetaObject::invokeMethod(this, "Slot_Set_Motor_Status_handler");

        return true;
    }

    return false;
}
void Motors::Slot_Set_Motor_Status_handler()
{
    if (!motorSync->isActive())
        motorSync->start(100);
}
void Motors::Slot_Set_Motor_Status()
{
    bool atEnd[16], active[16];
    for (int i = 0; i < 16; i++) {
        atEnd[i] = (Motor[i] == smLimitUp) || (Motor[i] == smLimitDown);
        active[i] = !atEnd[i] && ((Motor[i] == smRunningUp) || (Motor[i] == smRunningDown));
        emit motorStatusChanged(mapPhysToVirt(i+1), atEnd[i], active[i]);
        emit motorStatusChangedRealId(i, atEnd[i], active[i]);
    }

    updateMotorStatusWidget(std::min(motorStatusLimit, 16), atEnd, active);

    for (int i = 0; i < 4; i++)
        if (!active[i] && Motor_on[i] != active[i])
            MobileServer::sendData(DataID::Motor_1 + i, 0);
    if ((!active[0] || !active[1] || !active[2]) && (Motor_on[0] != active[0] || Motor_on[1] != active[1] || Motor_on[2] != active[2]))
        MobileServer::sendData(DataID::Motor_123, 0);

    for (int i = 0; i < 16; i++)
        Motor_on[i] = active[i];


    //land stop status
    QList<int> lland = getRoleMotors(Motors::RoleZ1Land);
    if (lland.size()) {
        for (int i = 0; i < lland.size(); i++)
            if (Motor_on[mapVirtToPhys(lland.at(i))-1])
                return;
    } else { //compat
        for (int i = 0; i < 3; i++)
            if (Motor_on[i])
                return;
    }

    emit mland_stopped();
}
void Motors::updateMotorStatusWidget(unsigned count, bool* atEnd, bool* active)
{
    QString status;
    bool deactivated = true;
    for (unsigned i = 0; i < count; i++) {
        int idx = mapVirtToPhys(i+1);

        bool _active = false;
        bool _atEnd = true;
        if (idx <= 16) {
            _active = active[idx-1];
            _atEnd = atEnd[idx-1];
        }

        status += motorStr(QString::number(i+1), _active, _atEnd);
        deactivated = deactivated && !_active;
    }
    SWmotor1->label->setText(status);

    if (deactivated) {
       SWmotor1->status->set(StatusWidget::Square, Qt::gray);
    } else {
        bool dir = false;
        for (unsigned i = 0; i < count; i++) {
            int idx = mapVirtToPhys(i+1);
            if (idx > 16)
                continue;
            dir = dir || (active[idx-1] && (Motor[idx-1] == (orient[i] ? smRunningDown : smRunningUp)));
        }
        SWmotor1->status->set(dir ? StatusWidget::Down : StatusWidget::Up, Qt::green);
    }
}


void Motors::set_raw_motors_control(unsigned char motor, unsigned short control)
{
    if (motor == 250) {
        set_raw_motors_control(1, control);
        set_raw_motors_control(2, control);
        set_raw_motors_control(3, control);
        return;
    }

    motor_cont_byte[motor] = control;
    send_control[motor] = true;
}
void Motors::set_step_motors_leveling(int motor, unsigned char motor_cont_byte_n, double motor_step_n, double motor_speed_n)
{
    if (motor_cont_byte_n == smStop && motor != 250) {
        set_raw_motors_control(motor, motor_cont_byte_n);
        return;
    }
    if (motor == 250)
        motor = 0;

    motor_step[motor] = motor_step_n;
    motor_speed[motor] = motor_speed_n;
    motor_cont_byte[motor] = motor_cont_byte_n;
    send_control[motor] = false;
    updMotor[motor] = true;
}

#include <i_eg3000.h>
bool Motors::sendData(const void *handle)
{
    bool wrote = false;

    if (sendReset.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_CONTROL, 0);
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(2) + REG_CONTROL, 0);
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(3) + REG_CONTROL, 0);
        eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(4) + REG_CONTROL, 0);
    }
    if (motorRst == 1) { //do rst
        motorRst = 2; //rdy
        if (updateCsbMotorAddr.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_CPU, CPU_MOTOR_CSB_ADDR, csb_motor_addr);
        }
        if (updateCsbMotorCount.claimMessage()) {
            wrote = true;
#ifdef _NEW_MOTORS_CHANGES
            eg3000_send_packet(handle, ADR_CPU, CPU_MOTOR_CSB_COUNT, csb_motor_count, 4);
#else
            eg3000_send_word_packet(handle, ADR_CPU, CPU_MOTOR_CSB_COUNT, csb_motor_count);
#endif
        }
    }
    if (motorRst == 2) { //ready
        if (sendBLspeed.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(1) + REG_BL_SPEED, motor_bl_speed); //head
            eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(8) + REG_BL_SPEED, motor_bl_speed); //second controller i2c-motor board
        }
        if (sendLandSafeCalib.claimMessage()) {
            wrote = true;
            eg3000_send_packet(handle, ADR_CPU, CPU_LANDING_MKM, &motor_land_safe_calib, sizeof(DWORD));
        }
        for (int i = 0; i < 16; i++) {
            if (update_inv_end[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS,
                                        SMD_MOTOR_ITOA(i + 1) + REG_INV_END, motor_inv_byte[i]);
            }
            if (send_bl_steps[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(i + 1) + REG_BL_STEPS_H, (motor_bl_steps[i] >> 16) & 0xFFFF);
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS, SMD_MOTOR_ITOA(i + 1) + REG_BL_STEPS_L, motor_bl_steps[i] & 0xFFFF);
            }
        }
        for (int i = 0; i < 16+1; i++) {
            if (updMotor[i].claimMessage()) {
                wrote = true;
                eg3000_set_motor_leveling((void *)handle, (i == 0 ? 250 : i), motor_cont_byte[i], motor_step[i], motor_speed[i]);
            }
            if (send_control[i].claimMessage()) {
                wrote = true;
                eg3000_send_word_packet(handle, ADR_STEP_MOTORS,
                                        SMD_MOTOR_ITOA(i) + REG_CONTROL, motor_cont_byte[i]);
            }
        }
        if (sendCustom.claimMessage()) {
            wrote = true;
            eg3000_send_word_packet(handle, ADR_STEP_MOTORS, rawmotoraddr, rawmotordata);
        }
    }

    return wrote;
}
#include <algorithm>
void Motors::resetWriter()
{
#ifdef _NEW_MOTORS_CHANGES
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    csb_motor_count[0] = settings_double(settings, "csb_motors_count", 255);
    csb_motor_count[1] = settings_double(settings, "csb_motors_count2", 0);
    csb_motor_count[2] = settings_double(settings, "csb_motors_count3", 0);
    csb_motor_count[3] = settings_double(settings, "csb_motors_count4", 0);


    std::fill_n(csb_motor_count_reported, 4, 0);
#endif
    motorRst = 0;
    reqsent = false;
    sendReset = true;

    sendBLspeed = false;
    sendLandSafeCalib = false;
    updateCsbMotorAddr = updateCsbMotorCount = false;
    updMotor[0] = false;
    send_control[0] = false;
    for (int i = 0; i < 16; i++) {
        update_inv_end[i] = false;
        send_bl_steps[i] = false;
        updMotor[i+1] = false;
        send_control[i+1] = false;
    }
    sendCustom = false;

    QMetaObject::invokeMethod(this, "init");
}

#ifdef _NEW_MOTORS_CHANGES

struct fw_flash_motor_cfg_1
{
    union {
        WORD data;
        struct {
            BYTE phys_remap_idx:4;
            BYTE dir_flag:1;
            BYTE inv_flag:1;
            BYTE unused:2;

            BYTE backlash_dist;
        };
    };
};

//do not modify record index
struct fw_flash_motor_cfg_2
{
    union {
        WORD data;
        struct {
            BYTE role:4;
            BYTE unused1:4;

            BYTE record_idx:4; //must be [1..N+1]
            BYTE unused2:4;
        };
    };
};

//do not modify, fw reports this initially
struct fw_flash_header_1
{
    union {
        WORD data;
        struct {
            BYTE csb_motor_count_local:4;
            BYTE csb_motor_addr:4;

            BYTE active_motor_count;
        };
    };
};

//do not modify record index
struct fw_flash_header_2
{
    union {
        WORD data;
        struct {
            BYTE unused1;

            BYTE record_idx:4; //must be 0
            BYTE unused2:4;
        };
    };
};

#endif

#ifdef _NEW_MOTORS_CHANGES

void Motors::capBits(int n, long bits)
{
    if (n == 5+1 || n == 6+1 || n == 7+1) {
        //hw uart channel. these messages may arrive out-of-order(e.g. ch2, ch3, ch1) and sparsely(e.g. ch3 only)
        //usually first connected channel reports(if saved) per-motor configuration for all channels
        int ch = n - (5+1);
        int record_idx = (bits & 0x0F000000) >> 24;
        qDebug() << ch << record_idx << (void *)bits << reqsent;
        int chNum = ch;
        if (chNum != 0 || bits != 0)
        {
            int numMotorsDebug = QString::number(bits, 16).toInt()%50;
            if (motorsDebugNumMot[chNum] == -1)
                motorsDebugNumMot[chNum] = numMotorsDebug;
            if (bits != 0)
                motorsDebug[chNum]->setText(tr("UART") + QString::number(chNum+1) + " = " + QString::number(motorsDebugNumMot[chNum]));
        }
        //parse data from ::save_device() and per-channel default config reporting
        if (record_idx == 0) {
            fw_flash_header_1 hdr1;
            fw_flash_header_2 hdr2;
            hdr1.data = bits & 0xFFFF;
            hdr2.data = bits >> 16;

            bool csbset = hdr1.csb_motor_count_local;
            int reported_count = std::accumulate(csb_motor_count_reported, csb_motor_count_reported + 3, 0);
            bool have_something = csbset || reported_count > 0;
            BSave->setVisible(have_something);
            BClearDevice->setVisible(have_something);
            if (!csbset) { //invalid config
                reqsent = false;
                sendCustom = false;
                return;
            }

            QSettings settings("Param/motor.conf", QSettings::IniFormat);
            settings.setValue("csb_motors_addr", hdr1.csb_motor_addr);

            if (hdr1.active_motor_count != 0)
                Enum->setValue(hdr1.active_motor_count);

            if (csb_motor_count[0] == 255) {
                std::fill_n(csb_motor_count, 4, 0);
            }
            csb_motor_count_reported[ch] = hdr1.csb_motor_count_local;
            csb_motor_count[ch] = csb_motor_count_reported[ch];
            updateCsbMotorCount = true;

            init();
        } else if (record_idx <= std::accumulate(csb_motor_count, csb_motor_count + 3, 0)) {
            fw_flash_motor_cfg_1 mcfg1;
            fw_flash_motor_cfg_2 mcfg2;
            mcfg1.data = bits & 0xFFFF;
            mcfg2.data = bits >> 16;

            CBdir[record_idx-1]->setChecked(mcfg1.dir_flag);
            CBinv[record_idx-1]->setChecked(mcfg1.inv_flag);
            if (mcfg1.phys_remap_idx == 0)
                mcfg1.phys_remap_idx = record_idx;
            Eidx[record_idx-1]->setValue(mcfg1.phys_remap_idx);
            EBLdist[record_idx-1]->setValue(float(mcfg1.backlash_dist) / 100); //to mm
            int axis_id = mcfg2.role;
            if (axis_id >= axes_names_count)
                axis_id = 0;
            LErole[record_idx-1]->setCurrentIndex(axis_id);
            Eidx[record_idx-1]->setNewValue();
            //EBLdist->setNewValue();
        } else { //if (idx > std::accumulate(csb_motor_count, csb_motor_count + 3, 0)) {
            qDebug() << "fixme! unexpected extra motor config! record_idx " << record_idx;
            return;
        }

        if (!reqsent || reqch > ch) {
            reqch = ch; //resend if sent to non-first connected device
            reqsent = true;
            while (sendCustom) {
                qApp->processEvents();
                SleepExport::msleep2(3);
            }

            rawmotoraddr = REG_FLASH_READ + SMD_MOTOR_ITOA(1);
            rawmotordata = 0;
            sendCustom = true;
        } else if (record_idx >= std::accumulate(csb_motor_count, csb_motor_count + 3, 0))
            reqsent = false;
    }
}

void Motors::write_motor_flash_cfg(BYTE idx, WORD data)
{
    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    rawmotoraddr = REG_FLASH_IDX + SMD_MOTOR_ITOA(1);
    rawmotordata = idx;
    sendCustom = true;

    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    rawmotoraddr = REG_FLASH_SET + SMD_MOTOR_ITOA(1);
    rawmotordata = data;
    sendCustom = true;
}

void Motors::save_device(int count_to_save)
{
    int count = Enum->value();
    if (count_to_save >= 0)
            count = count_to_save;

    BYTE first_csb_count = 0;
    for (int i = 0; i < 4; i++)
        if (csb_motor_count_reported[i]) {
            first_csb_count = csb_motor_count_reported[i];
            break;
        }

    //header
    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    fw_flash_header_1 hdr1;
    hdr1.active_motor_count = count;
    hdr1.csb_motor_count_local = first_csb_count;
    hdr1.csb_motor_addr = qMax(0, (int)settings_double(settings, "csb_motors_addr", 5));
    write_motor_flash_cfg(0, hdr1.data);

    fw_flash_header_2 hdr2;
    hdr2.record_idx = 0;
    write_motor_flash_cfg(1, hdr2.data);

    for (int i = 0; i < count; i++) {
        fw_flash_motor_cfg_1 mcfg1;
        mcfg1.phys_remap_idx = Eidx[i]->value();
        mcfg1.dir_flag = CBdir[i]->isChecked();
        mcfg1.inv_flag = CBinv[i]->isChecked();
        mcfg1.backlash_dist = EBLdist[i]->value() * 100;
        write_motor_flash_cfg((i+1)*2, mcfg1.data);

        //high-word
        fw_flash_motor_cfg_2 mcfg2;
        mcfg2.role = LErole[i]->currentIndex();
        mcfg2.record_idx = i + 1;
        write_motor_flash_cfg((i+1)*2 + 1, mcfg2.data);
    }

    // last "break" entry
    write_motor_flash_cfg((count+1)*2, 0xFFFF);
    write_motor_flash_cfg((count+1)*2 + 1, 0xFFFF);

    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    rawmotoraddr = REG_FLASH_WRITE + SMD_MOTOR_ITOA(1);
    rawmotordata = (count+1)*2;
    sendCustom = true;
}

#else

void Motors::capBits(int n, long bits)
{
    if (n == 5+1) {
        int idx = (bits & 0x0F000000) >> 24;
        //qDebug() << idx << (void *)bits << reqsent;
        //parse data from ::save_device()
        if (idx == 0) {
            int csbcount = bits & 0x0F;
            int count = (bits >> 8) & 0x0F;
            BSave->setVisible(csbcount != 0);
            BClearDevice->setVisible(csbcount != 0);
            if (csbcount == 0) {
                reqsent = false;
                sendCustom = false;
                return;
            }

            QSettings settings("Param/motor.conf", QSettings::IniFormat);
            settings.setValue("csb_motors_addr", int(bits & 0xF0) >> 4);

            if (count != 0)
                Enum->setValue(count);
            csb_motor_count = csbcount;
            //init();
        } else if (idx <= csb_motor_count) {
            CBdir[idx-1]->setChecked(bits & 0x10);
            CBinv[idx-1]->setChecked(bits & 0x20);
            int remap = bits & 0xF;
            if (remap == 0) remap = idx;
            Eidx[idx-1]->setValue(remap);
            EBLdist[idx-1]->setValue(float((bits & 0xFF00) >> 8) / 100); //to mm
            int axis_id = (bits & 0x0F0000) >> 16;
            if (axis_id >= axes_names_count) axis_id = 0;
            LErole[idx-1]->setCurrentIndex(axis_id);
            Eidx[idx-1]->setNewValue();
            //EBLdist->setNewValue();
        } else { //if (idx > csb_motor_count) {
            qDebug() << "fixme! excess motor config! idx " << idx;
            return;
        }

        if (!reqsent) {
            reqsent = true;
            while (sendCustom) {
                qApp->processEvents();
                SleepExport::msleep2(3);
            }

            rawmotoraddr = REG_FLASH_READ + SMD_MOTOR_ITOA(1);
            rawmotordata = 0;
            sendCustom = true;
        } else if (idx == csb_motor_count)
            reqsent = false;
    }
}

#define MFLASHCFG_WRITE(_idx, _data) \
{ \
    while (sendCustom) { \
        qApp->processEvents(); \
        SleepExport::msleep2(3); \
    } \
     \
    rawmotoraddr = REG_FLASH_IDX + SMD_MOTOR_ITOA(1); \
    rawmotordata = (_idx); \
    sendCustom = true; \
     \
    while (sendCustom) { \
        qApp->processEvents(); \
        SleepExport::msleep2(3); \
    } \
     \
    rawmotoraddr = REG_FLASH_SET + SMD_MOTOR_ITOA(1); \
    rawmotordata = (_data); \
    sendCustom = true; \
}

void Motors::save_device(int count_to_save)
{
    //TODO bitfield structs for data

    WORD data;
    int count = Enum->value();
    if (count_to_save >= 0)
            count = count_to_save;

    for (int i = 0; i < count; i++) {
        data = Eidx[i]->value() + (BYTE(EBLdist[i]->value() * 100) << 8) +
               (CBdir[i]->isChecked() ? 0x10 : 0) + (CBinv[i]->isChecked() ? 0x20 : 0);
        MFLASHCFG_WRITE((i+1)*2, data);

        //high-word
        data = ((i + 1) << 8) + LErole[i]->currentIndex();
        MFLASHCFG_WRITE((i+1)*2 + 1, data);
    }

    // last "break" entry
    MFLASHCFG_WRITE((count+1)*2, 0xFFFF);


    QSettings settings("Param/motor.conf", QSettings::IniFormat);
    data = (count << 8) + csb_motor_count + (qMax(0, (int)settings_double(settings, "csb_motors_addr", 5) << 4));
    MFLASHCFG_WRITE(0, data);


    while (sendCustom) {
        qApp->processEvents();
        SleepExport::msleep2(3);
    }

    rawmotoraddr = REG_FLASH_WRITE + SMD_MOTOR_ITOA(1);
    rawmotordata = (count+1)*2;
    sendCustom = true;
}

#endif



Motors::~Motors()
{

}

Motors* Motors::GetIt() noexcept
{
    auto motors = ModuleManager::instance->find1VoidPtr("motors");
    return static_cast<Motors*>(motors);
}
