#include "scanplot.h"
#include "defs.h"
#include <QCloseEvent>
#include "gui/status.hpp"
#include "gui/ttabwidget.hpp"
#include "mainwindow.h"
#include "spmtunning.h"
#include "oscilloscope.h"
#include "modules.hpp"
#include "hardware/EG3000.h"
#include <qwt_slider.h>

#include "flowlayout.hpp"
#include "3d/plot_3dplot.h"
#include "math/plot_MathProcessing.h"
#include <qwt_thermo.h>
#include <qwt_plot_marker.h>
#include <QCheckBox>
#include <QSettings>
#include <QTimer>
#include <QFormLayout>
#include <QApplication>
#include <QSplitter>
#include <QProgressBar>
#include <QMessageBox>
#include <QSizePolicy>
#include "gui/rectslider.h"
#include "3d/plot_3dplot.h"
#include "gui/t3dchooser.h"
#include "USBVideo/videoplayer.h"
#include "motors.hpp"
#include "widgets/threedoverlaybuttons.h"
#include "widgets/twodoverlaybuttons.h"
//#include "ccd/ccdscan.cpp"
#include "options_window.h"
#include "ccd/ccdplot.h"

#include <QDebug>

class plot_3dChooser;

static bool is_shown = false;
extern bool skip_pid_error; //autocalib_impl.cpp
void ScanPlot::s_scanError()
{
    static bool skiperr_1 = false;
    if (is_shown || !m_BStart->on_off)
        return;

    auto opt = options::GetIt();
    if (opt->optWin->chIgnoreScanSetPointError->isChecked())
        return;

    s_stopScan();

    if (skiperr_1)
        return;
    is_shown = true;
    skip_pid_error = true;

    int b1 = QMessageBox::warning(nullptr, tr("Controller"),
                                  tr("Scan x/y/tip PID error.") + "\n" +
                                  tr("Can't reach setpoint. Recalibrate sensors or change spm pid precision.") + "\n" +
                                  tr("Do you want to recalibrate sensors and restart scan now?"),
                                  QMessageBox::Yes, QMessageBox::No, QMessageBox::NoToAll);

    skip_pid_error = false;
    skiperr_1 = (QMessageBox::NoToAll == b1);
    if (QMessageBox::Yes != b1) {
        is_shown = false;
        return;
    }

    emit recalibrateSensors(scanner_scan);
}
void ScanPlot::recalibDone(bool ok)
{
    if (!is_shown)
        return;

    if (ok)
        m_BStart->change_state_man_with_sig(true);

    is_shown = false;
}

ScanPlot::~ScanPlot()
{
    for (int i = 0; i < 8; i++) {
        if (ScanDataPart[i])
            free(ScanDataPart[i]);
        if (ScanDataPart_lsb[i])
            free(ScanDataPart_lsb[i]);
    }

    foreach (IExternalScan *ptr, externalScans)
        delete ptr;
}
ScanPlot::ScanPlot(QWidget *parent): QWidget(parent)
{
    scan_lm_h = 0.;
    Step_X = 0.;
    Step_Y = 0.;
    scan_dir_st_point = 0;
    LmEn = false;
    LmHeight = 0.;
    scan_precise_mode = false;
    scanner_scan = 0;
    ccdNX = 0;
    ccdNY = 0;
    ccdMode = 0;
    m_spm_pid_precision = 0;
    m_scan_first_delay = 0;
    m_scan_line_delay = 0;
    m_scan_point_delay = 0;
    m_scan_fast_return = false;
    m_scan_sync_mode = 0;
    DACinitialValue = 0.;
    requested_scan_speed = 0.;
    user_scan_speed = 0.;
    scan_direction = false;
    showingVideoFrame = false;
    DACcurrentNum = 0;
    currExtScan = nullptr;
    isLaserSpotWindowIncreased = false;
    plotPointChanged = false;
    l2k_controller = false;
    l2k_controller_v22 = false;
    for (bool &isOptScan : is_optical_scanner)
        isOptScan = false;
    fast_scan_correction = 0;
    fast_scan_CLZ_disabled = false;
    Autosend = false;
    sectionEnabled = true;
    scanFailed = false;
    scan_bar = nullptr;
    ccd_scan_bar = nullptr;
    scan_on_off = 0;

    Xmax = Ymax = 100;
    Xmax2 = Ymax2 = 100;
    x = y = x2 = y2 = 0;

    Nx_points = Ny_points = 300;
    xLeft = yBottom = 0;
    xRight = yTop = 100.0;
    stPointX = 0;
    stPointY = 0;

    for (int i = 0; i < 4; i++) {
        scan_ch[i] = 0;
        ready[i] = true;
        nlcupd[i] = 0;
        refresh_3d_limit_timer[i].invalidate();
        refresh_limit_last_elapsed[i] = 0;
    }
    channelsActive = 1;
    Ch_activated = 0;
    for (int i = 0; i < 8; i++) {
        ScanDataPart[i] = nullptr;
        ScanDataPart_lsb[i] = nullptr;
    }

    GUICreate();


    plot::ResizeOwner(Section_data, 4);
    const Qt::GlobalColor colormap[4] = {Qt::black, Qt::red, Qt::green, Qt::blue};
    for (int i = 0; i < 4; i++)
        Section_data[i]->Pen.setColor(colormap[i]);

    dataScan_src = dataScan = new plot::T3dDataOwner;
    plot::ResizeOwner(*dataScan_src, 4);

    dataScan_X = new plot::T2dArray;
    dataScan_Xb = new plot::T2dArray;
    dataScan_Y = new plot::T2dArray;
    dataScan_Yb = new plot::T2dArray;

    AutosaveTimer = new QTimer(this);
    AutosaveTimer->setSingleShot(true);
    connect(AutosaveTimer, SIGNAL(timeout()), this, SLOT(slAutosaveData()));
    connect(AutosaveTimer, SIGNAL(timeout()), this, SLOT(sl_dacscan_advance()));

    RefreshTimer4d = new QTimer(this);
    RefreshTimer4d->setSingleShot(true);
    connect(RefreshTimer4d, SIGNAL(timeout()), this, SLOT(sl_ext_scan_update()));

    setWindowTitle(tr("Scan window - NSpec"));

    QMetaObject::connectSlotsByName(this);

    installEventFilter(this);

    connect(Plot, SIGNAL(AddDataToContainer(plot::T3dDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::T3dDataOwner,plot::TMetaStorage)));
    connect(Plot, SIGNAL(AddDataToContainer(plot::TCurvesDataOwner,plot::TMetaStorage)), SLOT(AddDataToContainer_slot(plot::TCurvesDataOwner,plot::TMetaStorage)));

    connect(Plot->TwoDPlot->Pure2dPlot, SIGNAL(MathUpdated()), this, SLOT(Send2dPlotData()));
    connect(Plot->ThreeDPlots.at(0), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData1()));
    connect(Plot->ThreeDPlots.at(1), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData2()));
    connect(Plot->ThreeDPlots.at(2), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData3()));
    connect(Plot->ThreeDPlots.at(3), SIGNAL(MathUpdated()), this, SLOT(Send3dPlotData4()));

    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        bool head_on = settings.value("display/oscilloscope/head", true).toBool();
        bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
        if (!head_on || !stage_on) {
            CBScanner->setCurrentIndex(head_on ? 0 : 1);
            while (scanControlWidgets.contains(CBScanner))
                scanControlWidgets.remove(scanControlWidgets.indexOf(CBScanner));
            CBScanner->hide();
            
            #ifdef _OLD_EXTRA_CONTROL
            LScanner->hide();
#endif
        }
        CBScanner->setEnabled(head_on && stage_on);
    }

    /*if (!show_additional_gui() && !show_advanced_gui()) {
        connect(CBStartPos, SIGNAL(currentIndexChanged(int)), this, SLOT(sl_CB_filters_update()));
        for (int i = 0; i < 4; i++)
            connect(CBChannel[i], SIGNAL(currentIndexChanged(int)), this, SLOT(sl_CB_filters_update()));
    }*/

    CBChannel[0]->setCurrentIndex(1);
    CBSec[0]->setChecked(true);

    updateDirection();
    ritem = new plot::ImageItem;

#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (settings.value("usb", false).toBool()) {
        QTimer *t = new QTimer;
        connect(t, SIGNAL(timeout()), this, SLOT(injectTestData()));
        t->start(10);
    }
#endif

    videoChannelAxisScaleXMax[0] = videoChannelAxisScaleXMax[1] = Xmax * 1e-6;
    videoChannelAxisScaleYMax[0] = videoChannelAxisScaleYMax[1] = Ymax * 1e-6;
}


void ScanPlot::restoreAttributes(const plot::TMetaStorage &attrib)
{
    if (attrib.contains("Velocity, Hz")) {
        rectSlFbp->setValue(attrib.value("Velocity, Hz").toDouble());
    }
    if (attrib.contains("Lift height, nm")) {
        LEd_LFmodeHeight->setValue(attrib.value("Lift height, nm").toDouble());
        LEd_LFmodeHeight->setNewValue(true);
    }
}

void ScanPlot::FillDataAttributes2()
{
    scan_attributes.insert("Lift mode", cbLiftMode->isChecked());
    if (cbLiftMode->isChecked())
        scan_attributes.insert("Lift height, nm", LEd_LFmodeHeight->text());

    scan_attributes.insert("Velocity, Hz", rectSlFbp->Value());

    if (is_dacscan()) {
        auto osc = Oscilloscope::GetIt();
        scan_attributes.insert("DAC(Z)scan channel", CBDACscanCh->currentText());
        if (osc)
            scan_attributes.insert("DAC(Z)scan value", osc->DacsWidget->get_dacs_channel(CBDACscanCh)->value());
    }

    if (auto spm = SPMTunning::GetIt())
        spm->appendAttrib(scan_attributes);
}
void ScanPlot::FillDataAttributes()
{
    QVariant gid_prev = scan_attributes.value("GroupId");

    scan_attributes.clear();

    QString date = TContainer::getDateHash();
    scan_attributes.insert("Date", date);

    if (is_dacscan() && DACcurrentNum > 0)
        scan_attributes.insert("GroupId", gid_prev);
    else {
        quint16 gid = qChecksum(date.toStdString().data(), date.length());
        scan_attributes.insert("GroupId", QString::number(gid, 16).toUpper());
    }

    scan_attributes.insert("Source", "Scan - " + CBScanner->currentText());

    scan_attributes.insert("Ch A", CBChannel[0]->currentText());
    scan_attributes.insert("Ch B", CBChannel[1]->currentText());
    scan_attributes.insert("Ch C", CBChannel[2]->currentText());
    scan_attributes.insert("Ch D", CBChannel[3]->currentText());

    if (cbLiftMode->isChecked()) {
        scan_attributes.insert("Ch A lift", LmSwitch[0]->isChecked());
        scan_attributes.insert("Ch B lift", LmSwitch[1]->isChecked());
        scan_attributes.insert("Ch C lift", LmSwitch[2]->isChecked());
        scan_attributes.insert("Ch D lift", LmSwitch[3]->isChecked());
    }

    scan_attributes.insert(QString("Scan width, ")+QChar(181)+QString("m"), LEdDeltaX->value());
    scan_attributes.insert(QString("Scan height, ")+QChar(181)+QString("m"), LEdDeltaY->value());

    scan_attributes.insert("Scan x points", CBPointsOX->currentText());
    scan_attributes.insert("Scan y points", CBPointsOY->currentText());
    scan_attributes.insert("Scan direction", scan_dir_st_point);

    auto osc = Oscilloscope::GetIt();
    if (!osc)
        return;
    if (CBScanner->currentIndex() == scStage) {
        scan_attributes.insert(QString("Head X, ")+QChar(181)+QString("m"), osc->DacsWidget->X->value());
        scan_attributes.insert(QString("Head Y, ")+QChar(181)+QString("m"), osc->DacsWidget->Y->value());
    } else {
        scan_attributes.insert(QString("Stage X, ")+QChar(181)+QString("m"), osc->DacsWidget->X2->value());
        scan_attributes.insert(QString("Stage Y, ")+QChar(181)+QString("m"), osc->DacsWidget->Y2->value());
    }

    scan_attributes.insert("XY PID", (CBScanner->currentIndex() == scHead ? osc->CalibWidget->CL_status_xy : osc->CalibWidget->CL2_status_xy));
}

void ScanPlot::AddDataToContainer_slot(plot::TCurvesDataOwner src, plot::TMetaStorage attributes)
{
    attributes.insert("Name", tr("Scan section"));

    QString date = TContainer::getDateHash();
    quint16 id = qChecksum((date + attributes.value("Name").toString() + "Scan").toStdString().data(), (date + attributes.value("Name").toString() + "Scan").length());
    attributes.insert("Id", QString::number(id, 16).toUpper());

    int j = 0;
    while (j < Plot->ThreeDPlots.size() && !Plot->ThreeDPlots.at(j)->BCrossectionIsChecked)
        j++;

    if (j < Plot->ThreeDPlots.size()) {
        QString date = scan_attributes.value("Date").toString();

        QString str = date;
        if (j == videoChannelId && showingVideoFrame)
            str += CBChannel[videoChannelId]->currentText();
        else
            str += plotLegends[j];
        str += "Scan";
        quint16 gid = qChecksum(str.toStdString().data(), str.length());
        attributes.insert("Parent", QString::number(gid, 16).toUpper());
    }

    attributes.insert("GroupId", scan_attributes.value("GroupId"));

    attributes.insert("Source", "Scan");

    if (j < Plot->ThreeDPlots.size()) {
		auto plot = Plot->ThreeDPlots.at(Plot->ActivePlot);
        attributes.insert("Marker A x", plot->MarkerA->xValue()*1e6);
        attributes.insert("Marker A y", plot->MarkerA->yValue()*1e6);

        attributes.insert("Marker B x", plot->MarkerB->xValue()*1e6);
        attributes.insert("Marker B y", plot->MarkerB->yValue()*1e6);
    }

    emit AddDataToContainer(src, attributes);

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
    {
        qDebug()<<"CLICKED---------------------------";
        mob->setPropIdVar(DataID::addedScanDataToContainer_2d, QVariant::fromValue(attributes));
    }
}

void ScanPlot::AddDataToContainer_slot(plot::T3dDataOwner src, plot::TMetaStorage attributes)
{
    int plotId = Plot->ActivePlot;
    FillContainer(plotId, &src, &attributes);

    emit AddDataToContainer(src, attributes);
}

void ScanPlot::FillContainer(int plotId, plot::T3dDataOwner *src, plot::TMetaStorage *attributes)
{
    QString name = plotLegends[plotId];
    attributes->insert("Name", name);

    QString date = scan_attributes.value("Date").toString();

    quint16 id = qChecksum((date + name + "Scan").toStdString().data(), (date + name + "Scan").length());
    attributes->insert("Id", QString::number(id, 16).toUpper());

    attributes->unite(scan_attributes);
}

void ScanPlot::sl_CB_update()
{
    if (sender() == CBChannel[videoChannelId])
    {
        ritem->detach();
        if (CBChannel[videoChannelId]->currentIndex() == 0)
        {
            BMakeScreen->setEnabled(true);
        }
        else
        {
            if (showingVideoFrame)
                BMakeScreen->click();
            BMakeScreen->setEnabled(false);
        }
    }
    for (int i = 0; i < numScanChannel; i++) {

        if (i == videoChannelId && showingVideoFrame)
            plotLegends[i] = comboVideoProfile->currentText();
        else
            plotLegends[i] = CBChannel[i]->currentText();
        QString str = "";
        if (!ratisMode)
            str = getLegendNameChannel(i) + plotLegends[i];
        str += getLegendCurrPoint();
        Plot->SetTitle(str, i);

        int idx = CBChannel[i]->itemData(CBChannel[i]->currentIndex()).toInt();

        QString units;
        if (i == videoChannelId && currExtScan)
            units = currExtScan->attributes.value("ZUnit", QObject::tr("Counts")).toString();
        else
            units = input_units(idx);

        Plot->SetZDescAndSIUnit(QString(), units, i);
    }
    Plot->UpdateMinMaxInSlider();
    update2DPlotScanMode();
}
QString ScanPlot::getLegendNameChannel(int i)
{
    QString str = tr("Channel") + " ";

    if (i == 0)
        str += tr("A");
    else if (i == 1)
        str += tr("B");
    else if (i == 2)
        str += tr("C");
    else
        str += tr("D");

    str += ": ";

    return str;
}
void ScanPlot::sl_CB_filters_update()
{
    if (sender() == CBChannel[videoChannelId])
    {
        if (CBChannel[videoChannelId]->currentIndex() == 0)
        {
            BMakeScreen->setEnabled(true);
        }
        else
        {
            if (showingVideoFrame)
                BMakeScreen->click();
            BMakeScreen->setEnabled(false);
        }
    }
    for (int i = 0; i < numScanChannel; i++) {
        int idx = CBChannel[i]->itemData(CBChannel[i]->currentIndex()).toInt();
        if (!idx)
            continue;

        bool want_math = false;
        if (idx == srcHeight || idx == srcDACZ || idx == srcDACZ2 || idx == srcDACZ3
                || idx == srcCZ || idx == srcCZ2 || idx == srcCZ3)
            want_math = true;

        bool ydir = (CBStartPos->currentIndex() % 2 != 0);

        if (!want_math || Plot->ThreeDPlots.at(i)->CurrentSubtrOrder < 1) {
            Plot->ThreeDPlots[i]->CurrentSubtrOrder = (want_math ? 2 : 0) - 1;
            Plot->ThreeDPlots[i]->Refresh();
            if (Plot->ActivePlot == i)
                Plot->PlotControl->AGSubtr->actions()[want_math ? 2 : 0]->trigger();
        }

        Plot->ThreeDPlots[i]->MakeCorrectLineByMedianaXDir(want_math && !ydir);
        Plot->ThreeDPlots[i]->MakeCorrectLineByMedianaYDir(want_math && ydir);
        if (Plot->ActivePlot == i) {
            Plot->PlotControl->MMathProcessing->actions()[0]->setChecked(want_math && !ydir);
            Plot->PlotControl->MMathProcessing->actions()[1]->setChecked(want_math && ydir);
        }
    }
}

void ScanPlot::sl_set_section_enabled(bool on)
{
    sectionEnabled = on;
    Plot->TwoDPlot->Pure2dPlot->MathProcessing->Work[plot::math::TMathProcessing::Scale2d].arg = sectionEnabled ? Ch_activated : -1;
    Plot->TwoDPlotVertical->Pure2dPlot->MathProcessing->Work[plot::math::TMathProcessing::Scale2d].arg = sectionEnabled ? Ch_activated : -1;
    SetSectionUnits();

//    Plot->TwoDPlot->Pure2dPlot->SourceData.clear();
//    Plot->TwoDPlotVertical->Pure2dPlot->SourceData.clear();
//    if (sectionEnabled)
//    {
//        Plot->TwoDPlot->Pure2dPlot->SourceData.resize(4); //TODO show?
//        Plot->TwoDPlotVertical->Pure2dPlot->SourceData.resize(4);
//    }
}

void ScanPlot::sl_section_legend(int value)
{
    Ch_activated = (unsigned int)value;
    Plot->TwoDPlot->Pure2dPlot->MathProcessing->Work[plot::math::TMathProcessing::Scale2d].arg = sectionEnabled ? Ch_activated : -1;
    Plot->TwoDPlotVertical->Pure2dPlot->MathProcessing->Work[plot::math::TMathProcessing::Scale2d].arg = sectionEnabled ? Ch_activated : -1;
    SetSectionUnits();

    char literChannel = 'A' + value;
    auto title = tr("Active channel:") + " " + literChannel;
    Plot->TwoDPlot->SetTitle(title);
    Plot->TwoDPlotVertical->SetTitle(title);

    static bool prevStates[numScanChannel];
    static bool prevChannelCCD = false;
    if (CBChannel[value]->currentData() == srcCCD)
    {
        for (int i =0; i<numScanChannel; i++)
        {
            prevStates[i] = CBSec[i]->isChecked();
            CBSec[i]->setChecked(i == value);
        }
        prevChannelCCD = true;
    }
    else if (prevChannelCCD)
    {
        int idCCDChannel = -1;
        for (int i =0; i<numScanChannel; i++)
            if (CBChannel[i]->currentData() == srcCCD)
                idCCDChannel = i;

        if (idCCDChannel != -1)
        {
            for (int i =0; i<numScanChannel; i++)
                CBSec[i]->setChecked(prevStates[i]);
            CBSec[idCCDChannel]->setChecked(false);
        }
        prevChannelCCD = false;
    }
}

void ScanPlot::SetSectionUnits()
{
    int idx = CBChannel[Ch_activated]->itemData(CBChannel[Ch_activated]->currentIndex()).toInt();

    QString units = input_units(idx);
    double maxLim = 100000;
    if (idx == srcCCD)
        units = QObject::tr("Counts");
    else if (idx == srcHeight)
        maxLim = 1;
    else
        maxLim = 50;

    if (currExtScan && (!sectionEnabled || Ch_activated == 1)) {
        Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(currExtScan->attributes.value("XLabel").toString(), currExtScan->attributes.value("XUnit").toString());
        Plot->TwoDPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(currExtScan->attributes.value("YLabel").toString(), currExtScan->attributes.value("YUnit").toString());
        Plot->TwoDPlot->Pure2dPlot->SetDefaultYLimits(-100000, 100000);

        Plot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(currExtScan->attributes.value("XLabel").toString(), currExtScan->attributes.value("XUnit").toString());
        Plot->TwoDPlotVertical->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(currExtScan->attributes.value("YLabel").toString(), currExtScan->attributes.value("YUnit").toString());
        Plot->TwoDPlotVertical->Pure2dPlot->SetDefaultXLimits(-100000, 100000);
        return;
    }
    Plot->TwoDPlot->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QString(), units);
    Plot->TwoDPlot->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
    Plot->TwoDPlot->Pure2dPlot->SetDefaultYLimits(-maxLim, maxLim);

    Plot->TwoDPlotVertical->Pure2dPlot->XAxisUnit.SetDescAndSIUnit(QString(), units);
    Plot->TwoDPlotVertical->Pure2dPlot->YAxisUnit.SetDescAndSIUnit(QString(), tr("m"));
    Plot->TwoDPlotVertical->Pure2dPlot->SetDefaultXLimits(-maxLim, maxLim);
}

void ScanPlot::multSpmPidStageChanged(int item)
{
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    settings.setValue("mult_spm_pid_stage", velocityMulfactorValues[item]);

    auto spm = SPMTunning::GetIt();
    if (spm)
    {
        spm->sendPID = true;
    }
}
QString ScanPlot::getLegendCurrPoint()
{
    if (!plotPointChanged)
        return "";

    static int firstTime = 0;
    if (ratisMode && firstTime < 10)
    {
        Plot->ThreeDPlots.first()->ResizeQPlot();
        Plot->on_BRestoreScale_clicked();
        firstTime++;
    }
    QString str = "      pos: ";
    str += "x = " + QString::number(LEdX->value(), 'g', 6);
    str += ", ";
    str += "y = " + QString::number(LEdY->value(), 'g', 6);
    return str;
}
void ScanPlot::materialThinChangedSave(int index)
{
    comboMaterialThin->setCurrentIndex(index);
    videoChannelMaterialThin->setCurrentIndex(index);
    QSettings settings("Param/scan.conf", QSettings::IniFormat);
    settings.setValue("MaterialThinCBIndex", comboMaterialThin->currentIndex());
}
void ScanPlot::scan_direction_changed(swapDir::Direction dir)
{
    bool right = (dir == swapDir::Direction::dir_right);
    for (int i=0; i<numScanChannel; i++)
    {
        Plot->ThreeDPlots[i]->xScanning = (right ? 1:0);
    }
    CBStartPos->setCurrentIndex(right ? 0:1); //right
    auto options = options::GetIt();
    options->optWin->saveScanFilterSettings();
    m_BStart->changeDirection(dir, false);

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::scanDirectionChanged, dir);
}

void ScanPlot::zLandButtonSysOff_motorEvent(int mZLanding, bool /*atEnd*/, bool active)
{
    auto m = Motors::GetIt();
    //auto spm = SPMTunning::GetIt();
    if (mZLanding != m->getRoleMotor(Motors::MotorRoles::RoleZ1Land)) return;

    if (active) return;

    smartLandingBtn1->setChecked(false);
    qDebug()<<"Stopped";
}
void ScanPlot::zLandButtonSysOffHeadLanding_motorEvent(int mRealId, bool atEnd, bool active)
{
    auto m = Motors::GetIt();
    //auto spm = SPMTunning::GetIt();
    auto zLands = m->getRoleMotors(Motors::MotorRoles::RoleZ1Land);
    int zMotor = m->mapPhysToVirt(mRealId+1);
    if (!zLands.contains(zMotor))
        return;

    zLandingMotorsStopped.replace(zLands.indexOf(zMotor), !active);
    if (zLandCountStoppedZLandingMotorsStopped() >= numZLandingMotors)
        smartLandingBtn1->setChecked(false);
}
void ScanPlot::zLandButtonSysOff(bool pressed)
{
    smartLandingBtn2->setEnabled(!pressed);
    auto m = Motors::GetIt();
    auto spm = SPMTunning::GetIt();
    auto mZLanding = m->getRoleMotor(Motors::MotorRoles::RoleZ1Land);
    if (pressed)
    {
        zLandClearZLandingMotorsStopped();
        if (spm->OptionsForMotor->headLanding)
        {
            auto zLands = m->getRoleMotors(Motors::MotorRoles::RoleZ1Land);
            for (int i=0; i<zLands.count(); i++)
            {
                connect(m, SIGNAL(motorStatusChangedRealId(int, bool, bool)), this, SLOT(zLandButtonSysOffHeadLanding_motorEvent(int, bool, bool)));
                m->goUpE(zLands.at(i), 1000);
            }
        }
        else
        {
            connect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(zLandButtonSysOff_motorEvent(int, bool, bool)));
            m->goDownE(mZLanding, 1000); //todo take speed from MAX speed table for that motor
        }
    }
    else
    {
        if (spm->OptionsForMotor->headLanding)
        {
            disconnect(m, SIGNAL(motorStatusChangedRealId(int, bool, bool)), this, SLOT(zLandButtonSysOffHeadLanding_motorEvent(int, bool, bool)));
            zLandingOff_stage = 0;
            auto zLands = m->getRoleMotors(Motors::MotorRoles::RoleZ1Land);
            for (int i=0; i<zLands.count(); i++)
                m->goStop(zLands.at(i));
        }
        else
        {
            disconnect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(zLandButtonSysOff_motorEvent(int, bool, bool)));
            m->goStop(mZLanding);
        }
    }

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::approachSOff, pressed);
}
void ScanPlot::zLandButtonSysOn_motorEvent(int mZLanding, bool atEnd, bool active)
{
    auto m = Motors::GetIt();
    auto spm = SPMTunning::GetIt();
    if (mZLanding != m->getRoleMotor(Motors::MotorRoles::RoleZ1Land)) return;

    qDebug()<<zLandingOff_stage<<atEnd<<active;
    if (zLandingOff_stage == 0)
    {
        if (active) return;

//        if (spm && spm->OptionsForMotor->sysOnMAF)
//        {
//            zLandingWaitsMAF = true;
//            disconnect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(zLandButtonSysOn_motorEvent(int, bool, bool)));
//            auto v = VideoPlayer::GetIt();
//            v->autoFocusStage = 0;
//            videoChannelAutoFocusButtonFast->change_state_man_with_sig(true);
//            return;
//        }
//        else
            zLandingOff_stage++;
    }
    if (zLandingOff_stage == 1)
    {
        zLandingWaitsMAF = false;
        if (active) return;

        zLandingOff_stage++;
        int dist = spm->OptionsForMotor->dist_ready - comboMaterialThin->itemData(comboMaterialThin->currentIndex()).toDouble()*1000;
        qDebug()<<"dist"<<dist;
        m->goUp(mZLanding, 1000, dist);
        return;
    }

    if (zLandingOff_stage == 2)
    {
        if (active) return;

        qDebug()<<"We at the end of race";
//        auto v = VideoPlayer::GetIt();
//        auto spm = SPMTunning::GetIt();
//        if (spm && spm->OptionsForMotor->sysOnMAF && (v->mafPrcsEnabled->isChecked() || v->mafFastEnabled->isChecked()))
//        {
//            videoChannelAutoFocusUpDownBtn->setEnabled(true);
//            videoChannelAutoFocusUpDownBtn->setVisible(videoChannelAutoFocusButtonPrec->isVisible());
//            videoChannelAutoFocusUpDownBtn->change_state(2);
//        }
        auto v = VideoPlayer::GetIt();
        auto spm = SPMTunning::GetIt();
        if (spm && spm->OptionsForMotor->sysOnMAF)
        {
            SleepExport::msleep_gui(100);
            spm->BLaserOnOff->change_state_man_with_sig(true);
            videoChannelAutoFocusButtonFast->change_state_man_with_sig(true);
            /*
            videoChannelAutoFocusButtonFast->change_state_man(true);
            videoChannelAutoFocusButtonPrec->setEnabled(false);
            videoChannelAutoFocusButtonPAF->setEnabled(false);
            BAutoFocus2->setEnabled(false);
            v->autoFocusDirection = 0;
            if (videoChannelAutoFocusUpDownBtn->isVisible())
            {
                videoChannelAutoFocusUpDownBtn->setEnabled(false);
                videoChannelAutoFocusUpDownBtn->setVisible(false);
            }
            if (!videoChannelPlayCamera->isOn)
                videoChannelPlayCamera->change_state_man_with_sig(true);
            videoChannelAutoFocusPrevMode = v->autoFocusScanVariant;
            v->autoFocusMode->setCurrentIndex(1);

            v->autoFocusDist1 = v->autoFocusDist1 - comboMaterialThin->itemData(comboMaterialThin->currentIndex()).toDouble()*1000 - v->tipThinkbness;
            BAutoFocus->change_state_man_with_sig(true);*/
        }
        smartLandingBtn2->setChecked(false);
        destructStatus(zLandingStatusBar);
        return;
    }
}
void ScanPlot::zLandClearZLandingMotorsStopped()
{
    if (zLandingMotorsStopped.count() < numZLandingMotors)
    {
        for (int i=0; i<numZLandingMotors; i++)
            zLandingMotorsStopped.append(false);
    }
    else
        std::fill(zLandingMotorsStopped.begin(), zLandingMotorsStopped.end(), false);
}
int ScanPlot::zLandCountStoppedZLandingMotorsStopped()
{
    int n = 0;
    for (int i=0; i<zLandingMotorsStopped.count(); i++)
        if (zLandingMotorsStopped.at(i))
            n++;
    return n;
}
void ScanPlot::zLandButtonSysOnHeadLanding_motorEvent(int mRealId, bool atEnd, bool active)
{
    auto m = Motors::GetIt();
    auto spm = SPMTunning::GetIt();
    auto zLands = m->getRoleMotors(Motors::MotorRoles::RoleZ1Land);
    int zMotor = m->mapPhysToVirt(mRealId+1);
    if (!zLands.contains(zMotor))
        return;

    qDebug()<<zLandingOff_stage<<atEnd<<active<<mRealId<<"("<<m->mapPhysToVirt(mRealId+1);

    if (zLandingOff_stage == -1) //HACK we need really sure that they on the end
    {
        zLandingMotorsStopped.replace(zLands.indexOf(zMotor), !active);
        if (zLandCountStoppedZLandingMotorsStopped() < numZLandingMotors)
            return;

        zLandClearZLandingMotorsStopped();
        zLandingOff_stage++;
        for (int i=0; i<zLands.count(); i++)
        {
            m->goUpE(zLands.at(i), 1000);
        }
        return;
    }

    if (zLandingOff_stage == 0)
    {
        zLandingMotorsStopped.replace(zLands.indexOf(zMotor), !active);
        if (zLandCountStoppedZLandingMotorsStopped() < numZLandingMotors)
            return;

        SleepExport::msleep_gui(100);

        if (!zLandingStatusBar.isNull())
            zLandingStatusBar->setValue(1);
        zLandClearZLandingMotorsStopped();
        zLandingOff_stage++;
        int dist = spm->OptionsForMotor->headn_land_exit_dist;
        qDebug()<<dist;
        if (dist > 0.0)
        {
            for (int i=0; i<zLands.count(); i++)
            {
                m->goDown(zLands.at(i), 1000, dist);
            }
            return;
        }
        else {
            zLandingMotorsStopped.replace(0, true);
            zLandingMotorsStopped.replace(1, true);
            zLandingMotorsStopped.replace(2, true);
        }
    }
    if (zLandingOff_stage == 1)
    {
        zLandingMotorsStopped.replace(zLands.indexOf(zMotor), !active);
        if (zLandCountStoppedZLandingMotorsStopped() < numZLandingMotors)
            return;

        SleepExport::msleep_gui(100);

        if (!zLandingStatusBar.isNull())
            zLandingStatusBar->setValue(2);
        zLandClearZLandingMotorsStopped();
        zLandingOff_stage++;
        double dist[3];
        dist[1] = spm->OptionsForMotor->head_land_offset_dist_1;
        dist[2] = spm->OptionsForMotor->head_land_offset_dist_3;
        for (int i=0; i<zLands.count(); i++)
        {
            if (i == 0 || dist[i] == 0.0)
            {
                qDebug()<<i<<zLands.at(i);
                zLandingMotorsStopped.replace(i, true);
                continue;
            }

            zLandingMotorsStopped.replace(i, false);
            if (dist[i] > 0.0)
                m->goUp(zLands.at(i), 1000, dist[i]);
            else
                m->goDown(zLands.at(i), 1000, -dist[i]);
            qDebug()<<zLands.at(i)<<dist[i];
        }
        if (zLandCountStoppedZLandingMotorsStopped() < numZLandingMotors)
            return;
    }
    if (zLandingOff_stage == 2)
    {
        zLandingMotorsStopped.replace(zLands.indexOf(zMotor), !active);
        if (zLandCountStoppedZLandingMotorsStopped() < numZLandingMotors)
            return;

        SleepExport::msleep_gui(100);

        if (!zLandingStatusBar.isNull())
            zLandingStatusBar->setValue(3);
        zLandClearZLandingMotorsStopped();
        zLandingOff_stage++;
        int dist = spm->OptionsForMotor->dist_ready - comboMaterialThin->itemData(comboMaterialThin->currentIndex()).toDouble()*1000 - spm->OptionsForMotor->headn_land_exit_dist;
        qDebug()<<"dist"<<dist;

        for (int i=0; i<zLands.count(); i++)
        {
            m->goDown(zLands.at(i), 1000, dist);
        }

        return;
    }
    if (zLandingOff_stage == 3)
    {
        zLandingMotorsStopped.replace(zLands.indexOf(zMotor), !active);
        if (zLandCountStoppedZLandingMotorsStopped() < numZLandingMotors)
            return;

        destructStatus(zLandingStatusBar);
        zLandClearZLandingMotorsStopped();
        smartLandingBtn2->setChecked(false);

        return;
    }
}
void ScanPlot::cancelZLandingSysOn()
{
    smartLandingBtn2->setChecked(false);
}
void ScanPlot::zLandButtonSysOn(bool pressed)
{
    smartLandingBtn1->setEnabled(!pressed);
    auto m = Motors::GetIt();
    auto spm = SPMTunning::GetIt();
    auto mZLanding = m->getRoleMotor(Motors::MotorRoles::RoleZ1Land);
    if (pressed)
    {
        zLandClearZLandingMotorsStopped();
        zLandingOff_stage = 0;
        if (spm->OptionsForMotor->headLanding)
        {
            zLandingStatusBar = new StatusBar(tr("Head landing"), 4);
            connect(zLandingStatusBar, SIGNAL(canceled()), SLOT(cancelZLandingSysOn()));
            zLandingStatusBar->add();
            zLandingOff_stage = -1;
            auto zLands = m->getRoleMotors(Motors::MotorRoles::RoleZ1Land);
            connect(m, SIGNAL(motorStatusChangedRealId(int, bool, bool)), this, SLOT(zLandButtonSysOnHeadLanding_motorEvent(int, bool, bool)));
            for (int i=0; i<zLands.count(); i++)
            {
                m->goUpE(zLands.at(i), 1000);
            }
        }
        else
        {
            connect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(zLandButtonSysOn_motorEvent(int, bool, bool)));
            m->goDownE(mZLanding, 1000);
        }
    }
    else
    {
        qDebug()<<"OFFEED";
        destructStatus(zLandingStatusBar);
        if (spm->OptionsForMotor->headLanding)
        {
            disconnect(m, SIGNAL(motorStatusChangedRealId(int, bool, bool)), this, SLOT(zLandButtonSysOnHeadLanding_motorEvent(int, bool, bool)));
            zLandingOff_stage = 0;
            auto zLands = m->getRoleMotors(Motors::MotorRoles::RoleZ1Land);
            for (int i=0; i<zLands.count(); i++)
                m->goStop(zLands.at(i));
        }
        else
        {
            disconnect(m, SIGNAL(motorStatusChanged(int, bool, bool)), this, SLOT(zLandButtonSysOn_motorEvent(int, bool, bool)));
            zLandingOff_stage = 1;
            m->goStop(mZLanding);
        }
    }

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::approachSOn, pressed);
}

void ScanPlot::Plot3DTabChanged(int tabId)
{
    Plot->blockVisibility2DPlot = false;
    static int prevId = 0;
    int devTabId = Plot->mainTab->indexOf(scrollAreaSchemeWid);
    if (tabId == devTabId)
    {
        Plot->blockVisibility2DPlot = true;
        Plot->SplitPlotTab->setSizes(QList<int>() << 1 << 0);
        auto osc = Oscilloscope::GetIt();

        if (osc)
        {
            if (!osc->start->isOn)
                osc->start->click();

            osc->CBChannel[0]->setCurrentIndex(osc->CBChannel[0]->findData(srcMag1));
            osc->CBChannel[1]->setCurrentIndex(osc->CBChannel[1]->findData(srcMagSin1));
            osc->CBChannel[2]->setCurrentIndex(osc->CBChannel[2]->findData(srcMagCos1));
            osc->CBChannel[3]->setCurrentIndex(osc->CBChannel[3]->findData(srcPhase1));
        }
    }
    else if (prevId == devTabId)
    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        int first = settings.value("scan/height_3dplot", 663).toInt();
        int second = settings.value("scan/height_2dplot", 250).toInt();
        Plot->SplitPlotTab->setSizes(QList<int>() << first << second);
    }
    prevId = tabId;
}

void ScanPlot::setFBoffon(bool checked)
{
    dynamic_cast<QPushButton*>(rectSlFbi->centralWidget)->setText(checked ? tr("FB on") : tr("FB off"));
}
void ScanPlot::setAutoTipPush(bool checked)
{
    rectAmplitude->enableControl = !checked;
    rectAmplitude->update();
    dynamic_cast<QPushButton*>(rectAmplitude->centralWidget)->setText(checked ? tr("auto") : tr("mnl"));

    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    settings.setValue("scan/auto_tip_push", checked);

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::rectTipDriveAutoControll, checked);
    
    if (checked)
    {
        auto spm = SPMTunning::GetIt();
        if (spm)
            spm->findMag();
    }
}

void ScanPlot::plotsChanged()
{
    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    QList<QVariant> order = QList<QVariant>();
    for (int i : *Plot->orderChannel)
    {
        order.append(i);
    }
    settings.setValue("scan/plot_order", order);
}

void ScanPlot::selectByRect_clicked()
{
    int n = Plot->ThreeDPlots.indexOf((plot::TPure3dPlot*)Plot->overlayButtons->BBackButton->parent());
    for (int i=0; i<Plot->ThreeDPlots.count(); i++)
        Plot->overlayButtons->visibilityMode[i] = plot::ThreeDOverlayButtons::svm_nothing_show;
    Plot->overlayButtons->visibilityMode[n] = plot::ThreeDOverlayButtons::svm_area_selection;
    Plot->SetActivePlot(n);
    #ifdef _SCAN_SLICE
    //scan_slice();
    auto opt = options::GetIt();
    bool enableSlice = opt->optWin->chEnableScanSlice->isChecked();
//    if (enableSlice)
//    {
//        for (int i=0; i<Plot->ThreeDPlots.count(); i++)
//        {
//            //Plot->ThreeDPlots[i]->sliceBounds = false;
//            //Plot->ThreeDPlots[i]->Enabled = false;
//        }
//    }
#endif

    if (selection_notification)
        Plot->PlotControl->BCancel->click();
    else
    {
        selection_notification = new Notification(tr("Press enter or esc buttons"), Notification::None, 9999);
    }

    Plot->overlayButtons->showingBackButtonPlot[n] = selection_notification;
    Plot->overlayButtons->showingApplyButtonPlot[n] = selection_notification;
    Plot->PlotControl->BAreaZzoom->change_state_man_with_sig(selection_notification);
    if (selection_notification)
        Plot->overlayButtons->scanButtonsSetVisibleGUI(false);
    Plot->overlayButtons->blockResizeTimes = 28;
    Plot->overlayButtons->scanButtonsSetVisibleGUI(true, Plot->ThreeDPlots[n]);

    Plot->PlotControl->BSelectByRect->setHidden(false);
    Plot->on_BSelectByRect_clicked();

    #ifdef _SCAN_SLICE
    if (enableSlice)
    {
        SleepExport::msleep_gui(100);
        if (copyX)
        {
            auto xr = getScanSliceRangeX();
            auto yr = getScanSliceRangeY();
            Plot->ThreeDPlots[n]->CurrentXLimits.arr[0] = xr->arr[0];
            Plot->ThreeDPlots[n]->CurrentXLimits.arr[1] = xr->arr[1];
            Plot->ThreeDPlots[n]->CurrentYLimits.arr[0] = yr->arr[0];
            Plot->ThreeDPlots[n]->CurrentYLimits.arr[1] = yr->arr[1];
            Plot->ThreeDPlots[n]->PlotsXLimits.arr[0] = xr->arr[0];
            Plot->ThreeDPlots[n]->PlotsXLimits.arr[1] = xr->arr[1];
            Plot->ThreeDPlots[n]->PlotsYLimits.arr[0] = yr->arr[0];
            Plot->ThreeDPlots[n]->PlotsYLimits.arr[1] = yr->arr[1];
        }
        for (int i=0; i<Plot->ThreeDPlots.count(); i++)
        {
            Plot->ThreeDPlots[i]->ZoomPlot(QPointF(0,0), -999999);

        }
        Plot->MakeCrossection();
    }
#endif
}
void ScanPlot::cancel_clicked(plot::ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    
}
void ScanPlot::cancelApply_clicked(plot::ThreeDOverlayButtons::SpecialVisibilityMode visibilityMode)
{
    int n = Plot->ThreeDPlots.indexOf((plot::TPure3dPlot*)Plot->overlayButtons->BBackButton->parent());
    Plot->SetActivePlot(n);
    auto opt = options::GetIt();

    destructStatus(selection_notification);
    if (showingVideoFrame)
    {
        updateVideoChannel();
    }
    //scanButtonsSetVisibleGUI(true, Plot->ThreeDPlots[n]);
    for (int i=0; i<Plot->ThreeDPlots.size(); i++)
    {
        if (Plot->overlayButtons->visibilityMode[i] == plot::ThreeDOverlayButtons::svm_nothing_show)
            Plot->overlayButtons->visibilityMode[i] = plot::ThreeDOverlayButtons::svm_none;
    }
    if (Plot->overlayButtons->BBackButton->isVisible())
    {
        Plot->overlayButtons->showingBackButtonPlot[n] = false;
        Plot->overlayButtons->showingApplyButtonPlot[n] = false;
        Plot->overlayButtons->BBackButton->hide();
        Plot->overlayButtons->BApplyButton->hide();
        Plot->overlayButtons->blockShowing3DPlotButtons = false;
        for (int i=0; i<Plot->ThreeDPlots.size(); i++)
            Plot->ThreeDPlots.at(i)->AreaZzoom = false;
        Plot->overlayButtons->scanButtonsSetVisibleGUI(true, Plot->ThreeDPlots[n]);
    }
    if (Plot->TwoDPlot->overlayButtons->B2DBack->isVisible()) //todo move from that class to his
        Plot->TwoDPlot->overlayButtons->B2DBack->click();

    #ifdef _SCAN_SLICE
    bool enableSlice = opt->optWin->chEnableScanSlice->isChecked();
    if (enableSlice)
    {
        for (int i=0; i<Plot->ThreeDPlots.count(); i++)
        {
            Plot->ThreeDPlots[i]->sliceBounds = false;
        }
    }
#endif

    if (
        #ifndef _OLD_EXTRA_CONTROL
        visibilityMode == plot::ThreeDOverlayButtons::svm_extra_control ||
        #endif
        visibilityMode == plot::ThreeDOverlayButtons::svm_palettes)
    {
        ChannelSetVisibleGUI(n, false, plot::ThreeDOverlayButtons::svm_none);
        Plot->overlayButtons->setVisibleIfUnderMouse(n);
    }

    if (!copyX)
    {
        copyX = new Range();
        copyMaxX = new Range();
    }
    SleepExport::msleep_gui(100);
    copyX->arr[0] = Plot->ThreeDPlots[n]->CurrentXLimits.arr[0];
    copyX->arr[1] = Plot->ThreeDPlots[n]->CurrentXLimits.arr[1];
    copyMaxX->arr[0] = Plot->ThreeDPlots[n]->PlotsXLimits.arr[0];
    copyMaxX->arr[1] = Plot->ThreeDPlots[n]->PlotsXLimits.arr[1];
}
void ScanPlot::feedBackClickedInSPM(bool val)
{
    spmMonitorFeedback->change_state_man(val);
    dynamic_cast<QPushButton*>(rectSlFbi->centralWidget)->setChecked(val);
}
void ScanPlot::tipDriveNewValue_FromSlider(double val)
{
    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::rectTipDrive, val);

    auto spm = SPMTunning::GetIt();
    if (!spm)
        return;

    double maxTipDriveAmplitude = 3.0;
    {
        QSettings settings("Param/tipMonitor.conf", QSettings::IniFormat);
        maxTipDriveAmplitude = settings_double(settings, "max_tip_drive_amplitude", 3.0);
    }

    deviceSchemeWid->oscilator->edAmp->setValue((val/100.0)*maxTipDriveAmplitude);
    spm->driveInt->setValue((val/100.0)*maxTipDriveAmplitude);
    spm->driveInt->setNewValue(true);
}
void ScanPlot::tipDriveNewValue(double val)
{
    double maxTipDriveAmplitude = 3.0;
    {
        QSettings settings("Param/tipMonitor.conf", QSettings::IniFormat);
        maxTipDriveAmplitude = settings_double(settings, "max_tip_drive_amplitude", 3.0);
    }

    rectAmplitude->setValue_man((val/maxTipDriveAmplitude)*100.0);
}

void ScanPlot::spm_SmartLandingButtonChangedMan(int oldSt, int newSt)
{
    if (oldSt == 1 && newSt == 2)
        Plot->setScanSPM2DPlot(false);
    else
        updateScan2DPlotVisible();

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::smartLandingChanged, newSt);
}
void ScanPlot::spm_SmartLandingButtonChanged(int oldSt, int newSt)
{
    auto spm = SPMTunning::GetIt();
    if (!spm)
        return;

    if (oldSt == 3 && newSt == 1)
    {
        spm->slot_DoMoveDown();
    }
    else if (oldSt == 1 && newSt == 2)
    {
        spm->scan_StartAutoLanding();
    }
    else if (oldSt == 2 && newSt == 3)
    {
        spm->slot_DoLanding(false);
    }
}

void ScanPlot::DCfbPNewValue(double val)
{
    DCfbPPerc->setValue(val*10.0);
    rectSlVelocity->setValue_man(val*10.0);
    DCfbPPerc->setNewValue(false);
}
void ScanPlot::DCfbPNewValue_emit(double val)
{
    DCfbPPerc->setValue(val*10.0);
    rectSlVelocity->setValue_man(val*10.0);
    DCfbPPerc->setNewValue(true);
}
void ScanPlot::DCfbPPercNewValue(double val)
{
    auto spm = SPMTunning::GetIt();
    if (!spm)
        return;

    spm->DCfbP->setValue(spm->DCfbP->maximum() * ((double)val/100.0));
    spm->DCfbP->setNewValue(true);
    deviceSchemeWid->pid->edPr->setValue(spm->DCfbP->maximum() * ((double)val/100.0));
}

void ScanPlot::rectSliderSlFbiChanged(double value)
{
    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
        mob->setPropIdVar(DataID::rectTipPush, value);

    auto spm = SPMTunning::GetIt();
    if (!spm)
        return;

    spm->DCfbSetPoint->setValue(oldMag - (value/100.0)*oldMag);
    spm->DCfbSetPoint->setNewValue(true);

    deviceSchemeWid->setPointEdit->setValue(oldMag - (value/100.0)*oldMag);
    deviceSchemeWid->setPointEdit->setNewValue(true);
}
//void ScanPlot::DCfbSetPointChanged(double value)
//{
//    rectSlFbi->setValue_man(((oldMag-value)*100.0)/oldMag);
//}
void ScanPlot::FeedBackChangedUpdateMag(bool val)
{
    if (val)
    {
        auto spm = SPMTunning::GetIt();
        if (!spm)
            return;

        oldMag = spm->nMAG;
        rectSliderSlFbiChanged(rectSlFbi->Value());
//        DCfbSetPointChanged(spm->DCfbSetPoint->value());
    }
}
void ScanPlot::setScanner(int)
{
    {
        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        settings.setValue("scanner_head_stage", CBScanner->currentIndex());
    }

    for (int i = 0; i < Plot->ThreeDPlots.size(); i++)
        Plot->ThreeDPlots[i]->SourceData = plot::getUnLock(scanData[CBScanner->currentIndex()][i]);
    Plot->Refresh();

    sl_set_XYZ_range_sync();

    if (CBScanner->currentIndex() == scHead) {
        LEdX->setValue(x);
        LEdY->setValue(y);
    } else {
        LEdX->setValue(x2);
        LEdY->setValue(y2);
    }
    sl_syncXY();

    check_speed();
}
void ScanPlot::updateDirection()
{
    bool prev_scan_direction = scan_direction;
    scan_direction = (CBStartPos->currentIndex() % 2 != 0); //0-fast axis x, 1-y

    if (prev_scan_direction == scan_direction)
        return;

    double active_range0 = !prev_scan_direction ? xRight - xLeft : yTop - yBottom;
    double active_range1 = !scan_direction ? xRight - xLeft : yTop - yBottom;
    if (active_range1 > 0) {
        user_scan_speed *= active_range0 / active_range1; //preserve tip speed
        if (rectSlFbp->Value() != user_scan_speed)
            rectSlFbp->setValue(user_scan_speed);
    }

    check_speed();
}

void ScanPlot::LmOnOff(bool value)
{
    for (auto &ls : LmSwitch)
        ls->setVisible(value);
    LblLiftMode->setVisible(value);
    LEd_LFmodeHeight->setVisible(value);
}


void ScanPlot::acquireNextPoint()
{
    if (!m_BStart->on_off || CBExternalMode->currentIndex() == 0)
        return;

    if (!currExtScan->acquireNextPoint()) {
        new Notification(tr("External scan") + ": " + tr("Scan error:") + " " + currExtScan->lastError, Notification::Bad, 60);
        doBeep();

        m_BStart->change_state_man_with_sig(false);
    }
}
void ScanPlot::sl_ext_scan_acquired()
{
    if (!m_BStart->on_off || CBExternalMode->currentIndex() == 0 || sender() != currExtScan)
        return;

    sendExtReady = true;

    if (ccd_scan_bar != nullptr)
        ccd_scan_bar->setValue(ccd_scan_bar->bar->value() + 1);

    sl_ext_scan_update();
}
void ScanPlot::sl_ext_scan_update()
{
    if (!refresh_3d_limit_timer[1].isValid() || refresh_3d_limit_timer[1].elapsed() > 100 + refresh_limit_last_elapsed[1]) { //~10 refreshes per second max
        refresh_3d_limit_timer[1].start();
        RefreshTimer4d->stop();

        currExtScan->updateData();
        refresh_limit_last_elapsed[1] = qMax<qint64>(refresh_limit_last_elapsed[1], refresh_3d_limit_timer[1].elapsed() / 0.8 * channelsActive); //time budget, %
    } else if (sender() != RefreshTimer4d && !RefreshTimer4d->isActive())
        RefreshTimer4d->start(200);
}
void ScanPlot::sl_ext_scan_refresh()
{
    if (CBExternalMode->currentIndex() == 0)
        return;

    Plot->SetZDescAndSIUnit(QString(), currExtScan->attributes.value("ZUnit", QObject::tr("Counts")).toString(), 1);

    SetSectionUnits();
    sl_scan_refresh_ch(1);
    sl_scan_refresh_ch(2);
    sl_scan_refresh_ch(3);
    sl_scan_refresh_ch(4);
}
void ScanPlot::sl_ext_mode()
{
    IExternalScan *scan = CBExternalMode->currentIndex() != 0 ?
                externalScans.at(CBExternalMode->currentIndex() - 1) : nullptr;
    if (currExtScan != scan) {
        if (currExtScan)
            currExtScan->clear();

        currExtScan = scan;
        Plot->TwoDPlot->overlayButtons->set4DScan(currExtScan ? currExtScan->name : "");

        CBChannel[1]->setCurrentIndex(CBChannel[1]->findData(currExtScan ? currExtScan->getDataChannel() : srcNone));
        //CBChannelB->setEnabled(!currExtScan); //TODO onoff sync

        for (int i = 0; i < externalScanWidgets.size(); i++)
        {
            bool visible = (i + 1) == CBExternalMode->currentIndex();
            for (auto w : *externalScanWidgets.at(i))
                w->setVisible(visible);
        }

        //LeftScroll->setWidget(LeftScroll->takeWidget()); //reset sizeHint
        //LeftScroll->updateGeometry(); //notify layout
    }

    //xy sync
    CBPointsOX->blockSignals(true);
    CBPointsOY->blockSignals(true);

    Set_CBPoints(CBPointsOX);
    Set_CBPoints(CBPointsOY);

    CBPointsOX->setCurrentIndex(CBPointsOXext->currentIndex());
    CBPointsOY->setCurrentIndex(CBPointsOYext->currentIndex());

    int nx = CBPointsOX->currentText().toInt();
    int ny = CBPointsOY->currentText().toInt();
    MobileServer::sendData(DataID::ScanNxNy, qSqrt(nx*ny));

    if (currExtScan && !BCCDXYLock->isOn) {
        int count = 0;
        do{
            if (CBPointsOX->itemText(count).toInt() % nx != 0)
                CBPointsOX->removeItem(count);
            else
                count++;
        } while (count < CBPointsOX->count());

        count = 0;
        do {
            if (CBPointsOY->itemText(count).toInt() % ny != 0)
                CBPointsOY->removeItem(count);
            else
                count++;
        } while (count < CBPointsOY->count());
    }

    CBPointsOX->blockSignals(false);
    CBPointsOY->blockSignals(false);

    if (BXYLock->isOn) {
        if (sender() == CBPointsOXext)
            CBPointsOYext->setCurrentIndex(CBPointsOXext->currentIndex());
        if (sender() == CBPointsOYext)
            CBPointsOXext->setCurrentIndex(CBPointsOYext->currentIndex());
    }

    gb4DButtons->setVisible(CBExternalMode->currentIndex() != 0 && show4dMenu);
    for (auto eScan : externalScans)
        eScan->delayInit();

    // restore big icon size
    smartLanding->setIconSize(QSize(64, 64));
}

void ScanPlot::addExternalScan(const QString &name, IExternalScan *iface)
{
    CBExternalMode->addItem(name);

    iface->displayPlot = Plot;

    externalScans.append(iface);
    externalScanWidgets.append(iface->createInterfaceWidget(lay4D, lay4DButtons));
    connect(iface, SIGNAL(acquired()), this, SLOT(sl_ext_scan_acquired()));
    connect(iface, SIGNAL(updated()), this, SLOT(sl_ext_scan_refresh()));
    connect(iface, SIGNAL(setSectionEnabled(bool)), this, SLOT(sl_set_section_enabled(bool)));

    if (externalScanWidgets.last()) {
        for (auto w : *externalScanWidgets.last())
        {
            w->hide();
        }
    }
}


bool ScanPlot::is_dacscan()
{
    return cbContScan->isChecked() &&
            CBDACscanCh->currentIndex() > 0;
}
void ScanPlot::sl_dacscan(bool toggled)
{
    DACcurrentNum = 0;

    auto osc = Oscilloscope::GetIt();
    if (!osc)
        return;

    CBDACscanCh->setVisible(toggled);
    EDACstep->setVisible(toggled);
    EDACcount->setVisible(toggled);

    CBDACscanCh->clear();
    CBDACscanCh->addItem(tr("None "), -1);
    osc->DacsWidget->fill_dacs_channel_CB(CBDACscanCh);
}
void ScanPlot::sl_dacscan_advance()
{
    if (!cbContScan->isChecked()) {
        DACcurrentNum = 0;
        return;
    }

    if (CBDACscanCh->currentIndex() <= 0)
        return;

    auto osc = Oscilloscope::GetIt();
    if (!osc) return;
    DoubleChanger *dc = osc->DacsWidget->get_dacs_channel(CBDACscanCh);
    if (DACcurrentNum == 0)
        DACinitialValue = dc->value();

    bool last = DACcurrentNum == EDACcount->value()-1;
    if (last) {
        dc->setValue(DACinitialValue);
        DACcurrentNum = 0;
    } else {
        dc->setValue(dc->value() + EDACstep->value());
        DACcurrentNum++;
    }
    dc->setNewValue(true);

    if (!last)
        m_BStart->change_state_man_with_sig(true);
}


void ScanPlot::onLEdDeltaX_editingFinished()
{
    Plot->SetSelectWidth(LEdDeltaX->value() / 1e6);
    if (BXYLock->isOn) {
        LEdDeltaY->setValue(LEdDeltaX->value());
        Plot->SetSelectHeight(LEdDeltaY->value() / 1e6);
    }

    auto mob = MobileServer::GetIt();
    if (mob && mob->need_send())
    {
        auto plot = Plot->ThreeDPlots[0];
        QList<QVariant> list;
        list << CBPointsOX->currentIndex();
        list << LEdDeltaX->value();
        list << LEdDeltaX->isEnabled();
        list << plot->XYAxisUnit.Legend;
        list << plot->XYAxisUnit.Multiplier;
        mob->setPropIdVar(DataID::plot3DChangeSize, list);
    }
}
void ScanPlot::onLEdDeltaY_editingFinished()
{
    Plot->SetSelectHeight(LEdDeltaY->value() / 1e6);
    if (BXYLock->isOn) {
        LEdDeltaX->setValue(LEdDeltaY->value());
        Plot->SetSelectWidth(LEdDeltaX->value() / 1e6);
    }
}

void ScanPlot::sendXYPos()
{
    if (CBScanner->currentIndex() == scHead)
        emit sig_XYPosition(LEdX->value(), LEdY->value());
    else
        emit sig_XYPosition2(LEdX->value(), LEdY->value());
}
void ScanPlot::sl_setCurrentX(double v)
{
    x = v;
    if (CBScanner->currentIndex() == scHead) {
        LEdX->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_setCurrentY(double v)
{
    y = v;
    if (CBScanner->currentIndex() == scHead) {
        LEdY->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_setCurrentX2(double v)
{
    x2 = v;
    if (CBScanner->currentIndex() == scStage) {
        LEdX->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_setCurrentY2(double v)
{
    y2 = v;
    if (CBScanner->currentIndex() == scStage) {
        LEdY->setValue(v);
        sl_syncXY();
    }
}
void ScanPlot::sl_syncXY()
{
    plotPointChanged = true;
    int active = Plot->ActivePlot;
    for (int i = 0; i < numScanChannel; i++) {
        Plot->ActivePlot = i;
        Plot->SetCrossCoords(LEdX->value() / 1e6, LEdY->value() / 1e6);
        QString str = "";
        if (!ratisMode)
            str = getLegendNameChannel(i) + plotLegends[i];
        str +=getLegendCurrPoint();
        Plot->SetTitle(str, i);
    }
    Plot->ActivePlot = active;
}
void ScanPlot::sl_Update_XYZ_pos()
{
    sendXYPos();
    sl_syncXY();
}
void ScanPlot::slUpdZscanner(double zpos)
{
    if (zpos < 0.33) Zposition->setFillBrush(Qt::red);
    else if (zpos > 0.66) Zposition->setFillBrush(Qt::green);
    else Zposition->setFillBrush(Qt::blue);
    Zposition->setValue(zpos);
    ZposValue->setText(tr("Z pos = ")+(QString::number((int)(100*zpos),'f',0))+tr("%"));
}

void ScanPlot::sl_set_XYZ_range(double xm, double ym, double)
{
    bool xyu = qFuzzyCompare(Xmax, xm) &&
            qFuzzyCompare(Ymax, ym);

    Xmax = xm;
    Ymax = ym;

    if (CBScanner->currentIndex() != scHead)
        return;

    if (!xyu)
        sl_set_XYZ_range_sync();

    check_speed();
}
void ScanPlot::sl_set_XYZ_range2(double xm, double ym, double)
{
    bool xyu = qFuzzyCompare(Xmax2, xm) &&
            qFuzzyCompare(Ymax2, ym);

    Xmax2 = xm;
    Ymax2 = ym;

    if (CBScanner->currentIndex() != scStage)
        return;

    if (!xyu)
        sl_set_XYZ_range_sync();

    check_speed();
}
void ScanPlot::sl_set_XYZ_range_sync()
{
    double _xLeft = xLeft / LEdX->maximum();
    double _xRight = xRight / LEdX->maximum();
    double _yTop = yTop / LEdY->maximum();
    double _yBottom = yBottom / LEdY->maximum();

    double xm = Xmax, ym = Ymax;
    if (CBScanner->currentIndex() == scStage) {
        xm = Xmax2; ym = Ymax2;
    }

    if (_xRight - _xLeft > 0.8 && _yTop - _yBottom > 0.8) {
        _xLeft = _yBottom = 0;
        _xRight = _yTop = 1;
    }
    _xLeft *= xm;
    _xRight *= xm;
    _yTop *= ym;
    _yBottom *= ym;

    LEdDeltaX->setMaximum(xm);
    LEdDeltaY->setMaximum(ym);
    LEdX->setMaximum(xm);
    LEdY->setMaximum(ym);


    Plot->SetAllXScales(0, xm * 1e-6);
    Plot->SetAllYScales(0, ym * 1e-6);
    for (int i = 0; i < numScanChannel; i++) {
        //if (i != videoChannelId || !showingVideoFrame)
        {
            Plot->ThreeDPlots[i]->SetPlotsXScales(_xLeft * 1e-6, _xRight * 1e-6);
            Plot->ThreeDPlots[i]->SetPlotsYScales(_yBottom * 1e-6, _yTop * 1e-6);
        }
        Plot->ThreeDPlots[i]->ResizeQPlot();
        Plot->ThreeDPlots[i]->PlotWasChanged();

//        if (i == videoChannelId)
//        {
//            videoChannelAxisScaleXMin = _xLeft * 1e-6;
//            videoChannelAxisScaleXMax[0] = _xRight * 1e-6;
//            videoChannelAxisScaleYMin = _yBottom * 1e-6;
//            videoChannelAxisScaleYMax[0] = _yTop * 1e-6;
//        }
    }
    Plot->Refresh();
    qApp->processEvents(); //flush event filter

    Plot->TwoDPlot->Pure2dPlot->SetDefaultXLimits(0, qMax(xm, ym) * 1e-6);
    Plot->TwoDPlotVertical->Pure2dPlot->SetDefaultYLimits(0, qMax(xm, ym) * 1e-6);

    xLeft = _xLeft;
    xRight = _xRight;
    yTop = _yTop;
    yBottom = _yBottom;
    LEdDeltaX->setValue(xRight - xLeft);
    LEdDeltaY->setValue(yTop - yBottom);

    check_speed();
}

void ScanPlot::sendXYsync(double x, double y, int sc)
{
    if (sc == scHead) {
        emit sync_scan_x(x);
        emit sync_scan_y(y);
    } else {
        emit sync_scan_x2(x);
        emit sync_scan_y2(y);
    }
}

void ScanPlot::lowVoltageModeChanged(bool pressed)
{
    auto osc = Oscilloscope::GetIt();
    if (!osc) return;

    auto scaner = osc->CalibWidget->Scanner_type->currentIndex();
    if (scaner < 2)
    {
        osc->CalibWidget->CBlvm[scaner]->setChecked(pressed);
        osc->CalibWidget->lvmClicked();
    }
}

void ScanPlot::setLowVoltageMode(bool pressed)
{
    #ifdef _OLD_EXTRA_CONTROL
    lowVoltageModeButton->change_state_man(pressed);
    Plot->overlayButtons->setButtonActive(lowVoltageModeButton, pressed);
#endif
}

void ScanPlot::setScannerLowVoltageMode(int scaner)
{
    auto osc = Oscilloscope::GetIt();
    if (!osc) return;

    setLowVoltageMode(osc->CalibWidget->CBlvm[scaner]->isChecked());
}

void ScanPlot::Send2dPlotData()
{
    send2dPlot(PlotID::ScanSection, Plot->TwoDPlot);
    //send2dPlot(PlotID::ScanSection, Plot->TwoDPlotVertical);
}
void ScanPlot::Send3dPlotData1()
{
    send3dPlot(PlotID::Scan1, Plot, 0);
    if (refresh_limit_need_math_update[0]) {
        refresh_limit_need_math_update[0] = false;
        refresh_limit_last_elapsed[0] = refresh_3d_limit_timer[0].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}
void ScanPlot::Send3dPlotData2()
{
    send3dPlot(PlotID::Scan2, Plot, 1, true, false, CBChannel[1]->currentData() == srcCCD);
    if (refresh_limit_need_math_update[1]) {
        refresh_limit_need_math_update[1] = false;
        refresh_limit_last_elapsed[1] = refresh_3d_limit_timer[1].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}
void ScanPlot::Send3dPlotData3()
{
    send3dPlot(PlotID::Scan3, Plot, 2);
    if (refresh_limit_need_math_update[2]) {
        refresh_limit_need_math_update[2] = false;
        refresh_limit_last_elapsed[2] = refresh_3d_limit_timer[2].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}
void ScanPlot::Send3dPlotData4()
{
    send3dPlot(PlotID::Scan4, Plot, 3);
    if (refresh_limit_need_math_update[3]) {
        refresh_limit_need_math_update[3] = false;
        refresh_limit_last_elapsed[3] = refresh_3d_limit_timer[3].elapsed() / 0.8 * channelsActive; //time budget, %
    }
}

#include <qwt_plot_canvas.h>
void ScanPlot::BSendToGraphEditor_clicked()
{
    emit SendBackground(QPixmap::grabWidget(Plot->ThreeDPlots.at(Plot->ActivePlot)->QPlot->canvas()).toImage());

//    if (Plot->ActivePlot == videoChannelId)
//        emit SendRect(plot::ConstructRectF(Range(videoChannelAxisScaleXMin, videoChannelAxisScaleXMax),
//                                    Range(videoChannelAxisScaleYMin, videoChannelAxisScaleYMax)));
//    else
        emit SendRect(plot::ConstructRectF(Plot->ThreeDPlots.at(Plot->ActivePlot)->CurrentXLimits,
                                    Plot->ThreeDPlots.at(Plot->ActivePlot)->CurrentYLimits));
}


void ScanPlot::scan_first_delay(double time)
{
    m_scan_first_delay = time*1000.;
    updateFirstDelay = true;
}
void ScanPlot::scan_line_delay(double time)
{
    m_scan_line_delay = time*1000.;
    updateLineDelay = true;
}
void ScanPlot::scan_point_delay(double time)
{
    m_scan_point_delay = time*1000.*10.;
    updatePointDelay = true;
}
void ScanPlot::scan_fast_return(bool on)
{
    m_scan_fast_return = on;
    updateFastReturn = true;
}
void ScanPlot::scan_sync_mode(int mode)
{
    m_scan_sync_mode = mode;
    updateSyncMode = true;
}
void ScanPlot::spm_pid_precision(double delta)
{
    m_spm_pid_precision = 65535. * delta / 20. ;
    updateSPMPrecision = true;
}
void ScanPlot::scan_lm_height(double h)
{
    scan_lm_h = h/1000.;
    sendLMheight = true;
}

#include <i_eg3000.h>
bool ScanPlot::sendData(const void *handle)
{
    bool wrote = false;

    if (updateFirstDelay.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_FIRST_DELAY, m_scan_first_delay);
    }
    if (updateLineDelay.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_LINE_DELAY, m_scan_line_delay);
    }
    if (updatePointDelay.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SCAN_POINT_DELAY, m_scan_point_delay);
    }
    if (updateFastReturn.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_FAST_RETURN, m_scan_fast_return ? 1 : 0);
    }
    if (updateSyncMode.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SCAN_WAIT_SYNC, m_scan_sync_mode);
    }
    if (updateSPMPrecision.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SPM_PRECISION, m_spm_pid_precision);
    }

    if (sendSpeed.claimMessage()) {
        wrote = true;
        eg3000_scan_speed((void *)handle, scanner_scan, requested_scan_speed);
    }
    if (sendLMheight.claimMessage()) {
        wrote = true;
        eg3000_send_packet(handle, ADR_CPU, CPU_SET_SCAN_LM_HEIGHT, &scan_lm_h, sizeof(float));
    }
    if (sendConfig.claimMessage()) {
        wrote = true;
        eg3000_scan_params((void *)handle, scanner_scan, Nx_points, Ny_points, requested_scan_speed,
                                Step_X,  Step_Y,  stPointX,  stPointY,
                                scan_dir_st_point, scan_ch[0], scan_ch[1], scan_ch[2], scan_ch[3],
                                LmEn, LmEn_ch[0], LmEn_ch[1], LmEn_ch[2], LmEn_ch[3], LmHeight, fast_scan_correction, scan_precise_mode);
    }
    if (sendExtConfig.claimMessage()) {
        wrote = true;
        eg3000_ccd_scan_params((void *)handle, scanner_scan, ccdMode, ccdNX, ccdNY, Step_X, Step_Y, stPointX, stPointY, scan_dir_st_point);
    }
    if (sendStartStop.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, (scanner_scan == scStage) ? CPU_SCAN_ON_OFF_STAGE : CPU_SCAN_ON_OFF, scan_on_off);
    }
    if (sendExtStartStop.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_SCAN_ON_OFF_CCD, scan_on_off);
    }
    if (sendExtReady.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_CCD_READY, 1);
    }

    return wrote;
}
void ScanPlot::resetWriter()
{
    updateSPMPrecision = false;
    updateFirstDelay = false;
    updateLineDelay = false;
    updatePointDelay = false;
    updateFastReturn = false;
    updateSyncMode = false;

    sendSpeed = false;
    sendLMheight = false;
    sendConfig = scan_on_off != 0;
    sendExtConfig = sendConfig && ccdMode != 0;
    sendStartStop = false;
    sendExtStartStop = false;
    sendExtReady = false;

    QMetaObject::invokeMethod(this, "reset");
}
void ScanPlot::reset()
{
    if (m_BStart->on_off)
        m_BStart->change_state_man_with_sig(false);

    spmPrecision->setNewValue(true);
    #ifdef _OLD_EXTRA_CONTROL
    firstDelay->setNewValue(true);
    lineDelay->setNewValue(true);
    pointDelay->setNewValue(true);
#else
    EdExScanFirstPDelay->setNewValue(true);
    EdExLineFirstPDelay->setNewValue(true);
    EdExSyncPointDelay->setNewValue(true);
#endif

    pointSync->setCurrentIndex(0);
    scan_sync_mode(0);
    scan_fast_return(fastReturn->isChecked());

    LEd_LFmodeHeight->setNewValue(true);

    for (int i = 0; i < numScanChannel; i++) {
        scanData[0][i].resize(qMin(scanData[0][i].size(), 1));
        scanData[1][i].resize(qMin(scanData[1][i].size(), 1));
    }
    Plot->Refresh();
}

void ScanPlot::capBits(int n, long bits)
{
    if (n == 0) {
        l2k_controller = bits & 4;
        l2k_controller_v22 = bits & 16;
        if (l2k_controller || l2k_controller_v22)
            check_speed();
    }
    if (n >= 2 && n <= 4) {
        is_optical_scanner[n-2] = bool(bits & 1);
    }
}

//emulator --------------------
#ifdef USE_EG3K_EMULATOR
static int sc_ch;
static bool ccdswitch;
static int ccdCounter;
static QElapsedTimer etmr;
static double test_plane_func(double xp, double yp)
{
    double a0=1000, a1=200, a2=200;
    return a0 + a1*xp + a2*yp;
}
#endif
void ScanPlot::reset_scan_test()
{
#ifdef USE_EG3K_EMULATOR
    QSettings settings("Param/Emulator.conf", QSettings::IniFormat);
    if (!settings.value("usb", false).toBool())
        return;

    sc_ch = 0;
    ccdswitch = ccdMode != 0;
    ccdCounter = 0;
    sendExtReady = true;
    etmr.start();
#endif
}
void ScanPlot::injectTestData()
{
#ifdef USE_EG3K_EMULATOR
    if (!scan_on_off)
        return;

    if (ccdswitch) {
        if (!sendExtReady)
            return;
        sendExtReady = false;
        if (ccdCounter < ccdNX*ccdNY) {
            processPacket(ADR_CPU, CPU_CCD_START, nullptr, 0);
            if (ccdCounter > 0 && ccdCounter % (scan_direction ? ccdNY : ccdNX) == 0)
                ccdswitch = false;
            ccdCounter++;
            if (ccdswitch)
                return;
        } else
            ccdswitch = false;
    }

    int scan_wait = qMax<int>(1, 1000/numScanChannel/user_scan_speed);
    if (etmr.elapsed() < scan_wait)
        return;
    etmr.restart();

    int prev_sc_ch = sc_ch;
    sc_ch++;
    if (sc_ch > 3) {
        sc_ch = 0;
        if (ccdMode != 0) {
            ccdswitch = true;
            sendExtReady = true;
        }
    }

    EG_ScanData Scan_Data;
    Scan_Data.Source = scan_ch[prev_sc_ch];
    if (Scan_Data.Source == srcNone)
        return;

    Scan_Data.Axis = scan_direction;
    Scan_Data.Direction = prev_sc_ch >= 2 ? dirBW : dirFW;
    Scan_Data.Channel = prev_sc_ch % 2;
    Scan_Data.Scanner = scanner_scan;
    Scan_Data.Line = line_number[prev_sc_ch];
    Scan_Data.Size = qMin<int>(1000, (scan_direction ? Ny_points : Nx_points));
    Scan_Data.ScanSize = qMin<int>(1000, (scan_direction ? Nx_points : Ny_points));
    WORD line_shift = (WORD)(rand() % 5000);
    for (int i = 0; i < Scan_Data.Size; i++)
    {
        double cur_x_point = xLeft;
        double cur_y_point = yBottom;
        if (scan_direction) {
            cur_x_point += line_number[prev_sc_ch] * Step_X;
            cur_y_point += i * Step_Y;
        } else {
            cur_x_point += i * Step_X;
            cur_y_point += line_number[prev_sc_ch] * Step_Y;
        }

        double ca[4] = {1, 0.5, 0.25, 0.125};
        double c = ca[prev_sc_ch];
        WORD func = 10000. *
                sin(c * cur_x_point) * sin(c * cur_x_point) *
                sin(c * cur_y_point) * sin(c * cur_y_point);
        func += test_plane_func(cur_x_point, cur_y_point) + line_shift;
        Scan_Data.Data[i] = func;
    }

    processPacket(ADR_CPU, CPU_SCANDATA, &Scan_Data, 0);
#endif
}

ScanPlot* ScanPlot::GetIt() noexcept
{
    auto scan = ModuleManager::instance->find1VoidPtr("scan");
    return static_cast<ScanPlot*>(scan);
}

void ScanPlot::on2DBackButton_clicked()
{
    Plot->TwoDPlot->Pure2dPlot->SetAutoScaleEnabled(m_BStart->on_off, false);
    Plot->TwoDPlotVertical->Pure2dPlot->SetAutoScaleEnabled(m_BStart->on_off, false);
}
