#include "xyzcalibrations.hpp"
#include "hardware/EG3000.h"
#include "gui/status.hpp"
#include "mainwindow.h"
#include "autocalib.hpp"
#include "oscilloscope.h"
#include <QVBoxLayout>
#include <QSplitter>
#include <QCheckBox>
#include <QSettings>
#include <QFormLayout>
#include <QFileDialog>
#include <QTextStream>
#include <QApplication>
#include "scanplot.h"
#include "spmtunning.h"
#include "spm_motors.hpp"
#include "datalistoverlaybuttons.h"
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
#include <QtMath>
#else
#include <qmath.h>
#include <qnumeric.h>
#endif


QLabel *newLabelCentered(QString text)
{
    QLabel *l = new QLabel(text);
    l->setAlignment(Qt::AlignCenter);
    return l;
}

XYZCalibrations::XYZCalibrations(QWidget *parent, Oscilloscope *_osc) : QWidget(parent)
{
    headConnected = false;
    stageConnected = false;
    SensorOFon = 0;
    updateSensorOF = 0;
    lvm_bits = 0;
    free_move_speed = 0.;
    osc = _osc;
    nlc_valid_cl = false;
    nlc_valid_ol = false;
    defaultFileName = "Param/default.par";

    csb_speed_extended = false;
    csb_speed_idx = 0;

    CL_status_xy = false;
    CL2_status_xy = false;
    CL_status_z = false;
    CL2_status_z = false;
    CL3_status_z = false;
    Nlc_status_cl_xy = false;
    Nlc2_status_cl_xy = false;
    Nlc_status_cl_z = false;
    Nlc2_status_cl_z = false;
    Nlc_status_ol_xy = false;
    Nlc2_status_ol_xy = false;
    Nlc_status_ol_z = false;
    Nlc2_status_ol_z = false;


    QVBoxLayout *mlay = new QVBoxLayout(this);
    mlay->setContentsMargins(0, 0, 0, 0);

    GrBCalibrations = new QGroupBox(this);
    vertCalibLay = new QVBoxLayout(GrBCalibrations);

    QHBoxLayout *Stage_test_layout = new QHBoxLayout();
    Stage_test_layout->setSpacing(6);
    Stage_test_layout->setContentsMargins(6, 6, 6, 6);
    vertCalibLay->addLayout(Stage_test_layout);

    mlay->addWidget(GrBCalibrations);

    // ----------------Stage calibration and test-------------------------
    QPushButton *Save_to_PC = new QPushButton(makeHugeIcon(":/container/save.png"), QString());
    Save_to_PC->setToolTip(tr("Save"));
    connect(Save_to_PC, SIGNAL(clicked()), this, SLOT(save_To_File()));
    DataListOverlayButtons::setButtonStyle(Save_to_PC);
    priborMenuButtons << Save_to_PC;

    QPushButton *Load_from_PC = new QPushButton(makeHugeIcon(":/container/open.png"), QString());
    Load_from_PC->setToolTip(tr("Load"));
    connect(Load_from_PC, SIGNAL(clicked()), this, SLOT(sl_Load_from_File()));
    DataListOverlayButtons::setButtonStyle(Load_from_PC);
    priborMenuButtons << Load_from_PC;

    BSensorAutocalib = new QPushButton(makeHugeIcon(":/AutoCalib/SensorsAutoAdjOf.png"), QString());
    BSensorAutocalib->setToolTip(tr("Automatic sensors calibration"));
    connect(BSensorAutocalib, SIGNAL(clicked()), this, SLOT(SensorAutocalib()));
    DataListOverlayButtons::setButtonStyle(BSensorAutocalib);
    priborMenuButtons << BSensorAutocalib;

    BPIDAutocalib = new QPushButton(makeHugeIcon(":/AutoCalib/PIDtune.png"), QString());
    BPIDAutocalib->setToolTip(tr("Automatic PI calibration"));
    connect(BPIDAutocalib, SIGNAL(clicked()), this, SLOT(PIAutocalib()));
    DataListOverlayButtons::setButtonStyle(BPIDAutocalib);
    priborMenuButtons << BPIDAutocalib;


    QPushButton *Save_to_Contr = new QPushButton(makeHugeIcon(":/Osc_window/saveDpsMemory.png"), QString());
    Save_to_Contr->setToolTip(tr("DSP Memory"));
    connect(Save_to_Contr, SIGNAL(clicked()), this, SLOT(WriteConfigToDeviceMem()));
    DataListOverlayButtons::setButtonStyle(Save_to_Contr);
    priborMenuButtons << Save_to_Contr;

    QPushButton *ProgramController = new QPushButton(makeHugeIcon(":/Osc_window/DSPwrite.png"), QString());
    ProgramController->setToolTip(tr("Program controller"));
    connect(ProgramController, SIGNAL(clicked()), this, SLOT(program_controller()));
    DataListOverlayButtons::setButtonStyle(ProgramController);
    priborMenuButtons << ProgramController;

    fb_on_off_xy = new TwoStateButton(":/Osc_window/CL_XY_off.png", ":/Osc_window/CL_XY_on.png");
    fb_on_off_xy->setToolTip(tr("Closed Loop XY"));
    connect(fb_on_off_xy, SIGNAL(state_changed(bool)), this, SLOT(s_setCLStatus(bool)));
    DataListOverlayButtons::setButtonStyle(fb_on_off_xy);
    priborMenuButtons << fb_on_off_xy;

    fb_on_off_z = new TwoStateButton(":/Osc_window/CL_Z_off.png", ":/Osc_window/CL_Z_on.png");
    fb_on_off_z->setToolTip(tr("Closed Loop Z"));
    connect(fb_on_off_z, SIGNAL(state_changed(bool)), this, SLOT(s_setCLStatusZ(bool)));
    DataListOverlayButtons::setButtonStyle(fb_on_off_z);
    priborMenuButtons << fb_on_off_z;

    DataListOverlayButtons::setButtonStyle(osc->BToCalib);
    priborMenuButtons << osc->BToCalib;

    QString nlc_full_string = tr("nonlinearity correction");
    nl_on_off_cl = new QCheckBox(tr("Sensor"));
    nl_on_off_cl->setToolTip(nlc_full_string);
    connect(nl_on_off_cl, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_cl(bool)));

    nl_on_off_ol = new QCheckBox(tr("Voltage"));
    nl_on_off_ol->setToolTip(nlc_full_string);
    connect(nl_on_off_ol, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_ol(bool)));

    nl_on_off_cl2 = new QCheckBox(tr("Sensor"));
    nl_on_off_cl2->setToolTip(nlc_full_string);
    connect(nl_on_off_cl2, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_cl(bool)));

    nl_on_off_ol2 = new QCheckBox(tr("Voltage"));
    nl_on_off_ol2->setToolTip(nlc_full_string);
    connect(nl_on_off_ol2, SIGNAL(clicked(bool)), this, SLOT(set_nlc_status_ol(bool)));

    QString noBluBoundStyle = "QLineEdit {border: 0px solid white; background-color: #FFFFFF;}";

    //xyz
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            linePID[i][j] = new DoubleEdit(0,100,7);
            linePID[i][j]->setStyleSheet(noBluBoundStyle);
            connect(linePID[i][j], SIGNAL(newValue()), this, SLOT(s_sendData()));
        }

        viewPiezoCalibration[i] = new NLCWidget;
        viewSensorCalibration[i] = new NLCWidget;
        connect(viewPiezoCalibration[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));
        connect(viewSensorCalibration[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));

        lineScannerRangeOL[i] = new DoubleEdit(1,1000,4);
        lineScannerRangeCL[i] = new DoubleEdit(1,1000,4);
        lineSensorScale[i] = new DoubleEdit(1,100000,0);
        lineSensorScale0[i] = new DoubleEdit(-32767,32767,0);

        lineScannerRangeOL[i]->setStyleSheet(noBluBoundStyle);
        lineScannerRangeCL[i]->setStyleSheet(noBluBoundStyle);
        lineSensorScale[i]->setStyleSheet(noBluBoundStyle);
        lineSensorScale0[i]->setStyleSheet(noBluBoundStyle);

        connect(lineScannerRangeOL[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineScannerRangeCL[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineSensorScale[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineSensorScale0[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
    }

    //xyz2
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            linePID2[i][j] = new DoubleEdit(0,100,7);
            linePID2[i][j]->setStyleSheet(noBluBoundStyle);
            connect(linePID2[i][j], SIGNAL(newValue()), this, SLOT(s_sendData()));
        }

        viewPiezoCalibration2[i] = new NLCWidget;
        viewSensorCalibration2[i] = new NLCWidget;
        connect(viewPiezoCalibration2[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));
        connect(viewSensorCalibration2[i], SIGNAL(clicked()), this, SLOT(nlcEdit()));

        lineScannerRangeOL2[i] = new DoubleEdit(1,1000,4);
        lineScannerRangeCL2[i] = new DoubleEdit(1,1000,4);
        lineSensorScale2[i] = new DoubleEdit(1,100000,0);
        lineSensorScale02[i] = new DoubleEdit(-32767,32767,0);

        lineScannerRangeOL2[i]->setStyleSheet(noBluBoundStyle);
        lineScannerRangeCL2[i]->setStyleSheet(noBluBoundStyle);
        lineSensorScale2[i]->setStyleSheet(noBluBoundStyle);
        lineSensorScale02[i]->setStyleSheet(noBluBoundStyle);

        connect(lineScannerRangeOL2[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineScannerRangeCL2[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineSensorScale2[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
        connect(lineSensorScale02[i], SIGNAL(newValue()), this, SLOT(s_sendData()));
    }

    //Z3
    linePIDProportionalZ3 = new DoubleEdit(0,100,7);
    linePIDIntegralZ3 = new DoubleEdit(0,100,7);

    connect(linePIDProportionalZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(linePIDIntegralZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));

    lineScannerRangeOLZ3 = new DoubleEdit(1,1000,4);
    lineScannerRangeCLZ3 = new DoubleEdit(1,1000,4);
    lineSensorScaleZ3 = new DoubleEdit(1,100000,0);
    lineSensorScaleZ03 = new DoubleEdit(-32767,32767,0);

    connect(lineScannerRangeOLZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(lineScannerRangeCLZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(lineSensorScaleZ3, SIGNAL(newValue()), this, SLOT(s_sendData()));
    connect(lineSensorScaleZ03, SIGNAL(newValue()), this, SLOT(s_sendData()));


    //lvm
    CBlvm[0] = new QCheckBox(tr("Low voltage mode"));
    CBlvm[1] = new QCheckBox(tr("Low voltage mode"));
    connect(CBlvm[0], SIGNAL(toggled(bool)), this, SLOT(lvmClicked()));
    connect(CBlvm[1], SIGNAL(toggled(bool)), this, SLOT(lvmClicked()));

    //nlc edit
    QGridLayout *nlcEditLay = new QGridLayout;
    nlcEditLay->addWidget(newLabelCentered("f(x) = a* x^4"),0, 0);
    nlcEditLay->addWidget(newLabelCentered(" + b* x^3"),0,1);
    nlcEditLay->addWidget(newLabelCentered(" + c* x^2"),0,2);
    nlcEditLay->addWidget(newLabelCentered(" + d* x"),0,3);
    nlcEditLay->addWidget(newLabelCentered(" + e"),0,4);
    //nlcEditLay->addWidget(newLabelCentered(" + f"),0,5);

    for (int i = 0; i < 6; i++) {
        lineNLCEdit[i] = new DoubleEdit(-1000000, 1000000, 10);
        connect(lineNLCEdit[i], SIGNAL(newValue()), this, SLOT(nlcEditUpd()));
        nlcEditLay->addWidget(lineNLCEdit[i], 1, i);
    }
    lineNLCEdit[5]->hide();

    viewNLCEdit = new NLCWidget;
    viewNLCEdit->setFixedWidth(lineNLCEdit[0]->width());
    nlcEditLay->addWidget(viewNLCEdit,0,6, 2, 1, Qt::AlignBottom);

    NLCEditValid = new QLabel();

    //QDialogButtonBox TODO
    QPushButton *nlcApply = new QPushButton(tr("Apply"));
    QPushButton *nlcCancel = new QPushButton(tr("Cancel"));
    QPushButton *nlcRst = new QPushButton(tr("Reset"));
    QHBoxLayout *nlcBLay = new QHBoxLayout;
    nlcCancel->setShortcut(QKeySequence(Qt::Key_Escape));
    nlcBLay->addWidget(nlcRst);
    nlcBLay->addWidget(NLCEditValid);
    nlcBLay->addStretch(1);
    nlcBLay->addWidget(nlcApply);
    nlcBLay->addWidget(nlcCancel);

    NLCEditWid = new QWidget(this, Qt::Window);
    NLCEditWid->setWindowTitle(tr("Function editor"));
    NLCEditWid->setWindowModality(Qt::WindowModal);

    QVBoxLayout *nlcLay = new QVBoxLayout(NLCEditWid);
    nlcLay->setSizeConstraint(QLayout::SetFixedSize);
    nlcLay->addLayout(nlcEditLay);
    nlcLay->addLayout(nlcBLay);
    connect(nlcRst, SIGNAL(clicked()), this, SLOT(nlcEditRst()));
    connect(nlcApply, SIGNAL(clicked()), this, SLOT(nlcEditApply()));
    connect(nlcCancel, SIGNAL(clicked()), NLCEditWid, SLOT(hide()));


    //-----------------------------------------------------------------
    QWidget *priborMenuWid = new QWidget();
    QVBoxLayout *Stage_Control_layout_0 = new QVBoxLayout(priborMenuWid);
    Stage_Control_layout_0->setContentsMargins(0, 0, 0, 0);
    for (int i=0; i<priborMenuButtons.size(); i++)
    {
        Stage_Control_layout_0->addWidget(priborMenuButtons.at(i), 0, Qt::AlignTop);
    }

    //
    headCalibWid = new QWidget;
    stageCalibWid = new QWidget;
    stageCalibWid->installEventFilter(this);
    QWidget *z3CalibWid = new QWidget;

    QSettings settings("Param/gui.conf", QSettings::IniFormat);

    Scanner_type = new QTabWidget;
    QSettings settings2("Param/gui.conf", QSettings::IniFormat);
    bool ratisMode = settings2.value("gui/Ratis_xyz_stage_mode", false).toBool();
    if (ratisMode)
    {
        Scanner_type->addTab(headCalibWid, "");
        Scanner_type->addTab(stageCalibWid, tr("Calibration data"));
    }
    else
    {
        Scanner_type->addTab(headCalibWid, makeHugeIcon(":Head_page.png"), tr("Head"));
        Scanner_type->addTab(stageCalibWid, makeHugeIcon(":stage_page.png"),tr("Stage"));
    }
    if (settings.value("display/oscilloscope/z3", false).toBool())
    Scanner_type->addTab(z3CalibWid, tr("Z3"));

    Scanner_type->setTabEnabled(0, settings.value("display/oscilloscope/head", true).toBool());
    Scanner_type->setTabEnabled(1, settings.value("display/oscilloscope/stage", true).toBool());
    //Scanner_type->setTabEnabled(2, settings.value("display/oscilloscope/z3", false).toBool());

    connect(Scanner_type, SIGNAL(currentChanged(int)), this, SLOT(CL_NLS_Lamps(int)));

    lineFreeSpeed = new DoubleEdit(0.1,1000,1);
    connect(lineFreeSpeed, SIGNAL(newValue()), this, SLOT(s_sendData()));
//    QHBoxLayout *free_speed_l = new QHBoxLayout;
//    free_speed_l->addWidget(new QLabel(tr("Free positioning speed") + ", " +
//                                          QObject::tr("micro")+QObject::tr("m" )+ "/" + QObject::tr("s") + ": "));
//    free_speed_l->addWidget(lineFreeSpeed);


    Stage_test_layout->addWidget(Scanner_type, 1);
    Stage_test_layout->addWidget(priborMenuWid, 1, Qt::AlignTop);
//    Stage_test_layout->addLayout(free_speed_l);


    QVBoxLayout *headLayout = new QVBoxLayout(headCalibWid);
    //
    calibVoltageGB[0] = new QGroupBox(tr("Voltages"));
    QGridLayout *layout_config = new QGridLayout(calibVoltageGB[0]);
    layout_config->setSpacing(6);
    layout_config->setContentsMargins(12, 9, 12, 9);


    stageLayout = new QVBoxLayout(stageCalibWid);


    calibVoltageGB[1] = new QGroupBox(tr("Voltages"));
    layout_config2 = new QGridLayout(calibVoltageGB[1]);
    layout_config2->setSpacing(6);
    layout_config2->setContentsMargins(12, 9, 12, 9);


    QFormLayout *layout_config3 = makeFormLayout(z3CalibWid);
    layout_config3->setSpacing(6);
    layout_config3->setContentsMargins(12, 9, 12, 9);

    calibTable[0] = new QTableWidget(9, 4);
    calibTable[1] = new QTableWidget(9, 4);

    headLayout->addWidget(calibTable[0]);
    if (!ratisMode)
        headLayout->addWidget(calibVoltageGB[0]);

    stageLayout->addWidget(calibTable[1]);
    if (!ratisMode)
        stageLayout->addWidget(calibVoltageGB[1]);


    for (int i=0; i<2; i++)
    {
        calibTable[i]->setDragEnabled(false);
        calibTable[i]->setFocusPolicy(Qt::NoFocus);
        calibTable[i]->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
        calibTable[i]->setSortingEnabled(false);
        //calibTable[i]->setAlternatingRowColors(true);
        calibTable[i]->verticalHeader()->setVisible(false);
        calibTable[i]->horizontalHeader()->setVisible(false);
        calibTable[i]->setSizePolicy(QSizePolicy::Policy::Minimum, QSizePolicy::Policy::Minimum);
        calibTable[i]->verticalScrollBar()->setEnabled(false);
        calibTable[i]->horizontalScrollBar()->setEnabled(false);
        calibTable[i]->verticalScrollBar()->setVisible(false);
        calibTable[i]->horizontalScrollBar()->setVisible(false);
        calibTable[i]->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        calibTable[i]->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        calibTable[i]->setSelectionMode(QAbstractItemView::SingleSelection);
        calibTable[i]->installEventFilter(this);
        //calibTable[i]->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
        calibTable[i]->horizontalHeader()->setResizeMode(0, QHeaderView::Fixed);
    }

    //xyz


    calibTable[0]->setCellWidget(0, 0, newLabelCentered(tr("Range")));
    calibTable[0]->setCellWidget(0, 1, newLabelCentered("X"));
    calibTable[0]->setCellWidget(0, 2, newLabelCentered("Y"));
    calibTable[0]->setCellWidget(0, 3, newLabelCentered("Z"));

    calibTable[0]->setCellWidget(1, 0, new QLabel(tr("OL, ")+QObject::tr("micro")+QObject::tr("m")));
    calibTable[0]->setCellWidget(2, 0, new QLabel(tr("CL, ")+QObject::tr("micro")+QObject::tr("m")));
    for (int i = 0; i < 3; i++) {
        calibTable[0]->setCellWidget(1, i + 1, lineScannerRangeOL[i]);
        calibTable[0]->setCellWidget(2, i + 1, lineScannerRangeCL[i]);
    }

    calibTable[0]->setCellWidget(3, 0, newLabelCentered(tr("Sensor")));
    calibTable[0]->setCellWidget(3, 1, newLabelCentered("X"));
    calibTable[0]->setCellWidget(3, 2, newLabelCentered("Y"));
    calibTable[0]->setCellWidget(3, 3, newLabelCentered("Z"));

    calibTable[0]->setCellWidget(4, 0, new QLabel(tr("Gain")));
    calibTable[0]->setCellWidget(5, 0, new QLabel(tr("Shift")));
    for (int i = 0; i < 3; i++) {
        calibTable[0]->setCellWidget(4, i + 1, lineSensorScale[i]);
        calibTable[0]->setCellWidget(5, i + 1, lineSensorScale0[i]);
    }

    calibTable[0]->setCellWidget(6, 0, newLabelCentered(tr("PI")));
    calibTable[0]->setCellWidget(6, 1, newLabelCentered("X"));
    calibTable[0]->setCellWidget(6, 2, newLabelCentered("Y"));
    calibTable[0]->setCellWidget(6, 3, newLabelCentered("Z"));

    calibTable[0]->setCellWidget(7, 0, new QLabel(tr("Proportional")));
    calibTable[0]->setCellWidget(8, 0, new QLabel(tr("Integral")));
    for (int i = 0; i < 3; i++) {
        calibTable[0]->setCellWidget(7, i + 1, linePID[0][i]);
        calibTable[0]->setCellWidget(8, i + 1, linePID[1][i]);
    }

    int c = 0;

    QLabel *l_nlc1;
    layout_config->addWidget(l_nlc1 = newLabelCentered("NLC"), c, 0);
    l_nlc1->setToolTip(nlc_full_string);
    layout_config->addWidget(newLabelCentered("X"), c, 1);
    layout_config->addWidget(newLabelCentered("Y"), c, 2);
    layout_config->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++)
        layout_config->addWidget(viewPiezoCalibration[i], c, i+1);
    layout_config->addWidget(nl_on_off_ol, c, 0);
    c++;

    for (int i = 0; i < 3; i++)
        layout_config->addWidget(viewSensorCalibration[i], c, i+1);
    layout_config->addWidget(nl_on_off_cl, c, 0);
    c++;

    layout_config->addWidget(CBlvm[0], c++, 0, 1, -1);
    layout_config->addItem(new QSpacerItem(0, 0, QSizePolicy::Minimum, QSizePolicy::Expanding), c++, 0, 1, -1);


    //xyz2

    calibTable[1]->setCellWidget(0, 0, newLabelCentered(tr("Range")));
    calibTable[1]->setCellWidget(0, 1, newLabelCentered("X"));
    calibTable[1]->setCellWidget(0, 2, newLabelCentered("Y"));
    calibTable[1]->setCellWidget(0, 3, newLabelCentered("Z"));

    calibTable[1]->setCellWidget(1, 0, new QLabel(tr("OL, ")+QObject::tr("micro")+QObject::tr("m")));
    calibTable[1]->setCellWidget(2, 0, new QLabel(tr("CL, ")+QObject::tr("micro")+QObject::tr("m")));
    for (int i = 0; i < 3; i++) {
        calibTable[1]->setCellWidget(1, i + 1, lineScannerRangeOL2[i]);
        calibTable[1]->setCellWidget(2, i + 1, lineScannerRangeCL2[i]);
    }

    calibTable[1]->setCellWidget(3, 0, newLabelCentered(tr("Sensor")));
    calibTable[1]->setCellWidget(3, 1, newLabelCentered("X"));
    calibTable[1]->setCellWidget(3, 2, newLabelCentered("Y"));
    calibTable[1]->setCellWidget(3, 3, newLabelCentered("Z"));

    calibTable[1]->setCellWidget(4, 0, new QLabel(tr("Gain")));
    calibTable[1]->setCellWidget(5, 0, new QLabel(tr("Shift")));
    for (int i = 0; i < 3; i++) {
        calibTable[1]->setCellWidget(4, i + 1, lineSensorScale2[i]);
        calibTable[1]->setCellWidget(5, i + 1, lineSensorScale02[i]);
    }

    calibTable[1]->setCellWidget(6, 0, newLabelCentered(tr("PI")));
    calibTable[1]->setCellWidget(6, 1, newLabelCentered("X"));
    calibTable[1]->setCellWidget(6, 2, newLabelCentered("Y"));
    calibTable[1]->setCellWidget(6, 3, newLabelCentered("Z"));

    calibTable[1]->setCellWidget(7, 0, new QLabel(tr("Proportional")));
    calibTable[1]->setCellWidget(8, 0, new QLabel(tr("Integral")));
    for (int i = 0; i < 3; i++) {
        calibTable[1]->setCellWidget(7, i + 1, linePID2[0][i]);
        calibTable[1]->setCellWidget(8, i + 1, linePID2[1][i]);
    }

    c = 0;
    QLabel *l_nlc2;
    layout_config2->addWidget(l_nlc2 = newLabelCentered("NLC"), c, 0);
    l_nlc2->setToolTip(nlc_full_string);
    layout_config2->addWidget(newLabelCentered("X"), c, 1);
    layout_config2->addWidget(newLabelCentered("Y"), c, 2);
    layout_config2->addWidget(newLabelCentered("Z"), c, 3);
    c++;
    for (int i = 0; i < 3; i++)
        layout_config2->addWidget(viewPiezoCalibration2[i], c, i+1);
    layout_config2->addWidget(nl_on_off_ol2, c, 0);
    c++;

    for (int i = 0; i < 3; i++)
        layout_config2->addWidget(viewSensorCalibration2[i], c, i+1);
    layout_config2->addWidget(nl_on_off_cl2, c, 0);
    c++;

    layout_config2->addWidget(CBlvm[1], c++, 0, 1, -1);
    layout_config2->addItem(new QSpacerItem(0, 0, QSizePolicy::Minimum, QSizePolicy::Expanding), c++, 0, 1, -1);

    //z3
    layout_config3->addRow(tr("Range OL: "), lineScannerRangeOLZ3);
    layout_config3->addRow(tr("Range CL: "), lineScannerRangeCLZ3);
    layout_config3->addRow(tr("Scale: "), lineSensorScaleZ3);
    layout_config3->addRow(tr("Shift: "), lineSensorScaleZ03);
    layout_config3->addRow(tr("P: "), linePIDProportionalZ3);
    layout_config3->addRow(tr("I: "), linePIDIntegralZ3);

    QColor grayTableColor(233, 231, 227);
    for (int j=0; j<2; j++)
    {
        int minHeight = 0;
        int minWidth = 0;
        for (int i=0; i<calibTable[j]->rowCount(); i++)
            minHeight += calibTable[j]->rowHeight(i);
        for (int i=0; i<calibTable[j]->columnCount(); i++)
            minWidth += calibTable[j]->columnWidth(i);
        minHeight += 3 + 60;
        calibTable[j]->setMinimumSize(minWidth + 8, minHeight);
        calibTable[j]->setColumnWidth(0, 145);

        for(int i=0;i<calibTable[j]->columnCount();++i)
        {
			auto model = calibTable[j]->model();
#ifdef _WIN32
            Q_ASSUME(model->setData(model->index(0, i), grayTableColor, Qt::BackgroundRole));
            Q_ASSUME(model->setData(model->index(3, i), grayTableColor, Qt::BackgroundRole));
            Q_ASSUME(model->setData(model->index(6, i), grayTableColor, Qt::BackgroundRole));
#endif

            ((QLabel*)calibTable[j]->cellWidget(0, i))->setStyleSheet("font-weight: bold;");
            ((QLabel*)calibTable[j]->cellWidget(3, i))->setStyleSheet("font-weight: bold;");
            ((QLabel*)calibTable[j]->cellWidget(6, i))->setStyleSheet("font-weight: bold;");
        }
        if (ratisMode)
        {
            int h = minHeight/calibTable[j]->rowCount();
            for (int i=0; i<calibTable[j]->rowCount(); i++)
            {
                calibTable[j]->setRowHeight(i, h);
            }
        }
    }
    installEventFilter(this);
    eventFilter(nullptr, new QEvent(QEvent::Resize));
}

bool XYZCalibrations::eventFilter(QObject *target, QEvent *event)
{
    if (event->type() == QEvent::Resize)
    {
        if (threeDPlot)
        {
            int freeSpace = threeDPlotGB->height();
            if (freeSpace > threeDPlotGB->width())
                freeSpace = threeDPlotGB->width();
            freeSpace -= 10;
            threeDPlot->fixedSize = freeSpace;
            threeDPlot->ResizeQPlot();
        }
        int UpButtonSize = 50;
        for (int i=0; i<priborMenuButtons.size(); i++)
        {
            priborMenuButtons.at(i)->setFixedSize(UpButtonSize, UpButtonSize);
            priborMenuButtons.at(i)->setIconSize(QSize(UpButtonSize, UpButtonSize));
        }
        int minWidth = 378;
        Scanner_type->setMinimumWidth(minWidth);
        headCalibWid->setMinimumWidth(minWidth);
        stageCalibWid->setMinimumWidth(minWidth);
        Scanner_type->setMaximumWidth(minWidth);
        headCalibWid->setMinimumWidth(minWidth);
        stageCalibWid->setMaximumWidth(minWidth);
    }

    return false;
}

void XYZCalibrations::SensorAutocalib()
{
    osc->acf->SensorReCalc(Scanner_type->currentIndex());
}

void XYZCalibrations::PIAutocalib()
{
    osc->acf->PICalcAuto(Scanner_type->currentIndex());
}


void XYZCalibrations::program_controller()
{
    QString SelectedFilter;
    QString FileName;

    FileName = QFileDialog::getOpenFileName(this,
                                            tr("Select firmware..."),
                                            QString(),
                                            "*.hex;;*.rpd",
                                            &SelectedFilter,
                                            QFileDialog::DontUseNativeDialog);

    if (FileName.isEmpty())
        return;

    if (!FileName.endsWith(SelectedFilter.remove(0, 1)))
        FileName += SelectedFilter;

    emit program_controller(FileName);
}


void XYZCalibrations::s_setCLStatus(bool status)
{
    switch (Scanner_type->currentIndex()) {
    case scHead:
        CL_status_xy = status;
        break;
    case scStage:
        CL2_status_xy = status;
        break;
    }
    s_sendData();
}
void XYZCalibrations::s_setCLStatusZ(bool status)
{
    switch (Scanner_type->currentIndex()) {
    case scHead:
        CL_status_z = status;
        break;
    case scStage:
        CL2_status_z = status;
        break;
    case scZ3:
        CL3_status_z = status;
        break;
    }
    s_sendData();
}
void XYZCalibrations::set_nlc_status_cl(bool st)
{
    switch (Scanner_type->currentIndex()) {
    case 0:
        Nlc_status_cl_z = Nlc_status_cl_xy = st;
        break;
    case 1:
        Nlc2_status_cl_z = Nlc2_status_cl_xy = st;
        break;
    }
    s_sendData();
}
void XYZCalibrations::set_nlc_status_ol(bool st)
{
    switch (Scanner_type->currentIndex()) {
    case 0:
        Nlc_status_ol_z = Nlc_status_ol_xy = st;
        break;
    case 1:
        Nlc2_status_ol_z = Nlc2_status_ol_xy = st;
        break;
    }
    s_sendData();
}
void XYZCalibrations::CL_NLS_Lamps(int scanner)
{
    switch (scanner) {
    case scHead:
        fb_on_off_xy->setEnabled(true);
        fb_on_off_xy->change_state_man(CL_status_xy);
        fb_on_off_z->change_state_man(CL_status_z);
        nl_on_off_cl->setChecked(Nlc_status_cl_xy && Nlc_status_cl_z);
        nl_on_off_ol->setChecked(Nlc_status_ol_xy && Nlc_status_ol_z);
        break;
    case scStage:
        fb_on_off_xy->setEnabled(true);
        fb_on_off_xy->change_state_man(CL2_status_xy);
        fb_on_off_z->change_state_man(CL2_status_z);
        nl_on_off_cl2->setChecked(Nlc2_status_cl_xy && Nlc2_status_cl_z);
        nl_on_off_ol2->setChecked(Nlc2_status_ol_xy && Nlc2_status_ol_z);
        break;
    case scZ3:
        fb_on_off_xy->setEnabled(false);
        fb_on_off_xy->change_state_man(false);
        fb_on_off_z->change_state_man(CL3_status_z);
        break;
    }

    auto scan = ScanPlot::GetIt();
    if (scan)
    {
        #ifndef _OLD_EXTRA_CONTROL
        scan->CBExLowVoltage->blockSignals(true);
        if (Scanner_type->currentIndex() < 2)
            scan->CBExLowVoltage->setChecked(CBlvm[Scanner_type->currentIndex()]->isChecked());
        scan->CBExLowVoltage->blockSignals(false);
        #endif
    }
}


void XYZCalibrations::LVMLoad()
{
    QSettings settings("Param/calib.conf", QSettings::IniFormat);
    CBlvm[0]->setChecked(settings.value("lvm/0").toBool());
    CBlvm[1]->setChecked(settings.value("lvm/1").toBool());
}
void XYZCalibrations::LVMSave()
{
    QSettings settings("Param/calib.conf", QSettings::IniFormat);
    settings.setValue("lvm/0", CBlvm[0]->isChecked());
    settings.setValue("lvm/1", CBlvm[1]->isChecked());
}


template<class T>
double NewtonResult(T arr[5], double x0, double value, double ac, DWORD max = 100)
{
    double x_new = x0;
    WORD N = 0;

    do {
        x0 = x_new;
        x_new = x0 -
                (((((arr[0] * x0 + arr[1]) * x0 + arr[2]) * x0 + arr[3]) * x0 + arr[4]) - value) /
                (((4*arr[0] * x0 + 3*arr[1]) * x0 + 2*arr[2]) * x0 + arr[3]);
        N++;
    } while ((fabs(x0 - x_new) > ac) && (N < max));

    return x_new;
}

inline
double polynom_at_pos(double pos, double coeff[5]) {
    double acc = coeff[0];
    for (int i = 1; i < 5; i++)
        acc = acc * pos + coeff[i];

    return acc;
}

QVector<double> buildNlcPic(double k[6])
{
    QVector<double> pic;

    double ShiftCL = NewtonResult(k, 10, 0, 0.001);
    double CalibRangeCL = NewtonResult(k, 50, 65535, 0.001) - ShiftCL;

    if (!qIsFinite(ShiftCL) || !qIsFinite(CalibRangeCL))
        return pic;

    pic.resize(100);
    for (int i = 0; i < 100; i++) {
        double x = i/(100. - 1);
        double x1 = x * CalibRangeCL + ShiftCL;
        double v1 = polynom_at_pos(x1, k) / 65535.;
        pic[i] = v1;
    }

    return pic;
}

class NLCCurveData : public QwtSeriesData<QPointF>
{
public:
    int size() const override {
        return yseq.size();
    }
    QPointF sample(int i) const override {
        return QPointF(xseq.at(0) + i * xseq.at(1), yseq.at(i));
    }
    QRectF boundingRect() const override { return QRectF(); }
    QVector<double> xseq;
    QVector<double> yseq;
};

void NLCWidget::paintEvent(QPaintEvent *)
{
    if (m_pic.size() < 2)
        return;

    QPainter p(this);
    p.setPen(Qt::red);
    p.setRenderHint(QPainter::Antialiasing, true);

    NLCCurveData *data = new NLCCurveData;
    data->xseq = QVector<double>() << 0 << 1. / (m_pic.size() - 1);
    data->yseq = m_pic;

    QwtPlotCurve c;
    c.setData(data);

    QwtScaleMap x, y;
    x.setScaleInterval(0, 1);
    y.setScaleInterval(1, 0);
    x.setPaintInterval(contentsRect().left(), contentsRect().right()+1);
    y.setPaintInterval(contentsRect().top(), contentsRect().bottom()+1);

    c.drawLines(&p, x, y, 0, m_pic.size()-1);
}

void XYZCalibrations::nlcEdit()
{
    double *source = nullptr;
    NLCWidget *status = qobject_cast<NLCWidget *>(sender());
    if (!status)
        return;

    for (int i = 0; !source && i < 3; i++)
        if (status == viewPiezoCalibration[i])
            source = linePiezoCalibration[i];
    for (int i = 0; !source && i < 3; i++)
        if (status == viewSensorCalibration[i])
            source = lineSensorCalibration[i];
    for (int i = 0; !source && i < 3; i++)
        if (status == viewPiezoCalibration2[i])
            source = linePiezoCalibration2[i];
    for (int i = 0; !source && i < 3; i++)
        if (status == viewSensorCalibration2[i])
            source = lineSensorCalibration2[i];

    if (!source)
        return;

    for (int i = 0; i < 6; i++)
        lineNLCEdit[i]->setValue(source[i]);
    nlcEditUpd(); //viewNLCEdit->SetPic(status->m_pic);

    NLCEditWid->setProperty("source", QVariant::fromValue((void *)source));
    NLCEditWid->setProperty("status", QVariant::fromValue((void *)status));
    NLCEditWid->layout()->activate();
    NLCEditWid->move(mapToGlobal(rect().center() - NLCEditWid->rect().center()));
    NLCEditWid->show();
    if (NLCEditWid->focusWidget())
        NLCEditWid->focusWidget()->clearFocus();
}
void XYZCalibrations::nlcEditApply()
{
    NLCEditWid->hide();

    double *source = (double *)NLCEditWid->property("source").value<void *>();
    NLCWidget *status = qobject_cast<NLCWidget *>((QObject *)NLCEditWid->property("status").value<void *>());
    if (!status || !source)
        return;

    for (int i = 0; i < 6; i++)
        source[i] = lineNLCEdit[i]->value();
    status->SetPic(viewNLCEdit->m_pic);

    s_sendData();
}
void XYZCalibrations::nlcEditRst()
{
    for (int i = 0; i < 6; i++)
        lineNLCEdit[i]->setValue(0);
    lineNLCEdit[3]->setValue(1);
    nlcEditUpd();
}
void XYZCalibrations::nlcEditUpd()
{
    double source[6];
    for (int i = 0; i < 6; i++)
        source[i] = lineNLCEdit[i]->value();
    viewNLCEdit->SetPic(buildNlcPic(source));

    bool ok = checkNLCform(source);
    NLCEditValid->setText(ok ? tr("Valid curve") : tr("Invalid curve"));
    NLCEditValid->setStyleSheet(ok ? "background-color: green;" : "background-color: red;");
}

bool XYZCalibrations::checkNLCform(double calib[6])
{
    //check newton converges
    double CalibRange[5];
    double Shift[5];
    const double acc = 0.001;

    int x[] = {1000, 2000, 5000, 10000, 50000};
    for (int i = 0; i < 5; i++) {
        int maxx = x[i];
        Shift[i] = NewtonResult(calib, 10, 0, acc, maxx);
        CalibRange[i] = NewtonResult(calib, 50, 65535, acc, maxx);
        if (i > 0) if (fabs(Shift[i-1] - Shift[i]) > acc || fabs(CalibRange[i-1] - CalibRange[i]) > acc)
            return false;
    }

    return true;
}

//---------------------

void XYZCalibrations::init_nlc_sensor()
{
    for (int j = 0; j < 2; j++) {
        float *calib = j == 0 ? Scanner_data_pid_cal_range : Scanner2_data_pid_cal_range;
        for (int i = 0; i < 3; i++) {
            //27-31.32 //33-37.38 //39-43.44
            ShiftCL[j][i] = NewtonResult(calib + 27 + 6 * i, 10, 0, 0.001);
            CalibRangeCL[j][i] = NewtonResult(calib + 27 + 6 * i, 50, 65535, 0.001) - ShiftCL[j][i];

            if (CalibRangeCL[j][i] != 0) CalibRangeCL[j][i] = 1. / CalibRangeCL[j][i];
            CalibRangeCL[j][i] *= calib[45 + i]; //to mkm
        }
    }

    nlc_valid_cl = true;
}

float XYZCalibrations::reverse_nlc_sensor(int axis, float c)
{
    BYTE sc = axis < AxisX2 ? scHead : scStage;
    BYTE ax = axis % AxisX2;
    float *calib = sc ? Scanner2_data_pid_cal_range : Scanner_data_pid_cal_range;

    if (ax != AxisZ ? !(sc ? Nlc2_status_cl_xy : Nlc_status_cl_xy) :
            !(sc ? Nlc2_status_cl_z : Nlc_status_cl_z))
        return c / 65535. * calib[45 + ax];

    if (!nlc_valid_cl)
        init_nlc_sensor();

    float C = NewtonResult(calib + 27 + 6 * ax, 10, c, 0.001);
    return (C - ShiftCL[sc][ax]) * CalibRangeCL[sc][ax];
}

//---------------------

void XYZCalibrations::init_nlc_voltage()
{
    for (int j = 0; j < 2; j++) {
        float *calib = j == 0 ? Scanner_data_pid_cal_range : Scanner2_data_pid_cal_range;
        for (int i = 0; i < 3; i++) {
            //9-13.14 //15-19.20 //21-25.26
            ShiftOL[j][i] = NewtonResult(calib + 9 + 6 * i, 10, 0, 0.001);
            CalibRangeOL[j][i] = NewtonResult(calib + 9 + 6 * i, 50, 65535, 0.001) - ShiftOL[j][i];

            if (CalibRangeOL[j][i] != 0) CalibRangeOL[j][i] = 1. / CalibRangeOL[j][i];
            CalibRangeOL[j][i] *= calib[45 + i]; //to mkm
        }
    }

    nlc_valid_ol = true;
}

float XYZCalibrations::reverse_nlc_voltage(int axis, float c)
{
    BYTE sc = axis < AxisX2 ? scHead : scStage;
    BYTE ax = axis % AxisX2;
    float *calib = sc ? Scanner2_data_pid_cal_range : Scanner_data_pid_cal_range;

    if (ax != AxisZ ? !(sc ? Nlc2_status_ol_xy : Nlc_status_ol_xy) :
            !(sc ? Nlc2_status_ol_z : Nlc_status_ol_z))
        return c / 65535. * calib[45 + ax];

    if (!nlc_valid_ol)
        init_nlc_voltage();

    float C = NewtonResult(calib + 9 + 6 * ax, 10, c, 0.001);
    return (C - ShiftOL[sc][ax]) * CalibRangeOL[sc][ax];
}

//---------------------

void XYZCalibrations::s_sendData()
{
    free_move_speed = lineFreeSpeed->value();

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            Scanner_data_pid_cal_range[i+j*3] = linePID[i][j]->value();

        Scanner_data_pid_cal_range[45+i] = ((i < 2 ? CL_status_xy : CL_status_z) ? lineScannerRangeCL : lineScannerRangeOL)[i]->value();
        Scanner_data_scale[i] = (DWORD)qRound(lineSensorScale[i]->value());
        Scanner_data_Shift[i] = qRound(lineSensorScale0[i]->value());

        for (int j = 0; j < 6; j++) {
            Scanner_data_pid_cal_range[9 + i * 6 + j] = linePiezoCalibration[i][j];
            Scanner_data_pid_cal_range[27 + i * 6 + j] = lineSensorCalibration[i][j];
        }
    }


    //------Second Scanner-------------------------------------------
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            Scanner2_data_pid_cal_range[i+j*3] = linePID2[i][j]->value();

        Scanner2_data_pid_cal_range[45+i] = ((i < 2 ? CL2_status_xy : CL2_status_z) ? lineScannerRangeCL2 : lineScannerRangeOL2)[i]->value();
        Scanner2_data_scale[i] = (DWORD)qRound(lineSensorScale2[i]->value());
        Scanner2_data_Shift[i] = qRound(lineSensorScale02[i]->value());

        for (int j = 0; j < 6; j++) {
            Scanner2_data_pid_cal_range[9 + i * 6 + j] = linePiezoCalibration2[i][j];
            Scanner2_data_pid_cal_range[27 + i * 6 + j] = lineSensorCalibration2[i][j];
        }
    }

    //Z3
    memset(Scanner3_data_pid_cal_range, 0, sizeof(float)*48);
    memset(Scanner3_data_scale, 0, sizeof(DWORD)*3);
    memset(Scanner3_data_Shift, 0, sizeof(long)*3);
    Scanner3_data_pid_cal_range[0] = linePIDProportionalZ3->value();
    Scanner3_data_pid_cal_range[1] = linePIDIntegralZ3->value();
    Scanner3_data_pid_cal_range[6] = linePIDProportionalZ3->value();
    Scanner3_data_pid_cal_range[7] = linePIDIntegralZ3->value();
    Scanner3_data_pid_cal_range[45] = Scanner3_data_pid_cal_range[47] =
            (CL3_status_z ? lineScannerRangeCLZ3 : lineScannerRangeOLZ3)->value();
    Scanner3_data_scale[2] = (DWORD)qRound(lineSensorScaleZ3->value());
    Scanner3_data_Shift[2] = qRound(lineSensorScaleZ03->value());

    doSendCalData = true;

    nlc_valid_cl = false;
    nlc_valid_ol = false;
    sl_XYZ_range();
}
void XYZCalibrations::updateCalibDisplay()
{
    lineFreeSpeed->setValue(free_move_speed);

    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            linePID[i][j]->setValue(Scanner_data_pid_cal_range[i+j*3]);

        if (Scanner_data_pid_cal_range[45+i] > 0) {
            lineScannerRangeOL[i]->setValue(Scanner_data_pid_cal_range[45+i]);
            lineScannerRangeCL[i]->setValue(Scanner_data_pid_cal_range[45+i]);
        } else
            Scanner_data_pid_cal_range[45+i] = ((i < 2 ? CL_status_xy : CL_status_z) ? lineScannerRangeCL : lineScannerRangeOL)[i]->value();

        lineSensorScale[i]->setValue(Scanner_data_scale[i]);
        lineSensorScale0[i]->setValue(Scanner_data_Shift[i]);

        for (int j = 0; j < 6; j++) {
            linePiezoCalibration[i][j] = Scanner_data_pid_cal_range[9 + i * 6 + j];
            lineSensorCalibration[i][j] = Scanner_data_pid_cal_range[27 + i * 6 + j];
        }
        viewPiezoCalibration[i]->SetPic(buildNlcPic(linePiezoCalibration[i]));
        viewSensorCalibration[i]->SetPic(buildNlcPic(lineSensorCalibration[i]));
    }

    //------------Scanner2-----------------------
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++)
            linePID2[i][j]->setValue(Scanner2_data_pid_cal_range[i+j*3]);

        if (Scanner2_data_pid_cal_range[45+i] > 0) {
            lineScannerRangeOL2[i]->setValue(Scanner2_data_pid_cal_range[45+i]);
            lineScannerRangeCL2[i]->setValue(Scanner2_data_pid_cal_range[45+i]);
        } else
            Scanner2_data_pid_cal_range[45+i] = ((i < 2 ? CL2_status_xy : CL2_status_z) ? lineScannerRangeCL2 : lineScannerRangeOL2)[i]->value();

        lineSensorScale2[i]->setValue(Scanner2_data_scale[i]);
        lineSensorScale02[i]->setValue(Scanner2_data_Shift[i]);

        for (int j = 0; j < 6; j++) {
            linePiezoCalibration2[i][j] = Scanner2_data_pid_cal_range[9 + i * 6 + j];
            lineSensorCalibration2[i][j] = Scanner2_data_pid_cal_range[27 + i * 6 + j];
        }
        viewPiezoCalibration2[i]->SetPic(buildNlcPic(linePiezoCalibration2[i]));
        viewSensorCalibration2[i]->SetPic(buildNlcPic(lineSensorCalibration2[i]));
    }

    CL_NLS_Lamps(Scanner_type->currentIndex());

    nlc_valid_cl = false;
    nlc_valid_ol = false;
    sl_XYZ_range();
}

bool XYZCalibrations::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD)
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_STATUS) {
        EG_Status *Deflectomter = (EG_Status *)data;

        headConnected = !(Deflectomter->ErrorCode & ecHeadNotFound);
        stageConnected = !(Deflectomter->ErrorCode & ecStageNotFound);
        scannersConnectedUpdated = true;

        if (Deflectomter->ErrorCode & ecScanErrorPID)
            emit sigScanError();
        else {
            if (Deflectomter->ErrorCode & ecAxisErrorPID)
                emit sigSetpointError(0);
            if (Deflectomter->ErrorCode & ecAxisErrorPID2)
                emit sigSetpointError(1);
        }

        return false;
    }

    return false;
}

#include <i_eg3000.h>
bool XYZCalibrations::sendData(const void *handle)
{
    bool wrote = false;

    if (sendStartup.claimMessage()) {
        wrote = true;

        eg3000_status_onoff((void *)handle, 1);
        eg3000_send_empty_packet(handle, ADR_CPU, CPU_CAP_BITS);

        QSettings settings("Param/controller.conf", QSettings::IniFormat);
        BYTE data = settings.value("double_c_output", 0).toInt();
        eg3000_csb_double_c_out((void *)handle, scStage, data);

        { //hackfix for scan in stage-only controllers
            QSettings settings("Param/gui.conf", QSettings::IniFormat);
            if (!settings.value("display/spm", true).toBool())
                if (!settings.value("display/oscilloscope/head", true).toBool())
                    eg3000_fb_pid_params((void *)handle, scStage, 0.1f, 0, 0, srcMag1, 0);
        }

        { //only z2->z3 for now
            float kfwd = settings.value("hvsplit/z2_k", 0).toDouble();
            float lpf_freq = settings.value("hvsplit/z2_freq", 1).toDouble();
            if (fabs(kfwd) > 0.01) { //enabled
                DWORD data[5];
                data[0] = AxisZ2; //axis
                data[1] = fabs(kfwd) > 0.01 ? 1 : 0; //on
                memcpy(&data[2], &kfwd, sizeof(float)); //kfwd
                float kbwd = 1/kfwd;
                memcpy(&data[3], &kbwd, sizeof(float)); //kbwd
                memcpy(&data[4], &lpf_freq, sizeof(float)); //lpf freq khz
                eg3000_send_packet(handle, ADR_CPU, CPU_HV_SPLIT_EN, data, 5*sizeof(DWORD));
            }
        }
    }

    if (updLVM.claimMessage()) {
        wrote = true;
        eg3000_send_word_packet(handle, ADR_CPU, CPU_HW_LVM, lvm_bits);
    }
    if (doWriteConfigToDeviceMem.claimMessage()) {
        wrote = true;
        eg3000_send_empty_packet(handle, ADR_CPU, CPU_SAVE_CALDATA);
    }
    if (doSendCalData.claimMessage()) {
        wrote = true;

        eg3000_send_caldata((void *)handle, scHead,
                            Scanner_data_pid_cal_range, Scanner_data_scale, Scanner_data_Shift,
                            Nlc_status_cl_xy, CL_status_xy, free_move_speed,
                            Nlc_status_cl_z, CL_status_z, Nlc_status_ol_xy, Nlc_status_ol_z);

        eg3000_send_caldata((void *)handle, scStage,
                            Scanner2_data_pid_cal_range, Scanner2_data_scale, Scanner2_data_Shift,
                            Nlc2_status_cl_xy, CL2_status_xy, free_move_speed,
                            Nlc2_status_cl_z, CL2_status_z, Nlc2_status_ol_xy, Nlc2_status_ol_z);

        {
            QSettings settings("Param/gui.conf", QSettings::IniFormat);
            if (settings.value("display/oscilloscope/z3", false).toBool())
                eg3000_send_caldata((void *)handle, scZ3,
                                    Scanner3_data_pid_cal_range, Scanner3_data_scale, Scanner3_data_Shift,
                                    0, CL3_status_z, free_move_speed,
                                    0, CL3_status_z, 0, 0);
        }
    }
    if (doSendCSBSpeed.claimMessage()) {
        wrote = true;

        BYTE idx = qMin(csb_speed_extended ? 7 : 3, csb_speed_idx);
        eg3000_send_packet(handle, ADR_CPU, CPU_SET_CSB_RATE, &idx, 1);
    }
    if (updateSensorOF) {
        BYTE v = updateSensorOF - 1;
        updateSensorOF = 0;
        eg3000_csb_overflow((void *)handle, v, SensorOFon);
    }

    return wrote;
}
void XYZCalibrations::WriteConfigToDeviceMem(void)
{
    doWriteConfigToDeviceMem = true;
}
void XYZCalibrations::setLVM(BYTE head, BYTE stage)
{
    lvm_bits = (head ? (0x3 << 0) : 0) + (stage ? (0x3 << 2) : 0);
    updLVM = true;
}

void XYZCalibrations::sl_Load_from_File()
{
    bool oldh = CBlvm[0]->isChecked();
    bool olds = CBlvm[1]->isChecked();
    LVMLoad();

    if (oldh == CBlvm[0]->isChecked() && olds == CBlvm[1]->isChecked())
        read_calib_par(defaultFileName);
    else
        lvmClicked();

    new Notification(tr("Calibrations loaded"), Notification::Good, 10);
}

template <class T>
int readValueArray(QTextStream &s, T *array, int size, T default_value = 0)
{
    int i = 0;
    while (!s.atEnd() && i < size) {
        QStringList l = s.readLine().split('=');
        if (l.size() != 2)
            continue;

        array[i++] = l.last().toDouble();
    }
    int j = i;
    while (j < size)
        array[j++] = default_value;
    return i;
}

void XYZCalibrations::read_calib_par(QString file)
{
    QFile dataFile(QFile::exists(file) ? file : ":/config/default.par");
    dataFile.open(QIODevice::ReadOnly|QIODevice::Text);
    QTextStream inData(&dataFile);

    bool AutLoad = false;
    readValueArray(inData, &AutLoad, 1, false);
    if (AutLoad)
        qDebug() << "config loading from controller not supported";

    bool loaded1 = false;
    bool loaded2 = false;
    QString name = file;
    name.truncate(name.lastIndexOf(".par"));
    float sz;
    {
        QString name1 = name + "_calib_head.bin";
        QString name2 = name + "_calib_head_lvm.bin";
        if (CBlvm[0]->isChecked()) {
            if (QFile::exists(name2)) name1 = name2;
        } else {
            if (!QFile::exists(name1)) name1 = name2;
        }
        if (QFile::exists(name1)) {
            QFile file(name1);
            file.open(QIODevice::ReadOnly);

            file.read((char *)Scanner_data_pid_cal_range, 48*sizeof(float));
            file.read((char *)Scanner_data_scale, 3*sizeof(DWORD));
            file.read((char *)Scanner_data_Shift, 3*sizeof(long));

            file.read((char *)&Nlc_status_ol_xy, 1); Nlc_status_ol_z = Nlc_status_ol_xy;
            file.read((char *)&Nlc_status_cl_xy, 1); Nlc_status_cl_z = Nlc_status_cl_xy;
            file.read((char *)&CL_status_xy, 1);
            file.read((char *)&CL_status_z, 1);

            for (int i = 0; i < 3; i++) {
                file.read((char *)&sz, sizeof(float));
                lineScannerRangeOL[i]->setValue(sz);
            }
            for (int i = 0; i < 3; i++) {
                file.read((char *)&sz, sizeof(float));
                lineScannerRangeCL[i]->setValue(sz);
            }
            for (int i = 0; i < 3; i++)
                Scanner_data_pid_cal_range[45+i] = -1;

            file.read((char *)&free_move_speed, sizeof(float));

            loaded1 = (file.error() == QFile::NoError);
            file.close();
        }
    }
    {
        QString name1 = name + "_calib_stage.bin";
        QString name2 = name + "_calib_stage_lvm.bin";
        if (CBlvm[1]->isChecked()) {
            if (QFile::exists(name2)) name1 = name2;
        } else {
            if (!QFile::exists(name1)) name1 = name2;
        }
        if (QFile::exists(name1)) {
            QFile file(name1);
            file.open(QIODevice::ReadOnly);

            file.read((char *)Scanner2_data_pid_cal_range, 48*sizeof(float));
            file.read((char *)Scanner2_data_scale, 3*sizeof(DWORD));
            file.read((char *)Scanner2_data_Shift, 3*sizeof(long));

            file.read((char *)&Nlc2_status_ol_xy, 1); Nlc2_status_ol_z = Nlc2_status_ol_xy;
            file.read((char *)&Nlc2_status_cl_xy, 1); Nlc2_status_cl_z = Nlc2_status_cl_xy;
            file.read((char *)&CL2_status_xy, 1);
            file.read((char *)&CL2_status_z, 1);

            for (int i = 0; i < 3; i++) {
                file.read((char *)&sz, sizeof(float));
                lineScannerRangeOL2[i]->setValue(sz);
            }
            for (int i = 0; i < 3; i++) {
                file.read((char *)&sz, sizeof(float));
                lineScannerRangeCL2[i]->setValue(sz);
            }
            for (int i = 0; i < 3; i++)
                Scanner2_data_pid_cal_range[45+i] = -1;

            file.read((char *)&free_move_speed, sizeof(float));

            loaded2 = (file.error() == QFile::NoError);
            file.close();
        }
    }


    if (!loaded1 || !loaded2) {
        readValueArray(inData, Scanner_data_pid_cal_range, 48);
        readValueArray(inData, Scanner_data_scale, 3, (DWORD)1);
        readValueArray(inData, Scanner_data_Shift, 3);
        for (int i = 0; i < 3; i++)
            if (Scanner_data_pid_cal_range[45+i] <= 0)
                Scanner_data_pid_cal_range[45+i] = 1;

        readValueArray(inData, &CL_status_xy, 1, false);

        readValueArray(inData, &Nlc_status_cl_xy, 1, false);
        Nlc_status_cl_z = Nlc_status_cl_xy;

        //-----------------
        readValueArray(inData, Scanner2_data_pid_cal_range, 48);
        readValueArray(inData, Scanner2_data_scale, 3, (DWORD)1);
        readValueArray(inData, Scanner2_data_Shift, 3);
        for (int i = 0; i < 3; i++)
            if (Scanner2_data_pid_cal_range[45+i] <= 0)
                Scanner2_data_pid_cal_range[45+i] = 1;

        readValueArray(inData, &CL2_status_xy, 1, false);

        readValueArray(inData, &Nlc2_status_cl_xy, 1, false);
        Nlc2_status_cl_z = Nlc2_status_cl_xy;

        //-----------------
        readValueArray(inData, &free_move_speed, 1, 100.f);
        readValueArray(inData, &free_move_speed, 1, 100.f);

        readValueArray(inData, &CL_status_z, 1, false);

        readValueArray(inData, &Nlc_status_ol_xy, 1, false);
        Nlc_status_ol_z = Nlc_status_ol_xy;

        readValueArray(inData, &CL2_status_z, 1, false);

        readValueArray(inData, &Nlc2_status_ol_xy, 1, false);
        Nlc2_status_ol_z = Nlc2_status_ol_xy;

        double val[3];

        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner_data_pid_cal_range[45+i]);
            lineScannerRangeOL[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner_data_pid_cal_range[45+i]);
            lineScannerRangeCL[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++)
            Scanner_data_pid_cal_range[45+i] = -1;

        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner2_data_pid_cal_range[45+i]);
            lineScannerRangeOL2[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++) {
            readValueArray<double>(inData, val + i, 1, Scanner2_data_pid_cal_range[45+i]);
            lineScannerRangeCL2[i]->setValue(val[i]);
        }
        for (int i = 0; i < 3; i++)
            Scanner2_data_pid_cal_range[45+i] = -1;
    }
    dataFile.close();


    QSettings settings("Param/gui.conf", QSettings::IniFormat);
    bool head_on = settings.value("display/oscilloscope/head", true).toBool();
    bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
    if (!head_on)
        CL_status_xy = CL_status_z = false;
    if (!stage_on)
        CL2_status_xy = CL2_status_z = false;


    if (!loaded1 && !loaded2) { //if no binary dumps
        bool oldh = CBlvm[0]->isChecked();
        bool olds = CBlvm[1]->isChecked();
        CBlvm[0]->setChecked(false);  //always non-lvm
        CBlvm[1]->setChecked(false);
        write_calib_par(file, false); //dump to bin so that lvm won't overwrite and lose original calibrations
        CBlvm[0]->setChecked(oldh);
        CBlvm[1]->setChecked(olds);
    }

    auto scan = ScanPlot::GetIt();
    if (scan)
        scan->sl_ScanCorrection();

    updateCalibDisplay();

    z3Load();
    s_sendData();
}

void XYZCalibrations::lvmClicked()
{
    QWidgetList cwl;
    cwl << this << osc->DacsWidget << osc->acp;
    foreach (QWidget *wid, cwl)
        wid->setEnabled(false);

    bool upd = false;
    bool upd2 = false;
    const double nullpos = 10./(150 + 10);
    if (sender() != CBlvm[1]) { //head
        upd2 = upd2 || CL_status_xy || CL_status_z;
        CL_status_xy = false;
        CL_status_z = false;
        osc->DacsWidget->X->setValue(Scanner_data_pid_cal_range[45] * nullpos);
        osc->DacsWidget->Y->setValue(Scanner_data_pid_cal_range[46] * nullpos);
        osc->DacsWidget->Z->setValue(Scanner_data_pid_cal_range[47] * nullpos);
        upd = true;
    }
    if (sender() != CBlvm[0]) { //stage
        upd2 = upd2 || CL2_status_xy || CL2_status_z;
        CL2_status_xy = false;
        CL2_status_z = false;
        osc->DacsWidget->X2->setValue(Scanner2_data_pid_cal_range[45] * nullpos);
        osc->DacsWidget->Y2->setValue(Scanner2_data_pid_cal_range[46] * nullpos);
        osc->DacsWidget->Z2->setValue(Scanner2_data_pid_cal_range[47] * nullpos);
        auto scan = ScanPlot::GetIt();
        if (scan)
        {
            scan->deviceSchemeWid->ZPiezoPosEdit[0]->setValue(Scanner2_data_pid_cal_range[45] * nullpos);
            scan->deviceSchemeWid->ZPiezoPosEdit[1]->setValue(Scanner2_data_pid_cal_range[46] * nullpos);
            scan->deviceSchemeWid->ZPiezoPosEdit[2]->setValue(Scanner2_data_pid_cal_range[47] * nullpos);
        }
        upd = true;
    }
    auto scan = ScanPlot::GetIt();
    if (scan)
    {
        #ifndef _OLD_EXTRA_CONTROL
        scan->CBExLowVoltage->blockSignals(true);
        scan->CBExLowVoltage->setEnabled(false);
        if (Scanner_type->currentIndex() < 2)
            scan->CBExLowVoltage->setChecked(CBlvm[Scanner_type->currentIndex()]->isChecked());
        scan->CBExLowVoltage->blockSignals(false);
        #endif
    }
    QElapsedTimer timer;
    if (upd2) {
        CL_NLS_Lamps(Scanner_type->currentIndex());
        s_sendData();

        timer.start();
        while (!timer.hasExpired(500)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }
    if (upd) {
        osc->DacsWidget->X->setNewValue(true);
        osc->DacsWidget->Y->setNewValue(true);
        osc->DacsWidget->Z->setNewValue(true);
        osc->DacsWidget->X2->setNewValue(true);
        osc->DacsWidget->Y2->setNewValue(true);
        osc->DacsWidget->Z2->setNewValue(true);

        double x_range = Scanner_data_pid_cal_range[45];//mkm
        double y_range = Scanner_data_pid_cal_range[46];//mkm
        double z_range = Scanner_data_pid_cal_range[47];//mkm
        double x_range2 = Scanner2_data_pid_cal_range[45];//mkm
        double y_range2 = Scanner2_data_pid_cal_range[46];//mkm
        double z_range2 = Scanner2_data_pid_cal_range[47];//mkm
        double range1 = qMax(x_range, qMax(y_range, z_range));
        double range2 = qMax(x_range2, qMax(y_range2, z_range2));
        double timem = qMax(range1, range2) / free_move_speed;

        timer.start();
        while (!timer.hasExpired(300) && osc->DacsWidget->sendSP[0] && osc->DacsWidget->sendSP[1]) {
            qApp->processEvents();
            SleepExport::msleep2(1);
        }

        int time = 100 + 1000 * qMin(10., timem);
        timer.start();
        while (!timer.hasExpired(time)) {
            qApp->processEvents();
            SleepExport::msleep2(10);
        }
    }

    setLVM(CBlvm[0]->isChecked(), CBlvm[1]->isChecked());
    timer.start();
    while (!timer.hasExpired(300) && updLVM) {
        qApp->processEvents();
        SleepExport::msleep2(1);
    }

    timer.start();
    while (!timer.hasExpired(100)) {
        qApp->processEvents();
        SleepExport::msleep2(10);
    }


    read_calib_par(defaultFileName);

    foreach (QWidget *wid, cwl)
        wid->setEnabled(true);

    apply_calib_data();

    if (scan)
    {
        #ifndef _OLD_EXTRA_CONTROL
        scan->CBExLowVoltage->setEnabled(true);
        #endif
    }
}

void XYZCalibrations::apply_calib_data()
{
    int time = 1500;
    QElapsedTimer timer;
    timer.start();
    while (!scannersConnectedUpdated && !timer.hasExpired(time)) {
        qApp->processEvents();
        SleepExport::msleep2(1);
    }
    if (timer.hasExpired(time))
        goto go_to_center;

    bool haveHead;
    bool haveStage;
    haveHead = headConnected; //yeah, fuck gcc
    haveStage = stageConnected;
    bool sensors, pid;
    {
        QSettings settings("Param/gui.conf", QSettings::IniFormat);
        bool head_on = settings.value("display/oscilloscope/head", true).toBool();
        bool stage_on = settings.value("display/oscilloscope/stage", true).toBool();
        if (!head_on && haveHead) {
            haveHead = false;
            new Notification(tr("Head is connected, but disabled in settings."), Notification::Warn, 20);
        }
        if (!stage_on && haveStage) {
            haveStage = false;
            new Notification(tr("Stage is connected, but disabled in settings."), Notification::Warn, 20);
        }

        QSettings settings2("Param/controller.conf", QSettings::IniFormat);
        pid = settings2.value("autocalib_pid", false).toBool();
        sensors = settings2.value("autocalib_sensors", false).toBool();
    }

    if (sensors) {
        if (haveHead)
            haveHead = osc->acf->SensorReCalc(scHead, true);
        if (haveStage)
            haveStage = osc->acf->SensorReCalc(scStage, true);
    }
    if (pid) {
        if (haveHead)
            osc->acf->PICalcAuto(scHead, true);
        if (haveStage)
            osc->acf->PICalcAuto(scStage, true);
    }


go_to_center:
    osc->DacsWidget->go_to_center(0);
    osc->DacsWidget->go_to_center(1);
    osc->DacsWidget->go_to_center(2);
}

void XYZCalibrations::AutoDataLoad()
{
    LVMLoad();
    setLVM(CBlvm[0]->isChecked(), CBlvm[1]->isChecked());

    read_calib_par(defaultFileName);
    apply_calib_data();
}

void XYZCalibrations::save_To_File()
{
    LVMSave();
    z3Save();

    write_calib_par(defaultFileName);

    new Notification(tr("Calibrations saved"), Notification::Good, 10);
}

template <class T>
void writeValueArray(QTextStream &s, T *array, int size)
{
    for (int i = 0; i < size; i++)
        s << "v=" << array[i] << endl;
}

void XYZCalibrations::write_calib_par(QString filename, bool need_old)
{
    if (need_old) {
        QFile dataFile(filename);
        dataFile.open(QIODevice::WriteOnly | QIODevice::Text);

        QTextStream outData(&dataFile);
        outData.setRealNumberPrecision(11);

        outData << "NanoScanTechnologies ltd." << endl;
        outData << "Compressed Parameter File for scanner parameters" << endl << endl;

        bool AutLoad = false;
        writeValueArray(outData, &AutLoad, 1);

        writeValueArray(outData, Scanner_data_pid_cal_range, 48);
        writeValueArray(outData, Scanner_data_scale, 3);
        writeValueArray(outData, Scanner_data_Shift, 3);

        writeValueArray(outData, &CL_status_xy, 1);

        if (Nlc_status_cl_z) Nlc_status_cl_xy = true;
        writeValueArray(outData, &Nlc_status_cl_xy, 1);

        //----------------
        writeValueArray(outData, Scanner2_data_pid_cal_range, 48);
        writeValueArray(outData, Scanner2_data_scale, 3);
        writeValueArray(outData, Scanner2_data_Shift, 3);

        writeValueArray(outData, &CL2_status_xy, 1);

        if (Nlc2_status_cl_z) Nlc2_status_cl_xy = true;
        writeValueArray(outData, &Nlc2_status_cl_xy, 1);

        //-----------------
        writeValueArray(outData, &free_move_speed, 1);
        writeValueArray(outData, &free_move_speed, 1);

        writeValueArray(outData, &CL_status_z, 1);

        if (Nlc_status_ol_z) Nlc_status_ol_xy = true;
        writeValueArray(outData, &Nlc_status_ol_xy, 1);

        writeValueArray(outData, &CL2_status_z, 1);

        if (Nlc2_status_ol_z) Nlc2_status_ol_xy = true;
        writeValueArray(outData, &Nlc2_status_ol_xy, 1);

        double tmp[3];

        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeOL[i]->value();
        writeValueArray(outData, tmp, 3);
        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeCL[i]->value();
        writeValueArray(outData, tmp, 3);

        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeOL2[i]->value();
        writeValueArray(outData, tmp, 3);
        for (int i = 0; i < 3; i++)
            tmp[i] = lineScannerRangeCL2[i]->value();
        writeValueArray(outData, tmp, 3);

        dataFile.close();
    }

    //binary dump
    QString name = filename;
    name.truncate(name.lastIndexOf(".par"));
    float sz;
    {
        QFile file(name + "_calib_head" + (!CBlvm[0]->isChecked() ? "" : "_lvm") + ".bin");
        file.open(QIODevice::WriteOnly | QIODevice::Truncate);
        file.write((char *)Scanner_data_pid_cal_range, 48*sizeof(float));
        file.write((char *)Scanner_data_scale, 3*sizeof(DWORD));
        file.write((char *)Scanner_data_Shift, 3*sizeof(long));

        file.write((char *)&Nlc_status_ol_xy, 1);
        file.write((char *)&Nlc_status_cl_xy, 1);
        file.write((char *)&CL_status_xy, 1);
        file.write((char *)&CL_status_z, 1);

        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeOL[i]->value();
            file.write((char *)&sz, sizeof(float));
        }
        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeCL[i]->value();
            file.write((char *)&sz, sizeof(float));
        }

        file.write((char *)&free_move_speed, sizeof(float));

        file.close();
    }
    {
        QFile file(name + "_calib_stage" + (!CBlvm[1]->isChecked() ? "" : "_lvm") + ".bin");
        file.open(QIODevice::WriteOnly | QIODevice::Truncate);
        file.write((char *)Scanner2_data_pid_cal_range, 48*sizeof(float));
        file.write((char *)Scanner2_data_scale, 3*sizeof(DWORD));
        file.write((char *)Scanner2_data_Shift, 3*sizeof(long));

        file.write((char *)&Nlc2_status_ol_xy, 1);
        file.write((char *)&Nlc2_status_cl_xy, 1);
        file.write((char *)&CL2_status_xy, 1);
        file.write((char *)&CL2_status_z, 1);

        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeOL2[i]->value();
            file.write((char *)&sz, sizeof(float));
        }
        for (int i = 0; i < 3; i++) {
            sz = lineScannerRangeCL2[i]->value();
            file.write((char *)&sz, sizeof(float));
        }

        file.write((char *)&free_move_speed, sizeof(float));

        file.close();
    }
}

void XYZCalibrations::z3Save()
{
    QSettings settings("Param/z3calib.conf", QSettings::IniFormat);

    settings.setValue("cl_z", CL3_status_z);
    settings.setValue("ol_speed", lineFreeSpeed->value());

    settings.setValue("pid_p_z", linePIDProportionalZ3->value());
    settings.setValue("pid_i_z", linePIDIntegralZ3->value());
    settings.setValue("range_z", lineScannerRangeOLZ3->value());
    settings.setValue("range_cl_z", lineScannerRangeCLZ3->value());
    settings.setValue("cap_gain_z", lineSensorScaleZ3->value());
    settings.setValue("cap_shift_z", lineSensorScaleZ03->value());
}
void XYZCalibrations::z3Load()
{
    QSettings settings("Param/z3calib.conf", QSettings::IniFormat);

    CL3_status_z = settings.value("cl_z", false).toBool();
    lineFreeSpeed->setValue(settings.value("ol_speed", 100).toDouble());

    linePIDProportionalZ3->setValue(settings.value("pid_p_z", 0.05).toDouble());
    linePIDIntegralZ3->setValue(settings.value("pid_i_z", 0.001).toDouble());
    lineScannerRangeOLZ3->setValue(settings.value("range_z", 100).toDouble());
    lineScannerRangeCLZ3->setValue(settings.value("range_cl_z", 100).toDouble());
    lineSensorScaleZ3->setValue(settings.value("cap_gain_z", 100).toDouble());
    lineSensorScaleZ03->setValue(settings.value("cap_shift_z", 0).toDouble());
}


void XYZCalibrations::sl_XYZ_range()
{
    emit sig_XYZ_range(Scanner_data_pid_cal_range[45], Scanner_data_pid_cal_range[46], Scanner_data_pid_cal_range[47]);
    emit sig_XYZ_range2(Scanner2_data_pid_cal_range[45], Scanner2_data_pid_cal_range[46], Scanner2_data_pid_cal_range[47]);
    emit sig_XYZ_range3(Scanner3_data_pid_cal_range[45], Scanner3_data_pid_cal_range[46], Scanner3_data_pid_cal_range[47]);

    sl_setCalibVC();
}

void XYZCalibrations::sl_setCalibVC()
{
    const int vch[3] = {srcDACX, srcDACX2, srcDACZ3-2};
    const int cch[3] = {srcCX, srcCX2, srcCZ3-2};
    /*bool * const vcl[3][3] = {{&CL_status_xy, &CL_status_xy, &CL_status_z},
                              {&CL2_status_xy, &CL2_status_xy, &CL2_status_z},
                              {&CL3_status_z, &CL3_status_z, &CL3_status_z}};*/
    DoubleEdit ** const vrange[3] = {lineScannerRangeOL, lineScannerRangeOL2, (&lineScannerRangeOLZ3)-2};
    DoubleEdit ** const crange[3] = {lineScannerRangeCL, lineScannerRangeCL2, (&lineScannerRangeCLZ3)-2};
    for (int i = 0; i < 3; i++) {
        for (int j = (i != 2 ? 0 : 2); j < 3; j++) {
            double vgain = (150+15) / 20.;
            double vshift = 10 - 20 * (15./(150+15)); //-15v shift
            const bool use_v_meters = false;// !(*vcl[i][j]);
            if (use_v_meters) {
                vgain = vrange[i][j]->value()*1e-6 / 20.;
                vshift = 10;
            }
            input_gain_set(vch[i] + j, vgain);
            input_shift_set(vch[i] + j, vshift);
            input_unit_set(vch[i] + j, use_v_meters ? QObject::tr("m") : QObject::tr("V"));

            double cgain = crange[i][j]->value()*1e-6 / 20.;
            double cshift = 10;
            input_gain_set(cch[i] + j, cgain);
            input_shift_set(cch[i] + j, cshift);
            input_unit_set(cch[i] + j, QObject::tr("m"));
        }
    }

    emit input_calib_updated();
}


double XYZCalibrations::get_csb_speed()
{
#define MAX_PID_ID 8 //number of entities in table
    double pid_settings[MAX_PID_ID][2] = {
        //csb_rate, sensors_rate
        {1, 10000},
        {2, 5000},
        {3, 3333.333},
        {5, 2000},
        {10, 1000},
        {20, 500},
        {30, 333.333},
        {100, 100}
    };
    BYTE idx = qBound(0, csb_speed_idx, csb_speed_extended ? 7 : 3);
    return pid_settings[idx][1];
}

void XYZCalibrations::capBits(int n, long bits)
{
    if (n == 0) {
        bool new_csb = bool(bits & 2);
        if (new_csb) {
            csb_speed_extended = true;
            doSendCSBSpeed = true;
        }
    }
}

void XYZCalibrations::setSensorOverflowMode(BYTE scanner, BYTE mode)
{
    SensorOFon = mode;
    updateSensorOF = scanner+1;
}
void XYZCalibrations::resetWriter()
{
    updateSensorOF = 0;
    doWriteConfigToDeviceMem = false;
    doSendCalData = false;
    updLVM = false;

    sendStartup = true;

    csb_speed_extended = false;
    QSettings settings("Param/controller.conf", QSettings::IniFormat);
    csb_speed_idx = settings.value("csb_speed_settings_idx", 0).toInt();
    doSendCSBSpeed = true;

    QMetaObject::invokeMethod(this, "reset");
}
void XYZCalibrations::reset()
{
    headConnected = true;
    stageConnected = true;
    scannersConnectedUpdated = false;

    AutoDataLoad();
}
